/*
**  MailboxesViewController.m
**
**  Copyright (c) 2002
**
**  Author: Ludovic Marcotte <ludovic@Sophos.ca>   
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#import "MailboxesViewController.h"

#import "FolderNode.h"
#import "FolderNodePopUpItem.h"
#import "GNUMailConstants.h"

#ifndef MACOSX
#import "MailboxesView.h"
#endif

#import "MailboxManagerController.h"
#import "Utilities.h"

#import <Pantomime/LocalStore.h>
#import <Pantomime/URLName.h>

static MailboxesViewController *singleInstance = nil;

//
//
//
@implementation MailboxesViewController

- (id) initWithNibName: (NSString *) theName
{
  self = [super init];

#ifdef MACOSX
  
  if ( ![NSBundle loadNibNamed: theName
		  owner: self] )
    {
      NSDebugLog(@"Fatal error occured while loading the MailboxesView nib file");
      AUTORELEASE(self);
      return nil;
    }

  RETAIN(view);

#else
  // We link our view
  view = [[MailboxesView alloc] initWithParent: self];
  [view layoutView];

  // We link our outlets
  compactMailboxButton = [view compactMailboxButton];
  reuseMailWindowButton = [view reuseMailWindowButton];

  defaultInboxPopUpButton = [view defaultInboxPopUpButton];
  defaultOutboxPopUpButton = [view defaultOutboxPopUpButton];
  defaultDraftsPopUpButton = [view defaultDraftsPopUpButton];
  defaultTrashPopUpButton = [view defaultTrashPopUpButton];

#endif

  // This MUST be the first step
  allNodes = RETAIN([Utilities initializeLocalAndIMAPFolderNodes]);

  // We get our defaults for this panel
  [self initializeFromDefaults];

  return self;
}


//
//
//
- (void) dealloc
{
  NSDebugLog(@"MailboxesViewController: -dealloc");

  singleInstance = nil;
  RELEASE(view);

  RELEASE(allNodes);

  [super dealloc];
}


//
// action methods
//






//
// access methods
//
- (NSImage *) image
{
  NSBundle *aBundle;
  
  aBundle = [NSBundle bundleForClass: [self class]];
  
  return AUTORELEASE([[NSImage alloc] initWithContentsOfFile:
					[aBundle pathForResource: @"MailIcon_mailboxes" ofType: @"tiff"]]);
}

- (NSString *) name
{
  return _(@"Mailboxes");
}

- (NSView *) view
{
  return view;
}

- (BOOL) hasChangesPending
{
  return YES;
}


//
//
//
- (void) initializeFromDefaults
{
  FolderNodePopUpItem *aPopUpItem;

  // We first populate our popup buttons
  [Utilities addItemsToPopUpButton: defaultInboxPopUpButton   usingFolderNodes: allNodes];
  [Utilities addItemsToPopUpButton: defaultOutboxPopUpButton  usingFolderNodes: allNodes];
  [Utilities addItemsToPopUpButton: defaultDraftsPopUpButton  usingFolderNodes: allNodes];
  [Utilities addItemsToPopUpButton: defaultTrashPopUpButton   usingFolderNodes: allNodes];

  
  //
  // Default Inbox
  //
  aPopUpItem = [Utilities folderNodePopUpItemForURLNameAsString: [[NSUserDefaults standardUserDefaults] 
								   objectForKey: @"INBOXFOLDERNAME"]
			  usingFolderNodes: allNodes
			  popUpButton: defaultInboxPopUpButton];
  
  if ( aPopUpItem )
    {
      [defaultInboxPopUpButton selectItem: aPopUpItem];
    }
  else
    {
      // FIXME
    }
  
  
  //
  // Default Outbox
  //
  aPopUpItem = [Utilities folderNodePopUpItemForURLNameAsString: [[NSUserDefaults standardUserDefaults] 
								   objectForKey: @"OUTBOXFOLDERNAME"]
			  usingFolderNodes: allNodes
			  popUpButton: defaultOutboxPopUpButton];
  
  if ( aPopUpItem )
    {
      [defaultOutboxPopUpButton selectItem: aPopUpItem];
    }
  else
    {
      // FIXME
    }
  

  //
  // Default Drafts
  //
  aPopUpItem = [Utilities folderNodePopUpItemForURLNameAsString: [[NSUserDefaults standardUserDefaults] 
								   objectForKey: @"DRAFTSFOLDERNAME"]
			  usingFolderNodes: allNodes
			  popUpButton: defaultDraftsPopUpButton];
  
  if ( aPopUpItem )
    {
      [defaultDraftsPopUpButton selectItem: aPopUpItem];
    }
  else
    {
      // FIXME
    }
  

  //
  // Default Trash
  //
  aPopUpItem = [Utilities folderNodePopUpItemForURLNameAsString: [[NSUserDefaults standardUserDefaults] 
								   objectForKey: @"TRASHFOLDERNAME"]
			  usingFolderNodes: allNodes
			  popUpButton: defaultTrashPopUpButton];
  
  if ( aPopUpItem )
    {
      [defaultTrashPopUpButton selectItem: aPopUpItem];
    }
  else
    {
      // FIXME
    }


  //
  // COMPACT_ON_CLOSE
  //
  if ( [[NSUserDefaults standardUserDefaults] objectForKey: @"COMPACT_MAILBOX_ON_CLOSE"] )
    {
      [compactMailboxButton setState: [[NSUserDefaults standardUserDefaults] 
					integerForKey: @"COMPACT_MAILBOX_ON_CLOSE"]];
    }
  else
    {
      [compactMailboxButton setState: NSOffState];
    }

  //
  // REUSE_MAILWINDOW
  //
  if ( [[NSUserDefaults standardUserDefaults] objectForKey: @"REUSE_MAILWINDOW"] )
    {
      [reuseMailWindowButton setState: [[NSUserDefaults standardUserDefaults] 
					 integerForKey: @"REUSE_MAILWINDOW"]];
    }
  else
    {
      [reuseMailWindowButton setState: NSOnState];
    }
}


//
//
//
- (void) saveChanges
{
  // We first synchronize our popup buttons
  [defaultInboxPopUpButton synchronizeTitleAndSelectedItem];
  [defaultOutboxPopUpButton synchronizeTitleAndSelectedItem];
  [defaultDraftsPopUpButton synchronizeTitleAndSelectedItem];
  [defaultTrashPopUpButton synchronizeTitleAndSelectedItem];


  [self _saveChangesForPopUpButton: defaultInboxPopUpButton
	name: @"INBOXFOLDERNAME"];
  
  [self _saveChangesForPopUpButton: defaultOutboxPopUpButton
	name: @"OUTBOXFOLDERNAME"];
  
  [self _saveChangesForPopUpButton: defaultDraftsPopUpButton
	name: @"DRAFTSFOLDERNAME"];
  
  [self _saveChangesForPopUpButton: defaultTrashPopUpButton
	name: @"TRASHFOLDERNAME"];

  [[NSUserDefaults standardUserDefaults] setInteger: [compactMailboxButton state]
					 forKey: @"COMPACT_MAILBOX_ON_CLOSE"];

  [[NSUserDefaults standardUserDefaults] setInteger: [reuseMailWindowButton state]
					 forKey: @"REUSE_MAILWINDOW"];
}


//
// class methods
//
+ (id) singleInstance
{
  if ( !singleInstance)
    {
      singleInstance = [[MailboxesViewController alloc] initWithNibName: @"MailboxesView"];
    }

  return singleInstance;
}

@end


//
//
//
@implementation MailboxesViewController (Private)

- (void) _saveChangesForPopUpButton: (NSPopUpButton *) thePopUpButton
			       name: (NSString *) theName
{
  FolderNode *aFolderNode;
  NSString *aString;
  
  //
  // FIXME: We must verify for, at least, local store if the folder is "selectable". 
  //
  aFolderNode = [(FolderNodePopUpItem *)[thePopUpButton selectedItem] folderNode];
  
  if ( [aFolderNode parent] == allNodes )
    {
      NSDebugLog(@"Selected a wrong mailbox, ignoring.");
      return;
    }

  aString = [Utilities stringValueOfURLNameFromFolderNode: aFolderNode];
  
  [[NSUserDefaults standardUserDefaults] setObject: aString
					 forKey: theName];
  
}

@end
