/*
**  IMAPViewController.m
**
**  Copyright (c) 2001, 2002
**
**  Author: Ludovic Marcotte <ludovic@Sophos.ca>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#import "IMAPViewController.h"

#import "FolderNode.h"
#import "GNUMailConstants.h"
#import "IMAPView.h"
#import "LabelWidget.h"
#import "NSStringExtensions.h"
#import "PasswordPanelController.h"
#import "ReceivingWindowController.h"
#import "Utilities.h"

#import <Pantomime/Constants.h>
#import <Pantomime/IMAPStore.h>
#import <Pantomime/IMAPFolder.h>

@implementation IMAPViewController

//
//
//
- (id) initWithParent: (id) theParent;
{
  self = [super init];
  
  parent = theParent;
  allFolders = [[FolderNode alloc] init];
  [allFolders setName: _(@"All Folders")];
  store = nil;

#ifdef MACOSX

  if ( ![NSBundle loadNibNamed: @"IMAPView" owner: self] )
    {
      NSDebugLog(@"Fatal error occured while loading the IMAPView nib file");
      AUTORELEASE(self);
      return nil;
    }
  
  RETAIN(imapView);
  mailboxColumn = [outlineView tableColumnWithIdentifier: @"Folders"];
  subscriptionColumn = [outlineView tableColumnWithIdentifier: @"Subscribed"];
  
  [outlineView setDoubleAction: @selector(doubleClickedOnNode:)];

#else
  // We link our view
  imapView = [[IMAPView alloc] initWithParent: self];

  retrieveMinutesField = [(IMAPView*)imapView retrieveMinutesField];
  outlineView = [(IMAPView*)imapView outlineView];

  mailboxColumn = [(IMAPView*)imapView mailboxColumn];
  subscriptionColumn = [(IMAPView*)imapView subscriptionColumn];
#endif

  return self;
}


//
//
//
- (void) dealloc
{
  //NSDebugLog(@"IMAPViewController: -dealloc");
    
  RELEASE(allFolders);
  RELEASE(imapView);

  [store close];
  RELEASE(store);

  [super dealloc];
}


//
// Datasource/delegate methods for the outline view
//
- (id) outlineView: (NSOutlineView *) outlineView
	     child: (int) index
	    ofItem: (id) item
{
  
  // root object
  if ( !item )
    {
      return allFolders;
    }
  
  
  if ( [item isKindOfClass: [FolderNode class]] )
    {
      return [(FolderNode *)item childAtIndex: index];
    }
  
  return nil;
}


//
//
//
- (BOOL) outlineView: (NSOutlineView *) outlineView
    isItemExpandable: (id) item
{
  if ( [item isKindOfClass: [FolderNode class]] )
    {
      if ( [(FolderNode *)item childCount] > 0 )
	{
	  return YES;
	}
      else
	{
	  return NO;
	}
    }

  return NO;
}


//
//
//
- (int)        outlineView: (NSOutlineView *) outlineView 
    numberOfChildrenOfItem: (id) item
{
  // Root, always one element
  if ( !item )
    {
      return 1;
    }

  if ( [item isKindOfClass: [FolderNode class]] )
    {
      return [(FolderNode *)item childCount];
    }

  return 0;
}


//
//
//
- (id)         outlineView: (NSOutlineView *) outlineView 
 objectValueForTableColumn: (NSTableColumn *) tableColumn 
		    byItem: (id) item
{
  if ( tableColumn == mailboxColumn )
    {
      return [(FolderNode *)item name];
    }
  else
    {
      return ([(FolderNode *)item subscribed] == YES ? _(@"Y") : @""); 
    }

  // Never reached...
  return nil;
}



//
// Action methods
//
- (IBAction) doubleClickedOnNode: (id) sender
{
  FolderNode *aFolderNode;
  int row;

  if ( !store )
    {
      NSBeep();
      return;
    }

  row = [outlineView selectedRow];
  aFolderNode = (FolderNode *)[outlineView itemAtRow: row];

  if ( [aFolderNode subscribed] )
    {
      [self unsubscribe: nil];
    }
  else
    {
      [self subscribe: nil];
    }
}


//
//
//
- (IBAction) list: (id) sender
{  
  NSEnumerator *theEnumerator;
  FolderNode *nodes;

  NSString *aPassword;
  
  [allFolders setChildren: nil];

  if ( [[[parent serverNameField] stringValue] length] == 0 || 
       [[[parent usernameField] stringValue] length] == 0)
    {
      NSBeep();
      return;
    }

  aPassword = [[[parent passwordSecureField] stringValue] stringByTrimmingWhiteSpaces];
  
  if ( [aPassword length] == 0 )
    {
      PasswordPanelController *theController; 
      int result;
      
      theController = [[PasswordPanelController alloc] initWithWindowNibName: @"PasswordPanel"];
      [[theController window] setTitle: [NSString stringWithFormat: @"%@ @ %@",
						  [[parent usernameField] stringValue],
						  [[parent serverNameField] stringValue]] ];
      
      result = [NSApp runModalForWindow: [theController window]];
      
      // If the user has entered a password...
      if (result == NSRunStoppedResponse)
	{
	  aPassword = [theController password];
	}
      else
	{
	  aPassword = @"";
	}
    }


  if ( !store )
    {
      if ( [[parent useSecureConnection] state] == NSOnState )
	{
	  store = [[IMAPStore alloc] initSSLWithName: [[parent serverNameField] stringValue]
				     port: [[parent serverPortField] intValue]];
	  
	}
      else
	{
	  store = [[IMAPStore alloc] initWithName: [[parent serverNameField] stringValue]
				     port: [[parent serverPortField] intValue]];
	  
	}
      if ( ![store authenticateWithUsername: [[parent usernameField] stringValue]
		   password: aPassword] )
	{
	  NSRunInformationalAlertPanel(_(@"Error!"),
				       _(@"Authentication error with the IMAP server."),
				       _(@"OK"),
				       NULL,
				       NULL,
				       NULL);
	  
	  [store close];
	  DESTROY(store);
	  
	  return;
	}
    }
  


  theEnumerator = [store folderEnumerator];
  
  if ( !theEnumerator )
    {
      NSRunInformationalAlertPanel(_(@"Error!"),
				   _(@"An error occured while listing the folders on this IMAP server."),
				   _(@"OK"),
				   NULL,
				   NULL,
				   NULL);
      return;
    }
  
  nodes = [Utilities folderNodesFromFolders: theEnumerator
		     separator: [store folderSeparator]];
  RETAIN(nodes);
  
  [allFolders setChildren: [nodes children]];
  RELEASE(nodes);
  
  [self _rebuildListOfSubscribedFolders];

  [outlineView reloadData];
}


//
//
//
- (IBAction) subscribe: (id) sender
{
  FolderNode *aFolderNode;
  NSString *aString;
  int row;

  if ( !store )
    {
      NSBeep();
      return;
    }

  row = [outlineView selectedRow];
  aFolderNode = (FolderNode *)[outlineView itemAtRow: row];

  if ( [aFolderNode subscribed] )
    {
      NSRunInformationalAlertPanel(_(@"Error!"),
				   _(@"You are already subscribed to this IMAP folder."),
				   _(@"OK"),
				   NULL,
				   NULL,
				   NULL);
      return;
    }

  aString = [Utilities pathOfFolderFromFolderNode: aFolderNode
		       separator: [store folderSeparator]];

  if ( aFolderNode == allFolders ||
       [store folderTypeForFolderName: aString] == HOLDS_FOLDERS )
    {
      NSRunInformationalAlertPanel(_(@"Error!"),
				   _(@"You cannot subscribe to this folder."),
				   _(@"OK"),
				   NULL,
				   NULL,
				   NULL);
      return;
    }
   
  if ( [store subscribeToFolderWithName: aString] )
    {
      [aFolderNode setSubscribed: YES];  
      [outlineView setNeedsDisplay: YES];
    }
  else
    {
      NSRunInformationalAlertPanel(_(@"Error!"),
				   _(@"An error occured while subscribing to folder:\n%@."),
				   _(@"OK"),
				   NULL,
				   NULL,
				   aString);
    }
}


//
//
//
- (IBAction) unsubscribe: (id) sender
{
  FolderNode *aFolderNode;
  NSString *aString;
  int row;

  if ( !store )
    {
      NSBeep();
      return;
    }

  row = [outlineView selectedRow];
  aFolderNode = (FolderNode *)[outlineView itemAtRow: row];

  if ( ![aFolderNode subscribed] )
    {
      NSRunInformationalAlertPanel(_(@"Error!"),
				   _(@"You are already unsubscribed from this IMAP folder."),
				   _(@"OK"),
				   NULL,
				   NULL,
				   NULL);
      return;
    }

  
  aString = [Utilities pathOfFolderFromFolderNode: aFolderNode
		       separator: [store folderSeparator]];
  
  if ( [store unsubscribeToFolderWithName: aString] )
    {
      NSString *pathToFile;

      [aFolderNode setSubscribed: NO];  
      [outlineView setNeedsDisplay: YES];

      // We remove the cache file.
      pathToFile = [NSString stringWithFormat: @"%@/IMAPCache_%@_%@", 
			     GNUMailUserLibraryPath(),
			     [store name],
			     [Utilities flattenPathFromString: aString
					separator: [store folderSeparator]] ];
      
      [[NSFileManager defaultManager] removeFileAtPath: pathToFile
				      handler: nil];
    }
  else
    {
      NSRunInformationalAlertPanel(_(@"Error!"),
				   _(@"An error occured while unsubscribing to folder:\n%@."),
				   _(@"OK"),
				   NULL,
				   NULL,
				   aString);
    }
}


//
// access / mutation methods
//
- (NSTextField *) retrieveMinutesField
{
  return retrieveMinutesField;
}

- (NSOutlineView *) outlineView
{
  return outlineView;
}

- (NSView *) imapView
{
  return imapView;
}

@end


//
// Private methods
//
@implementation IMAPViewController (Private)

- (void) _rebuildListOfSubscribedFolders
{
  NSEnumerator *theEnumerator;
  FolderNode *aFolderNode;
  NSString *aPath;
  
  theEnumerator = [store subscribedFolderEnumerator];

  while ( (aPath = [theEnumerator nextObject]) )
    { 
      aFolderNode = [Utilities folderNodeForPath: aPath
			       using: allFolders];

      if ( aFolderNode &&
	   aFolderNode != allFolders )
	{
	  [aFolderNode setSubscribed: YES];
	}
    }
}

@end
