/*
**  ReceivingViewController.m
**
**  Copyright (c) 2001, 2002
**
**  Author: Ludovic Marcotte <ludovic@Sophos.ca>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#import "ReceivingViewController.h"

#import "GNUMailConstants.h"
#import "MailboxManagerController.h"
#import "ReceivingWindowController.h"
#import "Utilities.h"

#ifndef MACOSX
#import "ReceivingView.h"
#endif

static ReceivingViewController *singleInstance = nil;

@implementation ReceivingViewController

- (id) initWithNibName: (NSString *) theName
{
  self = [super init];

#ifdef MACOSX
  
  if ( ![NSBundle loadNibNamed: theName
		  owner: self] )
    {
      NSDebugLog(@"Fatal error occured while loading the ReceivingView nib file");
      AUTORELEASE(self);
      return nil;
    }

  RETAIN(view);

  [tableView setTarget: self];
  [tableView setDoubleAction: @selector(edit:)];

#else
  // We link our view
  view = [[ReceivingView alloc] initWithParent: self];
  [view layoutView];

  // We link our outlets
  minutesField = [view minutesField];
  matrix = [(ReceivingView *)view matrix];

  showFilterPanelButton = [view showFilterPanelButton];
  showNoNewMessagesPanelButton = [view showNoNewMessagesPanelButton];

  tableView = [view tableView];

#endif
    
  // We get our defaults for this panel
  [self initializeFromDefaults];
  [self updateServersList];

  return self;
}


//
//
//
- (void) dealloc
{
  singleInstance = nil;

  // Cocoa bug?
#ifdef MACOSX
  [tableView setDataSource: nil];
#endif

  RELEASE(allServers);
  RELEASE(view);

  [super dealloc];
}


//
// delegate methods
//
- (int) numberOfRowsInTableView: (NSTableView *)aTableView
{
  return [allServers count];
}

- (id)           tableView: (NSTableView *)aTableView
 objectValueForTableColumn: (NSTableColumn *)aTableColumn
                       row:(int)rowIndex
{
  NSDictionary *allValues;
  NSArray *anArray;
  NSNumber *aNumber;

  anArray = [allServers allKeys];

  allValues = [allServers objectForKey: [anArray objectAtIndex: rowIndex] ];
  
  aNumber = [allValues objectForKey: @"SERVERTYPE"];

  if ( !aNumber )
    {
      aNumber = [NSNumber numberWithInt: POP3];
    }
  
  if ( [aNumber intValue] == POP3 || [aNumber intValue] == IMAP )
    {
      return [NSString stringWithFormat: @"%@ @ %@ [%@]",
		       [allValues objectForKey: @"USERNAME"],
		       ([allValues objectForKey: @"SERVERNAME"] ? [allValues objectForKey: @"SERVERNAME"] : 
			[anArray objectAtIndex: rowIndex]),
		       ([aNumber intValue] == POP3 ? _(@"POP3") : _(@"IMAP"))];
    }
  else
    {
      return [NSString stringWithFormat: _(@"%@ [UNIX]"),
		       [allValues objectForKey: @"MAILSPOOLFILE"]];
    }

  // Never reached.
  return nil;
}


//
// action methods 
//
- (IBAction) add: (id) sender
{
  ReceivingWindowController *receivingWindowController;
  int result;
 
  receivingWindowController = [[ReceivingWindowController alloc] initWithWindowNibName: @"ReceivingWindow"];
  [receivingWindowController setKey: nil];

  result = [NSApp runModalForWindow: [receivingWindowController window]];

  if (result == NSRunStoppedResponse)
    {
      [self updateServersList];
    }
}


- (IBAction) edit: (id) sender
{
  ReceivingWindowController *receivingWindowController;
  NSArray *anArray;
  NSDictionary *allValues;
  int result;

  if ([tableView selectedRow] < 0)
    {
      NSBeep();
      return;
    }
  
  anArray = [allServers allKeys];
  allValues = [allServers objectForKey: [anArray objectAtIndex:[tableView selectedRow]]];

  if ( [[MailboxManagerController singleInstance] storeForName: [allValues objectForKey: @"SERVERNAME"]
						  username: [allValues objectForKey: @"USERNAME"]] )
    {
      NSRunInformationalAlertPanel(_(@"Error!"),
				   _(@"To edit this account, you must first close the connection with the server."),
				   _(@"OK"),
				   NULL, 
				   NULL,
				   NULL);
      return;
    }

  receivingWindowController = [[ReceivingWindowController alloc] initWithWindowNibName: @"ReceivingWindow"];
  [receivingWindowController setKey: [anArray objectAtIndex:[tableView selectedRow]] ];
  [receivingWindowController initializeFromDefaults];

  result = [NSApp runModalForWindow: [receivingWindowController window]];
  
  if (result == NSRunStoppedResponse)
    {
      [self updateServersList];
    }
}


//
//
//
- (IBAction) delete: (id) sender
{
  int choice;
  NSString *aKey;
  NSArray *allKeys;
  NSDictionary *allValues;
  
  
  if ( [tableView selectedRow] < 0 ) 
    {
      NSBeep();
      return;
    }
  
  allKeys = [allServers allKeys];
  aKey = [allKeys objectAtIndex: [tableView selectedRow]];
  allValues = [allServers objectForKey: aKey];

  if ( [[MailboxManagerController singleInstance] storeForName: [allValues objectForKey: @"SERVERNAME"]
						  username: [allValues objectForKey: @"USERNAME"]] )
    {
      NSRunInformationalAlertPanel(_(@"Error!"),
				   _(@"To delete this account, you must first close the connection with the server."),
				   _(@"OK"),
				   NULL, 
				   NULL,
				   NULL);
      return;
    } 
  
  choice = NSRunAlertPanel(_(@"Delete..."),
			   _(@"Are you sure you want to delete this receiving server entry?"),
			   _(@"Cancel"), // default
			   _(@"Yes"),    // alternate
			   nil);

  if (choice == NSAlertAlternateReturn)
    {
      // We remove the entry from the user defaults db
      NSMutableDictionary *allPreferences;
      
      allPreferences = [[NSMutableDictionary alloc] init];
      [allPreferences addEntriesFromDictionary: [[NSUserDefaults standardUserDefaults] volatileDomainForName: @"PREFERENCES"]];
      allKeys = [allServers allKeys];
      aKey = [allKeys objectAtIndex: [tableView selectedRow]];

      // We remove our entry from our dictionary of servers
      [allServers removeObjectForKey: aKey];
      
      // We update the volatile user defaults with the new receiving servers list
      [allPreferences setObject: allServers forKey: @"RECEIVING"];
#ifndef MACOSX
      [[NSUserDefaults standardUserDefaults] removeVolatileDomainForName: @"PREFERENCES"];
#endif
      [[NSUserDefaults standardUserDefaults] setVolatileDomain: allPreferences
					     forName: @"PREFERENCES"];
      
      [self updateServersList];
      
      if ( [allServers count] > 0 ) 
	{
	  [tableView selectRow: 0 
		     byExtendingSelection: NO];
	}
    }
}


//
//
//
- (IBAction) setManually: (id) sender
{
  [minutesField setEditable: NO];
}


//
//
//
- (IBAction) setAutomatically: (id) sender
{
  [minutesField setEditable: YES];
}



//
// access methods
//
- (NSImage *) image
{
  NSBundle *aBundle;
  
  aBundle = [NSBundle bundleForClass: [self class]];
  
  return AUTORELEASE([[NSImage alloc] initWithContentsOfFile:
					[aBundle pathForResource: @"MailIcon_retrieve" ofType: @"tiff"]]);
}

- (NSString *) name
{
  return _(@"Receiving");
}

- (NSView *) view
{
  return view;
}

- (BOOL) hasChangesPending
{
  return YES;
}


//
//
//
- (void) initializeFromDefaults
{
  int retrieveMethod, retrieveMinutes;
    
  retrieveMethod = [[NSUserDefaults standardUserDefaults] integerForKey: @"RETRIEVEMETHOD"];
  retrieveMinutes = [[NSUserDefaults standardUserDefaults] integerForKey: @"RETRIEVEMINUTES"];

  // RETRIEVEMETHOD
  if ( retrieveMethod )
    {
      [matrix selectCellAtRow: 1 column: 0];
      [minutesField setEditable: YES];
    }
  else
    {
      [matrix selectCellAtRow: 0 column: 0];
      [minutesField setEditable: NO];
    }
  
  // RETRIEVEMINUTES
  if ( retrieveMinutes )
    {
      [minutesField setIntValue: retrieveMinutes];
    }
  else
    {
      [minutesField setStringValue: @"1"];
    }
  
    
  // SHOW_FILTER_PANEL
  if ( [[NSUserDefaults standardUserDefaults] objectForKey: @"SHOW_FILTER_PANEL"] )
    {
      [showFilterPanelButton setState: [[NSUserDefaults standardUserDefaults] 
					 integerForKey: @"SHOW_FILTER_PANEL"]];
    }
  else
    {
      [showFilterPanelButton setState: NSOnState];
    }

  // SHOW_NO_NEW_MESSAGES_PANEL
  if ( [[NSUserDefaults standardUserDefaults] objectForKey: @"SHOW_NO_NEW_MESSAGES_PANEL"] )
    {
      [showNoNewMessagesPanelButton setState: [[NSUserDefaults standardUserDefaults] 
						integerForKey: @"SHOW_NO_NEW_MESSAGES_PANEL"]];
    }
  else
    {
      [showNoNewMessagesPanelButton setState: NSOnState];
    }
}


//
//
//
- (void) saveChanges
{
  NSEnumerator *serversEnumerator, *filesEnumerator;
  NSString *aKey, *aString;
  
  // We compare the retrieiving method we had and the one we are about the save
  // to know if we must start or stop our timer in the local inbox window controller
  if ([[NSUserDefaults standardUserDefaults] 
	integerForKey: @"RETRIEVEMETHOD"] != [matrix selectedRow])
    {
      [[NSUserDefaults standardUserDefaults] setInteger: [minutesField intValue]
					     forKey: @"RETRIEVEMINUTES"];
      
      if ( [matrix selectedRow] )
	{
	  // We inform our LocalInboxWindowController to start the timer
	  NSDebugLog(@"Starting the timer!");
	  
	  [[NSNotificationCenter defaultCenter]
	    postNotificationName: ShouldStartLocalInboxTimer
	    object: nil
	    userInfo: nil];
	}
      else
	{
	  // We inform our LocalInboxWindowController to stop the timer
	  NSDebugLog(@"Stopping the timer!");

	  [[NSNotificationCenter defaultCenter]
	    postNotificationName: ShouldStopLocalInboxTimer
	    object: nil
	    userInfo: nil];
	}
    } 
  else if (([matrix selectedRow] == 1) 
	   && ([[NSUserDefaults standardUserDefaults] 
		 integerForKey:@"RETRIEVEMINUTES"] != [minutesField intValue]))
    {
      NSDebugLog(@"Restarting the timer!");
      [[NSUserDefaults standardUserDefaults] setInteger: [minutesField intValue]
					     forKey: @"RETRIEVEMINUTES"];
      
      [[NSNotificationCenter defaultCenter]
	postNotificationName: ShouldStopLocalInboxTimer
	object: nil
	userInfo: nil];
      [[NSNotificationCenter defaultCenter]
	postNotificationName: ShouldStartLocalInboxTimer
	object: nil
	userInfo: nil];
    }
  else
    {  
      [[NSUserDefaults standardUserDefaults] setInteger: [minutesField intValue]
					     forKey: @"RETRIEVEMINUTES"];
    }
  
  [[NSUserDefaults standardUserDefaults] setInteger: [matrix selectedRow]
					 forKey: @"RETRIEVEMETHOD"];
  
  [[NSUserDefaults standardUserDefaults] setInteger: [showFilterPanelButton state]
					 forKey: @"SHOW_FILTER_PANEL"];

  [[NSUserDefaults standardUserDefaults] setInteger: [showNoNewMessagesPanelButton state]
					 forKey: @"SHOW_NO_NEW_MESSAGES_PANEL"];
  
  
  // We verify which accounts have been renamed or deleted
  serversEnumerator = [(NSDictionary *)[[NSUserDefaults standardUserDefaults] objectForKey: @"RECEIVING"] keyEnumerator];
  
  while ((aKey = [serversEnumerator nextObject]))
    {
      if (! [allServers objectForKey: aKey])
	{
	  // We remove our cache files
	  NSDebugLog(@"Deleting cache files associated with %@", aKey);
	  
	  filesEnumerator = [[NSFileManager defaultManager] enumeratorAtPath: GNUMailUserLibraryPath()];
      
	  while ( (aString = [filesEnumerator nextObject]) )
	    {
	      // We verify for POP3Cache_server.name.com
	      if ( [aString isEqualToString: [NSString stringWithFormat: @"POP3Cache_%@", [Utilities flattenPathFromString: aKey
												     separator: @"/"]]] )
		{
		  [[NSFileManager defaultManager] removeFileAtPath: [NSString stringWithFormat: @"%@/%@",
									      GNUMailUserLibraryPath(), aString]
						  handler: nil];
	    }
	  
	      // We verify for IMAPCache_server.name.com_mailbox
	      if ( [aString hasPrefix: [NSString stringWithFormat: @"IMAPCache_%@_", [Utilities flattenPathFromString: aKey
												separator: @"/"]]] )
		{
		  [[NSFileManager defaultManager] removeFileAtPath: [NSString stringWithFormat: @"%@/%@",
									      GNUMailUserLibraryPath(), aString]
						  handler: nil];
		}
	    }
	}
    }

  [[NSUserDefaults standardUserDefaults] setObject: 
					   [[[NSUserDefaults standardUserDefaults] volatileDomainForName: @"PREFERENCES"] 
					     objectForKey: @"RECEIVING"]
					 forKey: @"RECEIVING"];

  // We inform that our receiving values have changed
  [[NSNotificationCenter defaultCenter]
    postNotificationName: ReceivingValuesHaveChanged
    object: nil
    userInfo: nil];
}


//
// other methods
//
- (void) updateServersList
{
  NSMutableDictionary *allPreferences;

  TEST_RELEASE(allServers);

  // We now get the servers list from the volatile user defaults
  allPreferences = [[NSMutableDictionary alloc] init];
  allServers = [[NSMutableDictionary alloc] init];
  
  [allPreferences addEntriesFromDictionary: [[NSUserDefaults standardUserDefaults] volatileDomainForName: @"PREFERENCES"]];

  if ( [allPreferences objectForKey: @"RECEIVING"] )
    {
      [allServers addEntriesFromDictionary: [allPreferences objectForKey: @"RECEIVING"]];
    }
  
  // We update the volatile user defaults with the new receiving servers list
  [allPreferences setObject: allServers forKey: @"RECEIVING"];
  // FIXME - This is causing a segfault under OS X
#ifndef MACOSX
  [[NSUserDefaults standardUserDefaults] removeVolatileDomainForName: @"PREFERENCES"];
#endif
  [[NSUserDefaults standardUserDefaults] setVolatileDomain: allPreferences
					 forName: @"PREFERENCES"];
  
  RELEASE(allPreferences);

  [tableView reloadData];
}


//
// class methods
//
+ (id) singleInstance
{
  if ( !singleInstance )
    {
      singleInstance = [[ReceivingViewController alloc] initWithNibName: @"ReceivingView"];
    }

  return singleInstance;
}

@end
