/*
**  ReceivingWindowController.m
**
**  Copyright (c) 2001, 2002
**
**  Author: Ludovic Marcotte <ludovic@Sophos.ca>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#import "ReceivingWindowController.h"

#import "FolderNode.h"
#import "FolderNodePopUpItem.h"
#import "GNUMailConstants.h"

#import "IMAPViewController.h"
#import "POP3ViewController.h"
#import "UNIXViewController.h"

#ifndef MACOSX
#import "ReceivingWindow.h"
#endif

#import "Utilities.h"

#define IMAP_SERVICE_PORT  143
#define POP3_SERVICE_PORT  110
#define IMAPS_SERVICE_PORT 993
#define POP3S_SERVICE_PORT 995

@implementation ReceivingWindowController

- (id) initWithWindowNibName: (NSString *) windowNibName
{
#ifdef MACOSX
  
  self = [super initWithWindowNibName: windowNibName];
  
#else
  ReceivingWindow *receivingWindow;

  receivingWindow = [[ReceivingWindow alloc] initWithContentRect: NSMakeRect(300,300,400,305)
					     styleMask: NSTitledWindowMask|NSResizableWindowMask
					     backing: NSBackingStoreRetained
					     defer: NO];

  [super initWithWindow: receivingWindow];

  [receivingWindow layoutWindow];
  [receivingWindow setDelegate: self];

  // We link our outlets
  tabView = [receivingWindow tabView];
  tabViewItem1 = [receivingWindow tabViewItem1];
  tabViewItem2 = [receivingWindow tabViewItem2];

  serverNameField = [receivingWindow serverNameField];
  serverPortField = [receivingWindow serverPortField];
  usernameField = [receivingWindow usernameField];
  passwordSecureField = [receivingWindow passwordSecureField];
  
  popup = [receivingWindow popup];
  rememberPassword = [receivingWindow rememberPassword];
  verifyOnStartup = [receivingWindow verifyOnStartup];
  useSecureConnection = [receivingWindow useSecureConnection];

  RELEASE(receivingWindow);
#endif

  // We initialize our IMAPViewController, POP3ViewController and UNIXViewController objects
  imapViewController = [[IMAPViewController alloc] initWithParent: self];
  pop3ViewController = [[POP3ViewController alloc] initWithParent: self];
  unixViewController = [[UNIXViewController alloc] initWithParent: self];

  [[self window] setTitle: _(@"Receiving preferences")];

  return self;
}


//
//
//
- (void) dealloc
{
  //NSDebugLog(@"ReceivingWindowController: -dealloc");

  RELEASE(imapViewController);
  RELEASE(pop3ViewController);

  TEST_RELEASE(key);
  
  [super dealloc];
}


//
//
//
- (void) initializeFromDefaults
{
  NSDictionary *allValues;
  NSNumber *serverTypeValue, *portValue;
  
  // We now get all the data from the volatile user defaults
  allValues = [[[[NSUserDefaults standardUserDefaults] volatileDomainForName: @"PREFERENCES"] 
		 objectForKey: @"RECEIVING"] objectForKey: [self key]];
  
  // We decode our type
  serverTypeValue = [allValues objectForKey: @"SERVERTYPE"];
  
  if (serverTypeValue && [serverTypeValue intValue] == IMAP)
    {
      [popup selectItemWithTitle: _(@"IMAP")];
    }
  else if (serverTypeValue && [serverTypeValue intValue] == POP3)
    {
      [popup selectItemWithTitle: _(@"POP3")];
    }
  else
    {
      [popup selectItemWithTitle: _(@"UNIX")];
    }
  
  // We decode our port
  portValue =  [allValues objectForKey:@"PORT"];
  
  if ( portValue )
    {
      [serverPortField setIntValue: [portValue intValue]];
    }
  else
    {
      if (serverTypeValue && [serverTypeValue intValue] == IMAP)
	{
	  [serverPortField setIntValue: IMAP_SERVICE_PORT];
	}
      else
	{
	  [serverPortField setIntValue: POP3_SERVICE_PORT];
	}
    }

  
  // We decode the rest of the informations. We begin with SERVERNAME.
  if ( allValues && [allValues objectForKey: @"SERVERNAME"] )
    {
      [serverNameField setStringValue: [allValues objectForKey: @"SERVERNAME"] ];
    }
  else if ( [self key] && [serverTypeValue intValue] != UNIX )
    {
      NSDebugLog(@"Server name is not defined; taking the key (%@).", [self key]);
      [serverNameField setStringValue: [self key]];
    }
  else
    {
      [serverNameField setStringValue: @""];
    }

  // USERNAME
  if ( allValues && [allValues objectForKey: @"USERNAME"] )
    {
      [usernameField setStringValue: [allValues objectForKey: @"USERNAME"] ];
    }
  else
    {
      [usernameField setStringValue: @""];
    }

  // We get our password, if we need to!
  if ( allValues && [allValues objectForKey: @"PASSWORD"] )
    {
      NSString *password;

      password = [Utilities decryptPassword: [allValues objectForKey: @"PASSWORD"]
			    withKey: [self key]];
      
      if ( !password )
	{
	  password = @"";
	}

      [passwordSecureField setStringValue: password];
    }
  else
    {
      [passwordSecureField setStringValue: @""];
    }
  
  // REMEMBERPASSWORD
  if ( allValues && [allValues objectForKey: @"REMEMBERPASSWORD"] )
    {
      [rememberPassword setState: [[allValues objectForKey: @"REMEMBERPASSWORD"] intValue]];
    }
  else
    {
      [rememberPassword setState: NSOffState];
    }

  // VERIFYONSTARTUP
  if ( allValues && [allValues objectForKey: @"VERIFYONSTARTUP"] )
    {
      [verifyOnStartup setState: [[allValues objectForKey: @"VERIFYONSTARTUP"] intValue]];
    }
  else
    {
      [rememberPassword setState: NSOffState];
    }

  // USESECURECONNECTION
  if ( allValues && [allValues objectForKey: @"USESECURECONNECTION"] )
    {
      [useSecureConnection setState: [[allValues objectForKey: @"USESECURECONNECTION"] intValue]];
    }
  else
    {
      [useSecureConnection setState: NSOffState];
    }

  // POP3 - LEAVEONSERVER
  if ( allValues && [allValues objectForKey: @"LEAVEONSERVER"] )
    {
      [[pop3ViewController leaveOnServer] setState: [[allValues objectForKey: @"LEAVEONSERVER"] intValue]];
    }
  else
    {
      [[pop3ViewController leaveOnServer] setState: NSOnState];
    }
  
  // POP3 - RETAINPERIOD
  if ( allValues && [allValues objectForKey: @"RETAINPERIOD"] )
    {
      [[pop3ViewController daysField] setIntValue: [[allValues objectForKey: @"RETAINPERIOD"] intValue]];
    }
  else
    {
      [[pop3ViewController daysField] setIntValue: 365];
    }
  
  // POP3 - USEAPOP
  if ( allValues && [allValues objectForKey: @"USEAPOP"] )
    {
      [[pop3ViewController useAPOP] setState: [[allValues objectForKey: @"USEAPOP"] intValue]];
    }
  else
    {
      [[pop3ViewController useAPOP] setState: NSOffState];
    }

  // POP3 OR UNIX - SAVE TO MAILBOX...
  if ( allValues && [allValues objectForKey: @"MAILBOXNAME"] )
    {
      FolderNodePopUpItem *aPopUpItem;
      id aController;

      if ( [serverTypeValue intValue] == POP3 )
	{
	  aController = pop3ViewController;
	}
      else
	{
	  aController = unixViewController;
	}
      
      aPopUpItem = [Utilities folderNodePopUpItemForURLNameAsString: [allValues objectForKey: @"MAILBOXNAME"]
			      usingFolderNodes: [aController allNodes]
			      popUpButton: [aController mailboxPopUpButton]];
      
      if ( aPopUpItem )
	{
	  [[aController mailboxPopUpButton] selectItem: aPopUpItem];
	}
      else
	{
	  // FIXME
	}
    }
  else
    {
      // FIXME
    }
  

  // IMAP - RETRIEVEMINUTES
  if ( allValues && [allValues objectForKey: @"RETRIEVEMINUTES"] )
    {
      int aValue;

      aValue = [[allValues objectForKey: @"RETRIEVEMINUTES"] intValue];

      if ( aValue <= 0 )
	{
	  aValue = 1;
	}

      [[imapViewController retrieveMinutesField] setIntValue: aValue];
    }
  else
    {
      [[imapViewController retrieveMinutesField] setIntValue: 1];
    }

  // UNIX - mail spool file
  if ( allValues && [allValues objectForKey: @"MAILSPOOLFILE"] )
    {
      NSString *aString;

      aString = [allValues objectForKey: @"MAILSPOOLFILE"];

   
      if ( [aString length] > 0 )
	{
	  [[unixViewController mailspoolFileField] setStringValue: aString];
	}
      else
	{
	  NSProcessInfo *processInfo;
	  
	  processInfo = [NSProcessInfo processInfo];
	  aString = [[processInfo environment] objectForKey: @"MAIL"];
	  
	  if ( aString )
	    {
	      [[unixViewController mailspoolFileField] setStringValue: aString];
	    }
	  else
	    {
	      [[unixViewController mailspoolFileField] setStringValue: @""];
	    }
	}
    }

  // We refresh our view
  [self setType: nil];
}


//
// delegate methods
//
- (void) windowWillClose: (NSNotification *) theNotification
{
  AUTORELEASE(self);
}


//
//
//
- (void) windowDidLoad
{
  // We select a default item in our combo
  [popup selectItemWithTitle: _(@"POP3")];

#ifdef MACOSX
  tabViewItem1 = [tabView tabViewItemAtIndex: 0];
  tabViewItem2 = [tabView tabViewItemAtIndex: 1];
#endif
  
  // We refresh our view
  [self setType: nil];
}


//
// action methods
//
- (IBAction) setType: (id) sender
{
  [popup synchronizeTitleAndSelectedItem];

  if ( [[popup titleOfSelectedItem] isEqualToString: _(@"POP3")] )
    {
      [self _setEnableUIElements: YES];

      [tabViewItem2 setLabel: _(@"POP3")];
      [tabViewItem2 setView: [pop3ViewController pop3View]];

      // If the number 'IMAP_SERVICE_PORT' was on the text field,
      // we set it to POP3_SERVICE_PORT. Else, we keep that custom port.
      if ( [serverPortField intValue] == IMAP_SERVICE_PORT &&
           [useSecureConnection state] == NSOffState )
	{
	  [serverPortField setIntValue: POP3_SERVICE_PORT];
	}
      else if ( [serverPortField intValue] == IMAPS_SERVICE_PORT &&
                [useSecureConnection state] == NSOnState )
	{
	  [serverPortField setIntValue: POP3S_SERVICE_PORT];
	}
    }
  else if ( [[popup titleOfSelectedItem] isEqualToString: _(@"IMAP")] )
    {
      [self _setEnableUIElements: YES];

      [tabViewItem2 setLabel: _(@"IMAP")];
      [tabViewItem2 setView: [imapViewController imapView]];

      // If the number 'POP3_SERVICE_PORT' was on the text field,
      // we set it to IMAP_SERVICE_PORT. Else, we keep that custom port.
      if ( [serverPortField intValue] == POP3_SERVICE_PORT &&
           [useSecureConnection state] == NSOffState )
	{
	  [serverPortField setIntValue: IMAP_SERVICE_PORT];
	}
      else if ( [serverPortField intValue] == POP3S_SERVICE_PORT &&
                [useSecureConnection state] == NSOnState )
	{
	  [serverPortField setIntValue: IMAPS_SERVICE_PORT];
	}
    }
  else
    {
      [self _setEnableUIElements: NO];

      [tabViewItem2 setLabel: _(@"UNIX")];
      [tabViewItem2 setView: [unixViewController unixView]];
    }

  // We best guess the mail spool file
  [self _bestGuessMailspoolFile];
  
  [tabView setNeedsDisplay: YES];
}


//
//
//
- (IBAction) okClicked: (id) sender
{
  NSMutableDictionary *allValues, *allServers, *allPreferences;
  NSString *theKey;

  NSNumber *serverTypeValue;
  id aController;
  
  FolderNode *aFolderNode;
  int aValue;
  
  
  allPreferences = [[NSMutableDictionary alloc] init];
  allServers = [[NSMutableDictionary alloc] init];
  
  [allPreferences addEntriesFromDictionary: [[NSUserDefaults standardUserDefaults] 
					      volatileDomainForName: @"PREFERENCES"]];
  [allServers addEntriesFromDictionary: [allPreferences objectForKey: @"RECEIVING"] ];

  
  // We init our dictionary of values
  allValues = [[NSMutableDictionary alloc] initWithCapacity: 8];

  // We set the type of our server and our port
  if ( [[popup titleOfSelectedItem] isEqualToString: _(@"POP3")] )
    {
      serverTypeValue = [NSNumber numberWithInt: POP3];
      aController = pop3ViewController;
      
      if ([serverPortField intValue] > 0)
	{
	  [allValues setObject: [NSNumber numberWithInt: [serverPortField intValue]]
		     forKey: @"PORT"];
	}
      else
	{
	  [allValues setObject: [NSNumber numberWithInt: POP3_SERVICE_PORT]
		     forKey: @"PORT"];
	}
    }
  else if ( [[popup titleOfSelectedItem] isEqualToString: _(@"IMAP")] )
    {
      serverTypeValue = [NSNumber numberWithInt: IMAP];
      
      // We save the value of POP3 even when we're in IMAP (for MAILBOXNAME) since 
      // we pretty much don't are when we are in IMAP about that value.
      aController = pop3ViewController; 

      if ([serverPortField intValue] > 0)
	{
	  [allValues setObject: [NSNumber numberWithInt: [serverPortField intValue]]
		     forKey: @"PORT"];
	}     
      else
	{
	  [allValues setObject: [NSNumber numberWithInt: IMAP_SERVICE_PORT]
		     forKey: @"PORT"];
	}
    }
  else
    {
      serverTypeValue = [NSNumber numberWithInt: UNIX];
      aController = unixViewController;
    }

  
  // We verify if the server name or username has changed
  if ( [serverTypeValue intValue] == UNIX )
    {
      theKey = [NSString stringWithFormat: @"local @ %@", 
			 [[unixViewController mailspoolFileField] stringValue]];
    }
  else
    {
      theKey = [NSString stringWithFormat: @"%@ @ %@", 
			 [usernameField stringValue],
			 [serverNameField stringValue]];
    }

  if ( ![theKey isEqualToString: [self key]] )
    {
      // We don't try to remove it if it's not in there!
      if ( [allServers objectForKey: [self key]] )
	{
	  [allServers removeObjectForKey: [self key]];
	}
      [self setKey: theKey];
    }

  // We set the SERVER TYPE pref value
  [allValues setObject: serverTypeValue 
	     forKey: @"SERVERTYPE"];
  
  // We set the rest of the informations
  [allValues setObject: [serverNameField stringValue]
	     forKey: @"SERVERNAME"];
  [allValues setObject: [usernameField stringValue]
	     forKey: @"USERNAME"];
  
  // We save the password or we remove it from the defaults database
  if ( [rememberPassword state] == NSOnState )
    {
      [allValues setObject:
                   [Utilities encryptPassword: [passwordSecureField stringValue]
                              withKey: theKey]
		 forKey:@"PASSWORD"];
    }
  else
    {
      // We don't try to remove it if it's not in there!
      if ( [allValues objectForKey: @"PASSWORD"] )
	{
	  [allValues removeObjectForKey: @"PASSWORD"];
	}
    }

  [allValues setObject: [NSNumber numberWithInt: [rememberPassword state]]
	     forKey: @"REMEMBERPASSWORD"];

  [allValues setObject: [NSNumber numberWithInt: [verifyOnStartup state]]
	     forKey: @"VERIFYONSTARTUP"];

  [allValues setObject: [NSNumber numberWithInt: [useSecureConnection state]]
	     forKey: @"USESECURECONNECTION"];

  // Our POP3 defaults
  [allValues setObject: [NSNumber numberWithInt: [[pop3ViewController leaveOnServer] state]] 
	     forKey: @"LEAVEONSERVER"];

  aValue = [[pop3ViewController daysField] intValue];
  [allValues setObject: [NSNumber numberWithInt: (aValue <= 0 ? 365 : aValue)]
	     forKey: @"RETAINPERIOD"];

  [allValues setObject: [NSNumber numberWithInt: [[pop3ViewController useAPOP] state]] 
	     forKey: @"USEAPOP"];

  [[aController mailboxPopUpButton] synchronizeTitleAndSelectedItem];
  aFolderNode = [(FolderNodePopUpItem *)[[aController  mailboxPopUpButton] selectedItem] folderNode];
  
  // FIXME, we should make this more efficient
  if ( [aFolderNode parent] == [aController allNodes] )
    {
      NSDebugLog(@"Selected a wrong mailbox, ignoring.");
    }
  else
    {
      NSString *aString;

      aString = [Utilities stringValueOfURLNameFromFolderNode: aFolderNode];

      [allValues setObject: aString
		 forKey: @"MAILBOXNAME"];
    }

  // Our IMAP defaults
  aValue = [[imapViewController retrieveMinutesField] intValue];
  [allValues setObject: [NSNumber numberWithInt: (aValue <= 0 ? 1 : aValue)]
	     forKey: @"RETRIEVEMINUTES"];
  

  // Our UNIX defaults
  [allValues setObject: [[unixViewController mailspoolFileField] stringValue]
	     forKey: @"MAILSPOOLFILE"];
  
  // We now save the new defaults pour this server
  [allServers setObject: allValues
	      forKey: theKey];

  // We update the volatile user defaults with the new receiving servers list
  [allPreferences setObject: allServers forKey: @"RECEIVING"];
#ifndef MACOSX
  [[NSUserDefaults standardUserDefaults] removeVolatileDomainForName: @"PREFERENCES"];
#endif
  [[NSUserDefaults standardUserDefaults] setVolatileDomain: allPreferences
					 forName: @"PREFERENCES"];

  RELEASE(allPreferences);
  RELEASE(allServers);
  RELEASE(allValues);

  [NSApp stopModal];
  [self close];
}


//
//
//
- (IBAction) cancelClicked: (id) sender
{
  [NSApp abortModal];
  [self close];
}


//
//
//
- (IBAction) sslClicked: (id) sender
{
  int newState = [useSecureConnection state];

  if ( [[popup titleOfSelectedItem] isEqualToString: _(@"POP3")] )
    {
      // POP3 server
      if ( NSOffState == newState &&
           POP3S_SERVICE_PORT == [serverPortField intValue] )
        {
          [serverPortField setIntValue: POP3_SERVICE_PORT];
        }
      else if ( NSOnState == newState &&
                POP3_SERVICE_PORT == [serverPortField intValue] )
        {
          [serverPortField setIntValue: POP3S_SERVICE_PORT];
        }
    }
  else
    {
      // IMAP server
      if ( NSOffState == newState &&
           IMAPS_SERVICE_PORT == [serverPortField intValue] )
        {
          [serverPortField setIntValue: IMAP_SERVICE_PORT];
        }
      else if ( NSOnState == newState &&
                IMAP_SERVICE_PORT == [serverPortField intValue] )
        {
          [serverPortField setIntValue: IMAPS_SERVICE_PORT];
        }
    }
}


//
// access/mutation methods
//
- (NSString *) key
{
  return key;
}


//
//
//
- (void) setKey: (NSString *) theKey
{
  RELEASE(key);

  if (theKey)
    {
      RETAIN(theKey);
      key = theKey;
    }
  else
    {
      key = nil;
      
      // We select a default item in our combo and we refresh our view
      [popup selectItemWithTitle: _(@"POP3")];
      [serverPortField setIntValue: POP3_SERVICE_PORT];
      [self setType: nil];
    }
}


//
//
//
- (NSTextField *) serverNameField
{
  return serverNameField;
}


//
//
//
- (NSTextField *) serverPortField
{
  return serverPortField;
}


//
//
//
- (NSTextField *) usernameField
{
  return usernameField;
}


//
//
//
- (NSSecureTextField *) passwordSecureField
{
  return passwordSecureField;
}


//
//
//
- (NSButton *) useSecureConnection
{
  return useSecureConnection;
}

@end


//
// private methods
//
@implementation ReceivingWindowController (Private)


//
// If the user HAS NOT specified one or the MAIL environment variable isn't defined,
// we try to guess the best path to the mail spool file a user would use.
//
- (void) _bestGuessMailspoolFile
{

  if ( [[[unixViewController mailspoolFileField] stringValue] isEqualToString: @""] )
    {
      BOOL isDir;
      
      if ( [[NSFileManager defaultManager] fileExistsAtPath: [NSString stringWithFormat: @"/var/mail/%@", 
    								       NSUserName()]
    					   isDirectory: &isDir] && !isDir )
	{
	  [[unixViewController mailspoolFileField] setStringValue: 
						     [NSString stringWithFormat: @"/var/mail/%@",
							       NSUserName()]];
	}
      else if ( [[NSFileManager defaultManager] fileExistsAtPath: [NSString stringWithFormat: @"/var/spool/mail/%@",
    									    NSUserName()]
    						isDirectory: &isDir] && ! isDir )
	{
	  [[unixViewController mailspoolFileField] setStringValue:
						     [NSString stringWithFormat: @"/var/spool/mail/%@",
							       NSUserName()]];
	}
      else if ( [[NSFileManager defaultManager] fileExistsAtPath: [NSString stringWithFormat: @"/usr/spool/mail/%@",
    									    NSUserName()]
    						isDirectory: &isDir] && ! isDir )
	{
	  [[unixViewController mailspoolFileField] setStringValue:
						     [NSString stringWithFormat: @"/usr/spool/mail/%@",
							       NSUserName()]];
	}
    }
}

- (void) _setEnableUIElements: (BOOL) aBOOL
{
  [serverNameField setEditable: aBOOL];
  [serverPortField setEditable: aBOOL];
  [usernameField setEditable: aBOOL];
  [passwordSecureField setEditable: aBOOL];

  [rememberPassword setEnabled: aBOOL];
  [verifyOnStartup setEnabled: aBOOL];
  [useSecureConnection setEnabled: aBOOL];
}

@end
