/*
**  EditWindowController.m
**
**  Copyright (c) 2001, 2002
**
**  Author: Ludovic Marcotte <ludovic@Sophos.ca>
**          Jonathan B. Leffert <jonathan@leffert.net>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#import "EditWindowController.h"

#import "Address.h"
#import "AddressBook.h"
#import "AddressBookController.h"

#ifndef MACOSX
#import "EditWindow.h"
#endif

#import "ExtendedAttachmentCell.h"
#import "Filter.h"
#import "FilterManager.h"
#import "GNUMail.h"
#import "GNUMail/GNUMailBundle.h"
#import "GNUMailConstants.h"
#import "LabelWidget.h"
#import "MailboxManagerController.h"
#import "MailWindowController.h"
#import "MimeTypeManager.h"
#import "MimeType.h"
#import "NSStringExtensions.h"
#import "PasswordPanelController.h"
#import "Utilities.h"

#import <Pantomime/Charset.h>
#import <Pantomime/Constants.h>
#import <Pantomime/LocalFolder.h>
#import <Pantomime/LocalStore.h>
#import <Pantomime/Message.h>
#import <Pantomime/MimeUtility.h>
#import <Pantomime/Part.h>
#import <Pantomime/Sendmail.h>
#import <Pantomime/SMTP.h>
#import <Pantomime/URLName.h>

#ifndef LONG_DELAY
#define LONG_DELAY 86400.0
#endif

@implementation EditWindowController

//
//
//
- (id) initWithWindowNibName: (NSString *) windowNibName
{
#ifndef MACOSX
  EditWindow *editWindow;
#endif

  // We first verify if we have at least one transport agent defined
  if ( ![[NSUserDefaults standardUserDefaults] objectForKey: @"SENDING"] ||
       ![(NSArray *)[[NSUserDefaults standardUserDefaults] objectForKey: @"SENDING"] count] )
    {
      NSRunAlertPanel(_(@"Error!"),
		      _(@"You must have at least one transport agent defined.\nSee Preferences -> Sending."),
		      _(@"OK"), // default
		      NULL,     // alternate
		      NULL);
      
      AUTORELEASE(self);
      return nil;
    }
  

#ifdef MACOSX
  
  self = [super initWithWindowNibName: windowNibName];
  
  // We create our mutable array and our mutable dictionary allowing
  // us to dynamically add toolbar items.
  allowedToolbarItemIdentifiers = [[NSMutableArray alloc] initWithObjects: 
							    NSToolbarSeparatorItemIdentifier,
							  NSToolbarSpaceItemIdentifier,
							  NSToolbarFlexibleSpaceItemIdentifier,
							  NSToolbarCustomizeToolbarItemIdentifier, 
							  @"send",
							  @"insert",
							  @"add_cc",
							  @"add_bcc",
							  @"addresses",
							  nil];
  
  additionalToolbarItems = [[NSMutableDictionary alloc] init];
  
#else
  
  editWindow = [[EditWindow alloc] initWithContentRect: NSMakeRect(50,75,750,520)
				   styleMask: (NSClosableWindowMask|NSTitledWindowMask|
					       NSMiniaturizableWindowMask|NSResizableWindowMask)
				   backing: NSBackingStoreRetained
				   defer: NO];
    
  self = [super initWithWindow: editWindow];
  [editWindow layoutWindow];
  [editWindow setDelegate: self];
  
  // We link our outlets
  subjectText = [editWindow subjectText];
  toText = [editWindow toText];
  ccText = [editWindow ccText];
  bccText = [editWindow bccText];

  subjectLabel = [editWindow subjectLabel];
  toLabel = [editWindow toLabel];
  ccLabel = [editWindow ccLabel];
  bccLabel = [editWindow bccLabel];

  sizeLabel = [editWindow sizeLabel];

  scrollView = [editWindow scrollView];
  textView = [editWindow textView];

  icon = [editWindow icon];
  send = [editWindow send];
  insert = [editWindow insert];
  addBcc = [editWindow addBcc];
  addCc = [editWindow addCc];
  addresses = [editWindow addresses];

  personalProfilePopUpButton = [editWindow personalProfilePopUpButton];
  transportMethodPopUpButton = [editWindow transportMethodPopUpButton];

  defaultCharsetPopUpButton = [editWindow defaultCharsetPopUpButton];

  RELEASE(editWindow);
#endif

  // We set our window title and edited attribute
  [[self window] setTitle: @""];
  [[self window] setDocumentEdited: NO];
  
  // We load our accessory views
  [self _loadAccessoryViews];

  // We load our personal profiles and our transport methods
  [Utilities loadPersonalProfilesInPopUpButton: personalProfilePopUpButton selectedProfile: nil];
  [Utilities loadTransportMethodsInPopUpButton: transportMethodPopUpButton];

  // We set some ivars
  forceEndEditingOfTextField = 0;
  personalProfilePopUpButtonIndex = -1;

  [self personalProfilesSelectionHasChanged: nil];

  [self _loadCharsets];

  // We finally set our autosave window frame name and restore the one from the user's defaults.
  [[self window] setFrameAutosaveName: @"EditWindow"];
  [[self window] setFrameUsingName: @"EditWindow"];

  return self;
}


//
//
//
#ifdef MACOSX
- (void) awakeFromNib
{
  NSToolbar *aToolbar;
  
  aToolbar = [[NSToolbar alloc] initWithIdentifier: @"EditWindowToolbar"];
  [aToolbar setDelegate: self];
  [aToolbar setAllowsUserCustomization: YES];
  [aToolbar setAutosavesConfiguration: YES];
  [[self window] setToolbar: aToolbar];

  RELEASE(aToolbar);
}
#endif


//
//
//
- (void) dealloc
{
  NSDebugLog(@"EditWindowController: -dealloc");
  
  // We remove our observer for our two notifications
  [[NSNotificationCenter defaultCenter]
    removeObserver: self
    name: PersonalProfilesHaveChanged
    object: nil];

  [[NSNotificationCenter defaultCenter]
    removeObserver: self
    name: TransportMethodsHaveChanged
    object: nil];

  TEST_RELEASE(message);
  TEST_RELEASE(unmodifiedMessage);

  // We release our passwordCache
  RELEASE(passwordCache);

#ifdef MACOSX
  RELEASE(allowedToolbarItemIdentifiers);
  RELEASE(additionalToolbarItems);
#endif

  [super dealloc];
}


//
// Implementation of the AddressTaker protocol.
//
- (void) takeToAddress: (Address *) theAddress
{
  [Utilities appendAddress: theAddress
	     toTextField: toText];
}


//
//
//
- (void) takeCcAddress: (Address *) theAddress
{
  if ( ![self showCc] )
    {
      [self showCc: self];
    }
  
  [Utilities appendAddress: theAddress
	     toTextField: ccText];
}


//
//
//
- (void) takeBccAddress: (Address *) theAddress
{
  if ( ![self showBcc] )
    {
      [self showBcc: self];
    }
  
  [Utilities appendAddress: theAddress
	     toTextField: bccText];
}


//
// action methods
//
- (IBAction) insertFile: (id) sender
{
  NSOpenPanel *oPanel;
  int result;
  
  oPanel = [NSOpenPanel openPanel];
  [oPanel setAllowsMultipleSelection:YES];
  result = [oPanel runModalForDirectory: [GNUMail currentWorkingPath] file:nil types:nil];
  
  if (result == NSOKButton)
    {
      ExtendedAttachmentCell *cell;
      NSTextAttachment *attachment;
      NSFileWrapper *filewrapper;
      NSEnumerator *filesToOpenEnumerator = [[oPanel filenames] objectEnumerator];
      NSString *theFilename;
      id encodedAttatchment;
      
      while ( (theFilename = [filesToOpenEnumerator nextObject]) )
	{
	  MimeType *aMimeType;

	  filewrapper = [[NSFileWrapper alloc] initWithPath: theFilename];
	  AUTORELEASE(filewrapper);

          // We save the path of the last file
          [GNUMail setCurrentWorkingPath: [theFilename stringByDeletingLastPathComponent]];
	  
	  // We set the icon of the attachment if the mime-type is found
	  aMimeType = [[MimeTypeManager singleInstance] mimeTypeForFileExtension:
							  [[theFilename lastPathComponent]
							    pathExtension]];
	  
	  // If we found a MIME-type for this attachment...
	  if ( aMimeType )
	    {
	      [filewrapper setIcon: [aMimeType icon]];
	    }
	  
	  // We now create our attachment from our filewrapper
	  attachment = [[NSTextAttachment alloc] initWithFileWrapper: filewrapper];

	  cell = [[ExtendedAttachmentCell alloc] initWithFilename: [theFilename lastPathComponent]
						 size: [[filewrapper regularFileContents] length]];
	  
	  [attachment setAttachmentCell: cell];
          
          // Cocoa bug
#ifdef MACOSX
          [cell setAttachment: attachment];
          [cell setImage: [filewrapper icon]];
#endif
      	  RELEASE(cell);

	  encodedAttatchment = [NSAttributedString attributedStringWithAttachment: attachment];
	  RELEASE(attachment);
	  
	  if ( encodedAttatchment )
	    {
	      [textView insertText: encodedAttatchment];
	    }
	}
      
      [[self window] makeFirstResponder: textView ];
    }
}


//
//
//
- (IBAction) showBcc: (id) sender
{    
  [self setShowBcc: ![self showBcc]];
  [[[self window] contentView] setNeedsDisplay: YES];
}


//
//
//
- (IBAction) showCc: (id) sender
{ 
  [self setShowCc: ![self showCc]];
  [[[self window] contentView] setNeedsDisplay: YES];
}


//
//
//
- (IBAction) personalProfilesSelectionHasChanged: (id) sender
{
  NSArray *anArray;
  int i;

  // We synchronize our from selection from both popups
  [personalProfilePopUpButton synchronizeTitleAndSelectedItem];
  [transportMethodPopUpButton synchronizeTitleAndSelectedItem];

  anArray = [[NSUserDefaults standardUserDefaults] objectForKey: @"SENDING"];
  
  for (i = 0; i < [anArray count]; i++)
    {
      NSDictionary *aDictionary;
      
      aDictionary = [anArray objectAtIndex: i];
      
      if ( [[aDictionary objectForKey: @"PERSONAL_PROFILE"] isEqualToString: 
							      [Utilities profileNameForItemTitle: 
									   [personalProfilePopUpButton titleOfSelectedItem]]] )
	{
	  [transportMethodPopUpButton selectItemAtIndex: i];
	  [self _replaceSignature];
	  return;
	}
    }
  
  [transportMethodPopUpButton selectItemAtIndex: 0];
  [self _replaceSignature];
}


//
//
//
- (IBAction) updateAnimatedIcon: (id) sender
{
  if (animation_index == 9)
    {
      animation_index = 1;
    }
  
  [icon setImage: [NSImage imageNamed: [NSString stringWithFormat: @"anim-logo-%d.tiff", animation_index]]];
  
  animation_index += 1;
}



//
//
//
- (IBAction) sendMessage: (id) sender
{
  NSDictionary *aDictionary;
  NSPort *port1, *port2;
  NSArray *ports;
  
  if ( animation && [animation isValid] )
    {
      mustStopTransfer = YES;
      return;
    }

  // We first try to initialize our message with all the information 
  if ( ![self updateMessageContentFromTextView: YES] )
    {
      return;
    }

  // We sync our popup
  [transportMethodPopUpButton synchronizeTitleAndSelectedItem];

  // We first ask for our password and we cache it (if we need too)
  aDictionary = [[[NSUserDefaults standardUserDefaults] objectForKey: @"SENDING"]
		  objectAtIndex: [transportMethodPopUpButton indexOfSelectedItem]];
  
  if ( [aDictionary objectForKey: @"SMTP_AUTH"] &&
       [[aDictionary objectForKey: @"SMTP_AUTH"] intValue] == NSOnState &&
       [[aDictionary objectForKey: @"TRANSPORT_METHOD"] intValue] == TRANSPORT_SMTP )
    {
      [self _passwordForServerName: [aDictionary objectForKey: @"SMTP_HOST"]
	    prompt: YES];
    }


  // We disable our buttons
#ifdef MACOSX
  [send setImage: [NSImage imageNamed: @"stop_32.tiff"]];
  [send setLabel: _(@"Stop")];
#else
  [send setImage: [NSImage imageNamed: @"stop_48.tiff"]];
  [send setTitle: _(@"Stop")];
#endif
  [insert setEnabled: NO];
  [addBcc setEnabled: NO];
  [addCc setEnabled: NO];
  [addresses setEnabled: NO];
  
  // We start our animation timer
  animation_index = 1;
  animation =  [NSTimer scheduledTimerWithTimeInterval: 0.1
			target: self
			selector: @selector(updateAnimatedIcon:)
			userInfo: nil
			repeats: YES];
  RETAIN(animation);

  mustStopTransfer = NO;

  // We initialize our NSConnection object and our ports
  port1 = [[NSPort alloc] init];
  port2 = [[NSPort alloc] init];
  
  connection = [[NSConnection alloc] initWithReceivePort: port1
				     sendPort: port2];
  [connection setRootObject: self];
  [connection setRequestTimeout: LONG_DELAY];
  [connection setReplyTimeout: LONG_DELAY];

  ports = [NSArray arrayWithObjects: port2, port1, nil];

  RELEASE(port1);
  RELEASE(port2);

  //
  // We detach our thread that will send the message using a SMTP
  // connection or a mailer like sendmail.
  //
  [[NSDistributedNotificationCenter defaultCenter]
    postNotificationName: GNUMailConsoleMessage
    object: nil
    userInfo: [NSDictionary dictionaryWithObject: _(@"Starting the thread for sending the E-Mail...")
			    forKey: @"Message"]];

  [NSThread detachNewThreadSelector: @selector(_sendMessageUsingPorts:)
	    toTarget: self
	    withObject: ports];
}


//
// delegate methods
//
- (BOOL)      control: (NSControl *) theControl 
 textShouldEndEditing: (NSText *) theFieldEditor
{
  if ( theControl == toText ||
       theControl == ccText ||
       theControl == bccText )
    {
      NSArray *allRecipientsFromString;
      NSMutableArray *allRecipients;
      NSString *aString;
      
      aString = [[theControl stringValue] stringByTrimmingWhiteSpaces];

      if ( [aString length] == 0 )
	{
	  return YES;
	}
      
      allRecipients = [NSMutableArray arrayWithArray: [self _recipientsFromString: aString]];

      if ( [allRecipients count] == 0 )
	{
	  return YES;
	}

      allRecipientsFromString = [[[AddressBookController singleInstance] addressBook] 
				  addressesWithPrefix: [allRecipients lastObject]];
      
      // If we have no match or more than one match, we don't give up the editing
      // unless the user really wants to.
      if ( [allRecipientsFromString count] != 1 )
	{
	  forceEndEditingOfTextField++;

	  if ( forceEndEditingOfTextField == 2 )
	    {
	      forceEndEditingOfTextField = 0;
	      return YES;
	    }
	  return NO;
	}

      // We delete the last "partially" entered address and we finally set the text of our field
      [allRecipients removeLastObject];
      [allRecipients addObject: [[allRecipientsFromString lastObject] formattedAddress]];

      [theControl setStringValue: [allRecipients componentsJoinedByString: @", "]];
    }

  return YES;
}


//
//
//
- (void) controlTextDidChange: (NSNotification *) aNotification
{
  if ( [aNotification object] == subjectText )
    {
      if ( [[subjectText stringValue] length] > 0 )
	{
	  [[self window] setTitle: [subjectText stringValue]];
	}
      else
	{
	  [[self window] setTitle: _(@"(no subject)")];
	}
    }
  
  [[self window] setDocumentEdited: YES];
}


//
//
//
- (void) textDidChange: (NSNotification *) aNotification
{
  [[self window] setDocumentEdited: YES];
  [self _updateSizeLabel];
}


//
//
//
- (BOOL) windowShouldClose: (id) sender
{
  // We first verify if the animation timer is running, if it is,
  // we are currently sending a mail so we inform the user about this.
  if ( animation && [animation isValid] )
    {
      NSRunInformationalAlertPanel(_(@"Closing..."),
				   _(@"A mail is currently being sent. Please wait."),
				   _(@"OK"), 
				   NULL, 
				   NULL);
      return NO;
    }


  if ( [[self window] isDocumentEdited] )
    {
      int choice;
      
      choice = NSRunAlertPanel(_(@"Closing..."),
			       _(@"Would you like to save this message in the Drafts folder?"),
			       _(@"Cancel"), // default
			       _(@"Yes"),    // alternate
			       _(@"No"));    // other return
      
      // We don't want to close the window
      if ( choice == NSAlertDefaultReturn )
	{
	  return NO;
	}
      // Yes we want to close it, and we also want to save the message to the Drafts folder.
      else if ( choice == NSAlertAlternateReturn )
	{
	  // We append the message to the Drafts folder
	  NSString *aFolderName;
	  URLName *theURLName;
	  
	  Store *aStore;
	  NSData *aData;
	  
	  // We update the current message content with the current content of the view
	  [self updateMessageContentFromTextView: NO];

	  theURLName = [[URLName alloc] initWithString: [[NSUserDefaults standardUserDefaults] objectForKey: @"DRAFTSFOLDERNAME"]
					path: [[NSUserDefaults standardUserDefaults] objectForKey: @"LOCALMAILDIR"]];
	  
	  aStore = [Utilities storeForURLName: theURLName];
	  
	  if ( !aStore )
	    {
	      NSRunAlertPanel(_(@"Error!"),
			      _(@"An error occured while attempting to obtain the message store for the Drafts folder."),
			      _(@"OK"),
			      NULL,
			      NULL);
	      return YES; 
	    }
	  
	  aFolderName = [theURLName foldername];
	  aData = [[self message] dataUsingSendingMode: SEND_TO_FOLDER];
	  
	  [[MailboxManagerController singleInstance] transferMessageFromRawSource: aData
						     flags: nil
						     toFolderWithName: aFolderName
						     orToFolder: nil
						     store: aStore];
	}
    }
  
  return YES;
}


//
//
//
- (void) windowWillClose: (NSNotification *) theNotification
{ 
  if ( [GNUMail lastAddressTakerWindowOnTop] == self )
    {
      [GNUMail setLastAddressTakerWindowOnTop: nil];
    }
      
  [[NSApp delegate] setEnableSaveInDraftsMenuItem: NO];

  AUTORELEASE(self);
}


//
//
//
- (void) windowDidBecomeMain: (NSNotification *) theNotification
{
#ifndef MACOSX
  [[[self window] contentView] setNeedsDisplay: YES];
#endif

  [GNUMail setLastAddressTakerWindowOnTop: self];

  [[NSApp delegate] setEnableSaveInDraftsMenuItem: YES];
}


//
//
//
- (void) windowDidLoad
{ 
#ifdef MACOSX
  [addresses setTarget: [NSApp delegate]];
  [addresses setAction: @selector(showAddressBook:)];
#endif

  connection = nil;

  // We initialize our password cache
  passwordCache = [[NSMutableDictionary alloc] init];

  // We initialize our variables
  [self setMessageWasSent: NO];
  [self setShowCc: NO];
  [self setShowBcc: NO];

  [self setUnmodifiedMessage: nil];

  [self setSignaturePosition: SIGNATURE_END];

  // We add our observer for our two notifications
  [[NSNotificationCenter defaultCenter]
    addObserver: self
    selector: @selector(_loadPersonalProfiles)
    name: PersonalProfilesHaveChanged
    object: nil];

  [[NSNotificationCenter defaultCenter]
    addObserver: self
    selector: @selector(_loadTransportMethods)
    name: TransportMethodsHaveChanged
    object: nil];
}


//
// access/mutation methods
//
- (BOOL) messageWasSent
{
  return messageWasSent;
}


//
//
//
- (void) setMessageWasSent: (BOOL) theBOOL
{
  messageWasSent = theBOOL;
}


//
//
//
- (Message *) message
{
  return message;
}


//
//
//
- (void) setMessage: (Message *) theMessage
{  
  if ( theMessage )
    {
      RETAIN(theMessage);
      RELEASE(message);
      message = theMessage;

      [self _updateViewWithMessage: message
      	    appendSignature: YES];
      [self _updateSizeLabel];
    }
  else
    {
      RELEASE(message);
      message = nil;
    }
}


//
//
//
- (Message *) unmodifiedMessage
{
  return unmodifiedMessage;
}


//
//
//
- (void) setUnmodifiedMessage: (Message *) theUnmodifiedMessage
{
  if ( theUnmodifiedMessage )
    {
      RETAIN(theUnmodifiedMessage);
      RELEASE(unmodifiedMessage);
      unmodifiedMessage = theUnmodifiedMessage;
    }
  else
    {
      RELEASE(unmodifiedMessage);
      unmodifiedMessage = nil;
    }
}


//
//
//
- (void) setMessageFromDraftsFolder: (Message *) theMessage
{
  if (theMessage)
    {
      RETAIN(theMessage);
      RELEASE(message);
      message = theMessage;

      [self _updateViewWithMessage: message
	    appendSignature: NO];
    }
  else
    {
      RELEASE(message);
      message = nil;
    }
}


//
//
//
- (BOOL) showCc
{
  return showCc;
}


//
//
//
- (void) setShowCc: (BOOL) theBOOL
{
  showCc = theBOOL;
  
  if (showCc)
    {
#ifdef MACOSX
      [ccText setEnabled: YES];
      [addCc setLabel: _(@"Remove Cc")];
      [addCc setImage: [NSImage imageNamed: @"remove_cc_32.tiff"]];
#else
      [[[self window] contentView] addSubview: ccLabel];
      [[[self window] contentView] addSubview: ccText];
      [addCc setTitle: _(@"Remove Cc")];
      [addCc setImage: [NSImage imageNamed: @"remove_cc_48.tiff"]];
#endif
    }
  else
    {
#ifdef MACOSX
      [ccText setEnabled: NO];
      [addCc setLabel: _(@"Add Cc")];
      [addCc setImage: [NSImage imageNamed: @"add_cc_32.tiff"]];
#else
      [ccLabel removeFromSuperviewWithoutNeedingDisplay];
      [ccText removeFromSuperviewWithoutNeedingDisplay];
      [addCc setTitle: _(@"Add Cc")];
      [addCc setImage: [NSImage imageNamed: @"add_cc_48.tiff"]];
#endif
    }
    
#ifndef MACOSX
  [self _adjustWidgetsPosition];
#endif

  [self _adjustNextKeyViews];
}


//
//
//
- (BOOL) showBcc
{
  return showBcc;
}


//
//
//
- (void) setShowBcc: (BOOL) theBOOL
{
  showBcc = theBOOL;
 
  if (showBcc)
    {
#ifdef MACOSX
      [bccText setEnabled: YES];
      [addBcc setLabel: _(@"Remove Bcc")];
      [addBcc setImage: [NSImage imageNamed: @"remove_bcc_32.tiff"]];
#else
      [[[self window] contentView] addSubview: bccLabel];
      [[[self window] contentView] addSubview: bccText];
      [addBcc setTitle: _(@"Remove Bcc")];
      [addBcc setImage: [NSImage imageNamed: @"remove_bcc_48.tiff"]];
#endif
    }
  else
    {
#ifdef MACOSX
      [bccText setEnabled: NO];
      [addBcc setLabel: _(@"Add Bcc")];
      [addBcc setImage: [NSImage imageNamed: @"add_bcc_32.tiff"]];
#else
      [bccLabel removeFromSuperviewWithoutNeedingDisplay];
      [bccText removeFromSuperviewWithoutNeedingDisplay];
      [addBcc setTitle: _(@"Add Bcc")];
      [addBcc setImage: [NSImage imageNamed: @"add_bcc_48.tiff"]];
#endif
    }

#ifndef MACOSX  
  [self _adjustWidgetsPosition];
#endif

  [self _adjustNextKeyViews];
}


//
//
//
- (int) signaturePosition
{
  return signaturePosition;
}


//
//
//
- (void) setSignaturePosition: (int) thePosition
{
  signaturePosition = thePosition;
}


//
//
//
- (id) mustStopTransfer
{
  //NSDebugLog(@"Polling to see if we must stop...");
  return [NSNumber numberWithBool: mustStopTransfer];
}



//
// other methods
//
- (void) authenticationFailedForServer: (NSString *) theServer
{
  NSRunAlertPanel(_(@"Error!"),
		  _(@"An error occured during the authentication with the SMTP server (%@)."),
		  _(@"OK"),
		  NULL,
		  NULL,
		  theServer);
  
  // We stop our animation
  [self stopAnimationAndRestoreButtonState];
}


//
//
//
- (void) errorOccuredWhenSendingMessage
{
  NSRunAlertPanel(_(@"Error!"),
		  _(@"An error occured while sending the E-Mail. It might be a network\nproblem or an error in your sending preferences."),
		  _(@"OK"),
		  NULL,
		  NULL);
  
  // We stop our animation
  [self stopAnimationAndRestoreButtonState];
}


//
// 
//
- (void) stopAnimationAndRestoreButtonState
{
  // We enable our UI elements
#ifdef MACOSX
  [send setImage: [NSImage imageNamed: @"send_32.tiff"]];
  [send setLabel: _(@"Send")];
#else
  [send setImage: [NSImage imageNamed: @"send_48.tiff"]];
  [send setTitle: _(@"Send")];
#endif
  [insert setEnabled: YES];
  [addBcc setEnabled: YES];
  [addCc setEnabled: YES];
  [addresses setEnabled: YES];
  
  // We stop our animation timer
  [animation invalidate];
  DESTROY(animation);
  [icon setImage: nil];

  if ( connection )
    {
      //[connection invalidate];
      DESTROY(connection);
    }
}


//
//
//
- (void) messageWasSentFromRAWSource: (NSData *) theRAWSource
{
  FilterManager *aFilterManager;
  NSString *aFolderName;

  URLName *theURLName;
  Store *aStore;
  
  aFilterManager = [FilterManager filtersFromDisk];
  theURLName = [aFilterManager matchedURLNameFromMessageAsRawSource: theRAWSource
			       type: TYPE_OUTGOING
			       key: nil];
  
  aFolderName = [theURLName foldername];
  aStore = [Utilities storeForURLName: theURLName];
  
  if ( !aStore )
    {
      NSRunAlertPanel(_(@"Error!"),
		      _(@"An error occured while attempting to obtain the message store for the Outbox folder.\nThe message nonetheless successfully sent."),
		      _(@"OK"),
		      NULL,
		      NULL);
      goto done;
    }
  
  // Only reached if we got a valid store.
  [[MailboxManagerController singleInstance] transferMessageFromRawSource: theRAWSource
					     flags: nil
					     toFolderWithName: aFolderName
					     orToFolder: nil
						 store: aStore];
  
 done:
  // We send the notification to inform our MailWindowController that the replied message has been
  // sucessfully sent so we can set the flag to 'ANSWERED'.
  if ( [self unmodifiedMessage] )
    {
      [[NSNotificationCenter defaultCenter]
	postNotificationName: ReplyToMessageWasSuccessful
	object: nil
	userInfo: [NSDictionary dictionaryWithObject: [self unmodifiedMessage] 
				forKey: @"Message"]];
    }

  // We stop our animation
  [self stopAnimationAndRestoreButtonState];
  
  [self close];
}


//
//
//
- (BOOL) updateMessageContentFromTextView: (BOOL) verifyRecipients
{

  NSArray *to_array, *cc_array, *bcc_array;
  NSTextStorage *textStorage;
 
  NSDictionary *allAdditionalHeaders, *aProfile;
 
  InternetAddress *anInternetAddress;
  BOOL hasFoundXMailerHeader;
  NSString *aString;
  int i;
  
  if ( verifyRecipients )
    {
      // We verify that at least one recipient has been defined in To:
      if ( [[[toText stringValue] stringByTrimmingWhiteSpaces] length] == 0 )
	{
	  NSRunAlertPanel(_(@"Error!"),
			  _(@"You must define a recipient in the field \"To\"."),
			  _(@"OK"), // default
			  NULL,     // alternate
			  NULL);
	  
	  return NO;
	}
    }  
  
  // We initialize our boolean value to false
  hasFoundXMailerHeader = NO;
  
  // We get the current selected profile when sending this email
  [personalProfilePopUpButton synchronizeTitleAndSelectedItem];
  
  // Then, we get our profile from our user defaults
  aProfile = [[[NSUserDefaults standardUserDefaults] objectForKey: @"PERSONAL"]
	       objectForKey: [Utilities profileNameForItemTitle: [personalProfilePopUpButton titleOfSelectedItem]] ];
  
  // We set the From header value
  anInternetAddress = [[InternetAddress alloc] initWithPersonal: [aProfile objectForKey: @"NAME"]
					       andAddress: [aProfile objectForKey: @"EMAILADDR"]];
  [message setFrom: anInternetAddress];
  RELEASE(anInternetAddress);
  
  // We set the Reply-To address (if we need too)
  aString = [aProfile objectForKey: @"REPLYTOADDR"];
  
  if (aString && [[aString stringByTrimmingWhiteSpaces] length] > 0)
    {
      anInternetAddress = [[InternetAddress alloc] initWithString: aString];
      
      if (anInternetAddress)
	{
	  [message setReplyTo: anInternetAddress];
	  RELEASE(anInternetAddress);
	}
    }
  else
    {
      [message setReplyTo: nil];
    }
  
  // We set the Organization header value (if we need to)
  aString = [aProfile objectForKey: @"ORGANIZATION"];
  
  if (aString && [[aString stringByTrimmingWhiteSpaces] length] > 0)
    {
      [message setOrganization: aString];
    }

  // For now, if there are recipients, we remove them in case a user add or changed them manually
  if ([message recipientsCount] > 0) 
    {
      [message removeAllRecipients];
    }

  // We decode our recipients, let's begin with the To field.
  if ( [[[toText stringValue] stringByTrimmingWhiteSpaces] length] > 0 )
    {
      to_array = [self _recipientsFromString: [toText stringValue] ];
      
      for (i = 0; i < [to_array count]; i++)
	{
	  anInternetAddress = [[InternetAddress alloc] initWithString: [to_array objectAtIndex:i]];
	  
	  if ( !anInternetAddress )
	    {
	      [self _showAlertPanelWithString: [to_array objectAtIndex: i] ];
	      return NO;
	    }
	  
	  [anInternetAddress setType: TO];
	  [message addToRecipients: anInternetAddress];
	  RELEASE(anInternetAddress);
	}
    }

  // We decode our Cc field, if we need to
  if ( showCc )
    {
      cc_array = [self _recipientsFromString: [ccText stringValue]];
      
      for (i = 0; i < [cc_array count]; i++)
	{
	  anInternetAddress = [[InternetAddress alloc] initWithString: [cc_array objectAtIndex:i]];

	  if ( !anInternetAddress )
	    {
	      [self _showAlertPanelWithString: [cc_array objectAtIndex: i] ];
	      return NO;
	    }

	  [anInternetAddress setType: CC];
	  [message addToRecipients: anInternetAddress];
	  RELEASE(anInternetAddress);
	}
      
    }

  // We decode our Bcc field, if we need to
  if ( showBcc )
    {
      bcc_array = [self _recipientsFromString: [bccText stringValue]];
      
      for (i = 0; i < [bcc_array count]; i++)
	{
	  anInternetAddress = [[InternetAddress alloc] initWithString: [bcc_array objectAtIndex:i]];
	  
	  if ( !anInternetAddress )
	    {
	      [self _showAlertPanelWithString: [bcc_array objectAtIndex: i] ];
	      return NO;
	    }

	  [anInternetAddress setType: BCC];
	  [message addToRecipients: anInternetAddress];
	  RELEASE(anInternetAddress);
	}
    }

  
  // We decode our Subject field
  [message setSubject: [subjectText stringValue]];
  
  // We finally add all our addition headers from the user defaults
  allAdditionalHeaders = [[NSUserDefaults standardUserDefaults] 
			   objectForKey: @"ADDITIONALOUTGOINGHEADERS"];
  
  if ( allAdditionalHeaders )
    {
      NSEnumerator *anEnumerator;
      NSString *aKey, *aValue;

      anEnumerator = [allAdditionalHeaders keyEnumerator];
      
      while ( (aKey = [anEnumerator nextObject]) )
	{
	  // We get our value
	  aValue = [allAdditionalHeaders objectForKey: aKey];
	  
	  // We add our X- if we don't have it
	  if ( ![aKey hasPrefix: @"X-"] )
	    {
	      aKey = [NSString stringWithFormat: @"X-%@", aKey];
	    }

	  if ( [aKey compare: @"X-Mailer"  options: NSCaseInsensitiveSearch] == NSOrderedSame )
	    {
	      hasFoundXMailerHeader = YES;
	    }
	  
	  [message addHeader: aKey
		   withValue: aValue];
	}
    }

  // We add our X-Mailer: GNUMail.app (Version ...) header if the user hasn't defined one
  if ( !hasFoundXMailerHeader )
    {
      [message addHeader: @"X-Mailer"
	       withValue: [NSString stringWithFormat: @"GNUMail.app (Version %@)", GNUMAIL_VERSION] ];
    }

  /// We get our text storage
  textStorage = [textView textStorage];
  
  // We now build our message body - considering all the attachments
  if ( ![textStorage containsAttachments] )
    {

      [self _setPlainTextContentFromString: [textView string]
	    inPart: message];
    }
  else
    {         
      MimeMultipart *aMimeMultipart;
      MimeBodyPart *aMimeBodyPart;      
      
      NSAutoreleasePool *pool;
      NSString *aString;

      // We create our local autorelease pool
      pool = [[NSAutoreleasePool alloc] init];

      aMimeMultipart = [[MimeMultipart alloc] init];
      
      // We first decode our text/plain body
      aMimeBodyPart = [[MimeBodyPart alloc] init];

      [self _setPlainTextContentFromString: [self _plainTextContentFromTextView]
	    inPart: aMimeBodyPart];

      // We add our new part
      [aMimeMultipart addBodyPart: aMimeBodyPart];
      RELEASE(aMimeBodyPart);

      // We get our text store string representation
      aString = [textStorage string];
      
      // We finally add all our attachments
      for (i = 0; i < [textStorage length]; i++)
	{
	  NSTextAttachment *aTextAttachment;
	  
	  if ( [aString characterAtIndex: i] != NSAttachmentCharacter )
	    {
	      continue;
	    }

	  aTextAttachment = [textStorage attribute: NSAttachmentAttributeName		      
					 atIndex: i 
					 effectiveRange: NULL];
	  if ( aTextAttachment ) 
	    {
	      Part *aPart;
	      
	      aPart = [(ExtendedAttachmentCell *)[aTextAttachment attachmentCell] part];
	      
	      if ( aPart )
		{
		  [aMimeMultipart addBodyPart: (MimeBodyPart *)aPart];
		}
	      else
		{
		  NSFileWrapper *aFileWrapper;
		  MimeType *aMimeType;
		  NSData *aData;
		  
		  aMimeBodyPart = [[MimeBodyPart alloc] init];
		  
		  aFileWrapper = [aTextAttachment fileWrapper];
		  [aMimeBodyPart setFilename: [[aFileWrapper filename] lastPathComponent]];
		  
		  // We search for the content-type to use 
		  aMimeType = [[MimeTypeManager singleInstance] bestMimeTypeForFileExtension:
								  [[[aFileWrapper filename] lastPathComponent]
								    pathExtension]];
		  if ( aMimeType )
		    {
		      [aMimeBodyPart setContentType: [aMimeType mimeType]];
		    }
		  else
		    {
		      [aMimeBodyPart setContentType: @"application/octet-stream"];
		    }
		  
		  [aMimeBodyPart setContentTransferEncoding: BASE64];   // always base64 encoding for now
		  [aMimeBodyPart setContentDisposition: @"attachment"]; // always attachment for now

		  aData = [aFileWrapper regularFileContents];
		  [aMimeBodyPart setContent: aData];
		  [aMimeMultipart addBodyPart: aMimeBodyPart];
		  RELEASE(aMimeBodyPart);
		}
	      
	    } // if ( aTextAttachment ) 
	  
	} // for (...)
      
      [message setContentTransferEncoding: NONE];
      [message setContentType: @"multipart/mixed"];     
      [message setContent: aMimeMultipart];
      
      // We generate a new boundary for our message
      [message setBoundary: [MimeUtility generateBoundary]];
      
      RELEASE(aMimeMultipart);
      RELEASE(pool);
    }

  //
  // We inform our bundles that our message WAS encoded
  //
  for (i = 0; i < [[GNUMail allBundles] count]; i++)
    {
      id<GNUMailBundle> aBundle;
      
      aBundle = [[GNUMail allBundles] objectAtIndex: i];
      
      if ( [aBundle respondsToSelector: @selector(messageWasEncoded:)] )
	{
	  // If AT LEAST ONE bundle failed to encode the message, we
	  // stop everything for now.
	  //
	  // FIXME: should use the returned copy successively instead of
	  //        modifying directly "message"
	  //
	  if ( ![aBundle messageWasEncoded: message] )
	    {
	      return NO;
	    }
	}
    }

  return YES;
}

//
// changes the profile in the popup button
//
- (void) setProfile: (NSString *) theProfile
{
  if ( theProfile )
    {
      [Utilities loadPersonalProfilesInPopUpButton: personalProfilePopUpButton
		 selectedProfile: theProfile];
      [self personalProfilesSelectionHasChanged: nil];
    }		
}


@end


//
// private methods
//
@implementation EditWindowController (Private)

- (void) _adjustNextKeyViews
{
  if (showCc && showBcc)
    {
      [toText setNextKeyView: ccText];
      [ccText setNextKeyView: bccText];
      [bccText setNextKeyView: subjectText];
    }	 
  else if (showCc && !showBcc)
    {
      [toText setNextKeyView: ccText];
      [ccText setNextKeyView: subjectText];
    }
  else if (!showCc && showBcc)
    {
      [toText setNextKeyView: bccText];
      [bccText setNextKeyView: subjectText];
    }
  else
    {
      [toText setNextKeyView: subjectText];
    }
}


//
//
//
- (void) _adjustWidgetsPosition
{  
  NSRect rectOfToText;
  float widthOfScrollView;

  rectOfToText = [toText frame];
  widthOfScrollView = [scrollView frame].size.width;
  
  if (showCc && showBcc)
    {
      NSDebugLog(@"showCc is YES showBcc is YES");
      
      // To - 25
      [ccLabel setFrame: NSMakeRect(5,rectOfToText.origin.y-25,50,21)];
      [ccText setFrame: NSMakeRect(65,rectOfToText.origin.y-25,widthOfScrollView-75,21)];
      
      // To - 50
      [bccLabel setFrame: NSMakeRect(5,rectOfToText.origin.y-50,50,21)];
      [bccText setFrame: NSMakeRect(65,rectOfToText.origin.y-50,widthOfScrollView-75,21)];
      
      // To - 75
      [subjectLabel setFrame: NSMakeRect(5,rectOfToText.origin.y-75,50,21)];
      [subjectText setFrame: NSMakeRect(65,rectOfToText.origin.y-75,widthOfScrollView-75,21)];
      
      // To - 100
      [sizeLabel setFrame: NSMakeRect(5,rectOfToText.origin.y-100,200,21)];
      
      // Space left...
      [scrollView setFrame: NSMakeRect(5,5,widthOfScrollView,rectOfToText.origin.y-105)];
      [[textView textContainer] setContainerSize: NSMakeSize([[scrollView contentView] frame].size.width, 
							     1E7)];
    }	 
  else if (showCc && !showBcc)
    {
      NSDebugLog(@"showCc is YES showBcc is NO");

      // To - 25
      [ccLabel setFrame: NSMakeRect(5,rectOfToText.origin.y-25,50,21)];
      [ccText setFrame: NSMakeRect(65,rectOfToText.origin.y-25,widthOfScrollView-75,21)];
      
      // To - 50
      [subjectLabel setFrame: NSMakeRect(5,rectOfToText.origin.y-50,50,21)];
      [subjectText setFrame: NSMakeRect(65,rectOfToText.origin.y-50,widthOfScrollView-75,21)];
      
      // To - 75
      [sizeLabel setFrame: NSMakeRect(5,rectOfToText.origin.y-75,200,21)];

      // Space left...
      [scrollView setFrame: NSMakeRect(5,5,widthOfScrollView,rectOfToText.origin.y-80)];
      [[textView textContainer] setContainerSize: NSMakeSize([[scrollView contentView] frame].size.width, 
							     1E7)];
    }
  else if (!showCc && showBcc)
    {
      NSDebugLog(@"showCc is NO showBcc is YES");

      // To - 25
      [bccLabel setFrame: NSMakeRect(5,rectOfToText.origin.y-25,50,21)];
      [bccText setFrame: NSMakeRect(65,rectOfToText.origin.y-25,widthOfScrollView-75,21)];
      
      // To - 50
      [subjectLabel setFrame: NSMakeRect(5,rectOfToText.origin.y-50,50,21)];
      [subjectText setFrame: NSMakeRect(65,rectOfToText.origin.y-50,widthOfScrollView-75,21)];
      
      // To - 75
      [sizeLabel setFrame: NSMakeRect(5,rectOfToText.origin.y-75,200,21)];

      // Space left...
      [scrollView setFrame: NSMakeRect(5,5,widthOfScrollView,rectOfToText.origin.y-80)];
      [[textView textContainer] setContainerSize: NSMakeSize([[scrollView contentView] frame].size.width, 
							     1E7)];
    }
  else
    {
      NSDebugLog(@"showCc is NO showBcc is NO");
     
      // To - 25
      [subjectLabel setFrame: NSMakeRect(5,rectOfToText.origin.y-25,50,21)];
      [subjectText setFrame: NSMakeRect(65,rectOfToText.origin.y-25,widthOfScrollView-75,21)];
      
      // To - 50
      [sizeLabel setFrame: NSMakeRect(5,rectOfToText.origin.y-50,200,21)];
     
      // Space left...
      [scrollView setFrame: NSMakeRect(5,5,widthOfScrollView,rectOfToText.origin.y-55)];
      [[textView textContainer] setContainerSize: NSMakeSize([[scrollView contentView] frame].size.width, 
							     1E7)];
    }
}


//
//
//
- (void) _loadAccessoryViews
{
  int i;

  for (i = 0; i < [[GNUMail allBundles] count]; i++)
    {
      id<GNUMailBundle> aBundle;
      
      aBundle = [[GNUMail allBundles] objectAtIndex: i];
      
      if ( [aBundle hasComposeViewAccessory] )
	{
#ifdef MACOSX
          NSToolbarItem *aToolbarItem;
          NSToolbar *aToolbar;
          id aView;
          
          aToolbarItem = [[NSToolbarItem alloc] initWithItemIdentifier: [aBundle name]];
          [allowedToolbarItemIdentifiers addObject: [aBundle name]];
          
          [additionalToolbarItems setObject: aToolbarItem
                                  forKey: [aBundle name]];
                                                      
          aView = [aBundle composeViewAccessory];
          [aToolbarItem setView: aView];
          [aToolbarItem setLabel: [aBundle name]];               // name
          [aToolbarItem setPaletteLabel: [aBundle description]]; // description
          [aToolbarItem setMinSize: [aView frame].size];
          [aToolbarItem setMaxSize: [aView frame].size];
          RELEASE(aToolbarItem);
          
          aToolbar = [[self window] toolbar];
          [aToolbar insertItemWithItemIdentifier: [aBundle name]
                    atIndex: [[aToolbar visibleItems] count]];
	  
#else    
          NSRect aRect;
	  id aView;
	  
	  aView = [aBundle composeViewAccessory];
	  aRect = [aView frame];      
	  aRect.origin.x = ([addresses frame].origin.x + 72);
	  aRect.origin.y = [addresses frame].origin.y;

	  [aView setFrame: aRect];
	  [aView setAutoresizingMask: NSViewMinYMargin];

	  [[[self window] contentView] addSubview: aView];
#endif
	}

      // We also set the current superview
      [aBundle setCurrentSuperview: [[self window] contentView]];
    }
}


//
//
//
- (void) _loadCharsets
{
  NSMutableArray *aMutableArray;
  
  
  // We first initialize our defaultCharsetPopUpButton with the right values.
  [defaultCharsetPopUpButton removeAllItems];
  
  aMutableArray = [[NSMutableArray alloc] init];
  [aMutableArray addObjectsFromArray: [[MimeUtility allCharsets] allValues]];
  [aMutableArray sortUsingSelector: @selector(compare:)];

  // We always add our "Automatic" item
  [aMutableArray insertObject: _(@"Automatic")
		 atIndex: 0];

  [defaultCharsetPopUpButton addItemsWithTitles: aMutableArray];
  RELEASE(aMutableArray);

  // We select the "default" charset
  if ( [[NSUserDefaults standardUserDefaults] objectForKey: @"DEFAULT_CHARSET"] )
    {
      NSString *aString;

      aString = [[MimeUtility allCharsets] objectForKey: [[[NSUserDefaults standardUserDefaults] 
							    objectForKey: @"DEFAULT_CHARSET"] lowercaseString]];
      
      if ( aString )
	{
	  [defaultCharsetPopUpButton selectItemWithTitle: aString];
	}
      else
	{
	  [defaultCharsetPopUpButton selectItemAtIndex: 0];
	}
    }
  else
    {
      [defaultCharsetPopUpButton selectItemAtIndex: 0];
    }
}


//
//
//
- (void) _loadPersonalProfiles
{
	[Utilities loadPersonalProfilesInPopUpButton: personalProfilePopUpButton selectedProfile: nil];
}


//
//
//
- (void) _loadTransportMethods
{
  [Utilities loadTransportMethodsInPopUpButton: transportMethodPopUpButton];
}


//
//
//
- (NSString *) _loadSignature
{
  NSString *aSignature;
  NSDictionary *aDictionary;

  aDictionary = [[[NSUserDefaults standardUserDefaults] objectForKey: @"PERSONAL"] 
		  objectForKey: [Utilities profileNameForItemTitle: 
					     [personalProfilePopUpButton itemTitleAtIndex: personalProfilePopUpButtonIndex]]];

  aSignature = nil;

  if ( [(NSNumber *)[aDictionary objectForKey: @"SIGNATURE_SOURCE"] intValue] == 0 )
    {
      aSignature = [NSString stringWithContentsOfFile:
			       [aDictionary objectForKey: @"SIGNATURE"]];
    }
  else
    {
      NSFileHandle *aFileHandle;
      NSTask *aTask;
      NSPipe *aPipe;
      NSData *aData;
      NSRange aRange;
      NSString *aString;
      
      // We get our program's name (and arguments, if any)
      aString = [aDictionary objectForKey: @"SIGNATURE"];

      // If a signature hasn't been set, let's return.
      if ( !aString )
	{
	  return nil;
	}
            
      aPipe = [NSPipe pipe];
      aFileHandle = [aPipe fileHandleForReading];
      
      aTask = [[NSTask alloc] init];
      [aTask setStandardOutput: aPipe];
      
      // We trim our string from any whitespaces
      aString = [aString stringByTrimmingWhiteSpaces];

      // We verify if our program to lauch has any arguments
      aRange = [aString rangeOfString: @" "];
      
      if ( aRange.length )
	{
	  [aTask setLaunchPath: [aString substringToIndex: aRange.location]];
	  [aTask setArguments: [NSArray arrayWithObjects: [aString substringFromIndex: (aRange.location + 1)],
					nil]];
	}
      else
	{
	  [aTask setLaunchPath: aString];
	}
      
      // We verify if our launch path points to an executable file
      if ( ![[NSFileManager defaultManager] isExecutableFileAtPath: [aTask launchPath]] )
	{
	  NSDebugLog(@"The signature's path doesn't point to an executable! Ignored.");
	  RELEASE(aTask);
	  return nil;
	}
      
      // We launch our task
      [aTask launch];
      
      while ( [aTask isRunning] )
	{
	  [[NSRunLoop currentRunLoop] runMode: NSDefaultRunLoopMode
				      beforeDate: [NSDate distantFuture]];
	}
      
      aData = [aFileHandle readDataToEndOfFile];
      
      aSignature = [[NSString alloc] initWithData: aData
				     encoding: NSASCIIStringEncoding];
      AUTORELEASE(aSignature);
      
      RELEASE(aTask);
    }
  
  if ( aSignature )
    {
      return [NSString stringWithFormat: @"\n\n-- \n%@", aSignature];
    }

  return nil;
}


//
//
//
- (void) _replaceSignature
{
  NSMutableDictionary *tAttr;                  // text attributes
  NSString *previousSignature, *newSignature;  // signatures

  if (personalProfilePopUpButtonIndex < 0)
    {
      personalProfilePopUpButtonIndex = 0;
      return;
    }
  
  tAttr = [[NSMutableDictionary alloc] init];
  
  [tAttr setObject: [Utilities fontFromFamilyName: [[NSUserDefaults standardUserDefaults] 
						     objectForKey: @"MESSAGE_FONT_NAME"]
			       trait: NSUnboldFontMask
			       size: [[NSUserDefaults standardUserDefaults] floatForKey: @"MESSAGE_FONT_SIZE"]]
	 forKey: NSFontAttributeName];

  // If we don't want a signature...
  if ( [self signaturePosition] == SIGNATURE_HIDDEN )
    {
      RELEASE(tAttr);
      return;
    }

  // We load the previous signature and remove it from the text storage
  previousSignature = [self _loadSignature];
  
  if ( previousSignature )
    {
      NSRange aRange;
      
      if ( [self signaturePosition] == SIGNATURE_BEGINNING )
	{
	  aRange = [[[textView textStorage] string] rangeOfString: previousSignature];
	}
      else if ( [self signaturePosition] == SIGNATURE_END )
	{
	  aRange = [[[textView textStorage] string] rangeOfString: previousSignature 
						    options: NSBackwardsSearch];
      }
      if ( aRange.length )
	{
	  [[textView textStorage] deleteCharactersInRange: aRange];
	}
    }

  // We get the index of the selected profile
  personalProfilePopUpButtonIndex = [personalProfilePopUpButton indexOfSelectedItem];

  // We load the new signature and insert it a the proper position
  newSignature = [self _loadSignature];

  if ( newSignature )
    {
      if ( [self signaturePosition] == SIGNATURE_BEGINNING )
  	{
	  NSAttributedString *anAttributedString;
	  
	  anAttributedString = [[NSAttributedString alloc] initWithString: newSignature
							   attributes: tAttr];
	  
	  [[textView textStorage] insertAttributedString: anAttributedString
				  atIndex: 0];
	  
	  RELEASE(anAttributedString);
	}
      else if ( [self signaturePosition] == SIGNATURE_END )
	{
	  NSAttributedString *anAttributedString;

	  anAttributedString = [[NSAttributedString alloc] initWithString: newSignature
							   attributes: tAttr];
							     
	  [[textView textStorage] appendAttributedString: anAttributedString];
	  
	  RELEASE(anAttributedString);
	}
    }

  // We update our size label
  [self _updateSizeLabel];

  RELEASE(tAttr);
}


//
//
//
- (NSString *) _passwordForServerName: (NSString *) theName
			       prompt: (BOOL) aBOOL
{
  NSDictionary *aDictionary;
  NSString *password;
  
  //
  // We get our right transport method
  // We don't need to verify if it's a SMTP transport because this method isn't called
  // if it's NOT SMTP.
  //
  aDictionary = [[[NSUserDefaults standardUserDefaults] objectForKey: @"SENDING"]
		  objectAtIndex: [transportMethodPopUpButton indexOfSelectedItem]];

  if ( aDictionary )
    {    
      password = nil;
      
      // We verify in the user defaults
      if ( [[aDictionary objectForKey: @"REMEMBERPASSWORD"] intValue] == NSOnState )
	{
	  password = [Utilities decryptPassword: [aDictionary objectForKey: @"SMTP_PASSWORD"] 
				withKey: theName];
	}

      // We verify in our cache
      if (! password )
	{
	  password = [passwordCache objectForKey: theName];
	}
      
      // If we must prompt for the password
      if (! password && aBOOL )
	{
	  PasswordPanelController *theController;
	  int result;
	  
	  theController = [[PasswordPanelController alloc] initWithWindowNibName: @"PasswordPanel"];
	  [[theController window] setTitle: theName];
      
	  result = [NSApp runModalForWindow: [theController window]];
	  
	  // If the user has entered a password...
	  if (result == NSRunStoppedResponse)
	    {
	      password = [theController password];
	      
	      // Let's cache this password...
	      [passwordCache setObject: password
			     forKey: theName];
	    }
	  else
	    {
	      password = nil;
	    }
	  
	  RELEASE(theController);
	}
    }
  else
    {
      password = nil;
    }
  
  return password;
}


//
//
//
- (NSString *) _plainTextContentFromTextView
{
  NSMutableString *aMutableString;
  NSTextStorage *textStorage;

  NSAutoreleasePool *pool;
  int i;
  
  textStorage = [textView textStorage];

  aMutableString = [[NSMutableString alloc] initWithString: [textStorage string]];
 
  pool = [[NSAutoreleasePool alloc] init];
  
  for (i = ([aMutableString length] - 1); i >= 0; i--)
    { 
      NSTextAttachment *aTextAttachment;     
      ExtendedAttachmentCell *cell;
 
      if ( [aMutableString characterAtIndex: i] != NSAttachmentCharacter )
	{
	  continue;
	}

      aTextAttachment = [textStorage attribute: NSAttachmentAttributeName
				     atIndex: i
				     effectiveRange: NULL];
      
      cell = (ExtendedAttachmentCell *)[aTextAttachment attachmentCell];
      
      if ( [cell part] && [[cell part] filename] )
	{
	  [aMutableString replaceCharactersInRange: NSMakeRange(i, 1)
			  withString: [NSString stringWithFormat: @"<%@>", [[cell part] filename]] ];
	}
      else if ( [cell part] )
	{
	  [aMutableString replaceCharactersInRange: NSMakeRange(i, 1)
			  withString: @"<unknown>"];
	}
      else
	{
	  [aMutableString replaceCharactersInRange: NSMakeRange(i, 1)
			  withString: [NSString stringWithFormat: @"<%@>", [[[aTextAttachment fileWrapper] filename]
									     lastPathComponent]] ];
	}
    }
  
  RELEASE(pool);
  
  return AUTORELEASE(aMutableString);
}


//
//
//
- (NSArray *) _recipientsFromString: (NSString *) theString
{
  NSMutableArray *pairsStack, *returnArray;
  NSMutableString *aString;
  unichar c;
  int i;

  returnArray = [NSMutableArray array];
  pairsStack = [NSMutableArray array];
  aString = (NSMutableString *)[NSMutableString string];
  
  for (i = 0; i < [theString length]; i++)
    {
      c = [theString characterAtIndex: i];
      switch (c)
	{
	case ',':
	  if ([pairsStack count] == 0 && [aString length])
	    {
	      [returnArray addObject: [NSString stringWithString: aString]];
	      [aString replaceCharactersInRange: NSMakeRange(0, [aString length])
		       withString: @""];
	      continue;
	    }
	  break;
	case ' ':
	  if (![aString length])
	    {
	      continue;
	    }
	  break;
	case '"':
	  if ([pairsStack count] == 0 || [(NSNumber*)[pairsStack lastObject] intValue] != '"')
	    {
	      [pairsStack addObject: [NSNumber numberWithChar: c]];
	    }
	  else
	    {
	      [pairsStack removeLastObject];
	    }
	  break;
	case '<':
	  [pairsStack addObject: [NSNumber numberWithChar: c]];
	  break;
	case '>':
	  if ([pairsStack count] && [(NSNumber*)[pairsStack lastObject] intValue] == '<')
	    {
	      [pairsStack removeLastObject];
	    }
	  break;
	case '(':
	  [pairsStack addObject: [NSNumber numberWithChar: c]];
	  break;
	case ')':
	  if ([pairsStack count] && [(NSNumber*)[pairsStack lastObject] intValue] == '(')
	    {
	      [pairsStack removeLastObject];
	    }
	  break;
	}
      [aString appendFormat: @"%c", c];
    }

  if ([pairsStack count] == 0 && [aString length])
    {
      [returnArray addObject: [NSString stringWithString: aString]];
    }

  return returnArray;
}


//
//
//
- (void) _sendMessageUsingPorts: (NSArray *) thePorts
{
  NSConnection *aServerConnection;
  NSAutoreleasePool *pool; 
  
  NSDictionary *aDictionary;
  BOOL manuallyStoppedTransfer;
  
  // We first create our autorealease pool for our thread
  pool = [[NSAutoreleasePool alloc] init];
  manuallyStoppedTransfer = NO;

  // We first get our connection to our main thread
  aServerConnection = [[NSConnection alloc] initWithReceivePort: [thePorts objectAtIndex: 0]
					    sendPort: [thePorts objectAtIndex: 1]];
  [aServerConnection setRequestTimeout: LONG_DELAY];
  [aServerConnection setReplyTimeout: LONG_DELAY];
 
  // Then we really send our message using the right delivery method!
  aDictionary = [[[NSUserDefaults standardUserDefaults] objectForKey: @"SENDING"]
		  objectAtIndex: [transportMethodPopUpButton indexOfSelectedItem]];

  if ( [[aDictionary objectForKey: @"TRANSPORT_METHOD"] intValue] == TRANSPORT_SMTP )
    {
      NSNumber *portValue;
      SMTP *aSMTP;
      
      aSMTP = nil;
      
      portValue = [aDictionary objectForKey: @"SMTP_PORT"];

      if ( !portValue )
	{
	  portValue = [NSNumber numberWithInt: 25];
	}

      NS_DURING
	  aSMTP = [[SMTP alloc] initWithName: [aDictionary objectForKey: @"SMTP_HOST"]
				port: [portValue intValue]];

	  [[aSMTP tcpConnection] setStopTarget: self];
	  [[aSMTP tcpConnection] setStopSelector: @selector(mustStopTransfer)];
	  
	  // We verify if we need to authenticate ourself to the SMTP sever
	  if ( [aDictionary objectForKey: @"SMTP_AUTH"] &&
	       [[aDictionary objectForKey: @"SMTP_AUTH"] intValue] == NSOnState )
	    {
	      NSString *password;
	      BOOL aBOOL;
	      
	      password = [self _passwordForServerName: [aDictionary objectForKey: @"SMTP_HOST"]
			       prompt: NO];
	      aBOOL = NO;
	      
	      if ( password )
		{
		  aBOOL = [aSMTP authenticateWithUsername: [aDictionary objectForKey: @"SMTP_USERNAME"]
				 password: password
				 mechanism: [aDictionary objectForKey: @"SMTP_AUTH_MECHANISM"]];
		}
	      
	      if ( !aBOOL )
		{
		  [aSMTP close];
		  RELEASE(aSMTP);
		  
		  [(id)[aServerConnection rootProxy] authenticationFailedForServer: [aDictionary objectForKey: @"SMTP_HOST"]];
		  goto done_handler;
		}
	    }
	  
	  if ( aSMTP )
	    {
	      NSDebugLog(@"Sending message...");
	      messageWasSent = [aSMTP sendMessage: message];
	      [aSMTP close];
	    }
	  else
	    {
	      messageWasSent = NO;
	    }
	  
	  RELEASE(aSMTP);

          done_handler:
	  aSMTP = nil;  

      NS_HANDLER
	  if ( [[localException name] isEqualToString: PantomimeStopException] )
	    {
	      [[NSDistributedNotificationCenter defaultCenter]
		postNotificationName: GNUMailConsoleMessage
		object: nil
		userInfo: [NSDictionary dictionaryWithObject: _(@"Manual stop of sending thread.")
					forKey: @"Message"]];
	      
	      manuallyStoppedTransfer = YES;
	    }
	  else
	    {
	      // We must inform the user that an error occured.
	      [(id)[aServerConnection rootProxy] errorOccuredWhenSendingMessage];
	    }
	  
	  if ( aSMTP )
	    {
	      NS_DURING
		  mustStopTransfer = NO;
		  //[aSMTP close]; // FIXME - Maybe set the TCP timeout to a very small value?
		  RELEASE(aSMTP);
		  aSMTP = nil;
 
	      NS_HANDLER
		  NSDebugLog(@"Error occured while closing the connection to the server...");
       	
	      NS_ENDHANDLER
	    }
	
      NS_ENDHANDLER
    }
  else 
    {
      Sendmail *aSendmail;

      aSendmail = [[Sendmail alloc] initWithPathToSendmail: [aDictionary objectForKey: @"MAILER_PATH"]];
      messageWasSent = [aSendmail sendMessage: message];

      RELEASE(aSendmail);
    }


  if ( messageWasSent )
    {
      [(id)[aServerConnection rootProxy] messageWasSentFromRAWSource:
	     [message dataUsingSendingMode: SEND_TO_FOLDER]];
    }
  else 
    {
      if ( !manuallyStoppedTransfer )
	{
	  [(id)[aServerConnection rootProxy] errorOccuredWhenSendingMessage];
	}
      else
	{
	  [(id)[aServerConnection rootProxy] stopAnimationAndRestoreButtonState];
	}
    }
  
  //[[aServerConnection sendPort] invalidate];
  //[[aServerConnection receivePort] invalidate];	
  //[aServerConnection invalidate];
  RELEASE(aServerConnection);

  [[NSDistributedNotificationCenter defaultCenter]
    postNotificationName: GNUMailConsoleMessage
    object: nil
    userInfo: [NSDictionary dictionaryWithObject: _(@"Exiting the thread used to send the E-Mail...")
			    forKey: @"Message"]];

  RELEASE(pool);

  //[NSThread exit];
}


//
//
//
- (void) _setPlainTextContentFromString: (NSString *) theString
                                 inPart: (Part *) thePart
{
  // We must first verify if our string contains any non-ascii character
  if ( [MimeUtility isASCIIString: theString] )
    {
      int wrapLimit;
      
      wrapLimit = [[NSUserDefaults standardUserDefaults] integerForKey: @"LINE_WRAP_LIMIT"];
      
      if ( !wrapLimit )
	{
	  wrapLimit = 998;
	}
      
      [thePart setContentType: @"text/plain"];                // Content-Type
      [thePart setContentTransferEncoding: NONE];             // encoding -> none
      [thePart setCharset: @"us-ascii"];
      [thePart setFormat: FORMAT_FLOWED];
      [thePart setLineLength: wrapLimit];
      [thePart setContent: theString];
    }
  else
    {
      NSString *aCharset;

      [defaultCharsetPopUpButton synchronizeTitleAndSelectedItem];
      
      // We verify if we are using automatic charset detection
      if ( [defaultCharsetPopUpButton indexOfSelectedItem] <= 0 )
	{
	  aCharset = [MimeUtility charsetForString: theString];
	}
      else
	{
	  NSArray *allKeys;

	  allKeys = [[MimeUtility allCharsets] allKeysForObject: [defaultCharsetPopUpButton titleOfSelectedItem]];
	  
	  if ( [allKeys count] )
	    {
	      aCharset = [allKeys objectAtIndex: 0];
	    }
	  else
	    {
	      aCharset = [MimeUtility charsetForString: theString];
	    }
	}
      
      [thePart setContentType: @"text/plain"];               // Content-Type
      [thePart setContentTransferEncoding: QUOTEDPRINTABLE]; // encoding -> quoted-printable
      [thePart setFormat: FORMAT_UNKNOWN];
      [thePart setCharset: aCharset];
      [thePart setContent: theString];
    }
}


//
//
//
- (void) _showAlertPanelWithString: (NSString *) theString
{
  NSRunInformationalAlertPanel(_(@"Error!"),
			       _(@"An error occured while decoding %@. Please fix this address."),
			       _(@"OK"),
			       NULL,
			       NULL,
			       theString);
}


//
//
//
- (void) _updateViewWithMessage: (Message *) theMessage
		appendSignature: (BOOL) aBOOL
{
  NSMutableString *aMutableString;
  InternetAddress *theRecipient;
  NSEnumerator *recipientsEnumerator;
  
  // We get all recipients and we set our To/Cc fields.
  aMutableString = [[NSMutableString alloc] init];
  recipientsEnumerator = [[theMessage recipients] objectEnumerator];
  
  while ( (theRecipient = [recipientsEnumerator nextObject]) )
    {  
      if ( [theRecipient type] == TO )
	{
	  [toText setStringValue:[theRecipient unicodeStringValue]];
	}
      else if ([theRecipient type] == CC) 
	{
	  [aMutableString appendString:[NSString stringWithFormat:@"%@, ", [theRecipient unicodeStringValue]]];
	}
    }
  
  if ([aMutableString length] > 0) 
    {
      [ccText setStringValue: [aMutableString substringToIndex:([aMutableString length] - 2)]];
    }

  // We are done with our mutable string, let's release it.
  RELEASE(aMutableString);

  // We set our subject
  if ( [theMessage subject] )
    {
      [subjectText setStringValue: [theMessage subject]];
    }
  else
    {
      [subjectText setStringValue: @""];
    }

  // We now set the content of the message (considering all attachments)
  // in case of a forward / reply
  if ( [theMessage content] )
    {
      [[textView textStorage]
	appendAttributedString: [Utilities formattedAttributedStringFromAttributedString:
					     [Utilities attributedStringFromContentForPart: theMessage]] ];
    }

  // We finally set the signature, if we need to
  if ( aBOOL )
    {
      [self _replaceSignature];
    }
}


//
//
//
- (void) _updateSizeLabel
{
  NSTextStorage *aTextStorage;
  NSString *aString;
  float size;

  size = [[textView string] length];
  size = (float)(size / (float)1024);

  aTextStorage = [textView textStorage];
  
  // FIXME
  // This is _very_ slow 
  if ( [aTextStorage containsAttachments] )
    {
      NSTextAttachment *aTextAttachment;
      int i;

      for (i = 0; i < [aTextStorage length]; i++)
	{  
	  aTextAttachment = [aTextStorage attribute: NSAttachmentAttributeName		      
					  atIndex: i 
					  effectiveRange: NULL];
	  if ( aTextAttachment ) 
	    {
	      Part *aPart;

	      aPart = [(ExtendedAttachmentCell *)[aTextAttachment attachmentCell] part];
	      
	      if ( aPart )
		{
		  size += (float)((float)[aPart size] / (float)1024);
		}
	      else
		{
		  NSFileWrapper *aFileWrapper;
		  
		  aFileWrapper = [aTextAttachment fileWrapper];

		  size += (float)((float)[[aFileWrapper regularFileContents] length] / (float)1024);
		}
	    }
	}
    }

  // We finally update the string value of our label
  aString = [NSString stringWithFormat: _(@"%0.1fKB (%d characters)"), size, [aTextStorage length]];
  
  [sizeLabel setStringValue: aString];
  [sizeLabel setNeedsDisplay: YES];
}

@end
