/*
**  FilterManager.m
**
**  Copyright (c) 2001, 2002
**
**  Author: Ludovic Marcotte <ludovic@Sophos.ca>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#import <AppKit/AppKit.h>

#import "AddressBook.h"
#import "AddressBookController.h"
#import "GNUMailConstants.h"
#import "FilterManager.h"
#import "Filter.h"
#import "NSRegEx.h"
#import "NSStringExtensions.h"
#import "MailboxManagerController.h"
#import "Utilities.h"

#import <Pantomime/InternetAddress.h>
#import <Pantomime/Message.h>
#import <Pantomime/NSDataExtensions.h>
#import <Pantomime/URLName.h>

NSString *PathToFilters()
{
  return [NSString stringWithFormat: @"%@/%@",
		   GNUMailUserLibraryPath(), @"Filters"];
}


//
//
//
@implementation FilterManager


//
//
//
- (id) init
{
  self = [super init];

  [self setFilters: [NSMutableArray array]];

  return self;
}


//
//
//
- (void) dealloc
{
  RELEASE(filters);

  [super dealloc];
}


//
//
//
- (BOOL) synchronize
{
  return [NSArchiver archiveRootObject: self 
		     toFile: PathToFilters()];
}


//
// NSCoding protocol
//
- (void) encodeWithCoder: (NSCoder *) theCoder
{
  [theCoder encodeObject: [self filters]];
}

- (id) initWithCoder: (NSCoder *) theCoder
{
  self = [super init];

  [self setFilters: [theCoder decodeObject]];

  return self;
}


//
// access/mutation methods
//
- (Filter *) filterAtIndex: (int) theIndex
{
  return [filters objectAtIndex: theIndex];
}


//
//
//
- (void) addFilter: (Filter *) theFilter
{
  [filters addObject: theFilter];
}


//
//
//
- (void) addFilter: (Filter *) theFilter
	   atIndex: (int) theIndex
{
  [filters insertObject: theFilter
	   atIndex: theIndex];
}


//
//
//
- (void) removeFilter: (Filter *) theFilter;
{
  [filters removeObject: theFilter];
}


//
//
//
- (NSArray *) filters
{
  return filters;
}


//
//
//
- (void) setFilters: (NSArray *) theFilters
{
  if ( theFilters )
    {
      NSMutableArray *newFilters;

      newFilters = [[NSMutableArray alloc] initWithArray: theFilters];
      RELEASE(filters);
      filters = newFilters;
    }
  else
    {
      RELEASE(filters);
      filters = nil;
    }
}


//
//
//
- (BOOL) matchExistsForFilter: (Filter *) theFilter
		      message: (Message *) theMessage
{
  NSAutoreleasePool *pool;
  Message *aMessage;
  
  BOOL aBOOL;

  if ( !theFilter || !theMessage )
    {
      return NO;
    }
  
  pool = [[NSAutoreleasePool alloc] init];

  //
  // FIXME: This can be CPU intensive.
  //
  if ( [theFilter useExternalProgram] )
    {
      aMessage = [self _messageFromExternalProgramUsingFilter: theFilter
		       message: theMessage];
      
      if ( !aMessage )
	{
	  RELEASE(pool);
	  return NO;
	}
    }
  else
    {
      aMessage = theMessage;
    }
  
  aBOOL = [self _matchCriteriasFromMessage: aMessage
		filter: theFilter];
  
  if ( aMessage != theMessage )
    {
      RELEASE(aMessage);
    }

  RELEASE(pool);
  
  return aBOOL;
}


//
// This method returns the *first* filter that matches
// for the specified message. If no filter matches, nil
// is returned.
//
- (Filter *) matchedFilterForMessage: (Message *) theMessage
				type: (int) theType
{
  NSAutoreleasePool *pool;
  Message *aMessage;
  int i;

  if ( !theMessage )
    {
      return nil;
    }

  pool = [[NSAutoreleasePool alloc] init];
  
  for (i = 0; i < [filters count]; i++)
    {
      Filter *aFilter;

      aFilter = [filters objectAtIndex: i];

      if ( [aFilter isActive] && [aFilter type] == theType )
	{  
	  //
	  // FIXME: This can be CPU intensive.
	  //
	  if ( [aFilter useExternalProgram] )
	    {
	      aMessage = [self _messageFromExternalProgramUsingFilter: aFilter
			       message: theMessage];
	      
	      if ( !aMessage )
		{
		  continue;
		}
	    }
	  else
	    {
	      aMessage = theMessage;
	    }

	  //
	  // If we've a filter that matches, we stop everything. We don't need
	  // to continue searching in the other filters.
	  //
	  if ( [self _matchCriteriasFromMessage: aMessage
		     filter: aFilter] )
	    {
	      if ( aMessage != theMessage )
		{
		  RELEASE(aMessage);
		}
	      
	      RELEASE(pool);
	      return aFilter;
	    }

	  if ( aMessage != theMessage )
	    {
	      RELEASE(aMessage);
	    }
	  
	} // if ( [aFilter isActive] )
    }
  
  RELEASE(pool);

  return nil;
}


//
//
//
- (NSColor *) colorForMessage: (Message *) theMessage
{
  NSAutoreleasePool *pool;
  Message *aMessage;
  int i;

  if ( !theMessage )
    {
      return nil;
    }
  
  pool = [[NSAutoreleasePool alloc] init];

  for (i = 0; i < [filters count]; i++)
    {
      Filter *aFilter;
      
      aFilter = [filters objectAtIndex: i];
      
      if ( [aFilter action] == SET_COLOR  && [aFilter isActive])
	{
	  //
	  // FIXME: Should we allow external programs for "coloring" filters? That
	  //        can be *very CPU intensive* 
	  // 
	  if ( [aFilter useExternalProgram] )
	    {
	      aMessage = [self _messageFromExternalProgramUsingFilter: aFilter
			       message: theMessage];
	      
	      if ( !aMessage )
		{
		  continue;
		}
	    }
	  else
	    {
	      aMessage = theMessage;
	    }
	  
	  //
	  // If we've a filter that matches, we stop everything. We don't need
	  // to continue searching in the other filters.
	  //
	  if ( [self _matchCriteriasFromMessage: aMessage
		     filter: aFilter] )
	    {
	      if ( aMessage != theMessage )
		{
		  RELEASE(aMessage);
		}
	      
	      RELEASE(pool);
	      return [aFilter actionColor];
	    }

	  if ( aMessage != theMessage )
	    {
	      RELEASE(aMessage);
	    }
	} // if ( [aFilter action] == SET_COLOR  && [aFilter isActive])
    }
  
  RELEASE(pool);
  return nil;
}


//
// theKey equals, for example: marcotte @ mac.com
//
//
- (URLName *) matchedURLNameFromMessage: (Message *) theMessage
				   type: (int) theType
				    key: (NSString *) theKey
{
  URLName *aURLName, *aDefaultURLName;
  NSString *aDefaultURLNameAsString;
  Filter *aFilter;

  aFilter = [self matchedFilterForMessage: theMessage
		  type: theType];
  
  if ( theType == TYPE_INCOMING )
    {
      // We first verify if we have a default mailbox name associated to this profile. If we do,
      // then we use that one!
      NSDictionary *allValues;
      
      allValues = [[[NSUserDefaults standardUserDefaults] objectForKey: @"RECEIVING"] 
		    objectForKey: theKey];

      if ( [allValues objectForKey: @"MAILBOXNAME"] )
	{
	  id aStore;

	  aDefaultURLNameAsString = [allValues objectForKey: @"MAILBOXNAME"];
	  
	  //
	  // FIXME: We verify if the mailbox exists if it's a LOCAL mailbox. That
	  //        should also be done in the future for an IMAP mailbox. If it
	  //        DOES NOT exist, we default to INBOXFOLDERNAME. Currently, it's
	  //        done for an IMAP mailbox only if the connection has been established
	  //        with the IMAP server before this method is invoked.
	  //
	  aURLName = [[URLName alloc] initWithString: aDefaultURLNameAsString
				      path: [[NSUserDefaults standardUserDefaults] objectForKey: @"LOCALMAILDIR"]];
	  
	  aStore = [Utilities storeForURLName: aURLName];
	  
	  if ( !aStore || 
	       ![[MailboxManagerController singleInstance] folderNameExists: [aURLName foldername]
							   forStore: aStore] )
	    {
	      aDefaultURLNameAsString = [[NSUserDefaults standardUserDefaults] objectForKey: @"INBOXFOLDERNAME"];
	    }
	  
	  RELEASE(aURLName);
	}
      else
	{
	  aDefaultURLNameAsString = [[NSUserDefaults standardUserDefaults] objectForKey: @"INBOXFOLDERNAME"];
	}
	
      aDefaultURLName = [[URLName alloc] initWithString: aDefaultURLNameAsString
					 path: [[NSUserDefaults standardUserDefaults] objectForKey: @"LOCALMAILDIR"]];
	
    }
  else
    {  
      aDefaultURLNameAsString = [[NSUserDefaults standardUserDefaults] objectForKey: @"OUTBOXFOLDERNAME"];
      aDefaultURLName = [[URLName alloc] initWithString: aDefaultURLNameAsString
					 path: [[NSUserDefaults standardUserDefaults] objectForKey: @"LOCALMAILDIR"]];
    }
  
  AUTORELEASE(aDefaultURLName);

  // If we have found a filter that matches our message
  if ( aFilter && [aFilter type] == theType )
    {
      // We verify if the operation is TRANSFER_TO_FOLDER and if folder DIFFERENT from our default folder
      if ( [aFilter action] == TRANSFER_TO_FOLDER &&
	   ![[aFilter actionFolderName] isEqualToString: aDefaultURLNameAsString] )
	{
	  aURLName = [[URLName alloc] initWithString: [aFilter actionFolderName]
				      path: [[NSUserDefaults standardUserDefaults] objectForKey: @"LOCALMAILDIR"]];
	  AUTORELEASE(aURLName);
	}
      // We verify if the operation is DELETE. If so, we transfer the message to the trash folder.
      else if ( [aFilter action] == DELETE )
	{      
	  aURLName = [[URLName alloc] initWithString: [[NSUserDefaults standardUserDefaults] objectForKey: @"TRASHFOLDERNAME"]
				      path: [[NSUserDefaults standardUserDefaults] objectForKey: @"LOCALMAILDIR"]];
	  AUTORELEASE(aURLName);
	}
      // We have an UNKNOWN action. We simply add the message to our Inbox (our Outbox) folder.
      else
	{
	  aURLName = aDefaultURLName;
	}
    }
  // We haven't found a filter, let's add it to our Inbox Folder
  else
    {
      aURLName = aDefaultURLName;
    }

  return aURLName;
}


//
//
//
- (URLName *) matchedURLNameFromMessageAsRawSource: (NSData *) theRawSource
					      type: (int) theType
                                               key: (NSString *) theKey
{
  URLName *aURLName;
  NSRange aRange;
 
  aRange = [theRawSource rangeOfCString: "\n\n"];
  aURLName = nil;

  if ( aRange.length )
    {
      Message *aMessage;
      
      aMessage = [[Message alloc] initWithHeadersFromData: [theRawSource subdataToIndex: aRange.location + 1] ];
      [aMessage setRawSource: theRawSource];

      aURLName = [self matchedURLNameFromMessage: aMessage
		       type: theType
		       key: theKey];
      
      RELEASE(aMessage);
      
    } // if ( aRange.length )
  else
    {
      if ( theType == TYPE_INCOMING )
	{
	  NSDebugLog(@"FilterManager: Corrupted message received for filtering. Returning default Inbox.");
	  aURLName = [[NSUserDefaults standardUserDefaults] objectForKey: @"INBOXFOLDERNAME"];
	}
      else
	{
	  NSDebugLog(@"FilterManager: Corrupted message received for filtering. Returning default Outbox.");
	  aURLName = [[NSUserDefaults standardUserDefaults] objectForKey: @"OUTBOXFOLDERNAME"];
	}
    }
  
  return aURLName;
}


//
// class methods
//
+ (id) filtersFromDisk
{
  return [NSUnarchiver unarchiveObjectWithFile: PathToFilters()];
}

@end


//
// private methods
//
@implementation FilterManager (Private)

//
// This method returns a message with a retain count of 1 to not overload
// the main autorelease pool.
//
- (Message *) _messageFromExternalProgramUsingFilter: (Filter *) theFilter
					     message: (Message *) theMessage
{
  NSFileHandle *aFileHandle, *theInputFileHandle;
  NSTask *aTask;
  NSPipe *aPipe;
  
  NSData *theRAWSource, *aData;
  Message *aMessage;

  NSString *aString, *aFilename;
  NSRange aRange;
    
  // We first write the raw source to a temporary file
  theRAWSource = [theMessage rawSource];
  
  if ( !theRAWSource )
    {
      NSDebugLog(@"Unable to obtain the raw source from ...");
      return nil;
    }

  NSDebugLog(@"Using external program for processing mail.");

  aFilename = [NSString stringWithFormat:@"%@/%d_%@", NSTemporaryDirectory(), 
			[[NSProcessInfo processInfo] processIdentifier],
			NSUserName()];

  if ( ![theRAWSource writeToFile: aFilename
		      atomically: YES] )
    {
      NSDebugLog(@"Unable to write the raw source of the message to %@. Aborting.", aFilename);
      return nil;
    }
  
  theInputFileHandle = [NSFileHandle fileHandleForReadingAtPath: aFilename];
  
  aPipe = [NSPipe pipe];
  aFileHandle = [aPipe fileHandleForReading];
  
  aTask = [[NSTask alloc] init];
  [aTask setStandardOutput: aPipe];
  [aTask setStandardInput: theInputFileHandle];
  
  // We build our right string
  aString = [[theFilter externalProgramName] stringByTrimmingWhiteSpaces];
    
  // We verify if our program to lauch has any arguments
  aRange = [aString rangeOfString: @" "];

  if ( aRange.length )
    {
      [aTask setLaunchPath: [aString substringToIndex: aRange.location]];      
      [aTask setArguments: [[aString substringFromIndex: (aRange.location + 1)] 
			     componentsSeparatedByString: @" "]];
    }
  else
    {
      [aTask setLaunchPath: aString];
    }
  
  // We launch our task
  [aTask launch];
  
  // FIXME - This doesn't work with SpamAssassin. It won't notice
  //         that the task has ended. Perhaps this is due to the
  //         pipe getting filled to rapidly and being write protected
  //         after so spamassassin waits until we read some data from
  //         it, but we wait until spamassassin to exit to do that.
  //
  // while ( [aTask isRunning] )
  //  {
  //    [[NSRunLoop currentRunLoop] runMode: NSDefaultRunLoopMode
  //	               		    beforeDate: [NSDate distantFuture]];
  //  }
  
  aData = [aFileHandle readDataToEndOfFile];

  aMessage = [[Message alloc] initWithData: aData];
  
  RELEASE(aTask);
  [theInputFileHandle closeFile];

  [[NSFileManager defaultManager] removeFileAtPath: aFilename
				  handler: nil];
  
  NSDebugLog(@"Done using external program.");
  
  return aMessage;
}


//
//
//
- (BOOL) _matchCriteriasFromMessage: (Message *) theMessage
			     filter: (Filter *) theFilter
{
  FilterCriteria *aFilterCriteria;
  BOOL aBOOL;
  

  //
  // First criteria
  //
  aFilterCriteria = [[theFilter allCriterias] objectAtIndex: 0];
  aBOOL = [self _matchStrings: [self _stringsFromMessage: theMessage  criteria: aFilterCriteria]
		operation: [aFilterCriteria criteriaFindOperation]
		criteria: [aFilterCriteria criteriaString]];
  
  //
  // Second criteria
  //
  aFilterCriteria = [[theFilter allCriterias] objectAtIndex: 1];
  
  if ( [aFilterCriteria criteriaSource] != NONE )
    {
      if ( [aFilterCriteria criteriaCondition] == AND )
	{
	  aBOOL = aBOOL && [self _matchStrings: [self _stringsFromMessage: theMessage  criteria: aFilterCriteria]
				 operation: [aFilterCriteria criteriaFindOperation]
				 criteria: [aFilterCriteria criteriaString]];
	}
      else
	{
	  aBOOL = aBOOL || [self _matchStrings: [self _stringsFromMessage: theMessage  criteria: aFilterCriteria]
				 operation: [aFilterCriteria criteriaFindOperation]
				 criteria: [aFilterCriteria criteriaString]];
	}
    }

  //
  // Third and last criteria
  //
  aFilterCriteria = [[theFilter allCriterias] objectAtIndex: 2];
  
  if ( [aFilterCriteria criteriaSource] != NONE )
    {
      if ( [aFilterCriteria criteriaCondition] == AND )
	{
	  aBOOL = aBOOL && [self _matchStrings: [self _stringsFromMessage: theMessage  criteria: aFilterCriteria]
				 operation: [aFilterCriteria criteriaFindOperation]
				 criteria: [aFilterCriteria criteriaString]];
	}
      else
	{
	  aBOOL = aBOOL || [self _matchStrings: [self _stringsFromMessage: theMessage  criteria: aFilterCriteria]
				 operation: [aFilterCriteria criteriaFindOperation]
				 criteria: [aFilterCriteria criteriaString]];
	}
    }

  return aBOOL;
}


//
//
//
- (BOOL) _matchStrings: (NSArray *) theStrings
	     operation: (int) theOperation
	      criteria: (NSString *) theCriteria
{
  // Variables used in this method
  NSArray *anArray;
  NSRange aRange;
  int i;

  // We must be sure to have a valid criteria.
  if ( theOperation != IS_IN_ADDRESS_BOOK &&
       (!theCriteria || [theCriteria length] == 0) )
    {
      return NO;
    }
  
  for (i = 0; i < [theStrings count]; i++)
    {
      NSString *theString;

      theString = [theStrings objectAtIndex: i];
      
      switch ( theOperation )
	{
	case CONTAINS:
	  aRange = [theString rangeOfString: theCriteria
			      options: NSCaseInsensitiveSearch];
	  
	  if ( aRange.length )
	    {
	      return YES;
	    }
	  break;
	  
	case IS_EQUAL:
	  if ( [theString caseInsensitiveCompare: theCriteria] == NSOrderedSame )
	    {
	      return YES;
	    }
	  break;
	  
	case HAS_PREFIX:
	  if ( [[theString lowercaseString] hasPrefix: [theCriteria lowercaseString]] )
	    {
	      return YES;
	    }
	  break;
	  
	case HAS_SUFFIX:
	  if ( [[theString lowercaseString] hasSuffix: [theCriteria lowercaseString]] )
	    {
	      return YES;
	    }
	  break;
	  
	case MATCH_REGEXP:
	  anArray = [NSRegEx matchString: theString
			     withPattern : theCriteria
			     isCaseSensitive: YES];
	  
	  if ( [anArray count] > 0 )
	    {
	      return YES;
	    }
	  break;
	  
	case IS_IN_ADDRESS_BOOK:
	  anArray = [[[AddressBookController singleInstance] addressBook]
		      addressesWithSubstring: theString];
	  
	  if ( [anArray count] > 0 )
	    {
	      return YES;
	    }
	  break;
	  
	default:
	  break;
	}
    } // for (...)

  return NO;
}


//
//
//
- (NSArray *) _stringsFromMessage: (Message *) theMessage
			 criteria: (FilterCriteria *) theFilterCriteria
{
  NSMutableArray *aMutableArray;
  NSArray *allRecipients;
  NSString *aString;
  int i, theSource;

  aMutableArray = [[NSMutableArray alloc] init];

  theSource = [theFilterCriteria criteriaSource];
  
  switch ( theSource )
    {
    case TO:
      allRecipients = [theMessage recipients];
      
      for (i = 0; i < [allRecipients count]; i++)
	{
	  InternetAddress *anInternetAddress;
	  
	  anInternetAddress = [allRecipients objectAtIndex: i];
	  
	  if ( [anInternetAddress type] == TO )
	    {
	      aString = [anInternetAddress unicodeStringValue];
	      
	      if ( aString )
		{
		  [aMutableArray addObject: aString];
		}
	    }
	  
	}
      break;
      
    case CC:
      allRecipients = [theMessage recipients];
      
      for (i = 0; i < [allRecipients count]; i++)
	{
	  InternetAddress *anInternetAddress;
	  
	  anInternetAddress = [allRecipients objectAtIndex: i];
	  
	  if ( [anInternetAddress type] == CC )
	    {
	      aString = [anInternetAddress unicodeStringValue];
	      
	      if ( aString )
		{
		  [aMutableArray addObject: aString];
		}
	    }
	  
	}
      break;
      
    case TO_OR_CC:
      allRecipients = [theMessage recipients];
      
      for (i = 0; i < [allRecipients count]; i++)
	{
	  InternetAddress *anInternetAddress;
	  
	  anInternetAddress = [allRecipients objectAtIndex: i];
	  
	  if ( [anInternetAddress type] == TO ||
	       [anInternetAddress type] == CC )
	    {
	      aString = [anInternetAddress unicodeStringValue];
	      
	      if ( aString )
		{
		  [aMutableArray addObject: aString];
		}
	    }
	  
	}
      break;
      
    case SUBJECT:
      aString = [theMessage subject];
      
      if ( aString )
	{
	  [aMutableArray addObject: aString];
	}
      break;
      
    case FROM:
      aString = [[theMessage from] unicodeStringValue];
      
      if ( aString )
	{
	  [aMutableArray addObject: aString];
	    }
      break;
      
    case EXPERT:
      if ( [theFilterCriteria criteriaHeaders] && 
	   [[theFilterCriteria criteriaHeaders] count] > 0)
	{
	  for (i = 0; i < [[theFilterCriteria criteriaHeaders] count]; i++)
	    {
	      aString = [theMessage headerValueForName: [[theFilterCriteria criteriaHeaders] objectAtIndex: i]];
	      
	      if ( aString )
		{
		  [aMutableArray addObject: aString];
		}
	    }
	}
      else
	{
	  // We do nothing.. so we won't have any matches.
	}
      
      break;
      
      // No criteria source, we just ignore it.
    case NONE:
    default:
      break;
    }

  return AUTORELEASE(aMutableArray);
}

@end
