/*
**  MailHeaderCell.m
**
**  Copyright (c) 2002
**
**  Author: Nicolas Roard <nicolas@roard.com>
**          Ludovic Marcotte <ludovic@Sophos.ca>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "MailHeaderCell.h"

#import "GNUMail.h"
#import "GNUMailConstants.h"
#import "NSUserDefaultsExtensions.h"

@implementation MailHeaderCell

//
//
//
- (id) init 
{
  self = [super init];

  width = 0;
  
  color = [[NSUserDefaults standardUserDefaults] colorForKey: @"MAILHEADERCELL_COLOR"];

  if ( !color )
    {
      color = [NSColor colorWithCalibratedRed: 0.9 
		       green: 0.9 
		       blue: 1.0 
		       alpha: 1.0];
    }

  RETAIN(color);

  allViews = [[NSMutableArray alloc] init];

  return self;
}


//
//
//
- (void) dealloc
{
  RELEASE(originalAttributedString);
  RELEASE(color);

  RELEASE(allViews);

  [super dealloc];
}



//
//
//
- (float) height
{
  return [[self attributedStringValue] size].height + 20;
}


//
//
//
- (float) width
{
  return width;
}


//
//
//
- (void) setWidth: (float) theWidth
{
  width = theWidth;
}


//
//
//
- (NSSize) cellSize
{
  return NSMakeSize (width, [self height]);
}


//
// other methods
//
- (void) addView: (id) theView
{
  if ( theView )
    {
      [allViews addObject: theView];
    }
}


//
// FIXME - Optimize.
//
- (void) resize: (id) sender
{
  NSRect aRect;
  
  aRect = [[[[GNUMail lastMailWindowOnTop] windowController] textView] frame];
  
  //if ( !NSEqualSizes(aRect.size, NSMakeSize(width, aRect.size.height)) )
  //  {
  [self setWidth: aRect.size.width];
  [super setAttributedStringValue: [self _fold]];
  //  }
}


//
//
//
- (void) setAttributedStringValue: (NSAttributedString *) theAttributedString
{
  originalAttributedString = RETAIN(theAttributedString);
  [super setAttributedStringValue: theAttributedString];
}


//
//
//
- (void) drawWithFrame: (NSRect) theFrame
		inView: (NSView *) theView
{  
  NSImage *aImage;
  NSView *aView;
  NSSize aSize;
  
  float current_x, current_y;
  int i;
  
  if ( ![theView window] )
    {
      return;
    }
  
  // We fill our header cell
  theFrame.origin.x += 5;
#ifdef MACOSX
  theFrame.size.width -= 20;
#else
  theFrame.size.width -= 10;
#endif
  [color set];
  NSRectFill (theFrame);

  // We draw its border
#ifdef MACOSX
  // no adjustments required
#else
  theFrame.size.width -= 5;
#endif  
  [[NSColor blackColor] set];
  theFrame.origin.y += 1;
  theFrame.size.height -= 2;
  NSFrameRect(theFrame);

  current_x = theFrame.origin.x + theFrame.size.width;
  [theView lockFocus];

  for (i = 0; i < [allViews count]; i++)
    {
      aView = [allViews objectAtIndex: i];

      if ( ![aView respondsToSelector: @selector(image)] )
	{
	  continue;
	}

      aImage = [(id)aView image];
      
      if ( !aImage )
	{
	  continue;
	}

      aSize = [aImage size];

      current_x = current_x - aSize.width - 10;
      current_y = theFrame.origin.y + aSize.height + (theFrame.size.height - aSize.height)/2;

      [aImage compositeToPoint: NSMakePoint(current_x, current_y)
	      operation: NSCompositeSourceAtop];
    }

  // We adjust our frame and we draw our attributed string
  theFrame.origin.x += 8; 
  theFrame.size.width -= 16;
  theFrame.origin.y += 10;
  theFrame.size.height -= 20;

  [[self attributedStringValue] drawInRect: theFrame];
  [theView unlockFocus];
}


@end



//
// Private interface
//
@implementation MailHeaderCell (Private)

- (NSAttributedString *) _fold
{
  NSMutableAttributedString *aMutableAttributedString;
  NSScanner *aScanner;

  int previousLocation, currentLocation;
  float currentLineWidth;
  
  if ( !originalAttributedString )
    {
      return nil;
    }

  aMutableAttributedString = [[NSMutableAttributedString alloc] init];

  aScanner = [NSScanner scannerWithString: [originalAttributedString string]];
  
  currentLocation = previousLocation = currentLineWidth = 0;

  // We tokenize our headers...
  while ( [aScanner scanUpToCharactersFromSet: [NSCharacterSet whitespaceAndNewlineCharacterSet]
		    intoString: NULL] )
    {
      NSString *aString;
      
      NSRange aRange;
      float wordWidth;
 
      currentLocation = [aScanner scanLocation];
      wordWidth = 0;

      aRange =  NSMakeRange(previousLocation, currentLocation - previousLocation);
      aString = [[aScanner string] substringWithRange: aRange];
 
      
      if ( [[aScanner string] length] > NSMaxRange(aRange) &&
	   [[aScanner string] characterAtIndex: NSMaxRange(aRange)] == '\n' &&
	   (currentLineWidth + wordWidth) <= ([self width] - 100) )
	{
	  [aMutableAttributedString appendAttributedString:
				      [originalAttributedString attributedSubstringFromRange: aRange]];
	  currentLineWidth = 0;
	  previousLocation = currentLocation;
	  continue;
	}

      wordWidth = [[originalAttributedString attributedSubstringFromRange: aRange] size].width;

      if ( (currentLineWidth + wordWidth) > ([self width] - 100) )
	{
	  [aMutableAttributedString appendAttributedString: 
	  			      AUTORELEASE([[NSAttributedString alloc] initWithString: @"\n"])];
	  [aMutableAttributedString appendAttributedString:
				      [originalAttributedString attributedSubstringFromRange: aRange]];
	  
	  currentLineWidth = wordWidth;
	}
      else
	{
	  [aMutableAttributedString appendAttributedString:
				      [originalAttributedString attributedSubstringFromRange: aRange]];
	  currentLineWidth += wordWidth;
	}
      
      previousLocation = currentLocation;
    }


  return AUTORELEASE(aMutableAttributedString);
}

@end
