/*
**  MailWindowController.m
**
**  Copyright (c) 2001, 2002
**
**  Author: Ludovic Marcotte <ludovic@Sophos.ca>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#import "MailWindowController.h"

#import "EditWindowController.h"
#import "ExtendedCell.h"
#import "GNUMail.h"
#import "GNUMail/GNUMailBundle.h"
#import "GNUMailConstants.h"
#import "LabelWidget.h"

#ifndef MACOSX
#import "MailWindow.h"
#endif

#import "Filter.h"
#import "FilterManager.h"
#import "LocalMailDelivery.h"
#import "MailboxManagerCache.h"
#import "MailboxManagerController.h"
#import "MailHeaderCell.h"
#import "MessageViewWindowController.h"
#import "MimeType.h"
#import "MimeTypeManager.h"
#import "NSRegEx.h"
#import "NSStringExtensions.h"
#import "NSUserDefaultsExtensions.h"
#import "Utilities.h"

#import <Pantomime/Folder.h>
#import <Pantomime/IMAPFolder.h>
#import <Pantomime/IMAPStore.h>
#import <Pantomime/LocalFolderCacheManager.h>
#import <Pantomime/LocalFolder.h>
#import <Pantomime/LocalStore.h>
#import <Pantomime/Message.h>
#import <Pantomime/POP3Folder.h>
#import <Pantomime/POP3Store.h>
#import <Pantomime/NSDataExtensions.h>
#import <Pantomime/POP3CacheManager.h>
#import <Pantomime/URLName.h>

#define LONG_DELAY 86400.0


@implementation MailWindowController

//
//
//
- (id) initWithWindowNibName: (NSString *) windowNibName
{
#ifdef MACOSX
  
  self = [super initWithWindowNibName: windowNibName];
  
#else
  MailWindow *mailWindow;
  
  mailWindow = [[MailWindow alloc] initWithContentRect: NSMakeRect(150,100,750,595)
				   styleMask: NSClosableWindowMask|NSTitledWindowMask|
				   NSMiniaturizableWindowMask|NSResizableWindowMask
				   backing: NSBackingStoreRetained
				   defer: NO];

  self = [super initWithWindow: mailWindow];
  
  [mailWindow layoutWindow];
  [mailWindow setDelegate: self];

  // We link our outlets
  tableScrollView = [mailWindow tableScrollView];
  textScrollView = [mailWindow textScrollView];

  splitView = [mailWindow splitView];
  tableView = [mailWindow tableView];
  textView = [mailWindow textView];
  
  icon = [mailWindow icon];
  next = [mailWindow next];
  previous = [mailWindow previous];
  delete = [mailWindow delete];
  mailboxes = [mailWindow mailboxes];
  compose = [mailWindow compose];
  forward = [mailWindow forward];
  reply = [mailWindow reply];
  addresses = [mailWindow reply];
  find = [mailWindow find];
  get = [mailWindow get];
  secureConnectionIcon = [mailWindow secureConnectionIcon];

  label = (NSTextField *)[mailWindow label];
  matrix = [mailWindow matrix];

  statusColumn = [mailWindow statusColumn];
  idColumn = [mailWindow idColumn];
  dateColumn = [mailWindow dateColumn];
  fromColumn = [mailWindow fromColumn];
  subjectColumn = [mailWindow subjectColumn];
  sizeColumn = [mailWindow sizeColumn];

  RELEASE(mailWindow);
#endif

  // We set our window title
  [[self window] setTitle: @""];

  // We must retain our columns under OS X after the outlets are linked (we release them in -dealloc)
#ifdef MACOSX
  RETAIN(label);
  RETAIN(statusColumn);
  RETAIN(idColumn);
  RETAIN(dateColumn);
  RETAIN(fromColumn);
  RETAIN(subjectColumn);
  RETAIN(sizeColumn);
  
  [matrix setCellSize: NSMakeSize(20,20)];
  [matrix removeColumn: 0];  
  [matrix setPrototype: [[[NSImageCell alloc] init] autorelease]];
  [matrix addColumn];
#endif

  // We create our mail header cell
  mailHeaderCell = [[MailHeaderCell alloc] init];
  
  [[NSNotificationCenter defaultCenter] addObserver: mailHeaderCell
					selector: @selector (resize:) 
					name: @"NSViewFrameDidChangeNotification" 
					object: textView];


  // We register the table view for dragged types
  [tableView registerForDraggedTypes: [NSArray arrayWithObject: MessagePboardType]];

  // We set any vertical mouse motion has being dragging
  [tableView setVerticalMotionCanBeginDrag: NO];

  // We set our custom cell
  [statusColumn setDataCell: AUTORELEASE([[ExtendedCell alloc] init])];

  // We load our accessory views
  [self _loadAccessoryViews];

  // We restore our sorting order
  [self _restoreSortingOrder];

  // We restore our split view knob position
  [self _restoreSplitViewSize];

  // We load the right set of columns
  [self reloadTableColumns];

  // We set our autosave window frame name and restore the one from the user's defaults.
  [[self window] setFrameAutosaveName: @"MailWindow"];
  [[self window] setFrameUsingName: @"MailWindow"];

  // We set our autosave name for our table view
  [tableView setAutosaveName: @"MailWindow"];
  [tableView setAutosaveTableColumns: YES];

  // We set our table view background color
  if ( [[NSUserDefaults standardUserDefaults] colorForKey: @"MAILWINDOW_TABLE_COLOR"] )
    {
      [tableView setBackgroundColor: [[NSUserDefaults standardUserDefaults]
				       colorForKey: @"MAILWINDOW_TABLE_COLOR"]];
      [tableScrollView setBackgroundColor: [[NSUserDefaults standardUserDefaults]
					     colorForKey: @"MAILWINDOW_TABLE_COLOR"]];
    }
  
  // We finally create our lock
  lock = [[NSLock alloc] init];

  return self;
}


//
//
//
#ifdef MACOSX
- (void) awakeFromNib
{
  NSToolbar *toolbar;
  
  toolbar = [[NSToolbar alloc] initWithIdentifier: @"MailWindowToolbar"];
  [toolbar setDelegate: self];
  [toolbar setAllowsUserCustomization: YES];
  [toolbar setAutosavesConfiguration: YES];
  [[self window] setToolbar: toolbar];
  RELEASE(toolbar);
}
#endif


//
//
//
- (void) dealloc
{
  NSDebugLog(@"MailWindowController: -dealloc");
  
  // We remove our timers
  [[NSNotificationCenter defaultCenter]
    removeObserver: self
    name: ShouldStartLocalInboxTimer
    object: nil];
  
  [[NSNotificationCenter defaultCenter]
    removeObserver: self
    name: ShouldStopLocalInboxTimer
    object: nil];
  
  // We now release...
  RELEASE(allMessageViewWindowControllers);
  RELEASE(filteredMessageBoxes);
  RELEASE(allConnections);

  TEST_RELEASE(timer);

  // We release our lock and our folder
  RELEASE(lock);
  RELEASE(folder);

  [[NSNotificationCenter defaultCenter] removeObserver: mailHeaderCell
					name: @"NSViewFrameDidChangeNotification" 
					object: textView];

  RELEASE(mailHeaderCell);

  TEST_RELEASE(filterManager);
  TEST_RELEASE(allMessages);
  
  TEST_RELEASE(currentOrder);
  TEST_RELEASE(previousOrder);

  // We release our retained columns under OS X
#ifdef MACOSX
  //[[self window] setToolbar: nil];
  RELEASE(label);
  RETAIN(statusColumn);
  RETAIN(idColumn);
  RETAIN(dateColumn);
  RETAIN(fromColumn);
  RETAIN(subjectColumn);
  RETAIN(sizeColumn);
#endif

  [[NSNotificationCenter defaultCenter]
    removeObserver: self
    name: FiltersHaveChanged
    object: nil];

  [[NSNotificationCenter defaultCenter]
    removeObserver: self
    name: ReplyToMessageWasSuccessful
    object: nil];

  [super dealloc];
}


//
// action methods
//
- (IBAction) doubleClickedOnTableView: (id) sender
{
  // If we are in the Draft folder, we re-opened the selected mail for editing
  if ( [Utilities URLWithString: [[NSUserDefaults standardUserDefaults] objectForKey: @"DRAFTSFOLDERNAME"]
		  matchFolder: [self folder]] )
    {
      [[NSApp delegate] restoreDraft: nil];
    }
  // Or.. we just 'reply' to the mail...
  else
    {
      if ( [[NSUserDefaults standardUserDefaults] objectForKey: @"DOUBLECLICKACTION"] && 
	   [[NSUserDefaults standardUserDefaults] integerForKey: @"DOUBLECLICKACTION"] == 0 )
      	{
	  [self replyMessage: nil];
	}
      else
      	{
	  [self viewMessageInWindow: nil];
	}
    }
}


//
//
//
- (IBAction) deleteMessage: (id) sender
{
  // If we have no element (or no selection), we return!
  if ([folder count] == 0 || [tableView numberOfSelectedRows] == 0)
    {
      NSBeep();
      return;
    } 
  else
    {
      NSEnumerator *anEnumerator;
      NSNumber *aRow;
      Message *theMessage;
      Flags *theFlags;

      int last_row, nb;
      BOOL firstFlagOfList;
      
      anEnumerator = [tableView selectedRowEnumerator];
      last_row = nb = 0;
      firstFlagOfList = NO;

      while ( (aRow = [anEnumerator nextObject]) )
	{
	  theMessage = [allMessages objectAtIndex: [aRow intValue]];
	  
	  // We set the flag Deleted (or not) to the message
	  theFlags = [theMessage flags];

          if ( nb == 0)
            {
              // This is the first message of the list we want to {un}delete
              // We must save the flag.
              if ( [theFlags contain: DELETED] )
                {
                  [theFlags remove: DELETED];
                  firstFlagOfList = NO;
                }
              else
                {
                  [theFlags add: DELETED];
                  firstFlagOfList = YES;
                }

	      [(GNUMail*)[NSApp delegate] updateMenuItemsForMessage: theMessage];
            }
          else
            {
              if ( (! firstFlagOfList) && [theFlags contain: DELETED] )
                { 
                  [theFlags remove: DELETED];
                }
              else if ( firstFlagOfList && (! [theFlags contain: DELETED]) )
                {
                  [theFlags add: DELETED];
                }
            }
	  
	  last_row = [aRow intValue];
          nb++;
	}
      
      // We now select the row right after the message(s) beeing deleted
      // If there is only one selected message
      // We do this operation IIF sender is the button 'delete'
      if (sender == delete)
	{
	  int count, old_selection;
	  
	  count = [[[self folder] allMessages] count];
	  old_selection = NSNotFound;

	  // If we show the deleted mails, we 'jump' to the next mail
	  // Else, we just try to show the same index again.
	  if ( [[self folder] showDeleted] )
	    {
	      last_row = last_row + 1;
	    }
	  else
	    {
	      // If we were 'deleting' the last row, let's shift
	      // our index by one.
	      if ( last_row >= count )
		{
		  last_row = (count - 1);
		}

	      old_selection = [tableView selectedRow];
	    }
	  
	  // We verify that we aren't off bounds
	  if ( count > last_row 
	       && (nb == 1) )
	    {
	      if (old_selection == last_row)
		{
		  [self tableViewSelectionDidChange: nil];
		}
	      else
		{
		  [tableView selectRow: last_row 
			     byExtendingSelection: NO];
		  [tableView scrollRowToVisible: last_row];
		}
	    }
	  else
	    {
	      // We at least post our notification
	      [[NSNotificationCenter defaultCenter]
		postNotificationName: SelectionOfMessageHasChanged
		object: nil
		userInfo: nil];
	    }
	}
      
      // We always refresh our tableView after a delete operation
      if ( [[self folder] showDeleted] )
	{
	  NSDebugLog(@"Showing deleted messages...");
	  [tableView setNeedsDisplay: YES];  
	}
      else
	{
	  NSDebugLog(@"NOT Showing deleted messages...");
	  [[self folder] updateCache];
	  [self tableViewShouldReloadData];
	}

      // We update the status label
      [self updateStatusLabel];
    }
}


//
// This method selects the message after the current
// selected message and displays it.
//
- (IBAction) nextMessage: (id) sender
{
  int i = [tableView selectedRow];

  if (i < [folder count] - 1)
    {
      [tableView selectRow: (i+1) byExtendingSelection: NO];
      [tableView scrollRowToVisible: (i+1)];
      [tableView setNeedsDisplay: YES];
    }
  else
    {
      NSBeep();
    }
}


//
//
//
- (IBAction) firstMessage: (id) sender
{
  if ([folder count] > 0)
    {
      [tableView selectRow: 0 byExtendingSelection: NO];
      [tableView scrollRowToVisible: 0];
      [tableView setNeedsDisplay: YES];
    }
}


//
//
//
- (IBAction) lastMessage: (id) sender
{
  if ([folder count] > 0)
    {
      [tableView selectRow: ([folder count] - 1) byExtendingSelection: NO];
      [tableView scrollRowToVisible: ([folder count] - 1)];
      [tableView setNeedsDisplay: YES];
    }
}


//
// This method selects the message before the current
// selected message and displays it.
//
- (IBAction) previousMessage: (id) sender
{
  int i = [tableView selectedRow];
  
  if (i > 0)
    {
      [tableView selectRow: (i-1) byExtendingSelection: NO];
      [tableView scrollRowToVisible: (i-1)];
      [tableView setNeedsDisplay: YES];
    }
  else
    {
      NSBeep();
    }
}


//
//
//
- (IBAction) composeMessage: (id) sender
{
  EditWindowController *controller;
  
  controller = [[EditWindowController alloc] initWithWindowNibName: @"EditWindow"];
  
  if ( controller )
    {
      Message *aMessage;

      // We create an empty message
      aMessage = [[Message alloc] init];

      [[controller window] setTitle: _(@"New message...")];
      [controller setMessage: aMessage];
      [controller setShowCc: NO];
      
      RELEASE(aMessage);
      
      [controller showWindow: self];
    }
}


//
//
//
- (IBAction) forwardMessage: (id) sender
{
  Message *aMessage;
  
  if ( [tableView selectedRow] < 0 ) 
    {
      NSBeep();
      return;
    }
  
  // We obtain the selected entry in the table for the forward information
  aMessage = [allMessages objectAtIndex: [tableView selectedRow]];
  
  // Forward this message
  [Utilities forwardMessage: aMessage];
}


//
//
//
- (IBAction) replyMessage: (id) sender
{
  Message *aMessage;
  
  if ( [tableView selectedRow] < 0 ) 
    {
      NSBeep();
      return;
    }
  
  // We obtain the selected entry in the table for the reply informations
  aMessage = [allMessages objectAtIndex: [tableView selectedRow]];
  
  // Invoke the message utility class method to reply to this message.
  [Utilities replyToSender: aMessage
	     folder: [self folder]];  
}


//
//
//
- (IBAction) replyAllMessage: (id) sender
{
  Message *aMessage;
  
  if ( [tableView selectedRow] < 0 ) 
    {
      NSBeep();
      return;
    }
  
  // We obtain the selected entry in the table for the reply informations
  aMessage = [allMessages objectAtIndex: [tableView selectedRow]];
  
  // Invoke the message utility class method to reply to this message.
  [Utilities replyToAll: aMessage
	     folder: [self folder]];
}


//
// This opens a new window and displays the message in it.
//
- (IBAction) viewMessageInWindow: (id) sender
{
  MessageViewWindowController *aViewWindowController;
  Message *aMessage;
  
  if ( [tableView selectedRow] < 0 )
    {
      NSBeep();
      return;
    }
  
  // We obtain the selected entry in the table for the display informations
  aMessage = [allMessages objectAtIndex: [tableView selectedRow]];
  
  // We create our window controller
  aViewWindowController = [[MessageViewWindowController alloc] initWithWindowNibName: @"MessageViewWindow"];
  
  // set our message and folder
  [aViewWindowController setMessage: aMessage];
  [aViewWindowController setFolder: [self folder]];
  [aViewWindowController setMailWindowController: self];
  [aViewWindowController setListIndex: [tableView selectedRow]];
  
  
  // show the window and the message
  [aViewWindowController showWindow: self];
  [allMessageViewWindowControllers addObject: aViewWindowController];

  [Utilities showMessage: aMessage
	     target: [aViewWindowController textView]
	     showAllHeaders: NO];

  // On MacOS X, if the mail window is not active, double-clicking on an unselected message
  // causes the message not to draw itself in the new window that pops up.
  // Let's force it to draw itself.
#ifdef MACOSX
  [[aViewWindowController textView] setNeedsDisplay: YES];
#endif

}


//
// This method returns the folder associated to this MailWindow.
//
- (Folder *) folder
{
  return folder;
}


//
// This method sets the folder associated to this MailWindow.
//
// NOTE: This method DOES NOT close the folder. It'll release it
//       but the folder SHOULD BE CLOSED FIRST.
//
- (void) setFolder: (Folder *) theFolder
{
  RETAIN(theFolder);
  RELEASE(folder);
  folder = theFolder;

  // We close all MessageViewWindows
  [self _closeAllMessageViewWindows];

  // We now set the window title
  if ( [folder isKindOfClass: [LocalFolder class]] )
    {
      [[self window] setTitle: [NSString stringWithFormat: _(@"Local - %@"), [folder name]] ];
    }
  else
    {
      IMAPStore *aStore;
      
      aStore = (IMAPStore *)[folder store];
      
      // We set the delegate on our folder
      [(IMAPFolder *)folder setDelegate: self];

      [[self window] setTitle: [NSString stringWithFormat: _(@"IMAP on %@ - %@"), [aStore name], 
					 [folder name]] ];
    }
  
  if ( [folder count] == 0 )
    {
      NSRunInformationalAlertPanel(_(@"No Messages!"),
				   _(@"There are no messages in this Mailbox!"),
				   _(@"OK"),
				   NULL,
				   NULL); 
    }
  else
    {
      int i, count;

      // We reload the data our of table (since it now has some)
      [self tableViewShouldReloadData];
      
      count = [folder count];
      
      // We search for the first message with the RECENT flag and we select it
      for (i = 0; i < count; i++)
	{
	  Message *aMessage;

	  aMessage = [allMessages objectAtIndex: i];

	  if ( [[aMessage flags] contain: RECENT] )
	    {
	      break;
	    }
	}

      if (i == count)
	{
	  if ( isReverseOrder )
	    {
	      i = 0;
	    }
	  else
	    {
	      i--;
	    }
	}

      // We scroll to and select our found row
      [tableView scrollRowToVisible: i];
      [tableView selectRow: i 
		 byExtendingSelection: NO];
    }

  // We always update our status label, if we have or not messages in our folder
  [self updateStatusLabel];
}



//
// delegate methods
//
- (int) numberOfRowsInTableView: (NSTableView *)aTableView
{
  return [folder count];
}


//
//
//
- (id)           tableView: (NSTableView *) aTableView
 objectValueForTableColumn: (NSTableColumn *) aTableColumn
                       row: (int) rowIndex
{
  InternetAddress *anInternetAddress;
  Message *aMessage;

  aMessage = [allMessages objectAtIndex: rowIndex];

  // We then provide the real data for our row
  if (aTableColumn == statusColumn)
    {
      return @"";
    }
  else if (aTableColumn == idColumn)
    {
      return [NSString stringWithFormat:@"%d", [aMessage messageNumber]];
    }
  else if (aTableColumn == dateColumn)
    {   
      NSUserDefaults *aUserDefaults; 
      NSString *aString;
      int day, today;

      aUserDefaults = [NSUserDefaults standardUserDefaults];      
      day = [[aMessage receivedDate] dayOfCommonEra];
      today = [[NSCalendarDate date] dayOfCommonEra];
      
      if ( day == today )
	{
	  aString = [aUserDefaults objectForKey: NSTimeFormatString];
	}
      else if ( day == today-1 )
	{
	  aString = [NSString stringWithFormat: @"%@ %@",
			      [[aUserDefaults objectForKey: NSPriorDayDesignations] objectAtIndex: 0],
			      [aUserDefaults objectForKey: NSTimeFormatString]];
	}
      else
	{ 
	  aString = [aUserDefaults objectForKey: NSShortDateFormatString];
	}
      
      if ( !aString )
	{
	  aString = @"%b %d %Y";
	}
      
      return [[aMessage receivedDate] descriptionWithCalendarFormat: aString
				      timeZone: [(NSCalendarDate*)[aMessage receivedDate] timeZone]
				      locale: nil];
    }
  else if (aTableColumn == fromColumn)
    {
      anInternetAddress = [aMessage from];
    
      if ([anInternetAddress personal] == nil || [[anInternetAddress personal] length] == 0)
	{
	  return [anInternetAddress address];
	}
      else
	{
	  return [anInternetAddress personal];
	}
    }
  else if (aTableColumn == subjectColumn)
    {
      return [aMessage subject];
    }
  else if (aTableColumn == sizeColumn) 
    {
      return [NSString stringWithFormat: @"%.1fKB ", ((float)[aMessage size]/(float)1024)];
    }
  else 
    {
      return _(@"< unknown >");
    }

  // Never reached.
  return nil;
}


//
//
//
- (void) tableView: (NSTableView *)aTableView
   willDisplayCell: (id)aCell
    forTableColumn: (NSTableColumn *)aTableColumn
               row: (int)rowIndex
{
  // For our filters
  NSColor *aColor;
  
  // For the rest
  Message *aMessage;
  Flags *theFlags;

  aMessage = [allMessages objectAtIndex: rowIndex];
  theFlags = [aMessage flags]; 

  // We verify for a filter
  aColor = [filterManager colorForMessage: aMessage];
 
  // We if have a special color coming from our filter, we set it for this cell
  if ( aColor )
    {
      [aCell setDrawsBackground: YES];
      [aCell setBackgroundColor: aColor];
    }
  else
    {
      [aCell setDrawsBackground: NO];
    }

  // If it's a new message, we set the cell's text to bold
  if ( [theFlags contain: SEEN] )
    {
#ifdef MACOSX
      [aCell setFont: [NSFont systemFontOfSize: [NSFont smallSystemFontSize]]];
#else
      [aCell setFont: [NSFont systemFontOfSize: 0]];
#endif
    }
  else
    {
#ifdef MACOSX
      [aCell setFont: [NSFont boldSystemFontOfSize: [NSFont smallSystemFontSize]]];
#else
      [aCell setFont: [NSFont boldSystemFontOfSize: 0]];
#endif
    }

  // If it's a deleted message, we set the cell's text to italic
  if ( [theFlags contain: DELETED] )
    {
      [aCell setTextColor: [NSColor darkGrayColor]];
#ifdef MACOSX
      [aCell setFont: [[NSFontManager sharedFontManager] convertFont: [NSFont systemFontOfSize:
										[NSFont smallSystemFontSize]]
							 toHaveTrait: NSItalicFontMask]];
#else
      [aCell setFont: [[NSFontManager sharedFontManager] convertFont: [NSFont systemFontOfSize: 0]
							 toHaveTrait: NSItalicFontMask]];
#endif
    }
  else
    {
      [aCell setTextColor: [NSColor blackColor]];
    }

  // We set the right aligment for our last (ie., Size) column.
  if ( aTableColumn == sizeColumn )
    {
      [aCell setAlignment: NSRightTextAlignment];
    }
  else
    {
      [aCell setAlignment: NSLeftTextAlignment];
    }

  // We set the image of our status cell
  if ( aTableColumn == statusColumn )
    {
      ExtendedCell *cell;
      
      cell = (ExtendedCell *)[statusColumn dataCell];
      [cell setFlags: [[allMessages objectAtIndex: rowIndex] flags]];
      [cell setIsMimeMessage: ([aMessage mimeVersion] == nil ? NO : YES)];
    }
}


//
//
//
- (void) tableViewSelectionDidChange: (NSNotification *) aNotification
{
  if (isReloading)
    {
      return;
    }
  
  if ( [tableView numberOfSelectedRows] > 1 )
    {
      [textView setString: @""];
    }
  else 
    {
      // If we have no selection, we simply return.
      if ([tableView selectedRow] < 0) 
	{
	  return;
	}
      else
	{
	  // We obtain our selected message
	  Message *aMessage;
	  
	  aMessage = [allMessages objectAtIndex: [tableView selectedRow]];
	  
	  // We show our message!
	  [Utilities showMessage: aMessage
		     target: [self textView]
		     showAllHeaders: NO];
	  [matrix setNeedsDisplay: YES];
	}
    }

  [self updateStatusLabel];

  [[NSNotificationCenter defaultCenter]
    postNotificationName: SelectionOfMessageHasChanged
    object: nil
    userInfo: nil];

}


//
//
//
- (void)   tableView: (NSTableView *)aTableView
 didClickTableColumn: (NSTableColumn *)aTableColumn
{
  NSString *newOrder;

  newOrder = [aTableColumn identifier];
  
  if ( ([newOrder isEqualToString: @"Id"] == NO)
       && ([newOrder isEqualToString: @"Date"] == NO)
       && ([newOrder isEqualToString: @"From"] == NO)
       && ([newOrder isEqualToString: @"Subject"] == NO)
       && ([newOrder isEqualToString: @"Size"] == NO) )
    {
      return;
    }
  
  [aTableView setHighlightedTableColumn: aTableColumn];
  [self setPreviousOrder: [self currentOrder]];

  if ([[self currentOrder] isEqualToString: newOrder])
    {
      if (isReverseOrder == YES)
	{
	  NSDebugLog(@"non reverse");
	}
      else
	{
	  NSDebugLog(@"reverse");
	}
      
      isReverseOrder = !isReverseOrder;
    }
  else
    {
      [self setCurrentOrder: newOrder];
      isReverseOrder = NO;
    }
  
  [[NSUserDefaults standardUserDefaults] setObject: [self currentOrder]
					 forKey: @"SORTINGORDER"];

  [[NSUserDefaults standardUserDefaults] setInteger: isReverseOrder
					 forKey: @"SORTINGSTATE"];

  [self tableViewShouldReloadData];
}


//
//
//
- (void)              tableView: (NSTableView *)aTableView
 mouseDownInHeaderOfTableColumn: (NSTableColumn *)aTableColumn
{
  NSDebugLog(@"mouse down %@", [aTableColumn identifier]);
}


//
//
//
-  (void) textView: (NSTextView *) aTextView
     clickedOnCell: (id <NSTextAttachmentCell>) attachmentCell
	    inRect: (NSRect) cellFrame
	   atIndex: (unsigned) charIndex
  
{
  [Utilities clickedOnCell: attachmentCell
	     inRect: cellFrame
	     atIndex: charIndex];
}


//
//
//
- (BOOL) textView: (NSTextView *) textView
    clickedOnLink: (id) link 
	  atIndex: (unsigned) charIndex
{
  NSDebugLog(@"Opening %@...", [link description]);
  return [[NSWorkspace sharedWorkspace] openURL: link];
}


//
// NSTableDataSource Drag and drop
//
- (BOOL) tableView: (NSTableView *) aTableView
	 writeRows: (NSArray *) rows
      toPasteboard: (NSPasteboard *) pboard
{
  NSMutableArray *propertyList;
  int i, row;
  Message *aMessage;
  
  propertyList = [[NSMutableArray alloc] initWithCapacity: [rows count]];
  
  for (i = 0; i < [rows count]; i++)
    {
      // For each implicated row, retrieve message and insert its
      // raw representation in a dictionary with two keys: 'flags' and 'message'
      
      NSMutableDictionary *aDictionary = [[NSMutableDictionary alloc] initWithCapacity: 2];
      NSData *flags, *rawSource;
      
      row = [[rows objectAtIndex: i] intValue];
      aMessage = [allMessages objectAtIndex: row];
      
      flags = [NSArchiver archivedDataWithRootObject: [aMessage flags]];
      [aDictionary setObject: flags forKey: @"flags"];

      rawSource = [NSData dataWithData: [aMessage rawSource]];
      [aDictionary setObject: rawSource forKey: @"message"];

      [propertyList addObject: aDictionary];
      RELEASE(aDictionary);
    }

  // Set property list of paste board
  [pboard declareTypes: [NSArray arrayWithObject: MessagePboardType] owner: self];
  [pboard setPropertyList: propertyList forType: MessagePboardType];
  RELEASE(propertyList);
  
  return YES;
}


//
// NSTableDataSource Drag and drop
//
- (NSDragOperation) tableView: (NSTableView *) aTableView
		 validateDrop: (id <NSDraggingInfo>) info
		  proposedRow: (int) row
	proposedDropOperation: (NSTableViewDropOperation) operation

{
  if ([info draggingSource] == tableView)
    {
      // We don't allow drag'n'drop to the same tableView
      return NSDragOperationNone;
    }

  if ([info draggingSourceOperationMask] & NSDragOperationGeneric)
    {
      return NSDragOperationGeneric;
    }
  else if ([info draggingSourceOperationMask] & NSDragOperationCopy)
    {
      return NSDragOperationCopy;
    }
  else
    {
      return NSDragOperationNone;
    }
}


//
// NSTableDataSource Drag and drop
//
- (BOOL) tableView: (NSTableView *) aTableView
	acceptDrop: (id <NSDraggingInfo>) info
	       row: (int) row
     dropOperation: (NSTableViewDropOperation) operation
{
  NSArray *propertyList;
  id sourceDelegate;
  int i, count;
  
  if ([info draggingSource] == tableView)
    {
      // We don't allow drag'n'drop to the same tableView
      return NO;
    }
  
  // We retrieve property list of messages from paste board
  propertyList = [[info draggingPasteboard] propertyListForType: MessagePboardType];

  if ( !propertyList )
  {
      return NO;
  }

  count = [propertyList count];
  
  for (i = 0; i < count; i++)
    {
      // We retrieve message from property list
      NSDictionary *aDictionary;
      Flags *flags;
      NSData *rawSource;
      
      aDictionary = (NSDictionary*)[propertyList objectAtIndex: i];
      
      rawSource = [aDictionary objectForKey: @"message"];
      
      flags = (Flags*)[NSUnarchiver unarchiveObjectWithData: (NSData*)[aDictionary objectForKey: @"flags"]];
      
      if (rawSource && flags)
	{
	  if ( ![[MailboxManagerController singleInstance] transferMessageFromRawSource: rawSource
							   flags: flags
							   toFolderWithName: [[self folder] name]
							   orToFolder: [self folder]
							   store: [[self folder] store]] )
	    {
	      // The transfer failed; we exit the loop
	      break;
	    }
	}
    }
  
  if (count > 0)
    {
      // We refresh the table view
      [self tableViewShouldReloadData];
      
      // We update the status label
      [self updateStatusLabel];
  
      // If the message is moved (and not copied),
      // we inform the dragging source that the selected messages have been transferred
      
      sourceDelegate = [[info draggingSource] delegate];
      
      // We mark messages as transferred if append succeed
      if (count == i &&
	  [info draggingSourceOperationMask] & NSDragOperationGeneric && 
	  [sourceDelegate respondsToSelector: @selector(setSelectedMessagesAsTransferred)])
	{
	  [sourceDelegate setSelectedMessagesAsTransferred];
	}
    }
  
  return YES;
}


//
//
//
- (BOOL) windowShouldClose: (id) sender
{
  int choice;

  // We first verify if the animation timer is running, if it is,
  // we are currently sending a mail so we inform the user about this.
  if ( animation && [animation isValid] )
    {
      NSRunInformationalAlertPanel(_(@"Closing..."),
				   _(@"GNUMail.app is currently receiving E-Mails. Please wait."),
				   _(@"OK"), 
				   NULL, 
				   NULL);
      return NO;
    }

  // If it is an IMAP folder, we offer the user the option to close the connection
  // to the IMAP store
  if ( [[self folder] isKindOfClass: [IMAPFolder class]] )
    {
      choice = NSRunAlertPanel(_(@"Closing..."),
			       _(@"Would you like to close the connection with the IMAP server?"),
			       _(@"Cancel"), // default
			       _(@"Yes"),    // alternate
			       _(@"No"));    // other return
      
      // We don't want to close the window
      if ( choice == NSAlertDefaultReturn )
	{
	  return NO;
	}
      // Yes we want to close it, and we also want to close the connection with the IMAP Store
      else if ( choice == NSAlertAlternateReturn )
	{
	  shouldCloseStore = YES;
	  return YES;
	}
      
    }
  
  shouldCloseStore = NO;
      
  return YES;
}


//
//
//
- (void) windowWillClose: (NSNotification *) not
{
  int i;
 
  // We stop our fetching timer
  [self stopTimer: nil];

  // We wait until all threads died before closing the window
  while ( threadCount > 0 )
    {
      [[NSRunLoop currentRunLoop] runMode: NSDefaultRunLoopMode
                                  beforeDate: [NSDate distantFuture]];
    }
 
  // We remove our notification observer
  [[NSNotificationCenter defaultCenter]
    removeObserver: self
    name: ShouldCloseMailbox
    object: nil];

  // We save the height of our textScrollView
  [[NSUserDefaults standardUserDefaults] setInteger: [textScrollView bounds].size.height
					 forKey: @"TEXTSCROLLVIEW_HEIGHT"];
  
  // We close all MessageViewWindows.
  [self _closeAllMessageViewWindows];

  // We update our last mail window on top if it was the current selected one
  // OR if we decided to re-use the MailWindow - we must set it to nil
  if ( [GNUMail lastMailWindowOnTop] == [self window] || 
       ([[NSUserDefaults standardUserDefaults] objectForKey: @"REUSE_MAILWINDOW"] &&
	[[[NSUserDefaults standardUserDefaults] objectForKey: @"REUSE_MAILWINDOW"] intValue] == NSOnState) )
    {
      [GNUMail setLastMailWindowOnTop: nil];
    }
  
  // We update our current super view for bundles (we set it to nil) and we
  // inform our bundles that the viewing view will be removed from the super view
  for (i = 0; i < [[GNUMail allBundles] count]; i++)
    {
      id<GNUMailBundle> aBundle;
      
      aBundle = [[GNUMail allBundles] objectAtIndex: i];
      
      if ( [aBundle hasViewingViewAccessory] )
	{
	  [aBundle setCurrentSuperview: nil];
	  [aBundle viewingViewAccessoryWillBeRemovedFromSuperview: [[self window] contentView]];
	}
    }

  
  // If we must compact the mbox on close, let's do it rigth now.
  if ( [[NSUserDefaults standardUserDefaults] integerForKey: @"COMPACT_MAILBOX_ON_CLOSE"] )
    {
      [folder expunge: NO];
      
      // We call this method just to update our cache, even if we do a little more work for nothing,
      // it's worth it :-)
      [self updateStatusLabel];
    }
  
  // We definitively close our folder.
  [folder close];

  // We add a message to our Console
  if ( [folder isKindOfClass: [LocalFolder class]] )
    {
      [[NSDistributedNotificationCenter defaultCenter]
	postNotificationName: GNUMailConsoleMessage
	object: nil
	userInfo: [NSDictionary dictionaryWithObject: [NSString stringWithFormat: _(@"Closed local folder %@."),
								[[self folder] name]]
				forKey: @"Message"]];
    }
  else
    {
      [[NSDistributedNotificationCenter defaultCenter]
	postNotificationName: GNUMailConsoleMessage
	object: nil
	userInfo: [NSDictionary dictionaryWithObject: [NSString stringWithFormat: _(@"Closed IMAP folder %@ on %@."),
								[[self folder] name], 
								[(IMAPStore *)[[self folder] store] name]]
				forKey: @"Message"]];
    }

  // We clear our 'Save Attachment' menu
  [(GNUMail *)[NSApp delegate] removeAllItemsFromMenu: [(GNUMail *)[NSApp delegate] saveAttachmentMenu]];

  // We remove our window from our list of opened windows
  [GNUMail removeMailWindow: [self window]];

  // We finally verify if we must close our IMAP Store.
  if ( shouldCloseStore )
    {
      NSString *aUsername, *aServerName;
      
      aServerName = [(IMAPStore *)[[self folder] store] name];
      aUsername = [(IMAPStore *)[[self folder] store] username];

      [[NSDistributedNotificationCenter defaultCenter]
	postNotificationName: GNUMailConsoleMessage
				  object: nil
				userInfo: [NSDictionary dictionaryWithObject:
					[NSString stringWithFormat: _(@"Closing IMAP connection on %@."),
					aServerName] forKey: @"Message"]];
	  
      [(IMAPStore *)[[self folder] store] close];
      [[MailboxManagerController singleInstance] setStore: nil
						 name: aServerName
						 username: aUsername];

      
      // FIXME - We should only collapse the item...
      [[[MailboxManagerController singleInstance] outlineView] reloadData];
    }

  AUTORELEASE(self);
}


//
//
//
- (void) windowDidLoad
{
  NSDebugLog(@"MailWindowController: -windowDidLoad");

#ifdef MACOSX
  [mailboxes setTarget: [NSApp delegate]];
  [mailboxes setAction: @selector(showMailboxManager:)];
  
  [addresses setTarget: [NSApp delegate]];
  [addresses setAction: @selector(showAddressBook:)];
  
  [find setTarget: [NSApp delegate]];
  [find setAction: @selector(showFindWindow:)];

  [tableView setDoubleAction: @selector(doubleClickedOnTableView:)];
#endif

  // We add our observer (same as above) for clean closing the Mailbox
  [[NSNotificationCenter defaultCenter]
    addObserver: self
    selector: @selector(close)
    name: ShouldCloseMailbox
    object: nil];

  // We finally add our observer if the filters changed
  [[NSNotificationCenter defaultCenter]
    addObserver: self
    selector: @selector(loadFilters:)
    name: FiltersHaveChanged
    object: nil];

  [[NSNotificationCenter defaultCenter]
    addObserver: self
    selector: @selector(updateMessageFlags:)
    name: ReplyToMessageWasSuccessful
    object: nil];
  

   // We load our filters and we initialize some values
  [self loadFilters: nil];

  [self setShowAllHeaders: NO];
  [GNUMail setLastMailWindowOnTop: [self window]];
  
  // We add our window from our list of opened window
  [GNUMail addMailWindow: [self window]];

  // We initialize our set of mutable arrays
  allMessageViewWindowControllers = [[NSMutableArray alloc] init];
  filteredMessageBoxes = [[NSMutableArray alloc] init];
  allConnections = [[NSMutableArray alloc] init];
  
  if ( [[NSUserDefaults standardUserDefaults] integerForKey: @"RETRIEVEMETHOD"] == AUTOMATICALLY)
    {  
      [self startTimer: nil];
    }
  
  // We add our notifications for knowing if we must start or stop the timer
  [[NSNotificationCenter defaultCenter]
    addObserver: self
    selector: @selector(startTimer:)
    name: ShouldStartLocalInboxTimer
    object: nil];

  [[NSNotificationCenter defaultCenter]
    addObserver: self
    selector: @selector(stopTimer:)
    name: ShouldStopLocalInboxTimer
    object: nil];
}


//
//
//
- (void) windowDidBecomeKey: (NSNotification *) aNotification
{
  int i;
  
  // We set the last window on top
  [GNUMail setLastMailWindowOnTop: [self window]];
  
  // We set the current superview of our bundle having providing
  // a viewing accessory view.
  for (i = 0; i < [[GNUMail allBundles] count]; i++)
    {
      id<GNUMailBundle> aBundle;
      
      aBundle = [[GNUMail allBundles] objectAtIndex: i];
      
      if ( [aBundle hasViewingViewAccessory] )
	{
	  [aBundle setCurrentSuperview: [[self window] contentView]];
	}
    }

  [[NSApp delegate] setEnableSaveInDraftsMenuItem: NO];
  [[NSApp delegate] setShowRawSourceMenuItem: ![self showRawSource]];
}


//
//
//
- (NSArray *) allMessages
{
  return allMessages;
}


//
//
//
- (Message *) selectedMessage
{
  int index;
  
  index = [tableView selectedRow];
  
  if (index < 0)
    {
      return nil;
    }

  return [allMessages objectAtIndex: index];
}


//
//
//
- (NSArray *) selectedMessages
{
  if ( [tableView numberOfSelectedRows] == 0 )
    {
      NSBeep();
    }
  else
    {
      NSMutableArray *aMutableArray;
      NSEnumerator *anEnumerator;
      NSNumber *aRow;;
      
      aMutableArray = [[NSMutableArray alloc] initWithCapacity: [tableView numberOfSelectedRows]];
      
      anEnumerator = [tableView selectedRowEnumerator];

      while ( (aRow = [anEnumerator nextObject]) )
	{
	  [aMutableArray addObject: [allMessages objectAtIndex: [aRow intValue]] ];
	}
      
      return AUTORELEASE(aMutableArray);
    }
  
  // Reached in case of error.
  return nil;
}


//
//
//
- (NSTableView *) tableView
{
  return tableView;
}


//
//
//
- (NSTextView *) textView
{
  return textView;
}


//
//
//
- (NSString *) currentOrder
{
  return currentOrder;
}


//
//
//
- (void) setCurrentOrder: (NSString *) theCurrentOrder
{
  RETAIN(theCurrentOrder);
  RELEASE(currentOrder);
  currentOrder = theCurrentOrder;
}


//
//
//
- (NSString *) previousOrder
{
  return previousOrder;
}


//
//
//
- (void) setPreviousOrder: (NSString *) thePreviousOrder
{
  RETAIN(thePreviousOrder);
  RELEASE(previousOrder);
  previousOrder = thePreviousOrder;
}


//
//
//
- (NSImageView *) secureConnectionIcon
{
  return secureConnectionIcon;
}


//
//
//
- (id) mustStopTransfer
{
  //NSDebugLog(@"Polling to see if we must stop...");
  return [NSNumber numberWithBool: mustStopTransfer];
}


//
//
//
- (MailHeaderCell *) mailHeaderCell
{
  return mailHeaderCell;
}


//
//
//
- (NSMutableArray *) allMessageViewWindowControllers
{
  return allMessageViewWindowControllers;
}


//
//
//
- (BOOL) showAllHeaders
{
  if ( [[NSUserDefaults standardUserDefaults] objectForKey: @"SHOWALLHEADERS"] )
    {
      return ([[[NSUserDefaults standardUserDefaults] objectForKey: @"SHOWALLHEADERS"] intValue] == NSOnState ? YES
	      : showAllHeaders);
    }

  return showAllHeaders;
}


//
//
//
- (void) setShowAllHeaders: (BOOL) aBOOL
{
  showAllHeaders = aBOOL;
}

- (BOOL) showRawSource
{
	return showRawSource;
}

- (void) setShowRawSource: (BOOL) aBool
{
	showRawSource = aBool;
}



//
//
//
- (BOOL) showStatusBar
{
  return showStatusBar;
}


//
//
//
- (void) setShowStatusBar: (BOOL) aBOOL
{
  NSRect aRect;

  showStatusBar = aBOOL;

  aRect = [tableView frame];
  
  if ( showStatusBar )
    {
      [[[self window] contentView] addSubview: label];
    }
  else
    {
      [label removeFromSuperviewWithoutNeedingDisplay];
    }
  
  [[[self window] contentView] setNeedsDisplay: YES];
}


//
//
//
- (IBAction) loadFilters : (id) sender
{
  TEST_RELEASE(filterManager);

  // We load our filters
  filterManager = [FilterManager filtersFromDisk];

  if ( filterManager )
    {
      RETAIN(filterManager);
    }

  [tableView setNeedsDisplay: YES];
}


//
//
//
- (IBAction) updateMessageFlags: (id) sender
{
  if (sender)
    {
      Message *aMessage;
      
      aMessage = [[sender userInfo] objectForKey: @"Message"];

      if (aMessage &&  [[[self folder] allMessages] containsObject: aMessage] )
	{
	  [[aMessage flags] add: ANSWERED];
	  [tableView setNeedsDisplay: YES];
	}
    }
}


//
//
//
- (IBAction) updateAnimatedIcon: (id) sender
{
  if (animation_index == 9)
    {
      animation_index = 1;
    }
  
  [icon setImage: [NSImage imageNamed: [NSString stringWithFormat: @"anim-logo-%d.tiff", animation_index]]];
  
  animation_index += 1;
}


//
//
//
- (IBAction) getNewMessages: (id) sender
{
  NSArray *allKeys;
  int i;
  
  // We verify 'who' initiated the call to this method, the timer or the 'Get' button.
  if ( sender == timer )
    {
      // If the timer calls again this selector, while the thread is running,
      // we don't stop but simply return.
      if ( animation && [animation isValid] )
	{
	  return;
	}

      initiatedFromTimer = YES;
    }
  else
    {
      // If the user clicks again on the Get button while the thread is running,
      // we simply set the thread to be stopped and we return quickly.
      if ( animation && [animation isValid] )
	{
	  mustStopTransfer = YES;
	  return;
	}
      
      initiatedFromTimer = NO;
    }
  
  //
  // If we have a IMAPFolder, we noop it to verify for new
  // messages.
  //
  if ( [[self folder] isKindOfClass: [IMAPFolder class]] )
    {
      [[NSDistributedNotificationCenter defaultCenter]
	postNotificationName: GNUMailConsoleMessage
	object: nil
	userInfo: [NSDictionary dictionaryWithObject: 
				  [NSString stringWithFormat: 
					      (@"Verifying for new messages on IMAP server %@, folder %@"),
					    [(IMAPStore *)[[self folder] store] name], [[self folder] name]]
				forKey: @"Message"]];
      
      [(IMAPFolder *)[self folder] noop];
    }
  
  // We initialize our thread count
  threadCount = 0;
  
  // We disable our automatic fetching timer, to avoid a deadlock
  [self stopTimer: nil];

  // We disable our Get button
#ifdef MACOSX
  [get setImage: [NSImage imageNamed: @"stop_32.tiff"]];
  [get setLabel: _(@"Stop")];
#else
  [get setImage: [NSImage imageNamed: @"stop_48.tiff"]];
  [get setTitle: _(@"Stop")];
#endif

  // We start our animation timer
  animation_index = 1;
  animation = [NSTimer scheduledTimerWithTimeInterval: 0.1
		       target: self
		       selector: @selector(updateAnimatedIcon:)
		       userInfo: nil
		       repeats: YES];
  RETAIN(animation);

  mustStopTransfer = NO;

  // We re-initialize our ivars
  transferredMessageCount = filteredMessageCount = 0;

  // We first remove all elements from our array of mailbox names
  [filteredMessageBoxes removeAllObjects];

  // We retrieve all keys of receiving accounts
  allKeys = [[[[NSUserDefaults standardUserDefaults] objectForKey: @"RECEIVING"] allKeys]
	      sortedArrayUsingSelector: @selector(compare:)];
  
  if ( sender != get &&
       [sender isKindOfClass: [NSMenuItem class]] &&
       [sender tag] >= 0 && 
       [sender tag] < [allKeys count] )
    {
      //
      // Only one POP3 account must be verified
      //
      allKeys = [NSArray arrayWithObject: [allKeys objectAtIndex: [sender tag]]];
      threadCount = threadCount + 1;
    }
  else
    {
      //
      // We transfer the emails from the local mail spool file
      //
      NSConnection *aConnection;
      NSPort *port1, *port2;
      NSArray *ports;
      
      //
      // We detach our thread that will fetch the messages from the local mail spool file
      //
      [[NSDistributedNotificationCenter defaultCenter]
	postNotificationName: GNUMailConsoleMessage
	object: nil
	userInfo: [NSDictionary dictionaryWithObject: (@"Starting the fetching thread for the local mail spool file...")
				forKey: @"Message"]];
      
      //
      // We increment the thread counter by 2, to avoid falling back to 0 before we start the fetch of our
      // POP3 folders. If we don't do that, this new thread might ends BEFORE a new one (for fetching mails from
      // a POP3 server) is created. That will cause the animation to be stopped while it was supposed to run,
      // breaking a lot of things. We will decrement it again at the end of this method, by calling
      // [self threadWillExit] - which is, totally safe.
      // We must be very careful about WHO initiated the method call to -getNewMessages. We act differently
      // if it's the menu since we could want to fetch the messages of only the local spool file or only
      // one POP3 server.
      //
      if ( ([sender isKindOfClass: [NSMenuItem class]] && [sender tag] == [allKeys count]) ||
	   [allKeys count] == 0)
	{
	  threadCount = threadCount + 1;
	}
      else
	{
	  threadCount = threadCount + 2;
	}
      
      // We initialize our NSConnection object and our ports
      port1 = [[NSPort alloc] init];
      port2 = [[NSPort alloc] init];
      
      aConnection = [[NSConnection alloc] initWithReceivePort: port1
					  sendPort: port2];
      [aConnection setRootObject: self];
      [aConnection setRequestTimeout: LONG_DELAY];
      [aConnection setReplyTimeout: LONG_DELAY];

      ports = [NSArray arrayWithObjects: port2, port1, nil];
      
      RELEASE(port1);
      RELEASE(port2);

      [allConnections addObject: aConnection];
      RELEASE(aConnection);
      
      [NSThread detachNewThreadSelector: @selector(_transferMessagesUsingInfoFromDictionary:)
		toTarget: self
		withObject: [NSDictionary dictionaryWithObjectsAndKeys: 
					    [NSNumber numberWithInt: (sender == self ? 1 : 0)], @"SENDER",
					  ports, @"PORTS",
					  nil]];
      
      // If the sender tag is equal to the keys count, only messages from the
      // mail spool file are transferred
      // FIXME: This won't work for multiple UNIX accounts.
      if ( [sender isKindOfClass: [NSMenuItem class]] &&
	   [sender tag] == [allKeys count] )
	{
	  return;
	}
    }
  
  //
  // We prompt for the passwords of all our POP3 servers and start the thread(s)
  //
  for (i = 0; i < [allKeys count]; i++)
    {
      NSDictionary *allValues;
      NSString *aKey;
      
      aKey = [allKeys objectAtIndex: i];
      allValues = [[[NSUserDefaults standardUserDefaults] objectForKey: @"RECEIVING"] objectForKey: aKey];
      
      if ( ![allValues objectForKey: @"SERVERTYPE"] ||
	   [[allValues objectForKey: @"SERVERTYPE"] intValue] == POP3 ) 
	{
	  // If the sender is self, that means that getNewMessages: was called by -verifyForNewMessageOnStartup.
	  // We MUST skip the server weren't not interested into, those that don't have the -verifyOnStartup flag.
	  if ( (sender == self &&
		[allValues objectForKey: @"VERIFYONSTARTUP"] &&
		[[allValues objectForKey: @"VERIFYONSTARTUP"] intValue] == NSOnState) ||
	       (sender != self) )
	    {
	      // We get the password for the POP3 server and we cache it.
	      if ( [Utilities passwordForKey: aKey
			      prompt: YES] )
		{
		  NSConnection *aConnection;
		  NSPort *port1, *port2;
		  NSArray *ports;

		  //
		  // We detach our thread that will fetch the messages from the POP3 servers
		  //
		  [[NSDistributedNotificationCenter defaultCenter]
		    postNotificationName: GNUMailConsoleMessage
		    object: nil
		    userInfo: [NSDictionary dictionaryWithObject: 
					      [NSString stringWithFormat: 
							  _(@"Starting the fetching thread for %@.."), aKey]
					    forKey: @"Message"]];
		  

		  // We increment the thread counter
		  threadCount = threadCount + 1;
		  

		  // We initialize our NSConnection object and our ports
		  port1 = [[NSPort alloc] init];
		  port2 = [[NSPort alloc] init];
		  
		  aConnection = [[NSConnection alloc] initWithReceivePort: port1
						      sendPort: port2];
		  [aConnection setRootObject: self];
		  [aConnection setRequestTimeout: LONG_DELAY];
		  [aConnection setReplyTimeout: LONG_DELAY];
		  
		  ports = [NSArray arrayWithObjects: port2, port1, nil];
		  
		  RELEASE(port1);
		  RELEASE(port2);

		  [allConnections addObject: aConnection];
		  RELEASE(aConnection);

		  [NSThread detachNewThreadSelector: @selector(_transferMessagesUsingInfoFromDictionary:)
			    toTarget: self
			    withObject: [NSDictionary dictionaryWithObjectsAndKeys: 
							[NSNumber numberWithInt: 
								    (sender == self ? 1 : 0)], @"SENDER",
						      ports, @"PORTS",
						      aKey, @"KEY",
						      nil]];
		}
	    }
	}
    } // for (...)

  // See comment at line ~275.
  if ( [allKeys count] >= 1)
    {
      [self threadWillExit];
    }
}


//
//
//
- (void) startTimer: (id) sender
{
  int nbOfSecondsToWait;
  
  if ([[NSUserDefaults standardUserDefaults] integerForKey: @"RETRIEVEMINUTES"] <= 0)
    {
      nbOfSecondsToWait = 60;
    }
  else
    {
      nbOfSecondsToWait = [[NSUserDefaults standardUserDefaults] integerForKey: @"RETRIEVEMINUTES"] *  60;
    }
  
  
  timer =  [NSTimer scheduledTimerWithTimeInterval: nbOfSecondsToWait
		    target: self
		    selector: @selector(getNewMessages:)
		    userInfo: nil
		    repeats: YES];
  RETAIN(timer);
}


//
//
//
- (void) stopTimer: (id) sender
{
  if ( timer )
    {
      [timer invalidate];
    }
  
  DESTROY(timer);
}


//
// Other methods
//
- (oneway void) communicationErrorOccuredOnServer: (NSString *) theServer
{
  NSRunInformationalAlertPanel(_(@"Error!"),
			       _(@"Unable to communicate with the POP3 server (%@)."),
			       _(@"OK"),
			       NULL,
			       NULL,
			       theServer);
}


//
//
//
- (oneway void) invalidPasswordOnServer: (NSString *) theServer
			       username: (NSString *) theUsername
{
  NSRunInformationalAlertPanel(_(@"Error!"),
			       _(@"Invalid password %@ @ %@."),
			       _(@"OK"),
			       NULL,
			       NULL,
			       theUsername,
			       theServer);
}


//
// FIXME: Why under OS X, it won't work if the method is oneway?
//
- (void) messageWasReceived: (NSData *) theMessage
		 forURLName: (URLName *) theURLName
{
  NSString *aFolderName;
  Store *aStore;

  if ( [[theURLName protocol] caseInsensitiveCompare: @"LOCAL"] == NSOrderedSame )
    {
      
      aFolderName = [theURLName foldername];
      aStore = [[MailboxManagerController singleInstance]
		 storeForName: @"GNUMAIL_LOCAL_STORE"
		 username: NSUserName()];
    }
  else
    {
      aFolderName = [theURLName foldername];
      aStore = [Utilities storeForURLName: theURLName];
    }

  if ( !aStore )
    {
      if ( ![[MailboxManagerController singleInstance] initializeIMAPStoreWithKey:
							 [NSString stringWithFormat: @"%@ @ %@", 
								   [theURLName username], 
								   [theURLName host]]] )
	{
	  URLName *aURLName;
	  
	  NSRunAlertPanel(_(@"Error!"),
			  _(@"An error occured while obtaining the IMAP Store to save the received message.\n%@"),
			  _(@"OK"),
			  NULL,
			  NULL,
			  _(@"The received message was saved in your local Inbox folder."));
	  
	  aURLName = [[URLName alloc] initWithString: [[NSUserDefaults standardUserDefaults] objectForKey: @"INBOXFOLDERNAME"]
				      path: [[NSUserDefaults standardUserDefaults] objectForKey: @"LOCALMAILDIR"]];
	  AUTORELEASE(aURLName);
	  
	  aFolderName = [aURLName foldername];
	  aStore = [[MailboxManagerController singleInstance]
		 storeForName: @"GNUMAIL_LOCAL_STORE"
		     username: NSUserName()];
	}
    }
  
  // We append the message to the folder
  [[MailboxManagerController singleInstance] transferMessageFromRawSource: theMessage
					     flags: nil
					     toFolderWithName: aFolderName
					     orToFolder: nil
					     store: aStore];

  // We update our table view and our label
  //[self tableViewShouldReloadData];
  [self updateStatusLabel];
  
  // We set the application icon to GNUMail_Full.tiff
  [NSApp setApplicationIconImage: [NSImage imageNamed: @"GNUMail_Full.tiff"]];

  // We finally force our window to upadte
  [[self window] update];
}


//
//
//
- (oneway void) messagesWereReceived
{
  if ( initiatedFromTimer )
    {
      NSBeep();
    }

  // We finally synchronize our local cache
  [[(LocalFolder *)[self folder] localFolderCacheManager] synchronize];
}


//
//
//
- (oneway void) noNewMessagesOnServer: (NSString *) theServer
			     username: (NSString *) theUsername
{
  NSRunInformationalAlertPanel(_(@"No New Messages..."),
			       _(@"There are no new messages on %@ @ %@."),
			       _(@"OK"),
			       NULL,
			       NULL,
			       theUsername,
			       theServer);
}


//
//
//
- (void) reloadTableColumns
{
  NSArray *shownTableColumns;
  int i;

  shownTableColumns = [[NSUserDefaults standardUserDefaults] objectForKey: @"SHOWNTABLECOLUMNS"];

  // If the value doesn't exist in the user's defaults, we show all table columns.
  if ( !shownTableColumns )
    {
      return;
    }

  RETAIN(statusColumn);
  RETAIN(idColumn);
  RETAIN(dateColumn);
  RETAIN(fromColumn);
  RETAIN(subjectColumn);
  RETAIN(sizeColumn);

  [[self tableView] removeTableColumn: statusColumn];
  [[self tableView] removeTableColumn: idColumn];
  [[self tableView] removeTableColumn: dateColumn];
  [[self tableView] removeTableColumn: fromColumn];
  [[self tableView] removeTableColumn: subjectColumn];
  [[self tableView] removeTableColumn: sizeColumn];

  for (i = 0; i < [shownTableColumns count]; i++)
    {
      NSString *column;

      column = [shownTableColumns objectAtIndex: i];

      if ( [column isEqualToString: @"Status"] )
	{
	  [[self tableView] addTableColumn: statusColumn];
	}
      else if ( [column isEqualToString: @"Id"] )
	{
	  [[self tableView] addTableColumn: idColumn];
	}
      else if ( [column isEqualToString: @"Date"] )
	{
	  [[self tableView] addTableColumn: dateColumn];
	}
      else if ( [column isEqualToString: @"From"] )
	{
	  [[self tableView] addTableColumn: fromColumn];
	}
      else if ( [column isEqualToString: @"Subject"] )
	{
	  [[self tableView] addTableColumn: subjectColumn];
	}
      else if ( [column isEqualToString: @"Size"] )
	{
	  [[self tableView] addTableColumn: sizeColumn];
	}
    }
      
      
  [[self tableView] sizeToFit];

  RELEASE(statusColumn);
  RELEASE(idColumn);
  RELEASE(dateColumn);
  RELEASE(fromColumn);
  RELEASE(subjectColumn);
  RELEASE(sizeColumn);

}


//
//
//
- (void) setSelectedMessagesAsTransferred

{
  NSEnumerator *anEnumerator;
  NSNumber *aRow;
  Message *theMessage;
  Flags *theFlags;
  
  anEnumerator = [tableView selectedRowEnumerator];
  
  while ( (aRow = [anEnumerator nextObject]) )
    {
      theMessage = [allMessages objectAtIndex: [aRow intValue]];
      theFlags = [theMessage flags];
      
      // We set the flag Transferred to the message
      [theFlags add: TRANSFERRED];
    }

  // We update the messages cache
  [[self folder] updateCache];

  // We refresh our tableView
  [self tableViewShouldReloadData];

  // We update the status label
  [self updateStatusLabel];
}


//
//
//
- (void) tableViewShouldReloadData
{
  NSArray *previousArray = allMessages;
  SEL sortingSel = NULL;

  if ( [self currentOrder] == nil )
    {
      [self setPreviousOrder: @"Id"];
      [self setCurrentOrder: @"Id"];
    }
  
  if ( [[self currentOrder] isEqualToString: @"Id"] )
    {
      //NSDebugLog(@"id");
      if (isReverseOrder == YES)
	{
	  sortingSel = @selector(reverseCompareAccordingToNumber:);
	  allMessages = 
	    RETAIN([[folder allMessages] 
		     sortedArrayUsingSelector:
		       @selector(reverseCompareAccordingToNumber:)]);
	}
      else
	{
	  sortingSel = @selector(compareAccordingToNumber:);
	  allMessages = [[NSArray alloc]
			  initWithArray: [folder allMessages]];
	}
    }
  else if ( [[self currentOrder] isEqualToString: @"Date"] )
    {
      //NSDebugLog(@"date");
      if (isReverseOrder == YES)
	{
	  sortingSel = @selector(reverseCompareAccordingToDate:);
	  allMessages = 
	    RETAIN([[folder allMessages] 
		     sortedArrayUsingSelector:
		       @selector(reverseCompareAccordingToDate:)]);
	}
      else
	{
	  sortingSel = @selector(compareAccordingToDate:);
	  allMessages = 
	    RETAIN([[folder allMessages] 
		     sortedArrayUsingSelector:
		       @selector(compareAccordingToDate:)]);
	}
    }
  else if ( [[self currentOrder] isEqualToString: @"From"] )
    {
      //NSDebugLog(@"from");
      if (isReverseOrder == YES)
	{
	  sortingSel = @selector(reverseCompareAccordingToSender:);
	  allMessages = 
	    RETAIN([[folder allMessages] 
		     sortedArrayUsingSelector:
		       @selector(reverseCompareAccordingToSender:)]);
	}
      else
	{
	  sortingSel = @selector(compareAccordingToSender:);
	  allMessages = 
	    RETAIN([[folder allMessages] 
		     sortedArrayUsingSelector:
		       @selector(compareAccordingToSender:)]);
	}
    }
  else if ( [[self currentOrder] isEqualToString: @"Subject"] )
    {
      //NSDebugLog(@"subject");
      if (isReverseOrder == YES)
	{
	  sortingSel = @selector(reverseCompareAccordingToSubject:);
	  allMessages = 
	    RETAIN([[folder allMessages] 
		     sortedArrayUsingSelector:
		       @selector(reverseCompareAccordingToSubject:)]);
	}
      else
	{
	  sortingSel = @selector(compareAccordingToSubject:);
	  allMessages = 
	    RETAIN([[folder allMessages] 
		     sortedArrayUsingSelector:
		       @selector(compareAccordingToSubject:)]);
	}
    }
  else if ( [[self currentOrder] isEqualToString: @"Size"] )
    {
      //NSDebugLog(@"subject");
      if (isReverseOrder == YES)
	{
	  sortingSel = @selector(reverseCompareAccordingToSize:);
	  allMessages = 
	    RETAIN([[folder allMessages] 
		     sortedArrayUsingSelector:
		       @selector(reverseCompareAccordingToSize:)]);
	}
      else
	{
	  sortingSel = @selector(compareAccordingToSize:);
	  allMessages = 
	    RETAIN([[folder allMessages] 
		     sortedArrayUsingSelector:
		       @selector(compareAccordingToSize:)]);
	}
    }
  
  if ( previousArray )
    {
      NSArray *sc = [[tableView selectedRowEnumerator] allObjects];
      NSMutableArray *sm;
      int selectedRow = [tableView selectedRow];
      int count = [sc count];
      id message;
      int index;
      int i;
      int newCount = [allMessages count];
      BOOL newSelection = NO;
      NSRange range = NSMakeRange(0, newCount);

      {
	sm = [[NSMutableArray alloc] initWithCapacity: newCount];
	for ( i = 0; i < count; i++)
	  {
	    [sm addObject: 
		  [previousArray objectAtIndex: 
			      [[sc objectAtIndex: i] intValue]]];
	  }
	[sm sortUsingSelector: sortingSel];
      }
      isReloading = YES;

      
      [tableView deselectAll: self];
      [tableView reloadData];

      for ( i = 0; i < count; i++ )
	{
	  message = [sm objectAtIndex: i];
			      
	  index = [allMessages indexOfObject: message
			       inRange: range];

	  if (index != NSNotFound)
	    {
	      [tableView selectRow: index
			 byExtendingSelection: YES];
	      range = NSMakeRange(index + 1, newCount - index - 1);
	    }
	  else
	    {
	      newSelection = YES;
	    }
	}

      if (selectedRow != -1)
	{
	  message = [previousArray objectAtIndex: selectedRow];
	  index = [allMessages indexOfObject: message];
	  
	  if (index != NSNotFound)
	    {
	      [tableView selectRow: index
			 byExtendingSelection: YES];
	    }
	}
      isReloading = NO;
      
      if (newSelection == YES)
	[self tableViewSelectionDidChange: nil];

      if ([tableView selectedRow] != -1)
	{
	  [tableView scrollRowToVisible: [tableView selectedRow]];
	}
    }

  //NSDebugLog(@"done");
  TEST_RELEASE(previousArray);

  [self setPreviousOrder: [self currentOrder]];

  [tableView reloadData];
  
  // We verify if we have at least one selected row, in case we don't, we just 
  // clear our textView
  if ( [tableView numberOfSelectedRows] != 1 )
    {
      [textView setString: @""];
    }
}


//
//
//
- (oneway void) threadWillExit
{
  // We decrement the thread counter
  threadCount = threadCount - 1;

  if ( threadCount == 0 )
    {
      int i;
      
      // We first stop our animation timer
      [animation invalidate];
      DESTROY(animation);
      [icon setImage: nil];
      
      // We finally re-enable our Get button
#ifdef MACOSX
      [get setImage: [NSImage imageNamed: @"retrieve_32.tiff"]];
      [get setLabel: _(@"Get")];
#else
      [get setImage: [NSImage imageNamed: @"retrieve_48.tiff"]];
      [get setTitle: _(@"Get")];
#endif     
       
      // We verify if we need to inform the user that received mails have been
      // transferred to folders by some filters.
      if ( !initiatedFromTimer && (filteredMessageCount > 0) &&
	   [[NSUserDefaults standardUserDefaults] integerForKey: @"SHOW_FILTER_PANEL"] == NSOnState )
	{
	  NSRunInformationalAlertPanel(_(@"Filtered messages..."),
				       _(@"%d messages have been transfered to the following folders:\n%@"),
				       _(@"OK"),
				       NULL,
				       NULL,
				       filteredMessageCount,
				       [filteredMessageBoxes componentsJoinedByString: @"\n"]);
	}

      // We restart our automatic fetching timer
      if ( [[NSUserDefaults standardUserDefaults] integerForKey: @"RETRIEVEMETHOD"] == AUTOMATICALLY )
	{  
	  [self startTimer: nil];
	}
      
      // We now invalidate all our connections
      for (i = 0; i < [allConnections count]; i++)
	{
	  id aConnection;
       
	  aConnection = [allConnections objectAtIndex: i];
	  
	  NS_DURING
	    [[aConnection receivePort] invalidate];
	    [[aConnection sendPort] invalidate];
	    [aConnection invalidate];
	  NS_HANDLER
	    NSDebugLog(@"LocalInboxWindowController: Exception occured while invalidating ports.");
	  NS_ENDHANDLER
	}

      [allConnections removeAllObjects];
    }
}


//
//
//
- (void) updateStatusLabel
{
  NSString *aString, *aStoreName, *aUsername;
  Message *aMessage;
  
  int i, totalSize, unreadCount, totalUnreadSize, messagesSize;
  
  NSEnumerator *enumerator;
  int numberOfSelectedRows;
  id anObject;
  
  totalSize = unreadCount = totalUnreadSize = 0;

  for (i = 0; i < [folder count]; i++)
    {
      aMessage = [allMessages objectAtIndex: i];
      totalSize += [aMessage size];
      if (![[aMessage flags] contain: SEEN])
	{
	  unreadCount++;
	  totalUnreadSize += [aMessage size];
	}
    }
  
  numberOfSelectedRows = [tableView numberOfSelectedRows];
  messagesSize = 0;
  
  if (numberOfSelectedRows > 0)
    {
      enumerator = [tableView selectedRowEnumerator];
      
      while ((anObject = [enumerator nextObject])) {
	aMessage = [allMessages objectAtIndex: [anObject intValue]];
	messagesSize += [aMessage size];
      }
    }
  
  aString = [NSString stringWithFormat: _(@"%d messages (%dKB) - %d unread (%dKB) - %d selected (%0.1fKB) - %d deleted"),
		      [folder count], 
		      (totalSize/1024), unreadCount, (totalUnreadSize/1024), 
		      numberOfSelectedRows, ((float)messagesSize/(float)1024),
		      [folder numberOfDeletedMessages]];
  
  [label setStringValue: aString];
  [label setNeedsDisplay: YES];


  // We finally update our cache
  if ( [(id<NSObject>)[[self folder] store] isKindOfClass: [LocalStore class]] )
    {
      aStoreName = @"GNUMAIL_LOCAL_STORE";
      aUsername = NSUserName();
    }
  else
    {
      aStoreName = [(IMAPStore *)[[self folder] store] name];
      aUsername = [(IMAPStore *)[[self folder] store] username];
    }
  
  [[[MailboxManagerController singleInstance] mailboxManagerCache]
    setAllValuesForStoreName: aStoreName
    folderName: [[self folder] name]
    username: aUsername
    nbOfMessages: [folder count]
    nbOfUnreadMessages: unreadCount
    totalSize: totalSize];

  [[[MailboxManagerController singleInstance] outlineView] setNeedsDisplay: YES];
}


//
//
//
- (void) verifyForNewMailOnStartup
{
  NSArray *allPOP3Keys;
  int i, aCount;

  //
  // We DO NOT verify for new mails on startup IF WE DON'T NEED TO! We simply count
  // the number of POP3Store we have to verify for new mails on startup and if the count
  // is more than 0, we can call -getNewMessages:
  //
  allPOP3Keys = [[[NSUserDefaults standardUserDefaults] objectForKey: @"RECEIVING"]
		  allKeys];
  aCount = 0;
  
  for (i = 0; i < [allPOP3Keys count]; i++)
    {
      NSDictionary *allPOP3Values;
      NSString *aKey;

      aKey = [allPOP3Keys objectAtIndex: i];
      allPOP3Values = [[[NSUserDefaults standardUserDefaults] objectForKey: @"RECEIVING"]
			objectForKey: aKey];
      
      if ( [allPOP3Values objectForKey: @"VERIFYONSTARTUP"] &&
	   [[allPOP3Values objectForKey: @"VERIFYONSTARTUP"] intValue] == NSOnState &&
	   ( ![allPOP3Values objectForKey: @"SERVERTYPE"] || [[allPOP3Values objectForKey: @"SERVERTYPE"] intValue] == POP3) )
	{
	  aCount++;
	}
    }

  if ( aCount > 0 )
    {
      [self getNewMessages: self];
    }
}


//
//
//
- (void) newMessagesWereReceived: (id) sender
{
  [self tableViewShouldReloadData];
  [self updateStatusLabel];
}

@end


//
// Private interface for MailWindowContrller
//
@implementation MailWindowController (Private)

//
//
//
- (void) _closeAllMessageViewWindows
{
  int i;

  // No need to actually remove the object from the array since that will
  // already be done in MessageViewWindowController: -windowWillClose.
  for (i = ([allMessageViewWindowControllers count] - 1); i >= 0; i--)
    {
      [[allMessageViewWindowControllers objectAtIndex: i] close];
    }

}

//
//
//
- (void) _launchExternalProgram: (NSDictionary *) theDictionary
{
  NSAutoreleasePool *pool;
  NSTask *aTask;
  
  pool = [[NSAutoreleasePool alloc] init];

  aTask = [[NSTask alloc] init];
  AUTORELEASE(aTask);
  
  [aTask setLaunchPath: [theDictionary objectForKey: @"command"]];
  [aTask setArguments: [theDictionary objectForKey: @"arguments"]];
  [aTask launch];
  [aTask waitUntilExit];
  
  [[NSFileManager defaultManager] removeFileAtPath: [theDictionary objectForKey: @"file"]
				  handler: nil];

  RELEASE(pool);
}


//
//
//
- (void) _loadAccessoryViews
{
  int i, index;

  index = 0;

  for (i = 0; i < [[GNUMail allBundles] count]; i++)
    {
      id<GNUMailBundle> aBundle;

      aBundle = [[GNUMail allBundles] objectAtIndex: i];
      
      if ( [aBundle hasViewingViewAccessory] )
	{
	  id aView;
	  
	  aView = [aBundle viewingViewAccessory];
	  
	  if ( [aBundle viewingViewAccessoryType] == ViewingViewTypeHeaderCell )
	    {
	      NSDebugLog(@"Adding ViewingViewTypeHeaderCell type of Bundle...");
	      [mailHeaderCell addView: aView];
	    }
	  else if ( [aBundle viewingViewAccessoryType] == ViewingViewTypeStatusBar )
	    {
	      NSImageCell *aCell;

	      NSDebugLog(@"Adding ViewingViewTypeStatusBar type of Bundle...");

	      aCell = (NSImageCell *)aView;
	      //[aCell setFrame: NSMakeRect(0,0,16,16)];
	      
	      [matrix addColumn];
	      [matrix putCell: aCell
		      atRow: 0
		      column: index];

	      index += 1;
	    }
	  else
	    {
	      NSRect aRect;
	      
	      aRect = [find frame];
	      aRect.origin.x += 72;

	      [aView setAutoresizingMask: NSViewMinYMargin];
	      [aView setFrame: aRect];
	      
	      [[[self window] contentView] addSubview: aView];
	    }
	}

      // We also set the current superview
      [aBundle setCurrentSuperview: [[self window] contentView]];
    }
}


//
//
//
- (void) _matchFilterRuleFromRawSource: (NSData *) theRawSource
			    connection: (NSConnection *) theConnection
				   key: (NSString *) theKey
{
  URLName *theURLName;
  
  while ( ![lock tryLock] )
    {
      [[NSRunLoop currentRunLoop] runMode: NSDefaultRunLoopMode
                                  beforeDate: [NSDate distantFuture]];
    }

  theURLName = [filterManager matchedURLNameFromMessageAsRawSource: theRawSource
			      type: TYPE_INCOMING
			      key: theKey];
  
  if ( ![Utilities URLWithString: [theURLName stringValue]
		   matchFolder: [self folder]] )
    {
      NSString *aFolderName;

      if ( [[theURLName protocol] caseInsensitiveCompare: @"LOCAL"] == NSOrderedSame )
	{
	  aFolderName = [NSString stringWithFormat: _(@"Local - %@"), [theURLName foldername]];
	}
      else
	{
	  aFolderName = [NSString stringWithFormat: _(@"IMAP %@ @ %@ - %@"), [theURLName username],
				  [theURLName host], [theURLName foldername]];
	}
      
      filteredMessageCount += 1;
      
      if ( ![filteredMessageBoxes containsObject: aFolderName] )
  	{
  	  [filteredMessageBoxes addObject: aFolderName];
  	}
    }
  
  // We release our lock
  [lock unlock];
  
  [(id)[theConnection rootProxy] messageWasReceived: theRawSource
       forURLName: theURLName];
}

//
//
//
- (void) _restoreSortingOrder
{
  // We get our default sorting order
  if ( [[NSUserDefaults standardUserDefaults] objectForKey: @"SORTINGORDER"] )
    {
      [self setCurrentOrder: [[NSUserDefaults standardUserDefaults] objectForKey: @"SORTINGORDER"]];
      
      isReverseOrder = [[NSUserDefaults standardUserDefaults] integerForKey: @"SORTINGSTATE"];

      if ( [[self currentOrder] isEqualToString: @"Date"] )
	{
	  [[self tableView] setHighlightedTableColumn: dateColumn];
	}
      else if ( [[self currentOrder] isEqualToString: @"From"] )
	{
	  [[self tableView] setHighlightedTableColumn: fromColumn];
	}
      else if ( [[self currentOrder] isEqualToString: @"Subject"] )
	{
	  [[self tableView] setHighlightedTableColumn: subjectColumn];
	}
      else if ( [[self currentOrder] isEqualToString: @"Size"] )
	{
	  [[self tableView] setHighlightedTableColumn: sizeColumn];
	}
      else
	{
	  [[self tableView] setHighlightedTableColumn: idColumn];
	}
    }
  else
    {
      [[self tableView] setHighlightedTableColumn: idColumn];
    }
}


//
//
//
- (void) _restoreSplitViewSize
{
  if ( [[NSUserDefaults standardUserDefaults] objectForKey: @"TEXTSCROLLVIEW_HEIGHT"] ) 
    {
      NSRect aRect;
      
      aRect = [textScrollView frame];
      aRect.size.height = [[NSUserDefaults standardUserDefaults] integerForKey: @"TEXTSCROLLVIEW_HEIGHT"];
      
      [textScrollView setFrame: aRect];
      
      [splitView adjustSubviews];
      [splitView setNeedsDisplay: YES];
    }
}


//
//
// 
- (void) _transferMessagesUsingInfoFromDictionary: (NSDictionary *) theDictionary
{
  NSDictionary *allPOP3Values;

  NSArray *thePorts;
  NSConnection *serverConnection;

  POP3Store *aStore;
  POP3Folder *aFolder;

  NSString *aKey;

  BOOL verifyingOnStartup;

  NSAutoreleasePool *threadpool;

  threadpool = [[NSAutoreleasePool alloc] init];

  thePorts = [theDictionary objectForKey: @"PORTS"];
  verifyingOnStartup = ([[theDictionary objectForKey: @"SENDER"] intValue] == 1 ? YES : NO);
  
  // We first get our connection to our main thread
  serverConnection = [[NSConnection alloc] initWithReceivePort: [thePorts objectAtIndex: 0]
					   sendPort: [thePorts objectAtIndex: 1]];
  [serverConnection setRequestTimeout: LONG_DELAY];
  [serverConnection setReplyTimeout: LONG_DELAY];

  // We get our key
  aKey = [theDictionary objectForKey: @"KEY"];

  if ( !aKey && [[NSUserDefaults standardUserDefaults] objectForKey: @"RECEIVING"] )
    {
      NSAutoreleasePool *pool;

      NSArray *allServers;
      int i;
      
      pool = [[NSAutoreleasePool alloc] init];

      allServers = [[[NSUserDefaults standardUserDefaults] objectForKey: @"RECEIVING"] allKeys];
      
      for (i = 0; i < [allServers count]; i++)
	{
	  NSMutableDictionary *allValues; 

	  allValues = [[[NSUserDefaults standardUserDefaults] objectForKey: @"RECEIVING"] 
			objectForKey: [allServers objectAtIndex: i]];

	  if ( [[allValues objectForKey: @"SERVERTYPE"] intValue] == UNIX )
	    {
	      LocalMailDelivery *aLocalMailDelivery;
	      NSAutoreleasePool *inner_pool;
	      NSArray *anArray;
	      int j;

	      aLocalMailDelivery = [[LocalMailDelivery alloc] initWithPathToSpool:
								[allValues objectForKey: @"MAILSPOOLFILE"]];

	      anArray = [aLocalMailDelivery messagesFromMailspoolFile];
	      transferredMessageCount += [anArray count];

	      inner_pool = nil;
	       
	      for (j = 0; j < [anArray count]; j++)
		{
		  if ( (j % 3) == 0 )
		    {
		      TEST_RELEASE(inner_pool);
		      inner_pool = [[NSAutoreleasePool alloc] init];
		    }
		  
		  [self _matchFilterRuleFromRawSource: [anArray objectAtIndex: j]
			connection: serverConnection
			key: [allServers objectAtIndex: i]];
		}

	      TEST_RELEASE(inner_pool);
	    }
	} // for (...)

      RELEASE(pool);
      
      aKey = [NSString stringWithString: @"the local mail spool file"];
    }
  else
    {
      NSNumber *serverTypeValue, *portValue, *retainPeriodValue;
      NSString *aServerName;

      // We get our values associated with the key
      allPOP3Values = [[[NSUserDefaults standardUserDefaults] objectForKey: @"RECEIVING"] objectForKey: aKey];
      
      serverTypeValue = [allPOP3Values objectForKey: @"SERVERTYPE"];
      aServerName = [allPOP3Values objectForKey: @"SERVERNAME"];
      portValue = [allPOP3Values objectForKey: @"PORT"];
      retainPeriodValue = [allPOP3Values objectForKey: @"RETAINPERIOD"];
      
      // We use the default POP3 port if it's not defined.
      if ( !portValue )
	{
	  portValue = [NSNumber numberWithInt: 110];
	}
      
      // We use the default retain period (365 days, if it's not defined)
      if ( !retainPeriodValue)
	{
	  retainPeriodValue = [NSNumber numberWithInt: 365];
	}
      
      if ( !serverTypeValue || [serverTypeValue intValue] == POP3 )
	{
	  NSString *aUsername, *aPassword;
	  
	  // We get our username
	  aUsername = [allPOP3Values objectForKey: @"USERNAME"];
	  
	  // We get our password. We prompt the user if we need to.
	  aPassword = [Utilities passwordForKey: aKey
				 prompt: NO];
	  
	  // 
	  if ( aPassword )
	    {
	      BOOL leaveOnServer;
	      int transferCount;
	      
	      leaveOnServer = ([[allPOP3Values objectForKey: @"LEAVEONSERVER"] intValue] == NSOnState ? YES : NO);
	      transferCount = 0;

	      // We expection exceptions...
	      NS_DURING
		  if ( [allPOP3Values objectForKey: @"USESECURECONNECTION"] &&
		       [[allPOP3Values objectForKey: @"USESECURECONNECTION"] intValue] == NSOnState )
		    {
		      aStore = (POP3Store *) [[POP3Store alloc] initSSLWithName: aServerName
								port: [portValue intValue]];
		    }
		  else
		    {
		      aStore = (POP3Store *) [[POP3Store alloc] initWithName: aServerName
								port: [portValue intValue]];
		      [[aStore tcpConnection] setStopTarget: self];
		      [[aStore tcpConnection] setStopSelector: @selector(mustStopTransfer)];
		    }
		  
		  // We verify that we got a valid connection with the POP3 server
		  if ( !aStore )
		    {
		      [(id)[serverConnection rootProxy] communicationErrorOccuredOnServer: aServerName];
		    }
		  else
		    {
		      // We verify if we should use APOP mode
		      if ( [allPOP3Values objectForKey: @"USEAPOP"] )
			{
			  [aStore setUseAPOP: ([[allPOP3Values objectForKey: @"USEAPOP"] intValue] == NSOnState ? YES : NO) ];
			}
		      
		      if ( [aStore authenticateWithUsername: aUsername
				   password: aPassword] )
			{
			  POP3CacheManager *aPOP3CacheManager;		      
			  
			  // We get our POP3 folder
			  aFolder = (POP3Folder *) [aStore folderForName: @"Inbox" prefetch: NO];
			  [aFolder setLeaveOnServer: leaveOnServer];
			  [aFolder setRetainPeriod: [retainPeriodValue intValue]];
			  
			  // We get our POP3 cache
			  aPOP3CacheManager = [POP3CacheManager pop3CacheObjectsFromDiskWithPath:
								  [NSString stringWithFormat: @"%@/POP3Cache_%@",
									    GNUMailUserLibraryPath(),
									    [Utilities flattenPathFromString: aKey
										       separator: @"/"]]];
			  [aFolder setPOP3CacheManager: aPOP3CacheManager];
			  
			  transferCount = [self _transferMessagesFromPOP3Folder: aFolder
						connection: serverConnection
						key: aKey];
			  
			  transferredMessageCount += transferCount;
			  
			  if ( transferCount == 0 &&
			       !initiatedFromTimer &&
			       !verifyingOnStartup )
			    {
			      // We show our "No New Messages" panel, if the user wants them!
			      if ( ![[NSUserDefaults standardUserDefaults] 
				      objectForKey: @"SHOW_NO_NEW_MESSAGES_PANEL"] ||
				   [[NSUserDefaults standardUserDefaults]
				     integerForKey: @"SHOW_NO_NEW_MESSAGES_PANEL"] == NSOnState )
				{
				  [(id)[serverConnection rootProxy] noNewMessagesOnServer: aServerName
				       username: aUsername];
				}
			      
			      [[NSDistributedNotificationCenter defaultCenter]
				postNotificationName: GNUMailConsoleMessage
				object: nil
				userInfo: [NSDictionary dictionaryWithObject: [NSString stringWithFormat: _(@"No new messages on server %@"),
											aServerName]
							forKey: @"Message"]];
			    }
			  
			  // We now close our folder, Pantomime does nothing for now with this.
			  [aFolder expunge: NO];
			  [aFolder close];
			  
			} // if ( [aStore authenticateWithUsername... ] )
		      else
			{
			  // We remove the password from the cache
			  [[Utilities passwordCache] removeObjectForKey: aKey];
			  
			  // We inform our main thread the the password isn't valid for this server
			  [(id)[serverConnection rootProxy] invalidPasswordOnServer: aServerName
			       username: aUsername];
			}
		  
		      
		      // We now close our store
		      [aStore close];
		      RELEASE(aStore);
		    } // else of if ( !aStore ) ...
		  
		  NS_HANDLER
		    [[NSDistributedNotificationCenter defaultCenter]
		      postNotificationName: GNUMailConsoleMessage
		      object: nil
		      userInfo: [NSDictionary dictionaryWithObject: [NSString stringWithFormat: _(@"Got a %@ while estabilishing (or closing) the POP3 connection to server %@."),
									      [localException name], aServerName]
					      forKey: @"Message"]];
		  NS_ENDHANDLER
		  
	    } // if ( password ) ...
	        
	} // if (!serverTypeValue || ... )
    }
  
  // If we have received some messages, and we are done fetching them, let's
  // inform our main thread from this situation
  if ( transferredMessageCount > 0 )
    {
      [(id)[serverConnection rootProxy] messagesWereReceived];
    }
  
  // We inform the main thread that this thread is about to exit
  [(id)[serverConnection rootProxy] threadWillExit];
  
  [[NSDistributedNotificationCenter defaultCenter]
    postNotificationName: GNUMailConsoleMessage
    object: nil
    userInfo: [NSDictionary dictionaryWithObject: [NSString stringWithFormat: _(@"Exiting the fetching thread for %@..."), aKey]
			    forKey: @"Message"]];
  
  // We clean our connection
  NS_DURING
    [serverConnection invalidate];
  NS_HANDLER
    NSDebugLog(@"LocalInboxWindowController: Exception occured while invalidating the thread's connection object.");
  NS_ENDHANDLER
  
  RELEASE(serverConnection);

  // We release our autorelease pool used in that thread and we exit that thread!
  RELEASE(threadpool);
  
  //[NSThread exit];
}


//
// This method is used to transfer message from a POP3 folder
// to a local folder.
//
// It verifies for various filters to be sure to transfer to the
// right folder.
//
- (int) _transferMessagesFromPOP3Folder: (POP3Folder *) theFolder
			     connection: (NSConnection *) theConnection
				    key: (NSString *) theKey
{
  NSData *aData;
 
  int i, count, transferCount;

  if ( !theFolder )
    {
      return 0;
    }

  //NSDebugLog(@"Entering...");
  count = [theFolder count];
  transferCount = 0;

  NS_DURING
    // We get all messages..
    for (i = 1; i <= count; i++)
      {
	aData = [theFolder prefetchMessageAtIndex: i];
	// If it's a new message (the string isn't nil when it's a new message)
	if ( aData )
	  {
	    [self _matchFilterRuleFromRawSource: aData
		  connection: theConnection
		  key: theKey];
	    
	    transferCount++;
	  }
      } // for 
  
    NS_HANDLER
      [[NSDistributedNotificationCenter defaultCenter]
	postNotificationName: GNUMailConsoleMessage
	object: nil
	userInfo: [NSDictionary dictionaryWithObject: 
				  [NSString stringWithFormat: _(@"Got a %@ while transferring mails from %@."),
					    [localException name], [(POP3Store*)[theFolder store] name]]
				forKey: @"Message"]];

    NS_ENDHANDLER

  return transferCount;
}


//
//
//
- (void) _updateMessageSelectionUsingIndex: (int) theIndex
{
  // We show the message
  if (theIndex > ([folder count] - 1) || 
      (theIndex - 1) >= 0)
    {
      theIndex = theIndex - 1;
    }
  else if ( (theIndex - 1) < 0 && 
	    [folder count] > 0)
    { 
      theIndex = 0;
    }
  
  if ([folder count] > 0)
    {
      [tableView selectRow: theIndex 
		 byExtendingSelection: NO];
      
      // We post a notification in case we transferred the first messages (index == 0);
      [self tableViewSelectionDidChange: [NSNotification 
             				   notificationWithName: NSTableViewSelectionDidChangeNotification
      					   object: tableView] ];
    }
  else
    {
      [textView setString: @""];
    }
}


@end
