/*
**  MailboxManagerController.m
**
**  Copyright (C) 2001, 2002
**
**  Author: Ludovic Marcotte <ludovic@Sophos.ca>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#import "MailboxManagerController.h"

#import "ExtendedOutlineView.h"
#import "GNUMail.h"
#import "GNUMailConstants.h"

#ifndef MACOSX
#import "MailboxManager.h"
#endif

#import "FolderNode.h"
#import "MailboxManagerCache.h"
#import "MailWindowController.h"
#import "MessageViewWindowController.h"
#import "NewMailboxPanelController.h"
#import "NSUserDefaultsExtensions.h"
#import "PasswordPanelController.h"
#import "Utilities.h"

#import <Pantomime/Constants.h>
#import <Pantomime/IMAPCacheManager.h>
#import <Pantomime/IMAPFolder.h>
#import <Pantomime/IMAPStore.h>
#import <Pantomime/LocalFolder.h>
#import <Pantomime/LocalStore.h>
#import <Pantomime/Message.h>
#import <Pantomime/NSDataExtensions.h>
#import <Pantomime/TCPConnection.h>
#import <Pantomime/URLName.h>


static MailboxManagerController *singleInstance = nil;

// 
// Here's how it does work:
//
// allFolders (NSArray) -> localNodes (FolderNode * - name == _("Local"))
//
//                      -> IMAP FolderNode 1 (FoderNode * - name == "username @ imap.server1.com")  
//                           
//                      -> IMAP FolderNode 2 (FoderNode * - name == "username @ imap.server2.com")
// 
//                      -> ...   
//                                 
//
@implementation MailboxManagerController

- (id) initWithWindowNibName: (NSString *) windowNibName
{
#ifdef MACOSX
  
  self = [super initWithWindowNibName: windowNibName];
  
#else
  MailboxManager *theWindow;
  
  theWindow = 
    [[MailboxManager alloc] initWithContentRect: NSMakeRect(200,200,400,435)
			    styleMask: (NSClosableWindowMask |
					NSTitledWindowMask |
					NSMiniaturizableWindowMask |
					NSResizableWindowMask)
			    backing: NSBackingStoreBuffered
			    defer: YES];

  self = [super initWithWindow: theWindow];

  [theWindow layoutWindow];
  [theWindow setDelegate: self];

  // We link our outlets
  move = [theWindow move];
  copy = [theWindow copy];
  outlineView = [theWindow outlineView];
  scrollView = [theWindow scrollView];
  openOnStartup = [theWindow openOnStartup];
  icon = [theWindow icon];

  RELEASE(theWindow);
#endif

  // We set the title of our window (causing it to be loaded under OS X)
  [[self window] setTitle: _(@"Mailboxes")];

  // We register the outline view for dragged types
  [outlineView registerForDraggedTypes: [NSArray arrayWithObject: MessagePboardType]];

  // We set our autosave window frame name and restore the one from the user's defaults.
  [[self window] setFrameAutosaveName: @"MailboxManager"];
  [[self window] setFrameUsingName: @"MailboxManager"]; 

  // We set our autosave name for our outline view
  [outlineView setAutosaveName: @"MailboxManager"];
  [outlineView setAutosaveTableColumns: YES];

  // We set our outline view background color
  if ( [[NSUserDefaults standardUserDefaults] colorForKey: @"MAILBOXMANAGER_OUTLINE_COLOR"] )
    {
      [outlineView setBackgroundColor: [[NSUserDefaults standardUserDefaults]
					 colorForKey: @"MAILBOXMANAGER_OUTLINE_COLOR"]];
      [scrollView setBackgroundColor: [[NSUserDefaults standardUserDefaults]
					colorForKey: @"MAILBOXMANAGER_OUTLINE_COLOR"]];
    }

  return self;
}


//
//
//
#ifdef MACOSX
- (void) awakeFromNib
{
  NSToolbar *toolbar;
  
  toolbar = [[NSToolbar alloc] initWithIdentifier: @"MailboxManagerToolbar"];
  [toolbar setDelegate: self];
  [toolbar setAllowsUserCustomization: YES];
  [toolbar setAutosavesConfiguration: YES];
  [[self window] setToolbar: toolbar];
  RELEASE(toolbar);
}
#endif

//
//
//
- (void) dealloc
{
  NSEnumerator *theEnumerator;
  id aStore;

  //NSDebugLog(@"MailboxManagerController: -dealloc");

  // Cocoa bug?
#ifdef MACOSX
  [outlineView setDataSource: nil];
  [[self window] setToolbar: nil];
#endif

  RELEASE(mailboxManagerCache);
  RELEASE(allFolders);
  
  NSDebugLog(@"Closing all Stores...");
  theEnumerator = [allStores objectEnumerator];

  while ( (aStore = [theEnumerator nextObject]) )
    {
      [aStore close];
    }

  RELEASE(allStores);

  [[NSNotificationCenter defaultCenter]
    removeObserver: self
    name: ReceivingValuesHaveChanged
    object: nil];
  
  [super dealloc];
}


//
// Datasource methods for the outline view
//
- (id) outlineView: (NSOutlineView *) outlineView
	     child: (int) index
	    ofItem: (id) item
{
  
  // root object
  if ( !item )
    {
      return allFolders;
    }
  
  if ( item == allFolders )
    {
      return [allFolders objectAtIndex: index];
    }

  if ( [item isKindOfClass: [FolderNode class]] )
    {
      return [(FolderNode *)item childAtIndex: index];
    }
 
  return nil;
}


//
//
//
- (BOOL) outlineView: (NSOutlineView *) outlineView
    isItemExpandable: (id) item
{
  if ( item == allFolders || [allFolders containsObject: item] )
    {
      return YES;
    }
  
  if ( [item isKindOfClass: [FolderNode class]] )
    {
      if ( [(FolderNode *)item childCount] > 0 )
	{
	  return YES;
	}
      else
	{
	  return NO;
	}
    }

  return NO;
}


//
//
//
- (int)        outlineView: (NSOutlineView *) outlineView 
    numberOfChildrenOfItem: (id) item
{
  // Root, always one element.
  if ( !item  )
    {
      return 1;
    }
 
  if ( item == allFolders )
    {
      return [allFolders count];
    }
 
  // Children of our root, the Local folder and all the IMAP folders, subfolders, etc.
  if ( [item isKindOfClass: [FolderNode class]] )
    {
      return [(FolderNode *)item childCount];
    }
  
  return 0;
}


//
//
//
- (id)         outlineView: (NSOutlineView *) outlineView 
 objectValueForTableColumn: (NSTableColumn *) tableColumn 
		    byItem: (id) item
{
  if ( [[[tableColumn headerCell] stringValue] isEqual: _(@"Mailbox")] )
    {
      if ( item == allFolders )
	{
	  return _(@"All Mailboxes");
	}
      
      if ( [item isKindOfClass: [FolderNode class]] )
	{
	  return [(FolderNode *)item name];
	}
    }
  
  if ( [item isKindOfClass: [FolderNode class]] && [item parent] )
    {
      int nbOfMessages, nbOfUnreadMessages, totalSize;
     
      [self _nbOfMessages: &nbOfMessages
	    nbOfUnreadMessages: &nbOfUnreadMessages
	    totalSize: &totalSize
	    forItem: item];
	   
      if ( [[[tableColumn headerCell] stringValue] isEqual: _(@"Size")] )
	{
	  if ( totalSize == 0 )
	    {
	      return @"";
	    }
	  else
	    {
	      return [NSString stringWithFormat: _(@"%d KB"), totalSize/1024];
	    }
	}
      else if ( [[[tableColumn headerCell] stringValue] isEqual: _(@"Messages")] )
	{
	  if ( nbOfMessages == 0 )
	    {
	      return @"";
	    }
	  else
	    {
	      return [NSString stringWithFormat: _(@"%d (%d unread)"), nbOfMessages, nbOfUnreadMessages];
	    }
	}
    }  
 
  return nil;
}


//
//
//
- (void) outlineView: (NSOutlineView *) aOutlineView
     willDisplayCell: (id) aCell
      forTableColumn: (NSTableColumn *) aTableColumn
                item: (id) item
{
  if ( [item isKindOfClass: [FolderNode class]] && [item parent] )
    {
      int nbOfMessages, nbOfUnreadMessages, totalSize;
      
      [self _nbOfMessages: &nbOfMessages
	    nbOfUnreadMessages: &nbOfUnreadMessages
	    totalSize: &totalSize
	    forItem: item];

      if ( nbOfUnreadMessages > 0 )
	{
#ifdef MACOSX
	  [aCell setFont: [NSFont boldSystemFontOfSize: [NSFont smallSystemFontSize]]];
#else
	  [aCell setFont: [NSFont boldSystemFontOfSize: 0]];
#endif
	  return;
	}
    }
  
  // We set our default font.
#ifdef MACOSX
  [aCell setFont: [NSFont systemFontOfSize: [NSFont smallSystemFontSize]]];
#else
  [aCell setFont: [NSFont systemFontOfSize: 0]];
#endif

  // We set the right text aligment
  if ( [[[aTableColumn headerCell] stringValue] isEqual: _(@"Mailbox")] )
    {
      [aCell setAlignment: NSLeftTextAlignment];
    }
  else
    {
      [aCell setAlignment: NSRightTextAlignment];
    }   
}


//
// Delegate methods
//
- (void) outlineViewSelectionIsChanging: (NSNotification *) aNotification
{
  NSMutableArray *foldersToOpen;
  NSString *aString, *pathOfFolder;
  id item;
  
  int row, level;

  NSDebugLog(@"outlineViewSelectionIsChanging... %d", [openOnStartup state]);
  
  row = [outlineView selectedRow];

  // We verify if we have at least one selected row
  if ( row < 0 )
    {
      [openOnStartup setState: NSOffState];
      [openOnStartup setEnabled: NO];
      return;
    }

  item = [outlineView itemAtRow: row];
  level = [outlineView levelForItem: item];

  // We verify that we have not more than a selected row and that we haven't
  // selected a store
  if ( [outlineView numberOfSelectedRows] != 1 ||
       level < 2 )
    {
      [openOnStartup setState: NSOffState];
      [openOnStartup setEnabled: NO];
      return;
    }

   // We create our mutable array from all folders to open
  foldersToOpen = [NSMutableArray arrayWithArray: [[NSUserDefaults standardUserDefaults] 
						    objectForKey: @"FOLDERS_TO_OPEN"]];

  pathOfFolder = [Utilities pathOfFolderFromFolderNode: item
			    separator: @"/"];

  // We verify if we have selected a local mailbox
  if ( [[Utilities completePathForFolderNode: item  separator: @"/"]
	 hasPrefix: [NSString stringWithFormat: @"/%@", _(@"Local")]] )
    { 
      aString = [NSString stringWithFormat: @"local://%@/%@",
			  [[NSUserDefaults standardUserDefaults] objectForKey: @"LOCALMAILDIR"],
			  pathOfFolder];
    }
  // The selected mailbox is an IMAP mailbox
  else
    {
      NSString *aUsername, *aServerName;
      
      [Utilities storeKeyForFolderNode: item
		 serverName: &aServerName
		 username: &aUsername];
     
      aString = [NSString stringWithFormat: @"imap://%@@%@/%@", aUsername, aServerName,
			  pathOfFolder];
    }
  
  // We update the state of our button
  if ( [foldersToOpen containsObject: aString] )
    {
      [openOnStartup setState: NSOnState];
    }
  else
    {
      [openOnStartup setState: NSOffState];
    }

  [openOnStartup setEnabled: YES];
}


//
//
//
- (BOOL) outlineView: (NSOutlineView *) outlineView
  shouldCollapseItem: (id) item
{
  if ( [allFolders containsObject: item] &&
       item != localNodes )
    {
      NSString *aUsername, *aServerName;
      IMAPStore *aStore;
      
      [Utilities storeKeyForFolderNode: item
		 serverName: &aServerName
		 username: &aUsername];
      
      aStore = [self storeForName: aServerName
		     username: aUsername];

      if ( [[NSUserDefaults standardUserDefaults] objectForKey: @"REUSE_MAILWINDOW"] &&
	   [[[NSUserDefaults standardUserDefaults] objectForKey: @"REUSE_MAILWINDOW"] intValue] == NSOnState &&
	   [[aStore tcpConnection] isConnected] )
	{
	  int choice;

	  choice = NSRunAlertPanel( _(@"Closing..."),
				    _(@"Would you like to close the connection with this IMAP server?"),
				    _(@"No"),  // default
				    _(@"Yes"), // alternate
				    NULL );
	  
	  if ( choice == NSAlertAlternateReturn )
	    {
	      [aStore close];
	      
	      [item setChildren: nil];

	      [allStores removeObjectForKey: [(FolderNode *)item name]];
	      
	      [[NSDistributedNotificationCenter defaultCenter]
		postNotificationName: GNUMailConsoleMessage
		object: nil
		userInfo: [NSDictionary dictionaryWithObject:
					  [NSString stringWithFormat: _(@"Closing IMAP connection on %@."),
						    aServerName] forKey: @"Message"]];
	    }
	}
      
    }

  return YES;
}


//
//
//
- (BOOL) outlineView: (NSOutlineView *) outlineView
    shouldExpandItem: (id) item
{
  if ( item == allFolders || item == localNodes )
    {
      return YES;
    }
  
  if ( [allFolders containsObject: item] )
    {
      NSString *aString;

      aString = [(FolderNode *)item name];    // returns username @ imap.server.com

      return [self initializeIMAPStoreWithKey: aString];
    }
  
  return YES;
}


//
// NSOutlineViewDataSource Drag and drop
//
- (NSDragOperation) outlineView: (NSOutlineView*) outlineView
		   validateDrop: (id <NSDraggingInfo>) info
		   proposedItem: (id) item
	     proposedChildIndex: (int) index
{
  if (! [item respondsToSelector: @selector(childCount)] ||
      [(FolderNode*)item childCount]  > 0)
    {
      return NSDragOperationNone;
    }

  if ([info draggingSourceOperationMask] & NSDragOperationGeneric)
    {
      return NSDragOperationGeneric;
    }
  else if ([info draggingSourceOperationMask] & NSDragOperationCopy)
    {
      return NSDragOperationCopy;
    }
  else
    {
      return NSDragOperationNone;
    }
}


//
// NSOutlineViewDataSource Drag and drop
//
- (BOOL) outlineView: (NSOutlineView*) outlineView
	  acceptDrop: (id <NSDraggingInfo>) info
		item: (id) item
	  childIndex: (int) index
{
  NSArray *propertyList;
  int count, i;
  
  FolderNode *aFolderNode;
  NSString *aFolderName;

  Folder *sourceFolder, *destinationFolder;
  Store *sourceStore, *destinationStore;
  MailWindowController *aMailWindowController;

  aFolderNode = (FolderNode*)item;
  aFolderName = [Utilities pathOfFolderFromFolderNode: aFolderNode
			   separator: @"/"];

  // We get our store
  if ( [[Utilities completePathForFolderNode: aFolderNode  separator: @"/"]
	 hasPrefix: [NSString stringWithFormat: @"/%@", _(@"Local")]] )
    {
      destinationStore = [self storeForName: @"GNUMAIL_LOCAL_STORE"
			       username: NSUserName()];
    }
  else
    {
      NSString *aServerName, *aUsername;
      
      [Utilities storeKeyForFolderNode: aFolderNode
		 serverName: &aServerName
		 username: &aUsername];
      
      destinationStore = [self storeForName: aServerName
			       username: aUsername];
    }

  // We get the MailWindowController source
  aMailWindowController = (MailWindowController *)[[info draggingSource] delegate];
  
  if ( !aMailWindowController || ![aMailWindowController isKindOfClass: [MailWindowController class]] )
    {
      NSBeep();
      return NO;
    }
  
  // We verify if we aren't trying to transfer to the current mbox!
  sourceFolder = [aMailWindowController folder];
  sourceStore = [sourceFolder store];
  
  if ( sourceStore == destinationStore && [[sourceFolder name] isEqualToString: aFolderName] )
    {
      NSRunInformationalAlertPanel(_(@"Transfer error!"),
				   _(@"You cannot transfer a message inside the same mailbox!"),
				   _(@"OK"),
				   NULL, 
				   NULL,
				   NULL);
      return NSDragOperationNone;      
    }

  // We get a reference to our destination folder, w/o parsing it if it's not already open.
  // or w/o selecting it if it's an IMAP store.
  if ( [(id<NSObject>)destinationStore isKindOfClass: [IMAPStore class]] )
    {
      destinationFolder = (Folder *)[(IMAPStore *)destinationStore folderForName: aFolderName
						  select: NO];
    }
  else
    {
      destinationFolder = (Folder *)[(LocalStore *)destinationStore folderForName: aFolderName];
    }

  propertyList = [[info draggingPasteboard] propertyListForType: MessagePboardType];
  count = [propertyList count];

  for (i = 0; i < count; i++)
    {
      NSDictionary *aDictionary;
      Flags *flags;
      NSData *rawSource;
      
      aDictionary = (NSDictionary*)[propertyList objectAtIndex: i];
      
      rawSource = [aDictionary objectForKey: @"message"];
      flags = (Flags*)[NSUnarchiver unarchiveObjectWithData: (NSData*)[aDictionary objectForKey: @"flags"]];

      if (rawSource && flags)
	{
	  if ( ![self  transferMessageFromRawSource: rawSource
		       flags: flags
		       toFolderWithName: aFolderName
		       orToFolder: destinationFolder
		       store: destinationStore] )
	    {
	      // The transfer failed; we exit the loop
	      break;
	    }
	}
    }

  // If the message is moved (and not copied),
  // we inform the dragging source that the selected messages have been transferred
  if ( count > 0 && 
       count == i && 
       [info draggingSourceOperationMask] & NSDragOperationGeneric )
    {
      // We mark messages as transferred if append succeed
      [aMailWindowController setSelectedMessagesAsTransferred];
    }

  // If the folder was not open, we just close it.
  if ( destinationFolder )
    {
      [destinationFolder close];
    }

  return YES;
}


//
//
//
- (void) windowWillClose: (NSNotification *) theNotification
{
  // We do nothing for now...
  NSDebugLog(@"Closing the MailboxWindow...");
}


//
//
//
- (void) windowDidLoad
{
  MailboxManagerCache *aMailboxManagerCache;

  NSDebugLog(@"MailboxManager window loaded!");
 
  // We initialize our mailbox manager cache
  aMailboxManagerCache = [MailboxManagerCache mailboxManagerCacheFromDisk];

  if ( !aMailboxManagerCache )
    {
      NSDebugLog(@"Creating a new Mailbox Manager cache.");
      aMailboxManagerCache = [[MailboxManagerCache alloc] init];
      [aMailboxManagerCache synchronize];

      AUTORELEASE(aMailboxManagerCache);
    }

  [self setMailboxManagerCache: aMailboxManagerCache];


  // We initialize our array containing all Stores and we load of folders
  allFolders = [[NSMutableArray alloc] init];
  
  // We initialize our dictionary containing all openend IMAPStores
  allStores = [[NSMutableDictionary alloc] init];

#ifdef MACOSX
  [outlineView setDoubleAction: @selector(open:)];
  [outlineView reloadData];
#endif

  [[NSNotificationCenter defaultCenter]
    addObserver: self
    selector: @selector(receivingValuesHaveChanged:)
    name: ReceivingValuesHaveChanged
    object: nil];
}


//
// action methods
//
- (IBAction) openOnStartupClicked: (id) sender
{
  NSMutableArray *foldersToOpen;
  NSString *aString, *pathOfFolder;
  id item;
  
  int row, level;

  NSDebugLog(@"openOnStartupClicked... %d", [openOnStartup state]);
  
  row = [outlineView selectedRow];

  // We verify if we have at least one selected row
  if ( row < 0 )
    {
      NSBeep();
      return;
    }

  item = [outlineView itemAtRow: row];
  level = [outlineView levelForItem: item];

  // We verify that we have not more than a selected row and that we haven't
  // selected a store
  if ( [outlineView numberOfSelectedRows] != 1 ||
       level < 2 )
    {
      NSBeep();
      [openOnStartup setState: NSOffState];
      return;
    }

  // We create our mutable array from all folders to open
  foldersToOpen = [NSMutableArray arrayWithArray: [[NSUserDefaults standardUserDefaults] 
						    objectForKey: @"FOLDERS_TO_OPEN"]];

  pathOfFolder = [Utilities pathOfFolderFromFolderNode: item
			    separator: @"/"];

  // We verify if we have selected a local mailbox
  if (  [[Utilities completePathForFolderNode: item  separator: @"/"]
	  hasPrefix: [NSString stringWithFormat: @"/%@/", _(@"Local")]] )
    {
      LocalStore *aStore;
      
      aStore = [self storeForName: @"GNUMAIL_LOCAL_STORE"
		     username: NSUserName()];

      if ( [aStore folderTypeForFolderName: pathOfFolder] == HOLDS_FOLDERS )
	{
	  NSBeep();
	  [openOnStartup setState: NSOffState];
	  return;
	}

      aString = [NSString stringWithFormat: @"local://%@/%@",
			  [[NSUserDefaults standardUserDefaults] objectForKey: @"LOCALMAILDIR"],
			  pathOfFolder];
    }
  // The selected mailbox is an IMAP mailbox
  else
    {
      NSString *aServerName, *aUsername;
      IMAPStore *aStore;

      [Utilities storeKeyForFolderNode: item
		 serverName: &aServerName
		 username: &aUsername];

      aStore = [self storeForName: aServerName
		     username: aUsername];

      if ( [aStore folderTypeForFolderName: pathOfFolder] == HOLDS_FOLDERS )
	{
	  NSBeep();
	  [openOnStartup setState: NSOffState];
	  return;
	}
      
      aString = [NSString stringWithFormat: @"imap://%@@%@/%@", aUsername, aServerName, 
			  pathOfFolder];
    }

  // We now add/remove the folder depending on the state of our button
  if ( [openOnStartup state] == NSOnState )
    {
      if ( ![foldersToOpen containsObject: aString] )
	{
	  [foldersToOpen addObject: aString];
	}
    }
  else
    {
      [foldersToOpen removeObject: aString];
    }

  // We save our FOLDERS_TO_OPEN value in the user's defaults
  [[NSUserDefaults standardUserDefaults] setObject: foldersToOpen
					 forKey: @"FOLDERS_TO_OPEN"];
  [[NSUserDefaults standardUserDefaults] synchronize];
}


//
//
//
- (IBAction) open: (id) sender
{
  id item;
  int row, level;
  
  row = [outlineView selectedRow];

  if ( row < 0 )
    {
      NSBeep();
      return;
    }

  item = [outlineView itemAtRow: row];
  level = [outlineView levelForItem: item];
  
  //
  // We must verify that:
  //
  // a) we have at least one selected row
  // b) we haven't selected our root, or a store (local or IMAP)
  //
  if ( [outlineView numberOfSelectedRows] != 1 )
    {
      NSRunInformationalAlertPanel(_(@"Mailbox error!"),
				   _(@"You must select a valid mailbox to open!"),
				   _(@"OK"),
				   NULL, 
				   NULL,
				   NULL);
      return;
    }
  else if ( level < 2 )
    {
      if ( [outlineView isItemExpanded: item] )
	{
	  [outlineView collapseItem: item];
	}
      else
	{
	  [outlineView expandItem: item];
	}
      return;
    }
 
  
  // We verify if it's a local folder
  if ( [[Utilities completePathForFolderNode: item  separator: @"/"] 
	 hasPrefix: [NSString stringWithFormat: @"/%@", _(@"Local")]] )
    { 
      NSString *aString;
      
      aString = [Utilities pathOfFolderFromFolderNode: (FolderNode *)item
			   separator: @"/"];

      [self openLocalFolderWithName: aString];
    }
  // It's an IMAP folder...
  else
    {
      NSString *aServerName, *aUsername, *aString;
      IMAPStore *aStore;
      
      [Utilities storeKeyForFolderNode: item
		 serverName: &aServerName
		 username: &aUsername];
      
      aStore = (IMAPStore *)[self storeForName: aServerName
				  username: aUsername];
      
      aString = [Utilities pathOfFolderFromFolderNode: (FolderNode *)item
			   separator: [aStore folderSeparator]];
      
      [self openIMAPFolderWithName: aString
	    store: aStore];
    }
}


//
//
//
- (IBAction) transfer: (id) sender
{
  Folder *destinationFolder, *sourceFolder;
  Store *destinationStore, *sourceStore;

  MailWindowController *aMailWindowController;
  NSString *aFolderName;
  
  NSArray *selectedMessages;
  
  int row, level, nbOfTransferredMessages;
  id item;
  
  nbOfTransferredMessages = 0;
  row = [outlineView selectedRow];
  
  if ( row < 0 )
    {
      NSBeep();
      return;
    }

  item = [outlineView itemAtRow: row];
  level = [outlineView levelForItem: item];
  
  if ( [outlineView numberOfSelectedRows] != 1 ||
       level < 2 )
    {
      NSRunInformationalAlertPanel(_(@"Mailbox error!"),
				   _(@"You must select a valid mailbox\nto transfer this E-Mail to!"),
				   _(@"OK"),
				   NULL, 
				   NULL,
				   NULL);
      return;
    }


  if ( [(FolderNode *)item childCount] > 0)
    {
      NSRunInformationalAlertPanel(_(@"Mailbox error!"),
				   _(@"You cannot transfer E-Mails to this folder!"),
				   _(@"OK"),
				   NULL, 
				   NULL,
				   NULL);
      return;
    }
  

  // We get our store!
  if ( [[Utilities completePathForFolderNode: item  separator: @"/"]
	 hasPrefix: [NSString stringWithFormat: @"/%@", _(@"Local")]] )
    {
      destinationStore = [self storeForName: @"GNUMAIL_LOCAL_STORE"
			       username: NSUserName()];
    }
  else
    {
      NSString *aServerName, *aUsername;
      
      [Utilities storeKeyForFolderNode: item
		 serverName: &aServerName
		 username: &aUsername];

      destinationStore = [self storeForName: aServerName
			       username: aUsername];
    }
  
  // We get our folder name, respecting the folder separator.
  aFolderName = [Utilities pathOfFolderFromFolderNode: (FolderNode *) item
			   separator: [(id<Store>)destinationStore folderSeparator]];
  
  aMailWindowController = (MailWindowController *)[[GNUMail lastMailWindowOnTop] delegate];

  if ( !aMailWindowController ) 
    {
      NSBeep();
      return;
    }
  
  // We verify if we aren't trying to transfer to the current mbox!
  sourceFolder = [aMailWindowController folder];
  sourceStore = [sourceFolder store];
  if ( sourceStore == destinationStore && [[sourceFolder name] isEqualToString: aFolderName] )
    {
      NSRunInformationalAlertPanel(_(@"Transfer error!"),
				   _(@"You cannot transfer a message inside the same mailbox!"),
				   _(@"OK"),
				   NULL, 
				   NULL,
				   NULL);
      return;      
    }
  
  // We get a reference to our destination folder, w/o parsing it if it's not already open.
  // or w/o selecting it if it's an IMAP store.
  if ( [(id<NSObject>)destinationStore isKindOfClass: [IMAPStore class]] )
    {
      destinationFolder = (Folder *)[(IMAPStore *)destinationStore folderForName: aFolderName
						  select: NO];
    }
  else
    {
      destinationFolder = (Folder *)[(LocalStore *)destinationStore folderForName: aFolderName];
    }
    
  
  // We transfer all the selected messages  
  selectedMessages = [aMailWindowController selectedMessages];
  
  if ( selectedMessages )
    {
      Message *aMessage;
      NSData *aData;
      int i;
      
      for (i = 0; i < [selectedMessages count]; i++)
	{
	  aMessage = [selectedMessages objectAtIndex: i];
	  aData = [aMessage rawSource];
	  
	  if ( aData )
	    {
	      Flags *theFlags;
	      
	      // We get our flags but we remove the TRANSFERRED flag from them
	      theFlags = [[aMessage flags] copy];
	      [theFlags remove: TRANSFERRED];
	      
	      if ( [self transferMessageFromRawSource: aData
			 flags: theFlags
			 toFolderWithName: aFolderName
			 orToFolder: destinationFolder
			 store: destinationStore] )
		{
		  // We verify if we flag it as TRANSFERRED or not (in case of a
		  // copy operation).
		  if ( [sender tag] == MOVE_MESSAGES )
		    {
		      [[aMessage flags] add: TRANSFERRED];
		      nbOfTransferredMessages++;
		    }
		}
	      RELEASE(theFlags);
	    }
	}
    }
  
  // We need to reload our tableView's data if we transferred some messages
  if ( nbOfTransferredMessages > 0 )
    {
      // We force the reload of our cache in Folder
      [[aMailWindowController folder] updateCache];
      [aMailWindowController tableViewShouldReloadData];
    }

  // If our folder wasn't open, we just close it
  if ( destinationFolder )
    {
      [destinationFolder close];
    }
}


//
//
//
- (IBAction) create: (id) sender
{
  NewMailboxPanelController *theController;
  id item;

  int row, level, result;
  
  row = [outlineView selectedRow];

  if ( row < 0 )
    {
      NSBeep();
      return;
    }

  item = [outlineView itemAtRow: row];
  level = [outlineView levelForItem: item];
  
  if ( [outlineView numberOfSelectedRows] != 1 ||
       level < 1  )
    {
      NSRunInformationalAlertPanel(_(@"Mailbox error!"),
				   _(@"You must select a valid root where to create this new mailbox."),
				   _(@"OK"),
				   NULL, 
				   NULL,
				   NULL);
      return;
    }
  
  
  // We create our NewMailboxPanelController object. It'll be automatically deallocated when the 
  // window will be closed.
  theController = [[NewMailboxPanelController alloc] initWithWindowNibName: @"NewMailboxPanel"
						     folderNode: (FolderNode *)item];
  [theController setMailboxManagerController: self];

  result = [NSApp runModalForWindow: [theController window]];
 
  if ( result == NSRunStoppedResponse )
    {
      NSString *aString;
      
      aString = [NSString stringWithFormat: @"%@/%@", [Utilities completePathForFolderNode: item
								 separator: @"/"],
			  [[[theController mailboxNameField] stringValue] stringByTrimmingWhiteSpaces] ];
      
      [self _reloadFoldersAndExpandParentsFromNode: item
	    selectNodeWithPath: aString];
    }
 
#ifndef MACOSX
  [[self window] makeKeyAndOrderFront: self];
#endif

  RELEASE(theController);
}


//
//
//
- (IBAction) delete: (id) sender
{
  int choice, row, level;
  id item;
  
  row = [outlineView selectedRow];

  if ( row < 0 )
    {
      NSBeep();
      return;
    }

  item = [outlineView itemAtRow: row];
  level = [outlineView levelForItem: item];
  
  if ( [outlineView numberOfSelectedRows] != 1 ||
       level < 2 )
    {
      NSRunInformationalAlertPanel(_(@"Mailbox error!"),
				   _(@"Please select the mailbox you would like to delete."),
				   _(@"OK"),
				   NULL, 
				   NULL,
				   NULL);
      return;
    }
  
  
  // We show our prompt panel
  choice = NSRunAlertPanel(_(@"Delete..."),
			   _(@"Are you sure you want to delete this mailbox?"),
			   _(@"Cancel"),  // default
			   _(@"Yes"),     // alternate
			   nil);
  
  if ( choice == NSAlertAlternateReturn )
    {
      NSString *aString;
      id aStore;
      
      // We verify if it's a local folder
      if ( [[Utilities completePathForFolderNode: item  separator: @"/"] 
	     hasPrefix: [NSString stringWithFormat: @"/%@", _(@"Local")]] )
	{
	  aStore = [self storeForName: @"GNUMAIL_LOCAL_STORE"
			 username: NSUserName()];
	}
      else
	{
	  NSString *aServerName, *aUsername;
	  
	  [Utilities storeKeyForFolderNode: item
		     serverName: &aServerName
		     username: &aUsername];
	  
	  aStore = [self storeForName: aServerName
			 username: aUsername];
	}
      
      // We get our folder name, respecting the folder separator
      aString = [Utilities pathOfFolderFromFolderNode: (FolderNode *)item
			   separator: [(id<Store>)aStore folderSeparator]];
      
      if ( [aStore folderForNameIsOpen: aString] )
	{
	  NSRunAlertPanel(_(@"Error while deleting!"),
  			  _(@"You can't delete an opened mailbox! Close it first."),
			  _(@"OK"),   // default
			  NULL,       // alternate
  			  NULL); 
	  return;
	}
      else
	{
	  if ( [aStore deleteFolderWithName: aString] )
	    {
	      // Delete cache files, ONLY if this is an IMAP folder!
	      if ( [aStore isKindOfClass: [IMAPStore class]] )
		{
		  int i;
		  FolderNode *node;
		  NSString *aKey, *cacheFilePath;
		  
		  aKey = [NSString stringWithFormat: @"%@ @ %@", 
				   [(IMAPStore *)aStore username], 
				   [(IMAPStore *)aStore name]];
		  
		  cacheFilePath = [NSString stringWithFormat: @"%@/IMAPCache_%@_%@", 
					    GNUMailUserLibraryPath(),
					    [Utilities flattenPathFromString: aKey
						   separator: @"/"],
					    [Utilities flattenPathFromString: [Utilities pathOfFolderFromFolderNode: (FolderNode *)item
											 separator: [(id<Store>)aStore folderSeparator]]
						       separator: @"/"]];
		  
		  NSDebugLog(@"(cache file: %@)", cacheFilePath);
		  if ( [[NSFileManager defaultManager] removeFileAtPath: cacheFilePath
						       handler: nil] )
		    {
		      NSDebugLog(@"Cache file %@ deleted.", cacheFilePath);
		    }
		  
		  for (i = 0; i < [[(FolderNode *)item parent] childCount]; i++)
		    {
		      node = [[(FolderNode *)item parent] childAtIndex: i];
		      cacheFilePath = [NSString stringWithFormat: @"%@/IMAPCache_%@_%@", 
						GNUMailUserLibraryPath(),
						[Utilities flattenPathFromString: aKey
							   separator: @"/"],
						[Utilities flattenPathFromString: [Utilities pathOfFolderFromFolderNode: node
											     separator: [(id<Store>)aStore folderSeparator]]
							   separator: @"/"]];
		      
		      NSDebugLog(@"(cache file: %@)", cacheFilePath);
		      if ( [[NSFileManager defaultManager] removeFileAtPath: cacheFilePath
							   handler: nil] )
			{
			  NSDebugLog(@"Cache file %@ deleted.", cacheFilePath);
			}
		    }
		}
	      
	      [self _reloadFoldersAndExpandParentsFromNode: [item parent]
		    selectNodeWithPath: [Utilities completePathForFolderNode: [item parent]
						   separator: @"/"]];
	    }
	  else
	    {
	      NSRunInformationalAlertPanel(_(@"Error!"),
					   _(@"The mailbox delete operation failed."),
					   _(@"OK"),
					   NULL,
					   NULL,
					   NULL);
	    }
	  
	}
    }
}


//
//
//
- (IBAction) rename: (id) sender
{
  NewMailboxPanelController *theController;
  Store *aStore;
  
  NSString *aFolderName;
  id item;

  int row, level, result;
  
  row = [outlineView selectedRow];

  if ( row < 0 )
    {
      NSBeep();
      return;
    }

  item = [outlineView itemAtRow: row];
  level = [outlineView levelForItem: item];
  
  if ( [outlineView numberOfSelectedRows] != 1 ||
       level < 2 )
    {
      NSRunInformationalAlertPanel(_(@"Mailbox error!"),
				   _(@"You must select a valid mailbox to rename!"),
				   _(@"OK"),
				   NULL, 
				   NULL,
				   NULL);
      return;
    }
     
  // We get our right store
  if ( [[Utilities completePathForFolderNode: item  separator: @"/"]
	 hasPrefix: [NSString stringWithFormat: @"/%@", _(@"Local")]] )
    {
      aStore = [self storeForName: @"GNUMAIL_LOCAL_STORE"
		     username: NSUserName()];
    }
  else
    {
      NSString *aServerName, *aUsername;
      
      [Utilities storeKeyForFolderNode: item
		 serverName: &aServerName
		 username: &aUsername];
      
      aStore = [self storeForName: aServerName
		     username: aUsername];
    }
  
  // We get our foldername, respecting the folder separator
  aFolderName = [Utilities pathOfFolderFromFolderNode: item
			   separator: [(id<Store>)aStore folderSeparator]];
  
  // We verify that the mailbox we want to rename is not open
  if ( [(id<Store>)aStore folderForNameIsOpen: aFolderName] )
    {
      NSRunAlertPanel(_(@"Error!"),
		      _(@"You can't rename an opened mailbox! Close it first."),
		      _(@"OK"),   // default
		      NULL,       // alternate
		      NULL); 
      return;
    }

  
  // We create our NewMailboxPanelController object. It'll be automatically 
  // deallocated when the window will be closed.
  theController = [[NewMailboxPanelController alloc] initWithWindowNibName: @"NewMailboxPanel"
						     folderNode: item];
  [theController setMailboxManagerController: self];
  [theController setName: [(FolderNode*)item name]];
  
  // We set the right window tile and label content
  [[theController window] setTitle: _(@"Edit Mailbox")];
  [[theController mailboxNameLabel] setStringValue: _(@"Please enter the new name of the Mailbox:")];

  result = [NSApp runModalForWindow: [theController window]];

  if ( result == NSRunStoppedResponse )
    {
      NSString *aString;

      aString = [NSString stringWithFormat: @"%@/%@", [Utilities completePathForFolderNode: [item parent]
								 separator: @"/"],
			  [[[theController mailboxNameField] stringValue] stringByTrimmingWhiteSpaces] ];
      
      [self _reloadFoldersAndExpandParentsFromNode: [item parent]
	    selectNodeWithPath: aString];
    }


#ifndef MACOSX
  [[self window] makeKeyAndOrderFront: self];
#endif

  RELEASE(theController);
}


//
//
//
- (void) receivingValuesHaveChanged: (id) sender
{
  [self reloadAllFolders];
}



//
// access / mutation methods
//
- (NSOutlineView *) outlineView
{
  return outlineView;
}


//
//
//
- (id) storeForName: (NSString *) theName
	   username: (NSString *) theUsername
{
  NSString *aString;
  
  aString = [NSString stringWithFormat: @"%@ @ %@", theUsername, theName];
  
  return [allStores objectForKey: aString];
}


//
//
//
- (void) setStore: (id) theStore
	     name: (NSString *) theName
	 username: (NSString *) theUsername
{
  NSString *aString;

  aString = [NSString stringWithFormat: @"%@ @ %@", theUsername, theName];

  // We verify if we want to remove an opened store.
  if ( !theStore && theName && theUsername )
    {
      FolderNode *aFolderNode;

      // For an IMAP store, we remove all children of our root node
      aFolderNode = [self _storeFolderNodeForName: aString];
      [aFolderNode setChildren: nil];
      [outlineView collapseItem: aFolderNode];
      [outlineView selectRow: [outlineView rowForItem: aFolderNode]
  		   byExtendingSelection: NO];

      [allStores removeObjectForKey: aString];
      return;
    }
  
  // We always first "remove" the object in case we call this method
  // multiple times with the same object.
  RETAIN(theStore);
  [allStores removeObjectForKey: aString];
      
  // We {re}add it to our dictionary.
  [allStores setObject: theStore
	     forKey: aString];
  RELEASE(theStore);
}


//
//
//
- (MailboxManagerCache *) mailboxManagerCache
{
  return mailboxManagerCache;
}


//
//
//
- (void) setMailboxManagerCache: (MailboxManagerCache *) theMailboxManagerCache
{
  RETAIN(theMailboxManagerCache);
  RELEASE(mailboxManagerCache);
  mailboxManagerCache = theMailboxManagerCache;
}


//
// class methods
//
+ (id) singleInstance
{
  if ( ! singleInstance )
    {
      singleInstance = [[self alloc] initWithWindowNibName: @"MailboxManager"];
    }
  
  return singleInstance;
}


//
// Other methods
//
- (void) openLocalFolderWithName: (NSString *) theFolderName
{
  MailWindowController *aMailWindowController;
  LocalStore *localStore;
  LocalFolder *aFolder;
  
  BOOL reusingLastMailWindowOnTop;

  // We get out local store and our folder.
  localStore = [self storeForName: @"GNUMAIL_LOCAL_STORE"
		     username: NSUserName()];
  
  
  // We first verify if the folder is still valid. For example, it could have been
  // deleted (the file) manually while GNUMail.app was running.
  if ( ![[NSFileManager defaultManager] 
	  fileExistsAtPath: [[localStore path] stringByAppendingPathComponent: theFolderName]] )
    {
      NSRunInformationalAlertPanel(_(@"Mailbox error!"),
				   _(@"The local mailbox %@ does not exist!"),
				   _(@"OK"),
				   NULL, 
				   NULL,
				   theFolderName);
      return;
    }

  // We now verify if it's not a directory (a folder holding folders)
  if ( [localStore folderTypeForFolderName: theFolderName] == HOLDS_FOLDERS )
    {
      NSRunInformationalAlertPanel(_(@"Mailbox error!"),
				   _(@"This folder (%@) holds folders, not messages!"),
				   _(@"OK"),
				   NULL, 
				   NULL,
				   theFolderName);
      return;
    }

  // If the folder is already open, we show our alert panel
  // and we return
  if ( [localStore folderForNameIsOpen: theFolderName] )
    {
      NSWindow *aWindow;

      aWindow = (NSWindow *)[Utilities windowForFolderName: theFolderName 
				       store: (Store *)localStore];
      if ( aWindow )
	{
	  [aWindow orderFrontRegardless];
	}
      
      return;
    }
  
  // The folder is not open, so we get it!
  aFolder = [localStore folderForName: theFolderName];

  // If we reuse our window controller...
  if ( [[NSUserDefaults standardUserDefaults] objectForKey: @"REUSE_MAILWINDOW"] &&
       [[[NSUserDefaults standardUserDefaults] objectForKey: @"REUSE_MAILWINDOW"] intValue] == NSOnState &&
       [GNUMail lastMailWindowOnTop] )
    {
      
      aMailWindowController = [[GNUMail lastMailWindowOnTop] windowController];
      reusingLastMailWindowOnTop = YES;

      // We must NOT assume that we got a MailWindowController
      if ( [aMailWindowController isKindOfClass: [MessageViewWindowController class]] )
	{
	  aMailWindowController = [(MessageViewWindowController *)aMailWindowController mailWindowController];
	}

      // We close the previous folder
      [[aMailWindowController folder] close];
    }
  else 
    {
      aMailWindowController = [[MailWindowController alloc] initWithWindowNibName: @"MailWindow"];
      reusingLastMailWindowOnTop = NO;
    }
      
  // We update the show/hide deleted/read
  [self _updateShowOrHideReadAndDeletedForFolder: aFolder];
  
  // We set the new folder
  [aMailWindowController setFolder: aFolder];

  // We we are reusing our window controller, we must always reload the table view
  if ( reusingLastMailWindowOnTop && [GNUMail lastMailWindowOnTop] )
    {
      [aMailWindowController tableViewShouldReloadData];
    }
  
  // We always remove the secure icon
  [[aMailWindowController secureConnectionIcon] setImage: nil];

  // And we show the window.. 
  [[aMailWindowController window] orderFrontRegardless];
  [[aMailWindowController window] makeKeyAndOrderFront: nil];

  [[NSDistributedNotificationCenter defaultCenter]
    postNotificationName: GNUMailConsoleMessage
    object: nil
    userInfo: [NSDictionary dictionaryWithObject: 
			      [NSString stringWithFormat: _(@"Local folder %@ opened."), theFolderName]
			    forKey: @"Message"]];
}


//
//
//
- (void) openIMAPFolderWithName: (NSString *) theFolderName
			  store: (IMAPStore *) theStore
{
  MailWindowController *aMailWindowController;
  IMAPCacheManager *anIMAPCacheManager;
  IMAPFolder *aFolder;
  NSString *aKey;
  
  
  if ( [[[theStore openedFoldersEnumerator] allObjects] count] > 0 &&
       ( ![[NSUserDefaults standardUserDefaults] objectForKey: @"REUSE_MAILWINDOW"] ||
	 [[[NSUserDefaults standardUserDefaults] objectForKey: @"REUSE_MAILWINDOW"] intValue] == NSOffState) )
    {
      NSRunInformationalAlertPanel(_(@"Mailbox error!"),
				   _(@"A folder (%@) is already open. Please close it first."),
				   _(@"OK"),
				   NULL, 
				   NULL,
				   [[theStore openedFoldersEnumerator] nextObject]);
      return;
    }
  
  

  // If we must reuse our window controller...
  if ( [[NSUserDefaults standardUserDefaults] objectForKey: @"REUSE_MAILWINDOW"] &&
       [[[NSUserDefaults standardUserDefaults] objectForKey: @"REUSE_MAILWINDOW"] intValue] == NSOnState &&
       [GNUMail lastMailWindowOnTop] )
    {
      // We must NOT assume that we got a MailWindowController
      if ( [[[GNUMail lastMailWindowOnTop] windowController] isKindOfClass: [MessageViewWindowController class]] )
	{
	  aMailWindowController = [[[GNUMail lastMailWindowOnTop] windowController] mailWindowController];
	}
      else
	{
	  aMailWindowController = [[GNUMail lastMailWindowOnTop] windowController];
	}
      
      // close the previous folder
      [[aMailWindowController folder] close];
    }
  else
    {
      aMailWindowController = [[MailWindowController alloc] initWithWindowNibName: @"MailWindow"];
    }

  // We send our message to the console saying we are about to open the IMAP folder
  [[NSDistributedNotificationCenter defaultCenter]
    postNotificationName: GNUMailConsoleMessage
    object: nil
    userInfo: [NSDictionary dictionaryWithObject: [NSString stringWithFormat: _(@"Opening IMAP folder %@ on %@..."), 
							    theFolderName, [theStore name]]
			    forKey: @"Message"]];
  
  // We get our cache manager for this server / folder
  aKey = [NSString stringWithFormat: @"%@ @ %@", [theStore username], [theStore name]];
  anIMAPCacheManager = [IMAPCacheManager imapCacheObjectsFromDiskWithPath:
					   [NSString stringWithFormat: @"%@/IMAPCache_%@_%@",
						     GNUMailUserLibraryPath(),
						     [Utilities flattenPathFromString: aKey
								separator: @"/"],
						     [Utilities flattenPathFromString: theFolderName
								separator: [theStore folderSeparator]]]];
  
  // We obtain our folder from the IMAP store
  aFolder = (IMAPFolder *)[theStore folderForName: theFolderName
				    withIMAPCacheManager: anIMAPCacheManager];
  

  // We verify if the server accepts messages in this folder
  if ( !aFolder)
    {
      NSRunInformationalAlertPanel(_(@"Mailbox error!"),
				   _(@"You must select a valid mailbox to open!"),
				   _(@"OK"),
				   NULL,
				   NULL,
				   NULL);
      return;
    }
  

  // We update the show/hide deleted/read
  [self _updateShowOrHideReadAndDeletedForFolder: aFolder];
  
  // We set the folder
  [aMailWindowController setFolder: aFolder];
  
  // We we are reusing our window controller, we must always reload the table view
  if ( [[NSUserDefaults standardUserDefaults] objectForKey: @"REUSE_MAILWINDOW"] &&
       [[[NSUserDefaults standardUserDefaults] objectForKey: @"REUSE_MAILWINDOW"] intValue] == NSOnState &&
       [GNUMail lastMailWindowOnTop] )
    {
      [aMailWindowController tableViewShouldReloadData];
    }

  // We verify if we are using a secure (SSL) connection or not
  if ( ![(id<NSObject>)[(id<Store>)[aFolder store] tcpConnection] isKindOfClass: [TCPConnection class]] )
    {
      [[aMailWindowController secureConnectionIcon] setImage: [NSImage imageNamed: @"pgp-mail-small.tiff"]];
    }
  else
    {
      [[aMailWindowController secureConnectionIcon] setImage: nil];
    }

  [[aMailWindowController window] orderFrontRegardless];
  [[aMailWindowController window] makeKeyAndOrderFront: nil];

  [[NSDistributedNotificationCenter defaultCenter]
    postNotificationName: GNUMailConsoleMessage
    object: nil
    userInfo: [NSDictionary dictionaryWithObject: [NSString stringWithFormat: _(@"IMAP folder %@ on %@ opened."), 
							    theFolderName, [theStore name]]
			    forKey: @"Message"]];
}


//
//
//
- (void) reloadAllFolders
{
  BOOL mustSynchronizeUserDefaults = NO;

  // We remove all our elements
  [allFolders removeAllObjects];

  // We always add our local folder
  localNodes = [Utilities folderNodesFromFolders: [[self storeForName: @"GNUMAIL_LOCAL_STORE"
							 username: NSUserName()] folderEnumerator]
			  separator: @"/"];

  [localNodes setName: _(@"Local")];
  [localNodes setParent: nil];
  [allFolders addObject: localNodes];

  // We verify if the RECEIVING preferences have been defined.
  if ( [[NSUserDefaults standardUserDefaults] objectForKey: @"RECEIVING"] )
    {
      NSMutableDictionary *allServers;
      NSDictionary *allValues;
      
      NSString *aKey, *aServerName;

      NSEnumerator *allKeys;
      
      allServers = [[NSMutableDictionary alloc] init];
      [allServers addEntriesFromDictionary: [[NSUserDefaults standardUserDefaults] objectForKey: @"RECEIVING"]];
      allKeys = [allServers keyEnumerator];
      
      // We build a correct subset of all our IMAP servers defined in RECEIVING
      while ( (aKey = [allKeys nextObject]) )
	{
	  allValues = [allServers objectForKey: aKey];
	  aServerName = [allValues objectForKey: @"SERVERNAME"];
	  
	  // We verify if we are using the old version of the preference values.
	  // If so, we convert them to the new format.
	  if ( !aServerName )
	    {
	      NSString *aServerType = nil;
	      
	      aServerName = [NSString stringWithString: aKey];

	      if ( [[allValues objectForKey: @"SERVERTYPE"] intValue] == IMAP )
		{
		  aServerType = [NSString stringWithString: @"IMAP"];
		}
	      else if ( [[allValues objectForKey: @"SERVERTYPE"] intValue] == POP3 )
		{
		  aServerType = [NSString stringWithString: @"POP3"];
		}
	      
	      if ( aServerType )
		{
		  NSMutableDictionary *newValues;
		  NSString *cacheFilePrefix, *cacheFileSuffix, *cacheFilePath;
		  NSArray *cacheFileNames;
		  int i;

		  newValues = [[NSMutableDictionary alloc] initWithCapacity: [allValues count] + 1];
		  
		  // We set the new values to the previous ones and we slightly modify them.
		  [newValues addEntriesFromDictionary: allValues];
		  
		  // We modify the password. previously, they were using the server name as the key. Now
		  // we use username @ servername so me must decode them, and re-encode them, if we need to.
		  if ( [[newValues objectForKey: @"REMEMBERPASSWORD"] intValue] == NSOnState )
		    {
		      NSString *aPassword;
		      
		      aPassword = [Utilities decryptPassword: [newValues objectForKey: @"PASSWORD"]
					     withKey: aServerName];
		      
		      aPassword = [Utilities encryptPassword: aPassword
					     withKey: [NSString stringWithFormat: @"%@ @ %@",
								[newValues objectForKey: @"USERNAME"],
								aServerName]];
		      
		      [newValues removeObjectForKey: @"PASSWORD"];
		      [newValues setObject: aPassword
				 forKey: @"PASSWORD"];
		    }

		  // We add the SERVERNAME
		  [newValues setObject: aServerName
			     forKey: @"SERVERNAME"];

		  [allServers removeObjectForKey: aKey];

		  aKey = [NSString stringWithFormat: @"%@ @ %@",
				   [newValues objectForKey: @"USERNAME"],
				   aServerName];

		  [allServers setObject: newValues
			      forKey: aKey];
		  
		  // We must now reename the cache file(s)
  		  cacheFilePrefix = [NSString stringWithFormat: @"%@Cache_%@",
					      aServerType,
					      aServerName];

		  cacheFileNames = [[NSFileManager defaultManager] directoryContentsAtPath: GNUMailUserLibraryPath()];

		  for (i = 0; i < [cacheFileNames count]; i++)
		    {
		      if ( [(NSString *)[cacheFileNames objectAtIndex: i] hasPrefix: cacheFilePrefix] )
			{
			  cacheFilePath = [NSString stringWithFormat: @"%@/%@", 
						    GNUMailUserLibraryPath(),
						    (NSString *)[cacheFileNames objectAtIndex: i]];
			  
			  if ( [cacheFilePrefix length] != [(NSString *)[cacheFileNames objectAtIndex: i] length] )
			    {
			      // Get the folder name
			      cacheFileSuffix = [(NSString *)[cacheFileNames objectAtIndex: i] 
							     substringFromIndex: [cacheFilePrefix length]];
			      NSDebugLog(@"\t(folder = %@)", cacheFileSuffix);
			    }
			  else
			    {
			      cacheFileSuffix = [NSString stringWithString: @""];
			    }

			  NSDebugLog(@"%@ => %@", cacheFilePath,
				[NSString stringWithFormat: @"%@/%@Cache_%@%@",
					  GNUMailUserLibraryPath(),
					  aServerType,
					  [Utilities flattenPathFromString: aKey
						     separator: @"/"],
					  cacheFileSuffix]);

			  if ( [[NSFileManager defaultManager] fileExistsAtPath: cacheFilePath] )
			    {
			      [[NSFileManager defaultManager] movePath: cacheFilePath
							      toPath: [NSString stringWithFormat: @"%@/%@Cache_%@%@",
										GNUMailUserLibraryPath(),
										aServerType,
										[Utilities flattenPathFromString: aKey
											   separator: @"/"],
										cacheFileSuffix]
							      handler: nil];
			    }
			}
		    }
		  
		  mustSynchronizeUserDefaults = YES;
		  RELEASE(newValues);
		} // if ( aServerType )...
	    } // if ( !aServerName ) ...
	  
	  // We add it only if it's an IMAP server...
	  if ( [[allValues objectForKey: @"SERVERTYPE"] intValue] == IMAP )
	    {
	      FolderNode *aFolderNode;

	      aFolderNode = [FolderNode folderNodeWithName: aKey
					parent: nil];
	      
	      [allFolders addObject: aFolderNode];
	      
	      // If our IMAP Store has been previously initialized, we re-initialize it in order to get
	      // the most recent values for the subscribed folders
	      if ( [self storeForName: aServerName
			 username: [allValues objectForKey: @"USERNAME"]] )
		{
		  [self _reloadFoldersForIMAPStoreWithName: aServerName
			username: [allValues objectForKey: @"USERNAME"]];
		}
	    }
	}
    
      // If we previously made some changes to the user defaults (move the configuration from the 
      // old format to the new one), we synchronize our defaults for the RECEIVING preferences.
      if ( mustSynchronizeUserDefaults )
	{
	  // We save our RECEIVING value in the user's defaults
	  [[NSUserDefaults standardUserDefaults] setObject: allServers
						 forKey: @"RECEIVING"];
	  [[NSUserDefaults standardUserDefaults] synchronize];
	}
      
      RELEASE(allServers);
    }
  
  // We inform our outline view to reload its data.
  [outlineView reloadData];

  // We alwas expand our root item
  [outlineView expandItem: allFolders];
  
  // We now select and expand the 'Local' folder if there's no IMAP folders defined
  if ( [allFolders count] == 1 )
    {
      [outlineView expandItem: localNodes];
      [outlineView selectRow: [outlineView rowForItem: localNodes]
  		   byExtendingSelection: NO];
   }
}


//
//
//
- (BOOL) initializeIMAPStoreWithKey: (NSString *) theKey
{
  NSString *aServerName, *aUsername, *aPassword;
  NSNumber *serverTypeValue, *portValue;  
  NSDictionary *allValues;
  
  // We begin by searching in our RECEIVING values for the right server.
  // Now, let's get all the receiving values
  allValues = [[[NSUserDefaults standardUserDefaults] objectForKey: @"RECEIVING"]
  		objectForKey: theKey];
  
  serverTypeValue = [allValues objectForKey: @"SERVERTYPE"];
  portValue =  [allValues objectForKey: @"PORT"];
	  
  // We use the default IMAP port if it's not defined.
  if ( !portValue )
    {
      portValue = [NSNumber numberWithInt: 143];
    }
  
  // We get our username
  aUsername = [allValues objectForKey: @"USERNAME"];

  // We get our servername
  aServerName = [allValues objectForKey: @"SERVERNAME"];
  
  // We first verify if we haven't already cached our store. If so,
  // we simply return since the Store has already been initialized.
  if ( [self storeForName: aServerName
	     username: aUsername] )
    {
      return YES;
    }

  // We get our password. We could prompt for it in case we need to.
  aPassword = [self passwordForKey: theKey];
  
  if ( aPassword )
    {		
      IMAPStore *aStore;
     
      // We verify if we must initiate a TCP/SSL or a normal TCP connectin.
      if ( [allValues objectForKey: @"USESECURECONNECTION"] &&
	   [[allValues objectForKey: @"USESECURECONNECTION"] intValue] == NSOnState )
	{
	  [[NSDistributedNotificationCenter defaultCenter]
	    postNotificationName: GNUMailConsoleMessage
	    object: nil
	    userInfo: [NSDictionary dictionaryWithObject: [NSString stringWithFormat: 
								      _(@"Connecting to IMAP server %@ using SSL"), 
								    aServerName]
				  forKey: @"Message"]];
	  
	  aStore = [[IMAPStore alloc] initSSLWithName: aServerName
				      port: [portValue intValue]];
	}
      else
	{
	  [[NSDistributedNotificationCenter defaultCenter]
	    postNotificationName: GNUMailConsoleMessage
	    object: nil
	    userInfo: [NSDictionary dictionaryWithObject: [NSString stringWithFormat: _(@"Connecting to IMAP server %@"),
								    aServerName]
				    forKey: @"Message"]];
	  
	  aStore = [[IMAPStore alloc] initWithName: aServerName
				      port: [portValue intValue]];
	}

      if ( !aStore )
	{
	  NSRunInformationalAlertPanel(_(@"Error!"),
				       _(@"Unable to communicate with the IMAP server (%@:%i)."),
				       _(@"OK"),
				       NULL,
				       NULL,
				       aServerName, 
				       [portValue intValue]);
	  return NO;
 	}
        else
	  {
	    [[NSDistributedNotificationCenter defaultCenter]
	      postNotificationName: GNUMailConsoleMessage
	      object: nil
	      userInfo: [NSDictionary dictionaryWithObject: [NSString stringWithFormat: _(@"Authenticating using %@..."),
								      aUsername]
				      forKey: @"Message"]];
	    
	    // We verify if the authentication is successful
	    if ( [aStore authenticateWithUsername: aUsername
			 password: aPassword] )
	      {
		// We cache our folder
		[self setStore: aStore
		      name: aServerName
		      username: aUsername];
		
		[self _reloadFoldersForIMAPStoreWithName: aServerName
		      username: aUsername];

		RELEASE(aStore);

		[[NSDistributedNotificationCenter defaultCenter]
		  postNotificationName: GNUMailConsoleMessage
		  object: nil
		  userInfo: [NSDictionary dictionaryWithObject: [NSString stringWithFormat: _(@"Connected to %@!"),
									  aServerName]
					  forKey: @"Message"]];

		return YES;
	      }
	    else
	      {
		NSRunInformationalAlertPanel(_(@"Error!"),
					     _(@"Invalid password %@ @ %@."), 
					     _(@"OK"),
					     NULL,
					     NULL,
					     aUsername, aServerName);
		[aStore close];
		RELEASE(aStore);
	      }

	  } // else of if ( !aStore ) ...
	
    } // if (password )
	
  return NO;
}


//
// This method is used to transfer a message (from raw source) to
// the specified folder.
//
// If the folder isn't opened, it is opened w/o parsing.
// If it was opened, we refresh the tableView that shows
// the messages to the user.
//
// Note: This method MUST NOT be called in a thread other than the
//       main thread.
//
- (BOOL) transferMessageFromRawSource: (NSData *) theRawSource
				flags: (Flags *) theFlags
		     toFolderWithName: (NSString *) theFolderName
			   orToFolder: (Folder *) theFolder
				store: (Store *) theStore
{
  NSString *aStoreName, *aUsername;
  Folder *aFolder;
  id aWindow;

  // We first verify if theRawSource is nil, if it is, we simply return NO.
  if ( !theRawSource || [theRawSource length] == 0 )
    {
      return NO;
    }
  
  aWindow = nil;

  // We get the store name
  if ( [(id<NSObject>)theStore isKindOfClass: [LocalStore class]] )
    {
      aStoreName = @"GNUMAIL_LOCAL_STORE";
      aUsername = NSUserName();
    }
  else
    {
      aStoreName = [(IMAPStore *)theStore name];
      aUsername = [(IMAPStore *)theStore username];
    }


  // We get a reference to our folder, w/o parsing it if it's not already open.
  if ( !theFolder )
    {
      if ( [(id<NSObject>)theStore isKindOfClass: [IMAPStore class]] )
	{
	  aFolder = [(IMAPStore *)theStore folderForName: theFolderName
				  select: NO];
	}
      else
	{
	  aFolder = [(id<Store>)theStore folderForName: theFolderName];
	}
    }
  else
    {
      aWindow = [Utilities windowForFolderName: [theFolder name]
			   store: theStore]; 
      aFolder = theFolder;
    }

  // Our folder is opened, let's get it from the list of opened windows.
  if ( !aFolder )
    {
      aWindow = [Utilities windowForFolderName: theFolderName
			   store: theStore]; 
      aFolder = (LocalFolder *)[[aWindow windowController] folder];
    } 
  
  // We transfer the new message
  NS_DURING
    {
      if ( theFlags )
	{
	  [aFolder appendMessageFromRawSource: theRawSource
		   flags: theFlags];
	}
      else
	{
	  [aFolder appendMessageFromRawSource: theRawSource];
	}

      // We update the Mailbox Manager cache, IIF the folder wasn't opened initially.
      // If it was open, we had its window and we are gonna refresh its cache in
      // -updateStatusLabel.
      if ( aWindow == nil )
	{
	  int nbOfMessages, nbOfUnreadMessages, totalSize;

	  nbOfMessages = [aFolder count];
	  nbOfUnreadMessages = [aFolder numberOfUnreadMessages];
	  totalSize = [aFolder size];
	  
	  [mailboxManagerCache setAllValuesForStoreName: aStoreName
			       folderName: [aFolder name]
			       username: aUsername
			       nbOfMessages: nbOfMessages
			       nbOfUnreadMessages: nbOfUnreadMessages
			       totalSize: totalSize];

	  [outlineView setNeedsDisplay: YES];
	}
    }
  NS_HANDLER
    {
      NSRunAlertPanel(_(@"Error!"),
		      _(@"A fatal error occured when appending the message to the %@ folder."),
		      _(@"OK"),
		      NULL,
		      NULL,
		      [aFolder name]);

      NS_VALUERETURN(NO, BOOL);
    }
  NS_ENDHANDLER
  
  // If the folder was open, we refresh its view. Otherwise, we just close it.
  if ( aWindow )
    {
      [[aWindow delegate] tableViewShouldReloadData];
      [[aWindow delegate] updateStatusLabel];
    }
  else
    {
      // If our folder wasn't first openend
      if ( !theFolder )
	{
	  [aFolder close];
	}
    }

  return YES;
}


//
// This method is used to verify if a folder with name "theFolderName"
// exists in the specified local store.
//
- (BOOL) folderNameExists: (NSString *) theFolderName
		 forStore: (Store *) theStore
{
  NSEnumerator *anEnumerator;
  NSString *aString;

  anEnumerator = [(id<Store>)theStore folderEnumerator];

  while ( (aString = [anEnumerator nextObject]) )
    {
      if ( [aString isEqualToString: theFolderName] )
	{
	  return YES;
	}
    }
  
  return NO;
}


//
//
//
- (NSString *) passwordForKey: (NSString *) theKey
{
  NSDictionary *allReceivingValues;
  NSString *aPassword;

  allReceivingValues = [[[NSUserDefaults standardUserDefaults] objectForKey: @"RECEIVING"] 
			 objectForKey: theKey];
  
  // We verify in the user defaults
  aPassword =  [Utilities decryptPassword: [allReceivingValues objectForKey: @"PASSWORD"]
			  withKey: theKey];
  
  // If we must prompt for the password
  if ( !aPassword )
    {
      PasswordPanelController *theController; 
      int result;
      
      theController = [[PasswordPanelController alloc] initWithWindowNibName: @"PasswordPanel"];
      [[theController window] setTitle: theKey];
      
      result = [NSApp runModalForWindow: [theController window]];
      
      // If the user has entered a password...
      if (result == NSRunStoppedResponse)
	{
	  aPassword = [theController password];
	}
      else
	{
	  aPassword = nil;
	}
      
      RELEASE(theController);

#ifndef MACOSX
      if ( [[self window] isVisible] )
	{
	  [[self window] makeKeyAndOrderFront: self];
	}
#endif
    }
  
  return aPassword;
}

@end


//
// Private methods
//
@implementation MailboxManagerController (Private)

//
//
//
- (void) _nbOfMessages: (int *) theNbOfMessages
    nbOfUnreadMessages: (int *) theNbOfUnreadMessages
             totalSize: (int *) theTotalSize
               forItem: (id) theItem
{
  NSString *aString, *aStoreName, *aFolderName, *aUsername;
  
  aString = [Utilities completePathForFolderNode: theItem
		       separator: @"/"];
  
  if ( [aString hasPrefix: [NSString stringWithFormat: @"/%@", _(@"Local")]] )
    {
      aStoreName = @"GNUMAIL_LOCAL_STORE";
      aFolderName = [Utilities pathOfFolderFromFolderNode: (FolderNode *)theItem
			       separator: @"/"];
      aUsername = NSUserName();
    }
  else
    {
      [Utilities storeKeyForFolderNode: theItem
		 serverName: &aStoreName
		 username: &aUsername];
      
      aFolderName = [Utilities pathOfFolderFromFolderNode: (FolderNode *)theItem
			       separator: @"/"];
    }
  
  [mailboxManagerCache allValuesForStoreName: aStoreName
		       folderName: aFolderName
		       username: aUsername
		       nbOfMessages: theNbOfMessages
		       nbOfUnreadMessages: theNbOfUnreadMessages
		       totalSize: theTotalSize];
}


//
//
//
- (void) _reloadFoldersForIMAPStoreWithName: (NSString *) theName
				   username: (NSString *) theUsername
{
  FolderNode *aFolderNode;
  IMAPStore *aStore;

  NSString *aKey;
  
  aKey =  [NSString stringWithFormat: @"%@ @ %@", theUsername, theName];

  aStore = [self storeForName: theName
		 username: theUsername];
  aFolderNode = [self _storeFolderNodeForName: aKey];
  
  if ( aFolderNode )
    {
      NSMutableDictionary *allServers, *allValues;
      NSArray *subscribedFolders;

      NSString *theSeparator;
      FolderNode *nodes;

      subscribedFolders = [[NSArray alloc] initWithArray: [[aStore subscribedFolderEnumerator] allObjects]];

      theSeparator = RETAIN([aStore folderSeparator]);

      nodes = [Utilities folderNodesFromFolders: [subscribedFolders objectEnumerator]
			 separator: theSeparator];
      RETAIN(nodes);
      
      [aFolderNode setChildren: [nodes children]];

      RELEASE(theSeparator);
      RELEASE(nodes);
      
      // FIXME - Use reloadItem under GNUstep too when the NSOutlineView is fixed.
#ifdef MACOSX
      [outlineView reloadItem: aFolderNode];
#else
      [outlineView reloadData];
#endif

      // We finally cache our subscribed folders in the user's defaults for this server.
      allServers = [[NSMutableDictionary alloc] initWithDictionary: [[NSUserDefaults standardUserDefaults] objectForKey: @"RECEIVING"]];

      allValues = [[NSMutableDictionary alloc] initWithDictionary: [allServers objectForKey: aKey]];

      [allValues setObject: subscribedFolders
		 forKey: @"SUBSCRIBED_FOLDERS"];
      
      [allServers setObject: allValues
		  forKey: aKey];
      
      [[NSUserDefaults standardUserDefaults] setObject: allServers
					     forKey: @"RECEIVING"];
      
      [[NSUserDefaults standardUserDefaults] synchronize];

      RELEASE(allServers);
      RELEASE(allValues);
      RELEASE(subscribedFolders);
    }
  else
    {
      NSDebugLog(@"Error getting the folder's root for IMAP folders tree.");
    }
}


//
//
//
- (FolderNode *) _storeFolderNodeForName: (NSString *) theName
{
  int i;

  for (i = 0; i < [allFolders count]; i++)
    {
      FolderNode *aFolderNode;
      
      aFolderNode = [allFolders objectAtIndex: i];

      if ( [theName isEqualToString: [aFolderNode name]] )
	{
	  return aFolderNode;
	}
    }

  return nil;
}


//
//
//
- (void) _reloadFoldersAndExpandParentsFromNode: (FolderNode *) theNode
			     selectNodeWithPath: (NSString *) thePath
{ 
  NSMutableArray *nodesToExpand;
  NSString *aString;
  
  id aParent, aRootNode;
  int i, aRow; 
  
  aString = [Utilities storeKeyForFolderNode: theNode
		       serverName: NULL
		       username: NULL];

  // We must refresh our outline view by reload its content
  [self reloadAllFolders];
  
  // We first get our root node
  if ( [thePath hasPrefix: [NSString stringWithFormat: @"/%@", _(@"Local")]] )
    {
      aRootNode = localNodes;
    }
  else
    {
      aRootNode = [self _storeFolderNodeForName: aString];
    }


  // We get our new node in our tree and also our new row index
  aParent = [Utilities folderNodeForPath: [thePath stringByDeletingLastPathComponent]
		       using: aRootNode];

  nodesToExpand = [[NSMutableArray alloc] init];

  // We expand all our parent, to make the row visible.
  while ( aParent )
    {
      [nodesToExpand addObject: aParent];
      aParent = [aParent parent];
    }
  
  // We must expand our nodes starting from the root to the children and not
  // the other way around. Otherwise, the NSOutlineView just borks.
  for (i = ([nodesToExpand count] - 1); i >= 0; i--)
    {
      [outlineView expandItem: [nodesToExpand objectAtIndex: i]];
    }

  RELEASE(nodesToExpand);

  // We now get our new node node (renamed or created). Since it's now shown on the screen,
  // we can now obtain the row for it and select it.
  aParent = [Utilities folderNodeForPath: thePath
		       using: aRootNode];

  aRow = [outlineView rowForItem: aParent];


  if ( aRow >= 0 && aRow < [outlineView numberOfRows] )
    {
      [outlineView selectRow: aRow
      		   byExtendingSelection: NO];
      [outlineView scrollRowToVisible: aRow];
    }
}


//
//
//
- (FolderNode *) _rootFolderNodeForNode: (FolderNode *) theNode
{
  FolderNode *aNode;

  aNode = theNode;

  while ( [aNode parent] )
    {
      aNode = [aNode parent];
    }

  return aNode;
}


//
//
//
- (void) _updateShowOrHideReadAndDeletedForFolder: (Folder *) theFolder
{
  
  // We now set the folder to show or hide our deleted messages from the folder
  // Showing them is the default behaviour.
  if ( ![[NSUserDefaults standardUserDefaults] objectForKey: @"HIDE_DELETED_MESSAGES"] ||
       [[[NSUserDefaults standardUserDefaults] objectForKey: @"HIDE_DELETED_MESSAGES"] intValue] == NSOffState)
    {
      [theFolder setShowDeleted: YES];
      [[NSApp delegate] updateShowOrHideDeletedMenuItem: YES];
    }
  else
    {
      [theFolder setShowDeleted: NO];
      [[NSApp delegate] updateShowOrHideDeletedMenuItem: NO];
    }

  // We now set the folder to show or hide read messages from the folder. Showing them
  // is the default behavior
  if ( ![[NSUserDefaults standardUserDefaults] objectForKey: @"HIDE_READ_MESSAGES"] ||
       [[[NSUserDefaults standardUserDefaults] objectForKey: @"HIDE_READ_MESSAGES"] intValue] == NSOffState)
    {
      [theFolder setShowRead: YES];
      [[NSApp delegate] updateShowOrHideReadMenuItem: YES];
    }
  else
    {
      [theFolder setShowRead: NO];
      [[NSApp delegate] updateShowOrHideReadMenuItem: NO];
    }
}

@end








