/*
**  NewMailboxPanelController.m
**
**  Copyright (c) 2001, 2002
**
**  Author: Ludovic Marcotte <ludovic@Sophos.ca>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#import "NewMailboxPanelController.h"

#import "Filter.h"
#import "FilterManager.h"
#import "GNUMailConstants.h"
#import "MailboxManager.h"
#import "MailboxManagerController.h"

#ifndef MACOSX
#import "NewMailboxPanel.h"
#endif

#import "NSStringExtensions.h"
#import "Utilities.h"

#import <Pantomime/Store.h>
#import <Pantomime/IMAPStore.h>
#import <Pantomime/LocalStore.h>
#import <Pantomime/URLName.h>

@implementation NewMailboxPanelController

//
//
//
- (id) initWithWindowNibName: (NSString *) windowNibName
		  folderNode: (FolderNode *) theFolderNode
{
#ifdef MACOSX
  
  self = [super initWithWindowNibName: windowNibName];
 
#else
  NewMailboxPanel *newMailboxPanel;

  newMailboxPanel =
    [[NewMailboxPanel alloc] initWithContentRect: NSMakeRect(200,200,345,150)
			     styleMask: (NSTitledWindowMask |
					 NSMiniaturizableWindowMask)
			     backing: NSBackingStoreBuffered
			     defer: YES];
  
  self = [super initWithWindow: newMailboxPanel];
  
  [newMailboxPanel layoutPanel];
  [newMailboxPanel setDelegate: self];
  
  // We link our outlets
  mailboxNameLabel = (NSTextField *)[newMailboxPanel mailboxNameLabel];
  mailboxNameField = [newMailboxPanel mailboxNameField];
  
  RELEASE(newMailboxPanel);
#endif

  // We link our ivar, no need to retain it
  folderNode = theFolderNode;

  // We set our window title
  [[self window] setTitle: _(@"New Mailbox")];

  return self;
}


//
//
//
- (void) dealloc
{
  NSDebugLog(@"NewMailboxPanelController: -dealloc");
  TEST_RELEASE(name);

  [super dealloc];
}


//
// action methods
//
- (IBAction) okClicked: (id) sender
{
  // If we are renaming a mailbox..
  if ( [self name] )
    {
      // We must be sure our names are different
      if ( ![[self name] isEqualToString: [[mailboxNameField stringValue] stringByTrimmingWhiteSpaces]] ) 
	{
	  // "name" is the old name of the mailbox while the content of our text field is the new
	  // mailbox name
	  NSString *pathOfFolder, *aSource, *aDestination;
	  Store *aStore;
	  
	  //
	  // We first get our right store
	  //
	  if ( [[Utilities completePathForFolderNode: folderNode  separator: @"/"]
		 hasPrefix: [NSString stringWithFormat: @"/%@", _(@"Local")]] )
	    {
	      aStore = [[MailboxManagerController singleInstance]
			 storeForName: @"GNUMAIL_LOCAL_STORE"
			 username: NSUserName()];
	    }
	  else
	    {
	      NSString *aServerName, *aUsername;

	      [Utilities storeKeyForFolderNode: folderNode
			 serverName: &aServerName
			 username: &aUsername];
	      
	      aStore = [[MailboxManagerController singleInstance] storeForName: aServerName
								  username: aUsername];
	    }


	  //
	  // pathOfFolder will hold a value like:  folderA
	  //                                       folderA/folderB
	  //                                       folderA/folderB/folderC
	  //                                  or:
	  //                                       folderA.folderB
	  //                                       folderA.folderB.folderC
	  //
	  pathOfFolder = [Utilities pathOfFolderFromFolderNode: folderNode
				    separator: [(id<Store>)aStore folderSeparator]];

	  if ( !pathOfFolder )
	    {
	      aSource = [self name];
	      aDestination = [[mailboxNameField stringValue] stringByTrimmingWhiteSpaces];
	    }
	  else 
	    {
	      NSRange aRange;

	      aSource = pathOfFolder;
	      
	      // We set the right destination folder, first by trimming the last path component
	      // using the right separator.
	      aRange = [pathOfFolder rangeOfString: [(id<Store>)aStore folderSeparator]
				     options: NSBackwardsSearch];
	      
	      if ( aRange.length )
		{
		  aDestination = [pathOfFolder substringToIndex: aRange.location];
		}
	      else
		{
		  aDestination = @"";
		}
	      
	      if ( [aDestination length] == 0 )
		{
		  aDestination = [[mailboxNameField stringValue] stringByTrimmingWhiteSpaces];
		}
	      else
		{
		  aDestination = [NSString stringWithFormat: @"%@%@%@", 
					   aDestination,
					   [(id<Store>)aStore folderSeparator],
					   [[mailboxNameField stringValue] stringByTrimmingWhiteSpaces]];
		}
	    }	      


	  if ( [(id<Store>)aStore renameFolderWithName: aSource
		       toName: aDestination] )
	    {
	      //
	      // We first verify if we need to change the target
	      // mailboxes for our filters.
	      //
	      NSString *aSourceURL, *aDestinationURL;
	      FilterManager *aFilterManager;
	      Filter *aFilter;
	      int i;
	      
	      // We build our right URLs
	      if ( [(id<NSObject>)aStore isKindOfClass: [LocalStore class]] )
		{
		  aSourceURL = [NSString stringWithFormat: @"local://%@/%@",
					 [[NSUserDefaults standardUserDefaults] objectForKey: @"LOCALMAILDIR"],
					 aSource];
		  aDestinationURL = [NSString stringWithFormat: @"local://%@/%@",
					      [[NSUserDefaults standardUserDefaults] objectForKey: @"LOCALMAILDIR"],
					      aDestination];
		}
	      else
		{
		  aSourceURL = [NSString stringWithFormat: @"imap://%@@%@/%@", 
					 [(IMAPStore *)aStore username], 
					 [(IMAPStore *)aStore name],
					 aSource];
		  aDestinationURL = [NSString stringWithFormat: @"imap://%@@%@/%@", 
					      [(IMAPStore *)aStore username], 
					      [(IMAPStore *)aStore name], 
					      aDestination];
		}

	      //NSDebugLog(@"aSourceURL = |%@| aDestinationURL = |%@|", aSourceURL, aDestinationURL); 
	      
	      aFilterManager = [FilterManager filtersFromDisk];
	      RETAIN(aFilterManager);
	      
	      for (i = 0; i < [[aFilterManager filters] count]; i++)
		{   
		  aFilter = [aFilterManager filterAtIndex: i];
		  
		  if ([aFilter action] == TRANSFER_TO_FOLDER)
			{
			  if ( [[aFilter actionFolderName] isEqualToString: aSourceURL] )
			    {
			      [aFilter setActionFolderName: aDestinationURL];
			    }
			}
		}
		  
	      [aFilterManager synchronize];
	      RELEASE(aFilterManager);
	   
	      
	      //
	      // Then, we must verify if we must rename our IMAP cache files.
	      //
	      if ( [(id<NSObject>)aStore isKindOfClass: [IMAPStore class]] )
		{
		  NSString *aKey, *aSourcePath, *aDestinationPath;
		  
		  aKey = [NSString stringWithFormat: @"%@ @ %@", 
				   [(IMAPStore *)aStore username], 
				   [(IMAPStore *)aStore name]];
		  
		  aSourcePath = [NSString stringWithFormat: @"%@/IMAPCache_%@_%@",
					  GNUMailUserLibraryPath(),
					  [Utilities flattenPathFromString: aKey
						     separator: @"/"],
					  [Utilities flattenPathFromString: aSource
						     separator: [(id<Store>)aStore folderSeparator]] ];
		  
		  aDestinationPath = [NSString stringWithFormat: @"%@/IMAPCache_%@_%@", 
					       GNUMailUserLibraryPath(),
					       [Utilities flattenPathFromString: aKey
							  separator: @"/"],
					       [Utilities flattenPathFromString: aDestination
							  separator: [(id<Store>)aStore folderSeparator]] ];
		  
		  [[NSFileManager defaultManager] movePath: aSourcePath
						  toPath: aDestinationPath
						  handler: nil];
		  
		}
	    }
	  else
	    {
	      NSRunInformationalAlertPanel(_(@"Error!"),
					   _(@"The mailbox rename operation failed."),
					   _(@"OK"),
					   NULL,
					   NULL,
					   NULL);
	    }
	}
    }
  //
  // We are creating a mailbox, Local or IMAP.
  //
  else
    {
      NSString *pathOfFolder, *aString;
      id aStore;
      
      //
      // We first get the right store.
      //
      if ( [[Utilities completePathForFolderNode: folderNode  separator: @"/"] 
	     hasPrefix: [NSString stringWithFormat: @"/%@", _(@"Local")]] )
	{
	  aStore = [[MailboxManagerController singleInstance]
		     storeForName: @"GNUMAIL_LOCAL_STORE"
		     username: NSUserName()];
	}
      else
	{
	  NSString *aServerName, *aUsername;
	  
	  [Utilities storeKeyForFolderNode: folderNode
		     serverName: &aServerName
		     username: &aUsername];
	  
	  aStore = [[MailboxManagerController singleInstance] storeForName: aServerName
							      username: aUsername];
	}

      
      pathOfFolder = [Utilities pathOfFolderFromFolderNode: folderNode
				separator: [aStore folderSeparator]];
      
      NSDebugLog(@"pathOfFolder = %@", pathOfFolder);

      if ( !pathOfFolder || [pathOfFolder length] == 0)
	{
	  aString = [[mailboxNameField stringValue] stringByTrimmingWhiteSpaces];
	}
      else
	{
	  aString = [NSString stringWithFormat: @"%@%@%@",
			      pathOfFolder,
			      [aStore folderSeparator],
			      [[mailboxNameField stringValue] stringByTrimmingWhiteSpaces]];
	}
      
      NSDebugLog(@"new pathOfFolder = %@", aString);

      // We can now proceed with the creation of our mbox		  
      if ( ![aStore createFolderWithName: aString] )
	{
	  NSRunInformationalAlertPanel(_(@"Error!"),
				       _(@"The mailbox create operation failed."),
				       _(@"OK"),
				       NULL,
				       NULL,
				       NULL);
	  [NSApp abortModal];
	  [self close];
	  return;
	}
    }
  
  [NSApp stopModal];
  [self close];
}


- (IBAction) cancelClicked: (id) sender
{
  [NSApp abortModal];
  [self close];
}

//
// delegate methods
//

- (void) windowWillClose: (NSNotification *) theNotification
{
  // We do nothing
}

- (void) windowDidLoad
{
  name = nil;
}

//
// access/mutation methods
//

- (MailboxManagerController *) mailboxManagerController
{
  return mailboxManagerController;
}

//
// We must NOT retain the mailbox manager controller here.
//
- (void) setMailboxManagerController: (MailboxManagerController *) theMailboxManagerController
{
  mailboxManagerController = theMailboxManagerController;
}

- (NSString *) name
{
  return name;
}

- (void) setName: (NSString *) theName
{
  if (theName)
    {
      RETAIN(theName);
      RELEASE(name);
      name = theName;
      
      [mailboxNameField setStringValue: name];
    }
  else
    {
      RELEASE(name);
      name = nil;

      [mailboxNameField setStringValue: @""];
    }
}


- (NSTextField *) mailboxNameLabel
{
  return mailboxNameLabel;
}

- (NSTextField *) mailboxNameField
{
  return mailboxNameField;
}

@end
