/*
**  Utilities.m
**
**  Copyright (c) 2001, 2002
**
**  Author: Ludovic Marcotte <ludovic@Sophos.ca>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#import "Utilities.h"

#import "Address.h"
#import "EditWindowController.h"
#import "ExtendedAttachmentCell.h"
#import "FolderNode.h"
#import "FolderNodePopUpItem.h"
#import "GNUMail.h"
#import "GNUMail/GNUMailBundle.h"
#import "GNUMailConstants.h"
#import "MailboxManagerController.h"
#import "MailHeaderCell.h"
#import "MailWindowController.h"
#import "MessageViewWindowController.h"
#import "MimeType.h"
#import "MimeTypeManager.h"
#import "NSRegEx.h"
#import "NSUserDefaultsExtensions.h"
#import "PasswordPanelController.h"
#import "TextEnrichedConsumer.h"

#import <Pantomime/Constants.h>
#import <Pantomime/Folder.h>
#import <Pantomime/IMAPStore.h>
#import <Pantomime/LocalFolder.h>
#import <Pantomime/LocalStore.h>
#import <Pantomime/Message.h>
#import <Pantomime/MimeBodyPart.h>
#import <Pantomime/MimeMultipart.h>
#import <Pantomime/MimeUtility.h>
#import <Pantomime/Store.h>
#import <Pantomime/URLName.h>
#import <Pantomime/InternetAddress.h>

#import <limits.h>

char ent(char **ref);
char *striphtml(char *s);


// Our static vars
static NSMutableDictionary *passwordCache = nil;


@implementation Utilities

//
//
//
+ (void) initialize
{
  if ( !passwordCache )
    {
      NSDebugLog(@"Initializing the password cache.");
      passwordCache = [[NSMutableDictionary alloc] init];
    }
}


//
//
//
+ (NSFont *) fontFromFamilyName: (NSString *) theName
			  trait: (int) theTrait
			   size: (int) theSize
{
  NSArray *allFontNames;
  NSString *aFontName;
  NSFont *aFont;
  int i;
  
  allFontNames = [[NSFontManager sharedFontManager] availableMembersOfFontFamily: theName];
  aFontName = nil;

  if ( theName )
    {
      for (i = 0; i < [allFontNames count]; i++)
	{
	  NSArray *attributes;
	  
	  attributes = [allFontNames objectAtIndex: i];
	  
	  // We verify if the font name has the trait we are looking for
	  if ( ([[attributes objectAtIndex: 3] intValue] & theTrait) == theTrait )
	    {
	      aFontName = [attributes objectAtIndex: 0];
	      break;
	    }
	}
    }

  if ( aFontName )
    {
      aFont = [NSFont fontWithName: aFontName
		      size: theSize];
    }
  else
    {
      switch ( theTrait )
	{
	case NSFixedPitchFontMask:
	  aFont = [NSFont userFixedPitchFontOfSize: theSize];
	  break;

	case NSBoldFontMask:
	  aFont = [NSFont boldSystemFontOfSize: theSize];
	  break;
	  
	case NSUnboldFontMask:
	default:
	  aFont = [NSFont systemFontOfSize: theSize];
	  break;
	}
    }

  return aFont;
}


//
//
//
+ (void) appendAddress: (Address *) theAddress
	   toTextField: (NSTextField *) theTextField
{
  NSRange aRange;

  aRange = [[theTextField stringValue] rangeOfString:[theAddress formattedAddress]
				       options: NSCaseInsensitiveSearch];
  
  if ( aRange.location != NSNotFound )
    {
      return; 
    }

  if ( [[theTextField stringValue] length] )
    {
      [theTextField setStringValue: 
		      [NSString stringWithFormat: @"%@, %@",
				[theTextField stringValue],
				[theAddress formattedAddress]]];
    }
  else
    {
      [theTextField setStringValue: [theAddress formattedAddress]];
    }
}


//
// This method returns a NSAttributedString object that has been built
// from the content of the message.
//
+ (NSAttributedString *) attributedStringFromContentForPart: (Part *) thePart
{
  NSMutableAttributedString *maStr;
  NSMutableDictionary *tAttr;
  
  tAttr = [[NSMutableDictionary alloc] init];
  
  [tAttr setObject: [Utilities fontFromFamilyName: [[NSUserDefaults standardUserDefaults] 
						     objectForKey: @"MESSAGE_FONT_NAME"]
			       trait: NSUnboldFontMask
			       size: [[NSUserDefaults standardUserDefaults] floatForKey: @"MESSAGE_FONT_SIZE"]]
	 forKey: NSFontAttributeName];
  
  maStr = [[NSMutableAttributedString alloc] init];
  
  if ( [[thePart content] isKindOfClass: [MimeMultipart class]] )
    {
      // We first verify if our multipart object is a multipart alternative.
      // If yes, we represent the best representation of the part.
      if ( [thePart isMimeType: @"multipart" : @"alternative"] )
	{
	  // We append our \n to separate our body part from the headers
	  [maStr appendAttributedString: [Utilities attributedStringWithString: @"\n" 
						    attributes: nil] ];
	  
	  // We then append the best representation from this multipart/alternative object
	  [maStr appendAttributedString: [Utilities bestRepresentationFromMultipartAlternative:
						      (MimeMultipart *)[thePart content]] ];
	}
      // Then we verify if our multipart object is a multipart appledouble.
      else if ( [thePart isMimeType: @"multipart" : @"appledouble"] )
	{
	  // We append our \n to separate our body part from the headers
	  [maStr appendAttributedString: [Utilities attributedStringWithString: @"\n" 
						    attributes: nil] ];
	  
	  // We then append the best representation from this multipart/appledouble object
	  [maStr appendAttributedString: [Utilities bestRepresentationFromMultipartAppleDouble:
						      (MimeMultipart *)[thePart content]] ];
	}
      // We have a multipart/mixed or multipart/related (or an other, unknown multipart/* object)
      else
	{
	  MimeMultipart *aMimeMultipart;
	  MimeBodyPart *aMimeBodyPart;
	  int i;
      
	  aMimeMultipart = (MimeMultipart*)[thePart content];
      
	  for (i = 0; i < [aMimeMultipart count]; i++)
	    {
	      // We get our part
	      aMimeBodyPart = [aMimeMultipart bodyPartAtIndex: i];
	      
	      // We recursevely call our method to show parts
	      [maStr appendAttributedString: [Utilities attributedStringFromContentForPart: aMimeBodyPart]];
	    }
	}
    }
  // We have a message with Content-Type: application/* OR audio/* OR image/* OR video/*
  // We can also have a text/* part that was base64 encoded, but, we skip it. It'll be
  // treated as a NSString object. (See below).
  else if ( [[thePart content] isKindOfClass:[NSData class]] )
    {
      NSTextAttachment *aTextAttachment;
      ExtendedAttachmentCell *cell;
      NSFileWrapper *aFileWrapper;
      NSImage *anImage;
      
      NSRect rectOfTextView;
      NSSize imageSize;
      
      
      aFileWrapper = [[NSFileWrapper alloc] initRegularFileWithContents: (NSData *)[thePart content]];
      
      if (! [thePart filename] )
	{
	  [aFileWrapper setPreferredFilename: @"unknown"];
	}
      else
	{
	  [aFileWrapper setPreferredFilename: [thePart filename]];
	}
      
      // If the image is a tiff, we create it directly from a NSData
      if ( [thePart isMimeType: @"image": @"tiff"] )
	{
	  anImage = [[NSImage alloc] initWithData:(NSData *)[thePart content]];
          AUTORELEASE(anImage);
	}
      // If not, we temporary write it to a file and we init everything from
      // the content of the file. Libwraster will be used here to load the image
      // from the file. We simply remove the file when we're done.
      //
      // NOTE: Currently, GNUstep doesn't allow creating images other than TIFF from a NSData object.
      //       This is a limitation that will be fixed in the future.
      //       We currently write the data object to a file and initialize an image from
      //       the content of this file. Libwraster's functions will be used by the xgps
      //       backend to decode formats other than TIFF.
      //       - Ludo
      else if ( [thePart isMimeType: @"image": @"*"] )
	{
	  NSString *aString;

	  aString = [NSString stringWithFormat:@"%@/%d_%@", NSTemporaryDirectory(), 
			      [[NSProcessInfo processInfo] processIdentifier],
			      [[thePart filename] lowercaseString]];

	  [(NSData *)[thePart content] writeToFile: aString
		     atomically: YES];
	  
	  anImage = [[NSImage alloc] initWithContentsOfFile: aString];
          AUTORELEASE(anImage);
	  
	  [[NSFileManager defaultManager] removeFileAtPath: aString
					  handler: nil];
	}
      //
      // It's not an image, let's query our MIME type manager in order if we can
      // get an icon from that attachment. If not, we just set a default one.
      //
      else
	{
	  NSString *aString;

	  aString = [[aFileWrapper preferredFilename] pathExtension];

	  if ( aString && [aString length] )
	    {
	      MimeType *aMimeType;

	      aMimeType = [[MimeTypeManager singleInstance] mimeTypeForFileExtension: aString];
	      
	      if (aMimeType && [aMimeType icon])
		{
		  anImage = [aMimeType icon];
		}
	      else
		{
#ifdef MACOSX
                  anImage = [[NSWorkspace sharedWorkspace] iconForFileType: aString];
#else
		  anImage = [NSImage imageNamed: @"common_Unknown.tiff"];
#endif
		}
	    }
	  else
	    {
#ifdef MACOSX
              anImage = [[NSWorkspace sharedWorkspace] iconForFileType: @""];
#else
	      anImage = [NSImage imageNamed: @"common_Unknown.tiff"];
#endif
	    }
	}
      
      // If the image has been loaded sucessfully, it become our icon for our filewrapper
      if ( anImage )
	{
	  [aFileWrapper setIcon: anImage];
	} 	
      
      // We now rescale the attachment if it doesn't fit in the text view. That could happen
      // very often for images.
      rectOfTextView = [[[[GNUMail lastMailWindowOnTop] windowController] textView] frame];
      imageSize = [[aFileWrapper icon] size];
      
      if ( imageSize.width > rectOfTextView.size.width )
	{
	  double delta =  1.0 / ( imageSize.width / rectOfTextView.size.width );
	  
	  [[aFileWrapper icon] setScalesWhenResized: YES];
	  [[aFileWrapper icon] setSize: NSMakeSize( (imageSize.width * delta), imageSize.height * delta)];
	}
      
      // We now create our text attachment with our file wrapper
      aTextAttachment = [[NSTextAttachment alloc] initWithFileWrapper: aFileWrapper];
      
      // We add this attachment to our 'Save Attachment' menu
      [(GNUMail *)[NSApp delegate] addItemToMenuFromTextAttachment: aTextAttachment];
        
      cell = [[ExtendedAttachmentCell alloc] initWithFilename: [aFileWrapper preferredFilename]
					     size: [(NSData *)[thePart content] length] ];
      [cell setPart: thePart];
      
      [aTextAttachment setAttachmentCell: cell];
      
      // Cocoa bug
#ifdef MACOSX
      [cell setAttachment: aTextAttachment];
      [cell setImage: [aFileWrapper icon]];
#endif
      RELEASE(cell);
      RELEASE(aFileWrapper);
      
      // We separate the text attachment from any other line previously shown
      [maStr appendAttributedString: [Utilities attributedStringWithString: @"\n" 
						attributes: nil] ];
      
      [maStr appendAttributedString: [NSAttributedString attributedStringWithAttachment: aTextAttachment]];
      RELEASE(aTextAttachment);
    }
  //
  // We have a message/rfc822 as the Content-Type.
  //
  else if ( [[thePart content] isKindOfClass:[Message class]] )
    {
      Message *aMessage;
      
      aMessage = (Message *)[thePart content];
      
      [maStr appendAttributedString: [Utilities attributedStringWithString: @"\n" 
						attributes: nil] ];
      
      // We must represent this message/rfc822 part as an attachment
      if ( [thePart contentDisposition] &&
	   ([[thePart contentDisposition] caseInsensitiveCompare: @"attachment"] == NSOrderedSame) )
	{
	  NSTextAttachment *aTextAttachment;
	  ExtendedAttachmentCell *cell;
	  NSFileWrapper *aFileWrapper;
	  NSData *aData;

	  aData = [aMessage rawSource];
	  
	  aFileWrapper = [[NSFileWrapper alloc] initRegularFileWithContents: aData];
	  
	  [aFileWrapper setPreferredFilename: @"message/rfc822 attachment"];
	  
	  aTextAttachment = [[NSTextAttachment alloc] initWithFileWrapper: aFileWrapper];
	  
	  // We add this attachment to our 'Save Attachment' menu
	  [(GNUMail *)[NSApp delegate] addItemToMenuFromTextAttachment: aTextAttachment];
          
	  cell = [[ExtendedAttachmentCell alloc] initWithFilename: [aFileWrapper preferredFilename]
						 size: [aData length] ];
	  [cell setPart: thePart];
	  
	  [aTextAttachment setAttachmentCell: cell];
          
          // Cocoa bug
#ifdef MACOSX
          [cell setAttachment: aTextAttachment];
          [cell setImage: [aFileWrapper icon]];
#endif
	  RELEASE(cell); 
	  RELEASE(aFileWrapper);
	  
	  [maStr appendAttributedString: [NSAttributedString attributedStringWithAttachment:
							       aTextAttachment]];
	  RELEASE(aTextAttachment);
	}
      // Its inline..
      else
	{	      
	  [maStr appendAttributedString: [Utilities attributedStringFromHeadersForMessage: aMessage
						    showAllHeaders: NO] ];
	  [maStr appendAttributedString: [Utilities attributedStringFromContentForPart: aMessage]];
	}
    }
  //
  // We have a message with Content-Type: text/*
  // or at least, a part that we can represent directly in our textView.
  // 
  else if ( [[thePart content] isKindOfClass:[NSString class]] )
    { 
      NSString *aString;
      
      aString = (NSString *)[thePart content];

      // We must represent our NSString as an text attachment in our textView.
      if ( [thePart contentDisposition] &&
	   ([[thePart contentDisposition] caseInsensitiveCompare: @"attachment"] == NSOrderedSame) )
	{
	  NSTextAttachment *aTextAttachment;
	  ExtendedAttachmentCell *cell;
	  NSFileWrapper *aFileWrapper;
	  NSData *aData;
	  
	  aData = [aString dataUsingEncoding: NSUTF8StringEncoding];
	  aFileWrapper = [[NSFileWrapper alloc] initRegularFileWithContents: aData];
	  
	  if (! [thePart filename] )
	    {
	      [aFileWrapper setPreferredFilename: @"unknown"];
	    }
	  else
	    {
	      [aFileWrapper setPreferredFilename: [thePart filename]];
	    }
	  
	  aTextAttachment = [[NSTextAttachment alloc] initWithFileWrapper: aFileWrapper];
	  
	  // We add this attachment to our 'Save Attachment' menu
	  [(GNUMail *)[NSApp delegate] addItemToMenuFromTextAttachment: aTextAttachment];

	  cell = [[ExtendedAttachmentCell alloc] initWithFilename: [aFileWrapper preferredFilename]
						 size: [aData length] ];
	  [cell setPart: thePart];
	  
	  [aTextAttachment setAttachmentCell: cell];
          
          // Cocoa bug
#ifdef MACOSX
          [cell setAttachment: aTextAttachment];
          [cell setImage: [aFileWrapper icon]];
#endif
	  RELEASE(cell);
	  RELEASE(aFileWrapper);
	  
	  // We separate the text attachment from any other line previously shown
	  [maStr appendAttributedString: [Utilities attributedStringWithString: @"\n" 
						    attributes: nil] ];
		      
	  [maStr appendAttributedString: [NSAttributedString attributedStringWithAttachment:
							       aTextAttachment]];
	  RELEASE(aTextAttachment);
	}
      // It's inline...
      else
	{
	  //
	  //
	  if ( [thePart isMimeType: @"text": @"html"] )
	    {
#ifdef MACOSX
	      NSAttributedString *anAttributedString;
	      NSData *aData;
              int encoding;
	      
              encoding = [MimeUtility stringEncodingForCharset: [[thePart charset] dataUsingEncoding: NSASCIIStringEncoding]];
              
              if ( encoding == -1 )
                {
                  encoding = NSASCIIStringEncoding;
                }
              
	      aData = [aString dataUsingEncoding: encoding];
	      
	      anAttributedString = [[NSAttributedString alloc] initWithHTML: aData
							       documentAttributes: nil];
	      
	      [maStr appendAttributedString: anAttributedString];
	      RELEASE(anAttributedString);
#else
	      char *cString;

	      NSDebugLog(@"Showing a text/html part. No formatting will be done for now.");

	      cString = striphtml( (char *)[aString lossyCString] );
	      [maStr appendAttributedString: [Utilities attributedStringWithString: [NSString stringWithCString: cString]
							attributes: tAttr] ];
#endif
	    }
	  else if ( [thePart isMimeType: @"text": @"enriched"] )
	    {
	      [maStr appendAttributedString: [TextEnrichedConsumer attributedStringFromTextEnrichedString: aString]];
	    }
	  else if ( [thePart isMimeType: @"text": @"rtf"] )
	    {
	      NSAttributedString *anAttributedString;
	      NSData *aData;
	      
	      aData = [aString dataUsingEncoding: NSASCIIStringEncoding];
	      
	      anAttributedString = [[NSAttributedString alloc] initWithRTF: aData
							       documentAttributes: NULL];
	      [maStr appendAttributedString: anAttributedString];
	      RELEASE(anAttributedString); 
	    }
	  // We surely got a text/plain part
	  else
	    {
	      NSMutableDictionary *plainTextMessageAttributes;
	      
	      if ( [[NSUserDefaults standardUserDefaults] 
		     objectForKey: @"USE_FIXED_FONT_FOR_TEXT_PLAIN_MESSAGES"] &&
		   [[NSUserDefaults standardUserDefaults] 
		     integerForKey: @"USE_FIXED_FONT_FOR_TEXT_PLAIN_MESSAGES"] == NSOnState )
		{
		  plainTextMessageAttributes = [[NSMutableDictionary alloc] init];
		  AUTORELEASE(plainTextMessageAttributes);

		  [plainTextMessageAttributes setObject:
						[Utilities fontFromFamilyName: 
							     [[NSUserDefaults standardUserDefaults] 
							       objectForKey: @"PLAIN_TEXT_MESSAGE_FONT_NAME"]
							   trait: NSFixedPitchFontMask
							   size: [[NSUserDefaults standardUserDefaults]
								   floatForKey: @"PLAIN_TEXT_MESSAGE_FONT_SIZE"]]
					      forKey: NSFontAttributeName];
		}
	      else
		{
		  plainTextMessageAttributes = tAttr;
		}
	      
	      [maStr appendAttributedString: [Utilities attributedStringWithString: aString
							attributes: plainTextMessageAttributes] ];
	      
	    }
	}
    }
  //
  // We have something that we probably can't display.
  // Let's inform the user about this situation.
  // 
  else
    {
      NSRunAlertPanel(_(@"Error!"),
		      _(@"A part of this E-Mail can't be represented. Please report this as a bug."),
		      _(@"OK"), // default
		      NULL,     // alternate
		      NULL);
    }
  
  RELEASE(tAttr);

  return AUTORELEASE(maStr);
}


//
// This method returns a NSAttributedString object that has been built
// from the headers of the message.
//
+ (NSAttributedString *) attributedStringFromHeadersForMessage: (Message *) theMessage
						showAllHeaders: (BOOL) aBOOL
{
  NSMutableDictionary *headerNameAttribute, *headerValueAttribute;
  NSMutableAttributedString *maStr; 

  NSMutableAttributedString *aMutableAttributedString;
  NSTextAttachment *aTextAttachment;

  MailHeaderCell *theMailHeaderCell;

  NSDictionary *allHeaders;
  NSArray *headersToShow;
  int i;

  maStr = [[NSMutableAttributedString alloc] init];
  

  // Attributes for our header names
  headerNameAttribute = [[NSMutableDictionary alloc] init];

  [headerNameAttribute setObject: [Utilities fontFromFamilyName: [[NSUserDefaults standardUserDefaults] 
								   objectForKey: @"HEADER_NAME_FONT_NAME"]
					     trait: NSBoldFontMask
					     size: [[NSUserDefaults standardUserDefaults] 
						     floatForKey: @"HEADER_NAME_FONT_SIZE"]]
		       forKey: NSFontAttributeName];


  // Attributes for our header values
  headerValueAttribute = [[NSMutableDictionary alloc] init];

  [headerValueAttribute setObject: [Utilities fontFromFamilyName: [[NSUserDefaults standardUserDefaults]
								    objectForKey: @"HEADER_VALUE_FONT_NAME"]
					      trait: NSUnboldFontMask
					      size: [[NSUserDefaults standardUserDefaults] 
						      floatForKey: @"HEADER_VALUE_FONT_SIZE"]]
			forKey: NSFontAttributeName];
			
  
  //
  allHeaders = [theMessage allHeaders];
  
  // We verify which headers of the message we show show.
  if ( aBOOL )
    {
      headersToShow = [allHeaders allKeys];
    }
  else
    {
      headersToShow = [[NSUserDefaults standardUserDefaults] objectForKey:@"SHOWNHEADERS"];
    }

  if ( headersToShow )
    {  
      for (i = 0; i < [headersToShow count]; i++)
	{
	  NSString *anHeader = [headersToShow objectAtIndex: i];
	  
	  if ( [anHeader caseInsensitiveCompare: @"Date"] == NSOrderedSame &&
	       [theMessage receivedDate] )
	    {
	      if ( [theMessage receivedDate] )
		{
		  [maStr appendAttributedString: [Utilities attributedStringWithString: @"Date: "
							    attributes: headerNameAttribute] ];
		  [maStr appendAttributedString: [Utilities attributedStringWithString: [[theMessage receivedDate] description]
							    attributes: headerValueAttribute] ];
		  [maStr appendAttributedString: [Utilities attributedStringWithString: @"\n"
							    attributes: nil] ];
		}
	    }
	  else if ( [anHeader caseInsensitiveCompare: @"From"] == NSOrderedSame &&
		    [theMessage from] )
	    {
	      [maStr appendAttributedString: [Utilities attributedStringWithString: @"From: "
							attributes: headerNameAttribute] ];
	      [maStr appendAttributedString: [Utilities attributedStringWithString: [[theMessage from] unicodeStringValue]
							attributes: headerValueAttribute] ];
	      [maStr appendAttributedString: [Utilities attributedStringWithString: @"\n"
							attributes: nil] ];
	    }
	  else if ( [anHeader caseInsensitiveCompare: @"Bcc"] == NSOrderedSame )
	    {
	      NSString *bccStr = [Utilities stringFromRecipients: [theMessage recipients]
					    type: BCC];
	      
	      if ( [bccStr length] > 0 )
		{
		  [maStr appendAttributedString: [Utilities attributedStringWithString: @"Bcc: "
							    attributes: headerNameAttribute] ];
		  [maStr appendAttributedString: [Utilities attributedStringWithString: [bccStr substringToIndex:([bccStr length] - 2)]
							    attributes: headerValueAttribute] ];
		  [maStr appendAttributedString: [Utilities attributedStringWithString: @"\n"
							    attributes: nil] ];
		}
	    }
	  else if ( [anHeader caseInsensitiveCompare: @"Cc"] == NSOrderedSame )
	    {
	      NSString *ccStr= [Utilities stringFromRecipients: [theMessage recipients]
					  type: CC];
	      
	      if ( [ccStr length] > 0 )
		{
		  [maStr appendAttributedString: [Utilities attributedStringWithString: @"Cc: "
							    attributes: headerNameAttribute] ];
		  [maStr appendAttributedString: [Utilities attributedStringWithString: [ccStr substringToIndex:([ccStr length] - 2)]
							    attributes: headerValueAttribute] ];
		  [maStr appendAttributedString: [Utilities attributedStringWithString: @"\n"
							    attributes: nil] ];		}
	    }
	  else if ( [anHeader caseInsensitiveCompare: @"Reply-To"] == NSOrderedSame )
	    { 
	      InternetAddress *anInternetAddress = [theMessage replyTo];
	      
	      if ( anInternetAddress )
		{
		  [maStr appendAttributedString: [Utilities attributedStringWithString: @"Reply-To: "
							    attributes: headerNameAttribute] ];
		  [maStr appendAttributedString: [Utilities attributedStringWithString: [[theMessage replyTo] unicodeStringValue]
							    attributes: headerValueAttribute] ];
		  [maStr appendAttributedString: [Utilities attributedStringWithString: @"\n"
							    attributes: nil] ];
		}	
	    }
	  else if ( [anHeader caseInsensitiveCompare: @"To"] == NSOrderedSame )
	    {
	      NSString *toStr = [Utilities stringFromRecipients: [theMessage recipients]
					   type: TO];
	      
	      if ( [toStr length] > 0 )
		{
		  [maStr appendAttributedString: [Utilities attributedStringWithString: @"To: "
							    attributes: headerNameAttribute] ];
		  [maStr appendAttributedString: [Utilities attributedStringWithString: [toStr substringToIndex:([toStr length] - 2)]
							    attributes: headerValueAttribute] ];
		  [maStr appendAttributedString: [Utilities attributedStringWithString: @"\n"
							    attributes: nil] ];
		}
	    }
	  else if ( [anHeader caseInsensitiveCompare: @"Content-Type"] == NSOrderedSame )
	    {
	      NSString *aString;
	      
	      if ( [theMessage charset] )
		{
		  aString = [NSString stringWithFormat: @"%@; charset=%@",
				      [theMessage contentType], [theMessage charset]];
		}
	      else
		{
		  aString = [theMessage contentType];
		}
	      
	      [maStr appendAttributedString: [Utilities attributedStringWithString: @"Content-Type: "
							attributes: headerNameAttribute] ];
	      [maStr appendAttributedString: [Utilities attributedStringWithString: aString
							attributes: headerValueAttribute] ];
	      [maStr appendAttributedString: [Utilities attributedStringWithString: @"\n"
							attributes: nil] ];
	    }
	  else
	    {
	      NSArray *allKeys = [allHeaders allKeys];
	      NSString *valueString = nil;
	      int j;
	      
	      for (j = 0; j < [allKeys count]; j++)
		{
		  if ( [[[allKeys objectAtIndex: j] uppercaseString] isEqualToString: [anHeader uppercaseString]] )
		    {
		      if ( [[allHeaders objectForKey:
					  [allKeys objectAtIndex: j]] isKindOfClass: [InternetAddress class]] )
			{
			  valueString = [[allHeaders objectForKey: [allKeys objectAtIndex: j]] unicodeStringValue];
			}
		      else
			{
			  valueString = [allHeaders objectForKey: [allKeys objectAtIndex: j]];
			}
		      break;
		    }
		}
		  
	      if ( valueString )
		{
		  [maStr appendAttributedString: [Utilities attributedStringWithString: [NSString stringWithFormat:@"%@: ", anHeader]
							    attributes: headerNameAttribute] ];
		  [maStr appendAttributedString: [Utilities attributedStringWithString: valueString
							    attributes: headerValueAttribute] ];
		  [maStr appendAttributedString: [Utilities attributedStringWithString: @"\n"
							    attributes: nil] ];
		}
	    }
	} // for (..) 
      
    } // if ( headersToShow )
  
  theMailHeaderCell = [[[GNUMail lastMailWindowOnTop] windowController] mailHeaderCell];
  [theMailHeaderCell setAttributedStringValue: maStr];
  [theMailHeaderCell resize: nil];
  RELEASE(maStr);

  // We now "embed" the header cell into a NSTextAttachment object
  // so we can add it to our mutable attributed string.
  aTextAttachment = [[NSTextAttachment alloc] init];
  [aTextAttachment setAttachmentCell: theMailHeaderCell];
  
  aMutableAttributedString = [[NSMutableAttributedString alloc] initWithString: @"\n"];
  
  [aMutableAttributedString appendAttributedString: 
			      [NSMutableAttributedString attributedStringWithAttachment: aTextAttachment]];
  RELEASE(aTextAttachment);
  
  [aMutableAttributedString appendAttributedString: [Utilities attributedStringWithString: @"\n\n" 
							       attributes: nil]];

  RELEASE(headerNameAttribute);
  RELEASE(headerValueAttribute);

  return AUTORELEASE(aMutableAttributedString);
}


//
//
//
+ (NSAttributedString *) attributedStringWithString: (NSString *) theString
					 attributes: (NSDictionary *) theAttributes
{
  if ( !theAttributes )
    {
      NSMutableDictionary *attributes;
      NSMutableDictionary *aMutableAttributedString;

      attributes = [[NSMutableDictionary alloc] init];
      [attributes setObject: [NSFont systemFontOfSize: 0]
		  forKey: NSFontAttributeName];
      
      aMutableAttributedString = [[NSAttributedString alloc] initWithString: theString
							     attributes: attributes];
      RELEASE(attributes);
      
      return AUTORELEASE( aMutableAttributedString );
    }
  else 
    {
      return AUTORELEASE( [[NSAttributedString alloc] initWithString: theString
						      attributes: theAttributes] );
    }
}


//
// FIXME: Currently, this method only search for the text/plain representation of a
// multipart/alternative object.
//
// In the future, we should try to represent the representation the user wants or, the
// more complex ones first (like text/html or text/enriched).
//
+ (NSAttributedString *) bestRepresentationFromMultipartAlternative: (MimeMultipart *) theMimeMultipart
{
  MimeBodyPart *aMimeBodyPart;
  int i;
  
  for (i = 0; i < [theMimeMultipart count]; i++)
    {
      aMimeBodyPart = [theMimeMultipart bodyPartAtIndex: i];
      
      if ( [aMimeBodyPart isMimeType: @"text": @"plain"] )
	{
	  NSMutableDictionary *tAttr;       // text attributes
	  id aContent;
	  
	  tAttr = [[NSMutableDictionary alloc] init];
	  AUTORELEASE(tAttr);
	  
	  [tAttr setObject: [Utilities fontFromFamilyName: [[NSUserDefaults standardUserDefaults] 
							     objectForKey: @"MESSAGE_FONT_NAME"]
				       trait: NSUnboldFontMask
				       size: [[NSUserDefaults standardUserDefaults] floatForKey: @"MESSAGE_FONT_SIZE"]]
		 forKey: NSFontAttributeName];
	  
          aContent = [aMimeBodyPart content];
	  
	  // We MUST be sure the content isn't nil and also that it's a valid NSString object.
	  // Pantomime will return the raw source (NSData object) if the decoding operation has failed.
	  if ( !aContent || ![aContent isKindOfClass: [NSString class]] )
	    {
	      break;
	    }

	  return [Utilities attributedStringWithString: (NSString *)aContent
			    attributes: tAttr];
	}
    }
  
  // We haven't found a text/plain part. Report this as a bug in GNUMail.app for not supporting
  // the other kind of parts.
  return [Utilities attributedStringWithString: _(@"No text/plain part found. Please report this bug since GNUMail.app doesn't support this other kind of parts.")
		    attributes: nil];
}


//
//
//
+ (NSAttributedString *) bestRepresentationFromMultipartAppleDouble: (MimeMultipart *) theMimeMultipart
{
  NSMutableAttributedString *aMutableAttributedString;
  NSMutableDictionary *attributes;
  
  MimeBodyPart *aMimeBodyPart;
  int i;
  
  // We create a set of attributes (base font, color red)
  attributes = [[NSMutableDictionary alloc] init];
  [attributes setObject: [NSColor redColor]
	      forKey: NSForegroundColorAttributeName];

  aMutableAttributedString = [[NSMutableAttributedString alloc] init];

  for (i = 0; i < [theMimeMultipart count]; i++)
    {
      aMimeBodyPart = [theMimeMultipart bodyPartAtIndex: i];
    
      if ( [aMimeBodyPart isMimeType: @"application": @"applefile"] )
	{
	  [aMutableAttributedString appendAttributedString:
				      [Utilities attributedStringWithString: @"(Decoded Apple file follow...)"
						 attributes: attributes]];
	}
      else
	{
	  // We first add a \n between our applefile description and the 'representaiton' of
	  // the attachment
	  [aMutableAttributedString appendAttributedString:
			      [Utilities attributedStringWithString: @"\n" attributes: nil]];
	  
	  // We add the representation of our attachment
	  [aMutableAttributedString appendAttributedString:
				      [Utilities attributedStringFromContentForPart: aMimeBodyPart]];
	} 
      
    }
  
  // We add a \n to separate everything.
  [aMutableAttributedString appendAttributedString:
			      [Utilities attributedStringWithString: @"\n" attributes: nil]];

  RELEASE(attributes);

  return AUTORELEASE(aMutableAttributedString);
}


//
//
//
+ (NSAttributedString *) formattedAttributedStringFromAttributedString: (NSAttributedString *) theAttributedString
{
  NSMutableAttributedString *aMutableAttributedString = nil;
  NSRange aRange;
  
  id attribute;
  int index;
  
  // If we only have one char (or less), we don't do anything.
  // This could append for example when we only have a image/tiff as the Content-Type of an E-Mail.
  if ( [theAttributedString length] <= 1 )
    {
      return theAttributedString;
    }
  
  index = 0;
  attribute = [theAttributedString attribute: NSAttachmentAttributeName
				   atIndex: index
				   effectiveRange: &aRange];
  
  // We if have NO text attachments at all... we just return our param
  if ( !attribute && 
       (aRange.length == [theAttributedString length]) )
    {
      return theAttributedString;
    }
  
  
  // We initialize our mutable string
  aMutableAttributedString = [[NSMutableAttributedString alloc] initWithAttributedString: theAttributedString];
 
  while ( index < [aMutableAttributedString length] )
    {
      // We found a text attachment...
      if ( attribute )
	{
	  // We search for <<filename>> OR <filename> and if we got one of them,
	  // we simply replace that string with the actual attachment.
	  ExtendedAttachmentCell *cell;
	  
	  cell = [attribute attachmentCell];

	  if ( [cell respondsToSelector: @selector(part)] )
	    {
	      NSString *aString;
	      NSRange r;
	      aString = [aMutableAttributedString string];
	      
	      // We first search for <<filename>>
	      r = [aString rangeOfString: [NSString stringWithFormat: @"<<%@>>", [[cell part] filename]]];
	      
	      // If not found, we search for <filename>
	      if ( r.length == 0 )
		{
		  r = [aString rangeOfString: [NSString stringWithFormat: @"<%@>", [[cell part] filename]]];
		}
	      
	      // We found something, let's replace <filename> (or <<filename>>) by the actual
	      // attachment cell.
	      if ( r.length )
		{
		  NSAttributedString *attachment;
		  
		  // We get the attributed that contains our attachment
		  attachment = [aMutableAttributedString attributedSubstringFromRange: aRange];
		  RETAIN(attachment);
	      
		  // FIXME - we should verify if the range of that attachment > of the range of
		  // the <filename in case the disposition of our E-Mail is 'wrong'.
		  
		  // We remove it from our string
		  [aMutableAttributedString deleteCharactersInRange: aRange];
		  
		  // We replace
		  [aMutableAttributedString replaceCharactersInRange: r
					    withAttributedString: attachment];
		  
		  RELEASE(attachment);
		  
		  // We adjust our index for the search
		  index = r.location + 1;
		}
	      else
		{
		  index = NSMaxRange(r);
		}
	    }
	  else
	    {
	      index = NSMaxRange(aRange);
	    }
	}
      else // of if ( attribute ) 
	{
	  index = NSMaxRange(aRange);
	}
      
      if (index >= [aMutableAttributedString length])
      	{
      	  break;
      	}

      attribute = [aMutableAttributedString attribute: NSAttachmentAttributeName
					    atIndex: index
					    effectiveRange: &aRange];
    }

  return aMutableAttributedString;
}


//
//
//
+ (NSString *) stringFromRecipients: (NSArray *) allRecipients
			       type: (int) recipientType
{
  InternetAddress *anInternetAddress;
  NSMutableString *aMutableString;
  int i;
  
  aMutableString = [[NSMutableString alloc] init];

  for (i = 0; i < [allRecipients count]; i++)
    {
      anInternetAddress = [allRecipients objectAtIndex: i];
      
      if ([anInternetAddress type] == recipientType)
	{
	  [aMutableString appendFormat: @"%@, ", [anInternetAddress unicodeStringValue]];
	}
      
    }
  
  return AUTORELEASE(aMutableString); 
}


//
//
//
+ (NSString *) encryptPassword: (NSString *) thePassword
                       withKey: (NSString *) theKey
{
  NSMutableString *key;
  NSMutableData *encryptedPassword;
  NSString *result;
  int i;
  unichar p, k, e;

  // The length of the key must be greater (or equal) than
  // the length of the password
  key = [[NSMutableString alloc] init];
  
  while ( [key length] < [thePassword length] )
    {
      [key appendString: theKey];
    }

  encryptedPassword = [[NSMutableData alloc] init];
 
  for ( i = 0; i < [thePassword length]; i++ )
    {
      p = [thePassword characterAtIndex: i];
      k = [key characterAtIndex: i];

      e = p ^ k;
      
      [encryptedPassword appendBytes: (void *)&e length: 2];
    }

  result = AUTORELEASE([[NSString alloc] initWithData: [MimeUtility encodeBase64: encryptedPassword lineLength: 0]
                             encoding: NSASCIIStringEncoding]);

  RELEASE(encryptedPassword);
  RELEASE(key);

  return result;
}


//
//
//
+ (NSString *) decryptPassword: (NSString *) thePassword
                       withKey: (NSString *) theKey
{
  NSMutableString *key;
  NSMutableString *password;
  NSData *dec;
  unsigned char *decryptedPassword;
  NSString *result;
  int i;
  unichar p, k, d;

  if ( nil == thePassword || nil == theKey )
    {
      return nil;
    }

  // We 'verify' if the password is not encoded in base64
  // We should not rely on this method but it's currently the best guess we could make
  if ( [thePassword length] == 0 || 
       ([thePassword length] & 0x03) || 
       [theKey length] == 0)
    {
      return thePassword;
    }

  // The length of the key must be greater (or equal) than
  // the length of the password
  key = [[NSMutableString alloc] init];
  
  while ( [key length] < [thePassword length] )
    {
      [key appendString: theKey];
    }

  password = [[NSMutableString alloc] init];

  dec = [MimeUtility decodeBase64: [thePassword dataUsingEncoding: NSASCIIStringEncoding]];
  decryptedPassword = (unsigned char *)[dec bytes];
  
  for ( i = 0; i < [dec length]; i += 2 )
    {
      d = decryptedPassword[i] | decryptedPassword[i+1];
      k = [key characterAtIndex: i/2];

      p = d ^ k;
      [password appendString: [NSString stringWithCharacters: &p length: 1]];
    }

  result = [[NSString alloc] initWithString: password];

  RELEASE(password);
  RELEASE(key);

  return AUTORELEASE(result);
}


//
//
//
+ (void) loadPersonalProfilesInPopUpButton: (NSPopUpButton *) thePopUpButton selectedProfile: (NSString *) profile
{
  NSDictionary *allPersonalProfiles;
  NSArray *allKeys;
  int i, index;
  
  NSString *defaultProfile = nil;

  if(profile)
	  defaultProfile = profile;
  else
	  defaultProfile = [[NSUserDefaults standardUserDefaults] objectForKey: @"DEFAULT_PERSONAL"];
  index = 0;
  
  allPersonalProfiles = [[NSUserDefaults standardUserDefaults] objectForKey: @"PERSONAL"];
  
  allKeys = [[allPersonalProfiles allKeys] sortedArrayUsingSelector: @selector(caseInsensitiveCompare:)];
  
  // We initialize our popup button
  [thePopUpButton removeAllItems];
  
  for (i = 0; i < [allKeys count]; i++)
    {
      NSDictionary *aDictionary;
      
      aDictionary = [allPersonalProfiles objectForKey: [allKeys objectAtIndex: i]];
      
      if ( defaultProfile && 
	   [[allKeys objectAtIndex: i] isEqualToString: defaultProfile] ) 
	{
	  index = i;
	}

      [thePopUpButton insertItemWithTitle: [NSString stringWithFormat: @"%@ (%@)", 
						     [aDictionary objectForKey: @"EMAILADDR"],
						     [allKeys objectAtIndex: i]]
		      atIndex: i];
    }
  
  [thePopUpButton selectItemAtIndex: index];

  [thePopUpButton synchronizeTitleAndSelectedItem];
}


//
//
//
+ (void) loadTransportMethodsInPopUpButton: (NSPopUpButton *) thePopUpButton
{
  NSArray *anArray;
  int i;

  // We initialize our popup button used to select the transport methods
  [thePopUpButton removeAllItems];

  anArray = [[NSUserDefaults standardUserDefaults] objectForKey: @"SENDING"];
  
  for (i = 0; i < [anArray count]; i++)
    {
      NSDictionary *aDictionary;
      NSString *aString;
      
      aDictionary = [anArray objectAtIndex: i];
      
      if ( [[aDictionary objectForKey: @"TRANSPORT_METHOD"] intValue] == TRANSPORT_SMTP)
	{
	  aString = [NSString stringWithFormat: @"%@ SMTP (%@)",
			      [aDictionary objectForKey: @"PERSONAL_PROFILE"],
			      [aDictionary objectForKey: @"SMTP_HOST"]];
	}
      else
	{
	  aString = [NSString stringWithFormat: @"%@ Mailer (%@)",
			      [aDictionary objectForKey: @"PERSONAL_PROFILE"],
			      [aDictionary objectForKey: @"MAILER_PATH"]];
	}
      
      
      [thePopUpButton insertItemWithTitle: aString
		      atIndex: i];
    }
}


//
//
//
+ (NSString *) profileNameForItemTitle: (NSString *) theString
{
  NSString *aProfileName;
  NSRange aRange;
  
  // We first extract the profile name from the popup title.
  aRange = [theString rangeOfString: @"("
		      options: NSBackwardsSearch];
  
  aProfileName = [theString substringWithRange: NSMakeRange(aRange.location + 1,
							    [theString length] - aRange.location - 2)];
  
  return aProfileName;
}


//
//
//
+ (id) windowForFolderName: (NSString *) theName
		     store: (Store *) theStore
{
  NSArray *allWindows;
	      
  // We get all opened windows
  allWindows = [GNUMail allMailWindows];
  
  if ( allWindows )
    {
      Folder *aFolder;
      id aWindow;
      int i;
      
      for (i = 0; i < [allWindows count]; i++)
	{
	  aWindow = [allWindows objectAtIndex: i];
	  aFolder = [(MailWindowController *)[aWindow windowController] folder];
	  
	  // If we found our opened folder
	  if ( [[aFolder name] isEqualToString: theName] && 
	       [aFolder store] == theStore )
	    {
	      return aWindow;
	    }
	}
    }
  
  return nil;
}


//
//
//
+ (FolderNode *) folderNodesFromFolders: (NSEnumerator *) theFolders
			      separator: (NSString *) theSeparator
{
  NSString *aString; 
  NSRange aRange;
  int length;
 
  FolderNode *root;
 
  root = [[FolderNode alloc] init];
  [root setParent: nil];

  if ( !theSeparator )
    {
      theSeparator = [NSString stringWithString: @"/"];
    }

  length = [theSeparator length];

  while ( ((aString = [theFolders nextObject])) )
    {
      aRange = [aString rangeOfString: theSeparator];
      
      if ( aRange.length > 0)
	{ 
	  FolderNode *parent;
	  int mark;
	  
	  parent = root;
	  mark = 0;
	  
	  while ( aRange.length > 0 )
	    {
	      if ( mark == aRange.location )
		{
		  mark += length; 
		}
	      else
		{
		  NSString *aName;
		  
		  aName = [aString substringWithRange: NSMakeRange(mark, aRange.location - mark)];
		  
		  if ( ![parent childWithName: aName] )
		    {
		      [parent addChild: [FolderNode folderNodeWithName: aName
						    parent: parent]];
		    }
		  
		  parent = [parent childWithName: aName];
		  mark = aRange.location + aRange.length;
		}
	      
	      aRange = [aString rangeOfString: theSeparator
				options: 0
				range: NSMakeRange(mark, [aString length] - mark)];
	    }
	  
	  [parent addChild: [FolderNode folderNodeWithName: [aString substringFromIndex: mark]
					parent: parent]];
	}
      else
	{
	  if ( ![root childWithName: aString] )
	    {
	      [root addChild: [FolderNode folderNodeWithName: aString
					  parent: root]];
	    }
	}
    }

  return AUTORELEASE(root);
}


//
// This method returns a node using a thePath to search
// and rootNode as the starting point in it's search.
//
// The path to the node MUST always use @"/" as the folder separator.
//
+ (FolderNode *) folderNodeForPath: (NSString *) thePath
			     using: (FolderNode *) rootNode
{
  NSArray *pathComponents;
  FolderNode *aFolderNode;
  int i, j;

  pathComponents = [thePath pathComponents];
  aFolderNode = rootNode;

  for (i = 0; i < [pathComponents count]; i++)
    {
      NSString *aPathComponent;

      aPathComponent = [pathComponents objectAtIndex: i];
      
      if ( [aPathComponent isEqualToString: @"/"] )
	{
	  continue;
	}

      for (j = 0; j < [aFolderNode childCount]; j++)
	{
	  if ( [[[aFolderNode childAtIndex: j] name] isEqualToString: aPathComponent] )
	    {
	      aFolderNode = [aFolderNode childAtIndex: j];
	      break;
	    }
	}
    }
  
  return aFolderNode;
}


//
// This method build a complete path of a node. It will stop
// building the path when parent == nil is reached.
//
// We always return a /<Store name>/folder/subfolder/subsubfolder
// or                 /<Store name>/folder.subfolder.subsubfolder
//
+ (NSString *) completePathForFolderNode: (FolderNode *) theFolderNode
			       separator: (NSString *) theSeparator
{
  NSMutableString *aMutableString;
  FolderNode *parent;

  aMutableString = [[NSMutableString alloc] init];
  parent = theFolderNode;

  if ( !theSeparator )
    {
      theSeparator = @"/";
    }

  // We don't loop for no reason if we root was passed as the parameter
  if ( ![parent parent] )
    {
      return [NSString stringWithFormat: @"/%@/", [parent name]];
    }
  
  while ( parent != nil )
    {
      [aMutableString insertString: [parent name]
		      atIndex: 0];
      
      // We verify if the parent of that node has a parent.
      // If it doesn't, that means we must add our /<Store name>/
      // and break the loop.
      if ( [[parent parent] parent] )
	{
	  [aMutableString insertString: theSeparator
			  atIndex: 0];
	}
      else
	{
	  [aMutableString insertString: [NSString stringWithFormat: @"/%@/", [[parent parent] name]]
			  atIndex: 0];
	  break;
	}
      
      parent = [parent parent];
    }
  
  return AUTORELEASE(aMutableString);
}


//
// We skip the first <foo bar> in  /<foo bar>/folder/subfolder/mbox
//                             or  /<foo bar>/folder.subfolder.mbox
// in order to only get the full path of the mbox.
//
// If the no mailbox was selected, we return nil.
//
+ (NSString *) pathOfFolderFromFolderNode: (FolderNode *) theFolderNode
				separator: (NSString *) theSeparator
{
  NSString *aString;
  NSRange aRange;
  
  if ( !theSeparator )
    {
      theSeparator = @"/";
    }

  // We build our full path (including /<Store>/) to our folder
  aString = [Utilities completePathForFolderNode: theFolderNode
		       separator: theSeparator];
  
  // We trim the /<Store>/ part.
  aRange = [aString rangeOfString: @"/"
		    options: 0
		    range: NSMakeRange(1, [aString length] - 1)];
  
  if ( aRange.length )
    {
      return [aString substringFromIndex: aRange.location + 1];
    }
  
  return nil;
}


//
//
//
+ (NSString *) flattenPathFromString: (NSString *) theString
			   separator: (NSString *) theSeparator
{
  NSMutableString *aMutableString;
  NSRange aRange;
  
  aMutableString = [[NSMutableString alloc] initWithString: theString];

  aRange = [aMutableString rangeOfString: theSeparator];

  while ( aRange.length )
    {
      [aMutableString replaceCharactersInRange: aRange
		      withString: @"_"];

      aRange = [aMutableString rangeOfString: theSeparator
			       options: 0
			       range: NSMakeRange(aRange.location + 1,
						  [aMutableString length] - aRange.location - 1)];
    }

  return AUTORELEASE(aMutableString);
}


//
// Calculates the store key from a node. The key returned is:
//
// <username> @ <store name>.
//
// The usage of this method only makes sense for FolderNode objects
// used to represent an IMAP store and its folders.
//
+ (NSString *) storeKeyForFolderNode: (FolderNode *) theFolderNode
			  serverName: (NSString **) theServerName
			    username: (NSString **) theUsername
{  
  NSString *aString = nil;
  
  if ( theFolderNode )
    {
      NSString *aServerName, *aUsername;
      NSRange aRange;
      
      aString = [Utilities completePathForFolderNode: theFolderNode
			   separator: @"/"];
      
      aRange = [aString rangeOfString: @"/"
			options: 0
			range: NSMakeRange(1, [aString length] - 1)];
  
      if ( aRange.length )
	{
	  aString = [aString substringWithRange: NSMakeRange(1, aRange.location - 1)];
	}
      else
	{
	  aString = [aString substringFromIndex: 1];
	}
      
      // aString now holds "<username> @ <store name>"
      aRange = [aString rangeOfString: @"@"
			options: NSBackwardsSearch];
      
      // We ensure that we really received a "IMAP value"
      if ( aRange.length )
	{
	  aServerName = [aString substringFromIndex: (aRange.location + 2)];
	  aUsername = [aString substringToIndex: (aRange.location - 1)];
	}
      else
	{
	  // Seems we got a Local value.
	  aServerName = nil;
	  aUsername = NSUserName();
	}
      
      if ( theServerName != NULL )
	{
	  *theServerName = aServerName;
	}
      
      if ( theUsername != NULL )
	{
	  *theUsername = aUsername;
	}
    }
  
  return aString;
}


//
//
//
+ (BOOL) URLWithString: (NSString *) theString
           matchFolder: (Folder *) theFolder
{
  URLName *theURLName;
  
  theURLName = [[URLName alloc] initWithString: theString
				path: [[NSUserDefaults standardUserDefaults] objectForKey: @"LOCALMAILDIR"]];

  if ( [[theFolder name] isEqualToString: [theURLName foldername]] )
    {
      // If's a local folder, we simply compare the protocol.
      if ( [theFolder isKindOfClass: [LocalFolder class]] )
	{
	  if ( [[theURLName protocol] caseInsensitiveCompare: @"LOCAL"] == NSOrderedSame )
	    {
	      RELEASE(theURLName);
	      return YES;
	    }
	}
      // It's an IMAP folder, we must compare the hostname and the username.
      else
	{
	  IMAPStore *aStore;

	  aStore = (IMAPStore *)[theFolder store];
	  
	  if ( [[aStore name] isEqualToString: [theURLName host]] &&
	       [[aStore username] isEqualToString: [theURLName username]] )
	    {
	      RELEASE(theURLName);
	      return YES;
	    }
	}
    }

  RELEASE(theURLName);
  return NO;
}


//
//
//
+ (id) storeForURLName: (URLName *) theURLName
{
  id aStore;
  
  if ( [[theURLName protocol] caseInsensitiveCompare: @"LOCAL"] == NSOrderedSame )
    {
      aStore = [[MailboxManagerController singleInstance]
		 storeForName: @"GNUMAIL_LOCAL_STORE"
		 username: NSUserName()];
    }
  else
    {
      if ( [[MailboxManagerController singleInstance] initializeIMAPStoreWithKey: [NSString stringWithFormat: @"%@ @ %@",
											    [theURLName username],
											    [theURLName host]]] )
	{
	  aStore = [[MailboxManagerController singleInstance]
		     storeForName: [theURLName host]
		     username: [theURLName username]];
	}
      else
	{
	  aStore = nil;
	}
    }
  
  return aStore;
}


//
//
//
+ (NSString *) stringValueOfURLNameFromFolderNode: (FolderNode *) theFolderNode
{
  NSString *aString;

  aString = [Utilities pathOfFolderFromFolderNode: theFolderNode
		       separator: @"/"];


  //
  // If it's a Local mailbox...
  //
  if ( [aString hasPrefix: _(@"Local Mailboxes")] )
    {
      // We have something like "Local Mailboxes/subfolderA/subfolderB/mbox"
      NSRange aRange;

      aRange = [aString rangeOfString: @"/"];

      // We build our URL
      aString = [NSString stringWithFormat: @"local://%@/%@",
			  [[NSUserDefaults standardUserDefaults] objectForKey: @"LOCALMAILDIR"],
			  [aString substringFromIndex: aRange.location + 1]];
    }
  //
  // It's an IMAP mailbox.
  //
  else
    {
      NSString *aServername, *aUsername, *aPathToFolder;
      NSRange r1, r2;
      
      // We have something like ludovic @ sophos.ca/subfolderA/subfolderB/mbox
      r1 = [aString rangeOfString: @"/"];  // we search for: sophos.ca/          <- the /
      r2 = [aString rangeOfString: @"@"    // we search for: ludovic@Sophos.ca/  <- the @
		    options: NSBackwardsSearch
		    range: NSMakeRange(0, r1.location)];
		    
      
      aUsername = [aString substringToIndex: r2.location - 1];
      aServername = [aString substringWithRange: NSMakeRange(r2.location + 2, r1.location - r2.location - 2)];
      aPathToFolder = [aString substringFromIndex: r1.location + 1];
      
      aString = [NSString stringWithFormat: @"imap://%@@%@/%@", aUsername, aServername, aPathToFolder];
    }
  
  return aString;
}


//
// 
//
+ (FolderNode *) initializeLocalAndIMAPFolderNodes
{
  NSDictionary *allServers;
  NSArray *allKeys;
  
  FolderNode *allNodes, *nodes;

  LocalStore *aStore;
  int i;

  allNodes = [[FolderNode alloc] init];
  
  //
  // Local
  //
  aStore = [[MailboxManagerController singleInstance] storeForName: @"GNUMAIL_LOCAL_STORE"
						      username: NSUserName()];  
  
  nodes = [Utilities folderNodesFromFolders: [aStore folderEnumerator]
		     separator: @"/"];
  [nodes setName: _(@"Local Mailboxes")];
  [allNodes addChild: nodes];
  [nodes setParent: allNodes];
    
  //
  // IMAP
  //
  allServers = [[NSUserDefaults standardUserDefaults] objectForKey: @"RECEIVING"];
  allKeys = [[allServers allKeys] sortedArrayUsingSelector: @selector(compare:)];
  
  for (i = 0; i < [allKeys count]; i++)
    {
      NSDictionary *allValues;
      NSArray *theArray;

      allValues = [allServers objectForKey: [allKeys objectAtIndex: i]];
      theArray = [allValues objectForKey: @"SUBSCRIBED_FOLDERS"];
      
      if ( theArray && [theArray count] > 0  )
	{
	  nodes = [Utilities folderNodesFromFolders: [theArray objectEnumerator]
			     separator: @"/"];
	  
	  [nodes setName: [allKeys objectAtIndex: i]];
	  
	  [allNodes addChild: nodes];
	  [nodes setParent: allNodes];
	}
    }
  
  return AUTORELEASE(allNodes);
}


//
// Usually, theFolderNodes will be a pointer to a FolderNode object
// representing the "tree" of all our Local and/or IMAP nodes.
// This "tree" is usually build using:
//
// Utilities: +initializeLocalAndIMAPFolderNodes
//
// We will ALWAYS have at least one node.
//
+ (void) addItemsToPopUpButton: (NSPopUpButton *) thePopUpButton
              usingFolderNodes: (FolderNode *) theFolderNodes
{
  int i;

  [thePopUpButton removeAllItems];
  [thePopUpButton setAutoenablesItems: NO];
    
  // We now add all our nodes
  for (i = 0; i < [theFolderNodes childCount]; i++)
    {
      [Utilities addItem: [theFolderNodes childAtIndex: i]
		 level: 0
		 toPopUpButton: thePopUpButton];
    }
  
  [thePopUpButton selectItemAtIndex: 0];
}


//
// This method add new items to the popup button.
// It respects the level parameter for indenting properly
// the items in order to represent a hierarchy.
//
+ (void) addItem: (FolderNode *) theFolderNode
	   level: (int) theLevel
   toPopUpButton: (NSPopUpButton *) thePopUpButton
{
  NSMutableString *aMutableString;
  FolderNodePopUpItem *theItem;
  int i;

  aMutableString = [[NSMutableString alloc] init];

  for (i = 0; i < theLevel; i++)
    {
      [aMutableString appendString: @"     "];
    }

  [aMutableString appendString: [theFolderNode name]];
  
  theItem = [[FolderNodePopUpItem alloc] initWithTitle: aMutableString
					 action: NULL
					 keyEquivalent: @""];
  [theItem setFolderNode: theFolderNode];
  RELEASE(aMutableString);

  // We enable / disable our item
  if ( [theFolderNode childCount] > 0 )
    {    
      [theItem setAction: NULL];
      [theItem setEnabled: NO];
    }
  else
    {
      [theItem setAction: @selector(foo:)];
      [theItem setEnabled: YES];
    }

  // We finally add our item
  [[thePopUpButton menu] addItem: theItem];
  RELEASE(theItem);

  for (i = 0; i < [theFolderNode childCount]; i++)
    {
      [Utilities addItem: [theFolderNode childAtIndex: i]
		 level: (theLevel + 1)
		 toPopUpButton: thePopUpButton];
    }
}


//
//
//
+ (FolderNodePopUpItem *) folderNodePopUpItemForFolderNode: (FolderNode *) theFolderNode
					       popUpButton: (NSPopUpButton *) thePopUpButton
{
  FolderNodePopUpItem *theItem;
  int i;

  for (i = 0; i < [thePopUpButton numberOfItems]; i++)
    {
      theItem = [thePopUpButton itemAtIndex: i];

      if ( [theItem folderNode] == theFolderNode )
	{
	  return theItem;
	}
    }

  return nil;
}


//
//
//
+ (FolderNodePopUpItem *) folderNodePopUpItemForURLNameAsString: (NSString *) theString
					       usingFolderNodes: (FolderNode *) theFolderNodes
						    popUpButton: (NSPopUpButton *) thePopUpButton
{
  FolderNodePopUpItem *aPopUpItem;
  FolderNode *aFolderNode;
  URLName *aURLName;

  aURLName = [[URLName alloc] initWithString: theString
			      path: [[NSUserDefaults standardUserDefaults] objectForKey: @"LOCALMAILDIR"]];
  
  if ( [[aURLName protocol] caseInsensitiveCompare: @"LOCAL"] == NSOrderedSame )
    {
      aFolderNode = [Utilities folderNodeForPath: [NSString stringWithFormat: @"%@/%@", _(@"Local Mailboxes"),
							    [aURLName foldername]]
			       
			       using: theFolderNodes];
    }
  else
    {
      aFolderNode = [Utilities folderNodeForPath: [NSString stringWithFormat: @"%@ @ %@/%@", [aURLName username],
							    [aURLName host], [aURLName foldername]]
			       
			       using: theFolderNodes];
    }
  
  aPopUpItem =  [Utilities folderNodePopUpItemForFolderNode: aFolderNode
			   popUpButton: thePopUpButton];

  RELEASE(aURLName);

  return aPopUpItem;
}


//
//
//
+ (NSString *) passwordForKey: (NSString *) theKey
		       prompt: (BOOL) aBOOL;
{
  NSDictionary *allPOP3Values;
  NSString *aPassword;

  allPOP3Values = [[[NSUserDefaults standardUserDefaults] objectForKey: @"RECEIVING"] 
		    objectForKey: theKey];
  
  // We verify in the user defaults
  aPassword = [Utilities decryptPassword: [allPOP3Values objectForKey: @"PASSWORD"]
			 withKey: theKey];
  
  // We verify in our cache
  if ( !aPassword )
    {
      aPassword = [passwordCache objectForKey: theKey];
    }

  // If we must prompt for the password
  if ( !aPassword && aBOOL )
    {
      PasswordPanelController *theController; 
      int result;
      
      theController = [[PasswordPanelController alloc] initWithWindowNibName: @"PasswordPanel"];
      [[theController window] setTitle: [NSString stringWithFormat: @"%@ @ %@",
						  [allPOP3Values objectForKey: @"USERNAME"],
						  [allPOP3Values objectForKey: @"SERVERNAME"]]];
      
      result = [NSApp runModalForWindow: [theController window]];
      
      // If the user has entered a password...
      if (result == NSRunStoppedResponse)
	{
	  aPassword = [theController password];
	  
	  // Let's cache this password...
	  [passwordCache setObject: aPassword
			 forKey: theKey];
	}
      else
	{
	  aPassword = nil;
	}
      
      RELEASE(theController);
    }
  
  return aPassword;
}


//
//
//
+ (NSMutableDictionary *) passwordCache
{
  return passwordCache;
}


//
// Creates a reply
//
+ (void) replyToSender: (Message *) theMessage
		folder: (Folder *) theFolder
{
  EditWindowController *editWindowController;
  NSString *aProfileName, *aProfileAddress;
  NSDictionary *allPersonalProfiles;
  Message *aMessage;
  
  BOOL shouldReplyToAll;
  int i;
  
  if ( !theMessage )
    {
      NSBeep();
      return;
    }

  // guess the profile we should be using
  aProfileName = [self _guessPersonalProfile: theMessage];

  if ( aProfileName )
    {
      allPersonalProfiles = [[NSUserDefaults standardUserDefaults] objectForKey: @"PERSONAL"];
      aProfileAddress = [[allPersonalProfiles objectForKey: aProfileName] objectForKey: @"EMAILADDR"];
    }
  else
    {
      aProfileAddress = nil;
    }

  NSDebugLog(@"profile = |%@|", aProfileName);
  NSDebugLog(@"profileAddress = |%@|", aProfileAddress);
  
  if ( [theMessage recipientsCount] > 1 )
    {
      int choice;
      
      choice = NSRunAlertPanel(_(@"Reply..."),
			       _(@"Would you like to reply to all recipients?"),
			       _(@"No"),  // default
			       _(@"Yes"), // alternate
			       nil);
      
      if ( choice == NSAlertAlternateReturn )
        {
	  shouldReplyToAll = YES;
	}
      else
	{
	  shouldReplyToAll = NO;
	}
    }
  else
    {
      shouldReplyToAll = NO;
    }
  
  
  // We create our window controller
  editWindowController = [[EditWindowController alloc] initWithWindowNibName: @"EditWindow"];

  if ( editWindowController )
    {
      [[editWindowController window] setTitle: _(@"Reply to a message...")];
      [editWindowController setSignaturePosition: 
			      [[NSUserDefaults standardUserDefaults]
				integerForKey: @"SIGNATURE_REPLY_POSITION"] ];
      [editWindowController setShowCc: shouldReplyToAll];
      
      // We set the original message
      [editWindowController setUnmodifiedMessage: theMessage];
      
      // We create a replied copy of our message and we retain it
      aMessage = [theMessage replyWithReplyToAll: shouldReplyToAll];
      RETAIN(aMessage);
      
      // If we are in the Outbox folder, we replace the recipient with the original recipients
      // of the messages.
      if ( [Utilities URLWithString: [[NSUserDefaults standardUserDefaults] objectForKey: @"OUTBOXFOLDERNAME"]
		      matchFolder: theFolder] )
	{
          [aMessage setRecipients: [[editWindowController unmodifiedMessage] recipients]];
	}
      
      // remove original recipient from recipient list
      if ( shouldReplyToAll && aProfileAddress )
	{
	  for (i = ([aMessage recipientsCount] - 1); i >= 0; i--)
	    {
	      NSDebugLog(@"Recipient is %@", [(InternetAddress*)[[aMessage recipients] objectAtIndex:i] address]);
	      
	      if( [[(InternetAddress*)[[aMessage recipients] objectAtIndex: i] address] 
		    caseInsensitiveCompare: aProfileAddress] == NSOrderedSame )
		{
		  [aMessage removeFromRecipients: (InternetAddress*)[[aMessage recipients] objectAtIndex: i]];
		  break;
		}
	    }		  
	}
      
      
      [editWindowController setMessage: aMessage];
      RELEASE(aMessage);
      
      // set the appropriate profile
      if ( aProfileName )
	{
	  [editWindowController setProfile: aProfileName];
	}
      
      [editWindowController showWindow: self];
    }
}


//
// Constructs a reply to all recipients
//
+ (void) replyToAll: (Message *) theMessage
	     folder: (Folder *) theFolder
{
  EditWindowController *editWindowController;
  NSString *aProfileName, *aProfileAddress;
  NSDictionary *allPersonalProfiles;
  Message *aMessage;
  
  int i;
  
  if ( !theMessage ) 
    {
      NSBeep();
      return;
    }

  // guess the profile we should be using
  aProfileName = [self _guessPersonalProfile: theMessage];
  
  if ( aProfileName )
    {
      allPersonalProfiles = [[NSUserDefaults standardUserDefaults] objectForKey: @"PERSONAL"];
      aProfileAddress = [[allPersonalProfiles objectForKey: aProfileName] objectForKey: @"EMAILADDR"];
    }
  else
    {
      aProfileAddress = nil;
    }

  NSDebugLog(@"profile = |%@|", aProfileName);
  NSDebugLog(@"profileAddress = |%@|", aProfileAddress);
  
  // We create our window controller
  editWindowController = [[EditWindowController alloc] initWithWindowNibName: @"EditWindow"];
  
  if ( editWindowController )
    {
      [[editWindowController window] setTitle: _(@"Reply to a message...")];
      [editWindowController setSignaturePosition: 
			      [[NSUserDefaults standardUserDefaults]
				integerForKey: @"SIGNATURE_REPLY_POSITION"] ];
      [editWindowController setShowCc: YES];
      
      // We set the original message
      [editWindowController setUnmodifiedMessage: theMessage];
      
      // We create a replied copy of our message and we retain it
      aMessage = [theMessage replyWithReplyToAll: YES];
      RETAIN(aMessage);
      
      // If we are in the Outbox folder, we replace the recipient with the original recipients
      // of the messages.
      if ( [Utilities URLWithString: [[NSUserDefaults standardUserDefaults] objectForKey: @"OUTBOXFOLDERNAME"]
		      matchFolder: theFolder] )
	{
	  [aMessage setRecipients: [[editWindowController unmodifiedMessage] recipients]];
	}
      
      // remove original recipient from recipient list
      if ( aProfileAddress )
	{
	  for (i = ([aMessage recipientsCount] - 1); i >= 0; i--)
	    {
	      NSDebugLog(@"Recipient is %@", [(InternetAddress*)[[aMessage recipients] objectAtIndex:i] address]);
	      
	      if ( [[(InternetAddress*)[[aMessage recipients] objectAtIndex:i] address]
		     caseInsensitiveCompare: aProfileAddress] == NSOrderedSame )
		{
		  [aMessage removeFromRecipients: (InternetAddress*)[[aMessage recipients] objectAtIndex: i]];
		  break;
		}
	    }
	}
      
	  
      [editWindowController setMessage: aMessage];
      RELEASE(aMessage);
      
      // set the appropriate profile
      if ( aProfileName )
	{
	  [editWindowController setProfile: aProfileName];	  
	}	
      
      [editWindowController showWindow: self];
    }
}


//
//
//
+ (void) forwardMessage: (Message *) theMessage
{
  EditWindowController *controller;
  Message *aMessage;
  NSString *profile;
  
  if ( !theMessage ) 
    {
      NSBeep();
      return;
    }

  // guess the profile we should be using
  profile = [self _guessPersonalProfile: theMessage];
  
  // We create a forwarded copy of our message and we retain it
  aMessage = [theMessage forward];
  RETAIN(aMessage);
  
  controller = [[EditWindowController alloc] initWithWindowNibName: @"EditWindow"];
  
  if ( controller )
    {
      [[controller window] setTitle: _(@"Forward a message...")];
      [controller setSignaturePosition: [[NSUserDefaults standardUserDefaults] 
					  integerForKey: @"SIGNATURE_FORWARD_POSITION"] ];
      [controller setMessage: aMessage];
      [controller setShowCc: NO];
      
      // set the appropriate profile
      if ( profile )
	{
	  [controller setProfile: profile];	  
	}
      
      [controller showWindow: self];
    }
  
  RELEASE(aMessage);
}


//
// This method displays a message in the textView.
//
+ (void) showMessage: (Message *) theMessage
	      target: (NSTextView *) theTextView 
      showAllHeaders: (BOOL) headersFlag
{
  if ( theMessage )
    {
      Flags *theFlags;
      id aDelegate;
      int i;
      
      
      // If the content of the message has neven been parsed before, we do it now!
      if ( ![theMessage isInitialized] )
	{
	  [theMessage setInitialized: YES];
	}
      
      // We clear our 'Save Attachment' menu
      [(GNUMail *)[NSApp delegate] removeAllItemsFromMenu: [(GNUMail *)[NSApp delegate] saveAttachmentMenu]];
      
      // We begin by clearing what we have in our text view
      [[theTextView textStorage] deleteCharactersInRange:
				    NSMakeRange(0, [[theTextView textStorage] length])];
      
      // We inform our bundles that the message WILL BE shown in the text view
      for (i = 0; i < [[GNUMail allBundles] count]; i++)
	{
	  id<GNUMailBundle> aBundle;
	  
	  aBundle = [[GNUMail allBundles] objectAtIndex: i];

	  if ( [aBundle respondsToSelector: @selector(messageWillBeDisplayed:inView:)] )
	    {
	      [aBundle messageWillBeDisplayed: theMessage
		       inView: theTextView];
	    }
	}

      // Then, we add our headers and our content
      [[theTextView textStorage] appendAttributedString: 
				   [Utilities attributedStringFromHeadersForMessage: theMessage
					      showAllHeaders: headersFlag] ];
      
      [[theTextView textStorage] appendAttributedString:
				   [Utilities _quoteMessageContentFromAttributedString:			     
						[Utilities formattedAttributedStringFromAttributedString:
							     [Utilities attributedStringFromContentForPart: 
									  theMessage]]] ]; 
      
      // We update the Flags of our message (we add SEEN and remove RECENT)
      theFlags = [theMessage flags];
      
      // If the message was a new one, let's change the app icon back to GNUMail.tiff
      if ( [theFlags contain: RECENT] )
	{
	  [NSApp setApplicationIconImage: [NSImage imageNamed: @"GNUMail.tiff"]];
	}
      
      if ( ![theFlags contain: SEEN] ) 
	{ 
	  [theFlags add: SEEN];
	}
      
      [[theMessage flags] remove: RECENT];
      
      // We ensure that the selected table row is properly refreshed after changing the flags
      // FIXME: find a more elegant solution.
      aDelegate = [[GNUMail lastMailWindowOnTop] delegate];
      
      if ( aDelegate )
	{
	  NSTableView *aTableView;
	  
	  aTableView = nil;

	  if ( [aDelegate isKindOfClass: [MailWindowController class]] )
	    {
	      aTableView = [aDelegate tableView];
	    }
	  else if ( [aDelegate isKindOfClass: [MessageViewWindowController class]] )
	    {
	      aTableView = [[aDelegate mailWindowController] tableView];
	    }
	  
	  if ( aTableView )
	    {
	      [aTableView setNeedsDisplayInRect: [aTableView rectOfRow: [aTableView selectedRow]]];		  
	    }
	}
      
      // We finally highlight the URLs in our message, if we want to
      if ( [[NSUserDefaults standardUserDefaults] objectForKey: @"HIGHLIGHT_URL"] &&
	   [[[NSUserDefaults standardUserDefaults] objectForKey: @"HIGHLIGHT_URL"] intValue] == NSOnState )
	{
          [Utilities _highlightAndActivateURLs: theTextView];
	}

      // We inform our bundles that the message HAS BEEN shown in the text view
      for (i = 0; i < [[GNUMail allBundles] count]; i++)
	{
	  id<GNUMailBundle> aBundle;
	  
	  aBundle = [[GNUMail allBundles] objectAtIndex: i];
	  
	  if ( [aBundle respondsToSelector: @selector(messageWasDisplayed:inView:)] )
	    {
	      [aBundle messageWasDisplayed: theMessage
		       inView: theTextView];
	    }
	}

      // If we have more than one attachement, we create a 'save all' menu at the top
      if ( [[(GNUMail *)[NSApp delegate] saveAttachmentMenu] numberOfItems] > 1 )
        {
          NSMenuItem *aMenuItem;

	  aMenuItem = [[NSMenuItem alloc] init];
          [aMenuItem setTitle: _(@"Save all")];
	  [aMenuItem setTarget: [NSApp delegate]];
          [aMenuItem setAction: @selector(saveAllAttachments:)];
          [aMenuItem setKeyEquivalent: @""];
          [[(GNUMail *)[NSApp delegate] saveAttachmentMenu] insertItem: aMenuItem
							    atIndex: 0];
	  
          RELEASE(aMenuItem);
        }
    }
  else
    {
      NSDebugLog(@"Unable to find the message in the hashtable!");
    }
  
  // We set the flag to hide our all our headers on the next click
  //[self setShowAllHeaders: NO];

  // We scroll to the beginning of the message and we remove any previous text selection
  [theTextView scrollRangeToVisible: NSMakeRange(0,0)];
  [theTextView setSelectedRange: NSMakeRange(0,0)];
}


//
//
//
+ (void) showMessageRawSource: (Message *) theMessage 
		       target: (NSTextView *) theTextView
{
  if ( theMessage && theTextView )
    {
      NSString *aString;
     
      aString = [[NSString alloc] initWithData: [theMessage rawSource]
				  encoding: NSASCIIStringEncoding];
      
      if ( aString )
	{
	  NSAttributedString *theAttributedString;
	  NSDictionary *attributes;
	  
	  attributes = [NSDictionary dictionaryWithObject: [NSFont userFixedPitchFontOfSize: 0]
				     forKey: NSFontAttributeName];
	  
	  theAttributedString = [[NSAttributedString alloc] initWithString: aString
							    attributes: attributes];
	  
	  [[theTextView textStorage] setAttributedString: theAttributedString];
	  
	  RELEASE(theAttributedString);
	  RELEASE(aString);
	}
    }
  else
    {
      NSBeep();
    }
}


//
//
//
+ (void) clickedOnCell: (id <NSTextAttachmentCell>) attachmentCell
	        inRect: (NSRect) cellFrame
               atIndex: (unsigned) charIndex
{
  NSTextAttachment *attachment;
  NSFileWrapper *filewrapper;
  MimeType *aMimeType;
  
  // If it's our header cell, we immediately return
  if ( [attachmentCell isKindOfClass: [MailHeaderCell class]] )
    {
      return;
    }
  
  attachment = [attachmentCell attachment];
  filewrapper = [attachment fileWrapper];
  aMimeType = nil;

  aMimeType = [[MimeTypeManager singleInstance] mimeTypeForFileExtension:
						  [[filewrapper preferredFilename] pathExtension]];
  
  if ( !aMimeType || [aMimeType action] == 0 )
    {
      NSSavePanel *aSavePanel;
      int aChoice;

      aSavePanel = [NSSavePanel savePanel];
      [aSavePanel setAccessoryView: nil];
      [aSavePanel setRequiredFileType: @""];

      aChoice = [aSavePanel runModalForDirectory: [GNUMail currentWorkingPath]
			    file: [filewrapper preferredFilename]];

      /* if successful, save file under designated name */
      if (aChoice == NSOKButton)
	{
	  if (![filewrapper writeToFile: [aSavePanel filename]
			    atomically: YES
			    updateFilenames: YES] )
	    {
	      NSBeep();
	    }

          [GNUMail setCurrentWorkingPath: [[aSavePanel filename] stringByDeletingLastPathComponent]];
	}
    }
  else
    {
      NSString *aString;

      if ( ![[NSFileManager defaultManager] fileExistsAtPath: [aMimeType dataHandlerCommand]] )
	{
	  
	  NSRunAlertPanel(_(@"Error!"),
			  _(@"The external program (%@) for opening this MIME-Type (%@) can't be found."),
			  _(@"OK"),
			  NULL,
			  NULL,
			  [aMimeType dataHandlerCommand], [aMimeType mimeType]);
	  return;
	}
      
      aString = [NSString stringWithFormat:@"%@/%d_%@", NSTemporaryDirectory(), 
			  [[NSProcessInfo processInfo] processIdentifier],
			  [filewrapper preferredFilename]];

      if ( [filewrapper writeToFile: aString
			atomically:YES
			updateFilenames: NO] )
	{
	  NSDictionary *aDictionary;

	  aDictionary = [NSDictionary dictionaryWithObjectsAndKeys: [aMimeType dataHandlerCommand], @"command",
				      [NSArray arrayWithObjects: aString, nil], @"arguments",
				      aString, @"file", nil];
	  
	  
	  [NSThread detachNewThreadSelector: @selector(_launchExternalProgram:)
		    toTarget: self
		    withObject: aDictionary];
	}
      else
	{
	  NSBeep();
	}
    }
}

@end



//
// Private implementation for Utilities
//
@implementation Utilities (Private)


//
//
//
+ (NSMutableAttributedString *) _quoteMessageContentFromAttributedString: (NSAttributedString *) theAttributedString
{
  NSMutableAttributedString *aMutableAttributedString;
  NSDictionary *attributes;
  NSArray *anArray;
  int i;

  aMutableAttributedString = [[NSMutableAttributedString alloc] initWithAttributedString: theAttributedString];
  
  // We get the range of all our 'quoted' strings
  // FIXME:
  // With the following regex we don't get the first line, 
  // so we should test the first characters.
  // "^" doesn't seem to work	
  anArray = [NSRegEx matchString: [theAttributedString string]
		     withPattern: @"\n[>]+[^\n]*" 
		     isCaseSensitive: NO];
    
  //
  // We use the '>' as the indicator of depth.
  // We count the number of '>' at the beginning
  // of a line. We don't care if there are spaces,
  // and we stop our count when we 1) reach the end
  // of the line 2) or get a character != of ' ' or '>'
  //
  if ( [anArray count] > 0 )
    {
      NSMutableArray *colors;
      
      colors = [[NSMutableArray alloc] init];
      
      // We look at the user preferences to get the colors...
      
      if ( [[NSUserDefaults standardUserDefaults] colorForKey: @"QUOTE_COLOR_LEVEL_1"] )
	{
	  [colors addObject: [[NSUserDefaults standardUserDefaults] colorForKey: @"QUOTE_COLOR_LEVEL_1"]];
	}	
      else
	{
	  [colors addObject: [NSColor blueColor]];
	}
      
      if ( [[NSUserDefaults standardUserDefaults] colorForKey: @"QUOTE_COLOR_LEVEL_2"] )
	{
	  [colors addObject: [[NSUserDefaults standardUserDefaults] colorForKey: @"QUOTE_COLOR_LEVEL_2"]];
	}	
      else
	{
	  [colors addObject: [NSColor redColor]];
	}
      
      if ( [[NSUserDefaults standardUserDefaults] colorForKey: @"QUOTE_COLOR_LEVEL_3"] )
	{
	  [colors addObject: [[NSUserDefaults standardUserDefaults] colorForKey: @"QUOTE_COLOR_LEVEL_3"]];
	}	
      else
	{
	  [colors addObject: [NSColor greenColor]];
	}
      
      if ( [[NSUserDefaults standardUserDefaults] colorForKey: @"QUOTE_COLOR_LEVEL_4"] )
	{
	  [colors addObject: [[NSUserDefaults standardUserDefaults] colorForKey: @"QUOTE_COLOR_LEVEL_4"]];
	}
      else
	{
	  [colors addObject: [NSColor cyanColor]];
	}
      
      for (i = ([anArray count] - 1); i >= -1; i--)
	{
	  int j, level, begin, end;
	  NSRange aRange;

	  level = begin = end = 0;
	  
	  if ( i == -1 ) // Hack to check the first line (don't want to duplicate the following loop)
	    {
	      aRange = [[anArray objectAtIndex: 0] rangeValue];
	      end = aRange.location;
	    }
	  else
	    {
	      aRange = [[anArray objectAtIndex: i] rangeValue];
	      begin = aRange.location+1; // don't take the '\n'
	      end = aRange.location + aRange.length;
	    }
	  
	  for ( j = begin; j < end; j++ )
	    {
	      unichar ch;
	      
	      ch = [[aMutableAttributedString string] characterAtIndex: j];
	      
	      if ( ch == '>')
		{
		  level++;
		}
	      else if ( ch != ' ')
		{
		  break;
		}
	    }
	  
	  // As we check the first line, level *could* be equal to zero
	  if ( level > 0 )
	    {
	      level--;
	      
	      // We cycle the colors...
	      level = level % [colors count];
	      
	      // We are on the first line (i==-1)
	      if (i == -1)
		{
		  aRange = NSMakeRange (0,aRange.location);
		}
	      
	      attributes = [NSDictionary dictionaryWithObjectsAndKeys: 
					   [colors objectAtIndex: level], 
					 NSForegroundColorAttributeName, 
					 nil];
	      
	      [aMutableAttributedString addAttributes: attributes range: aRange];
	    }
	}

      RELEASE(colors);
      
    } // if ( [anArray count] > 0 ) ...
  
  return AUTORELEASE(aMutableAttributedString);
}


//
//
//
+ (void) _highlightAndActivateURLs: (NSTextView *) targetView
{
  NSRange searchRange, foundRange;
  NSString *aString, *urlString;
  NSEnumerator *urlEnumerator;
  NSArray *urlArray;

  urlArray = [NSArray arrayWithObjects: @"http://", @"https://", @"ftp://", @"file://", nil];
  urlEnumerator = [urlArray objectEnumerator];
  
  [[targetView textStorage] beginEditing];
  
  while ( (urlString = (NSString *)[urlEnumerator nextObject]) )
    {
      aString = [targetView string];
      searchRange = NSMakeRange(0, [aString length]);
      
      do
	{
	  foundRange = [aString rangeOfString: urlString
				options: 0
				range: searchRange];
	  
	  
	  // If we found an URL...
	  if ( foundRange.length > 0 )
	    {
	      NSDictionary *linkAttributes;
	      NSRange endOfURLRange;
	      NSURL *anURL;
	      
	      // Restrict the searchRange so that it won't find the same string again
	      searchRange.location = foundRange.location + foundRange.length;
	      searchRange.length = [aString length] - searchRange.location;
	      
	      // We assume the URL ends with whitespace
	      endOfURLRange = [aString rangeOfCharacterFromSet: [NSCharacterSet whitespaceAndNewlineCharacterSet]
				       options: 0
				       range: searchRange];
	      
	      // The URL could also end at the end of the text.  The next line fixes it in case it does
	      if ( endOfURLRange.length == 0)
		{
		  endOfURLRange.location= [aString length];
		}
	      
	      // Set foundRange's length to the length of the URL
	      foundRange.length = endOfURLRange.location - foundRange.location;
	      
	      // We create our URL object from our substring
	      anURL = [NSURL URLWithString: [aString substringWithRange: foundRange]];
	      
	      
	      // Make the link attributes
	      linkAttributes= [NSDictionary dictionaryWithObjectsAndKeys: anURL, NSLinkAttributeName,
					    [NSNumber numberWithInt: NSSingleUnderlineStyle], NSUnderlineStyleAttributeName,
					    [NSColor blueColor], NSForegroundColorAttributeName,
					    NULL];
	      
	      // Finally, apply those attributes to the URL in the text
	      [[targetView textStorage] addAttributes: linkAttributes
					range: foundRange];
	    }
	  
	} while ( foundRange.length != 0 );
    }
  
  [[targetView textStorage] endEditing];
}


//
// Given a message, guesses the profile it is associated with by comparing
// recipients with profile addresses. Currently compares address domains
// only (string after '@') to handle email aliases. Once email alias support
// has been added to profiles, we will compare full email addresses.
//
+ (NSString *) _guessPersonalProfile: (Message *)theMessage
{
  NSDictionary *allPersonalProfiles, *thisProfile;
  NSArray *allProfileKeys, *recipients;
  InternetAddress *iAddress;
  NSString *profile;
  int i, j;
  
  profile = nil;

  // get all the message recipients
  recipients = [theMessage recipients];
  
  // get all our personal profiles
  allPersonalProfiles = [[NSUserDefaults standardUserDefaults] objectForKey: @"PERSONAL"];
  
  // get all the keys for the personal profiles
  allProfileKeys = [[allPersonalProfiles allKeys] sortedArrayUsingSelector: @selector(caseInsensitiveCompare:)];
  
  // walk through the profiles and check if the a message recipient matches a profile address.
  for (i = 0; i < [allProfileKeys count]; i++)
    {
      NSString *profileAddress, *profileDomain;
      NSRange aRange;
      
      // grab the profile corresponding to this key
      thisProfile = [allPersonalProfiles objectForKey:[allProfileKeys objectAtIndex:i]];
      
      if ( thisProfile )
	{
	  // grab the address for this profile
	  profileAddress = [thisProfile objectForKey: @"EMAILADDR"];
	  
	  if ( profileAddress && recipients )
	    {
	      // walk through the recipients and check against profile address
	      for (j = 0; j < [recipients count]; j++)
		{
		  // check only the domain part of profile address (after '@')
		  // this is to handle email aliases.
		  // this will of course break if to profiles have the same domains.
		  aRange = [profileAddress rangeOfString: @"@"
					   options: NSBackwardsSearch];
		  profileDomain = [profileAddress substringFromIndex: NSMaxRange(aRange)];
		  iAddress = (InternetAddress *)[recipients objectAtIndex: j];
		  aRange = [[iAddress address] rangeOfString: profileDomain];
		  
		  if ( aRange.length > 0 )
		    {
		      NSDebugLog(@"Profile to be used = %@", [allProfileKeys objectAtIndex: i]);
		      
		      // we found a matching profile; get out!
		      profile = [allProfileKeys objectAtIndex: i];
		      return profile;
		    }
		}
	    }
	}
    }
  
  // we did not find a atching profile, return nil
  return nil;
}


@end


//
//
//
NSString *GNUMailUserLibraryPath()
{
  NSString *aString;

  aString = [NSString stringWithFormat: @"%@/GNUMail", 
		      [NSSearchPathForDirectoriesInDomains(NSLibraryDirectory, NSUserDomainMask, YES)
							  objectAtIndex:0] ];

  return aString;
}


//
// This C function has been written by Abhijit Menon-Sen <ams@wiw.org>
// This code is in the public domain.
//
char *striphtml(char *s)
{
    int sgml = 0, tag = 0;
    char c, last = '\0', quote = '\0', *t, *text;

    if ((t = text = malloc(strlen(s)+1)) == NULL)
        return NULL;

    while ((c = *s++)) {
        if (c == quote) {
            if (c == '-' && last != '-')
                goto next;
            else
                last = '\0';
            quote = '\0';
        }
        else if (!quote) {
            switch (c) {
            case '<':
                tag = 1;
                if (*s++ == '!')
                    sgml = 1;
                break;
            case '>':
                if (tag)
                    sgml = tag = 0;
                break;
            case '-':
                if (sgml && last == '-')
                    quote = '-';
                break;
            /* case '"':
            case '\'':
                if (tag)
                    quote = c;
                break; */
            case '&':
                *t++ = ent(&s);
                break;
            default:
                if (!tag)
                    *t++ = c;
                break;
            }
        }

    next:
        last = c;
    }

    return text;
}


//
// This C function has been written by Abhijit Menon-Sen <ams@wiw.org>
// This code is in the public domain.
//
char ent(char **ref)
{
    int i;
    char c = ' ', *s = *ref, *t = s;

    struct {
        char *name;
        char chr;
    } refs[] = {
        { "lt"    , '<'       },
        { "gt"    , '>'       },
        { "amp"   , '&'       },
        { "quot"  , '"'       },
        { "nbsp"  , (char)160 },
        { "iexcl" , (char)161 },
        { "cent"  , (char)162 },
        { "pound" , (char)163 },
        { "curren", (char)164 },
        { "yen"   , (char)165 },
        { "brvbar", (char)166 },
        { "sect"  , (char)167 },
        { "uml"   , (char)168 },
        { "copy"  , (char)169 },
        { "ordf"  , (char)170 },
        { "laquo" , (char)171 },
        { "not"   , (char)172 },
        { "shy"   , (char)173 },
        { "reg"   , (char)174 },
        { "macr"  , (char)175 },
        { "deg"   , (char)176 },
        { "plusmn", (char)177 },
        { "sup2"  , (char)178 },
        { "sup3"  , (char)179 },
        { "acute" , (char)180 },
        { "micro" , (char)181 },
        { "para"  , (char)182 },
        { "middot", (char)183 },
        { "cedil" , (char)184 },
        { "sup1"  , (char)185 },
        { "ordm"  , (char)186 },
        { "raquo" , (char)187 },
        { "frac14", (char)188 },
        { "frac12", (char)189 },
        { "frac34", (char)190 },
        { "iquest", (char)191 },
        { "Agrave", (char)192 },
        { "Aacute", (char)193 },
        { "Acirc" , (char)194 },
        { "Atilde", (char)195 },
        { "Auml"  , (char)196 },
        { "Aring" , (char)197 },
        { "AElig" , (char)198 },
        { "Ccedil", (char)199 },
        { "Egrave", (char)200 },
        { "Eacute", (char)201 },
        { "Ecirc" , (char)202 },
        { "Euml"  , (char)203 },
        { "Igrave", (char)204 },
        { "Iacute", (char)205 },
        { "Icirc" , (char)206 },
        { "Iuml"  , (char)207 },
        { "ETH"   , (char)208 },
        { "Ntilde", (char)209 },
        { "Ograve", (char)210 },
        { "Oacute", (char)211 },
        { "Ocirc" , (char)212 },
        { "Otilde", (char)213 },
        { "Ouml"  , (char)214 },
        { "times" , (char)215 },
        { "Oslash", (char)216 },
        { "Ugrave", (char)217 },
        { "Uacute", (char)218 },
        { "Ucirc" , (char)219 },
        { "Uuml"  , (char)220 },
        { "Yacute", (char)221 },
        { "THORN" , (char)222 },
        { "szlig" , (char)223 },
        { "agrave", (char)224 },
        { "aacute", (char)225 },
        { "acirc" , (char)226 },
        { "atilde", (char)227 },
        { "auml"  , (char)228 },
        { "aring" , (char)229 },
        { "aelig" , (char)230 },
        { "ccedil", (char)231 },
        { "egrave", (char)232 },
        { "eacute", (char)233 },
        { "ecirc" , (char)234 },
        { "euml"  , (char)235 },
        { "igrave", (char)236 },
        { "iacute", (char)237 },
        { "icirc" , (char)238 },
        { "iuml"  , (char)239 },
        { "eth"   , (char)240 },
        { "ntilde", (char)241 },
        { "ograve", (char)242 },
        { "oacute", (char)243 },
        { "ocirc" , (char)244 },
        { "otilde", (char)245 },
        { "ouml"  , (char)246 },
        { "divide", (char)247 },
        { "oslash", (char)248 },
        { "ugrave", (char)249 },
        { "uacute", (char)250 },
        { "ucirc" , (char)251 },
        { "uuml"  , (char)252 },
        { "yacute", (char)253 },
        { "thorn" , (char)254 },
        { "yuml"  , (char)255 }
    };

    while (isalpha(*s) || isdigit(*s) || *s == '#')
        s++;

    for (i = 0; i < sizeof(refs)/sizeof(refs[0]); i++) {
        if (strncmp(refs[i].name, t, s-t) == 0) {
            c = refs[i].chr;
            break;
        }
    }

    if (*s == ';')
        s++;

    *ref = s;
    return c;
}
